// #define VERBOSE

#ifdef WIN32
    #define WINDOWS
#endif

// Defined by AmigaOS are: IMPORT, UBYTE, UWORD, SWORD, ULONG, LONG, TEXT, STRPTR.

#define AUTO                   auto   /* automatic variables */
#define TRANSIENT              auto   /* automatic variables */
#define MODULE                 static /* external static (file-scope) */
#define PERSIST                static /* internal static (function-scope) */
#define DISCARD                (void) /* discarded return values */
#define elif                   else if
#define acase                  break; case
#define adefault               break; default
#define EXPORT

#ifndef __LCLINT__
    #define ASM    __asm
    #define REG(x) register __ ## x
#endif

typedef signed char        FLAG;    /*  8-bit signed quantity (replaces BOOL) */
typedef signed char        SBYTE;   /*  8-bit signed quantity (replaces Amiga BYTE) */
typedef signed short       SWORD;   /* 16-bit signed quantity (replaces Amiga WORD) */
typedef signed long        SLONG;   /* 32-bit signed quantity (same as LONG) */
typedef unsigned long      UHUGE;
typedef long               SHUGE;

#ifdef AMIGA
    struct RECT
    {   SHORT left, top, right, bottom;
    };
    typedef struct RECT RECT;
#endif

// #define PROFILE
// whether or not to enable instruction profiling
//  (Slash (/) (AmiArcadia) or Ctrl+I (WinArcadia) to view profile report)
// #define SUPPRESSUNPROFILED
// if SUPPRESSUNPROFILED is defined, opcodes which have not been executed
//  at all will not be shown in the profile report.

// #define COVERAGE
// whether or not to enable coverage reporting
//  (Backslash (\) (AmiArcadia) or Ctrl+Shift+I (WinArcadia) to view
//  coverage report)
// #define SUPPRESSUNCOVERED
// if SUPPRESSUNCOVERED is defined, bytes which have not been covered
//  (accessed) at all will not be shown in the coverage report.
#define COVERAGE_ADDRESS 8
#define COVERAGE_EXECUTE 4
#define COVERAGE_READ    2
#define COVERAGE_WRITE   1

#ifdef AMIGA
    #define MAX_PATH       512 // you will have problems if this is too low!
#endif
// it's already built-in on IBM-PC

#define KILOBYTE          1024
#define ONE_MILLION    1000000
#define ONE_BILLION 1000000000

#define MAXBOXWIDTH  (BOXWIDTH  * 6)
#define MAXBOXHEIGHT (BOXHEIGHT * 3)

// true CPL is about 56.75
#define PVI_CPL 56 // 56.75 or more doesn't work for Elektor monitor software (missing a text row)

#define KIND_ILBM  0
#define KIND_ACBM  1
#define KIND_BMP   2
// screenshot kinds must be first, savescreenshot() depends on it
#define KIND_BIN   3
#define KIND_PGM   4
#define KIND_TVC   5
#define KIND_EAS   6
#define KIND_INS   7
#define KIND_EAR   8
#define KIND_INR   9
#define KIND_TVS  10
#define KIND_TVR  11
#define KINDS     12 // counting from 0

#define PMSK    0x1fff          /* mask page offset */
#define PLEN    0x2000          /* page length */
#define PAGE    0x6000          /* mask page */
#define AMSK    0x7fff          /* mask address range */

// don't change these!
#define REGION_NTSC 0
#define REGION_PAL  1

// bitfield
#define NOREAD       1
#define NOWRITE      2
#define READONCE     4
#define AUDIBLE      8
#define MIRRORED_R  16
#define MIRRORED_W  32

#ifdef AMIGA
    #define IMAGE_STOP       18
    #define IMAGE_PLAY       19
    #define IMAGE_RECORD     20
    #define RECMODE_NORMAL   IMAGE_STOP
    #define RECMODE_PLAY     IMAGE_PLAY
    #define RECMODE_RECORD   IMAGE_RECORD
#endif
#ifdef WIN32
    #define RECMODE_NORMAL   1000000000
    #define RECMODE_PLAY     1000000001
    #define RECMODE_RECORD   1000000002
#endif

#define NET_OFF           0
#define NET_CLIENT        1
#define NET_SERVER        2

#define NETPORT        6666

#define A_OFFSET_UPPERSCREEN 6144 // $1800
#define A_OFFSET_LOWERSCREEN 6656 // $1A00
#define A_OFFSET_SPRITES     6528 // $1980

#define SENSITIVITY_VLOW    1
#define SENSITIVITY_LOW     2
#define SENSITIVITY_MEDIUM  3
#define SENSITIVITY_HIGH    4
#define SENSITIVITY_VHIGH   5
#define SENSITIVITY_DEFAULT 3
#define SENSITIVITY_MAX     5

#define OUTOFRANGE 32768

#define CHECKINPUT if (trace) checkinput();

#ifdef AMIGA
    #define T1 ;
    #define T2 ;
#endif
#ifdef WIN32
    #define T1 if (!consoleopen) openconsole();
    #define T2 reactivate();
#endif

// (amiga1|amiga2|ibm).c
EXPORT void updatescreen(void);
EXPORT void redrawscreen(void);
EXPORT void readkybd(void);
EXPORT UBYTE KeyDown(UWORD scancode);
EXPORT void checkinput(void);
EXPORT void cleanexit(SBYTE rc);
EXPORT UWORD ReadJoystick(UWORD joynum);
EXPORT void playsound(FLAG force);
EXPORT void settitle(void);
EXPORT ULONG getsize(STRPTR filename);
EXPORT void pause(void);
EXPORT void unpause(void);
EXPORT void openconsole(void);
EXPORT void cripple(void);
EXPORT void uncripple(void);
EXPORT void arcadia(FLAG user);
EXPORT void interton(FLAG user);
EXPORT void elektor(FLAG user);
EXPORT void updatebiggads(void);
EXPORT void updatesmlgad(int which, ULONG state, FLAG now);
EXPORT void updatesmlgads(void);
EXPORT void tick(int which);
EXPORT void untick(int which);
EXPORT void ghost(int which);
EXPORT void unghost(int which);
EXPORT void beep(void);
EXPORT void close_monitor(void);
EXPORT void close_controls(void);
EXPORT void sound_off(FLAG force);
EXPORT void sound_on(void);
EXPORT void setbutton(int which, FLAG enabled, FLAG state);
EXPORT void say(STRPTR text, FLAG handlesound);
EXPORT void infowindow(void);
EXPORT void view_controls(void);
EXPORT void view_monitor(void);
EXPORT void changefilepart(STRPTR newfilepart);
EXPORT void debug_set(int kind);
EXPORT void debug_setrastn(void);
EXPORT void debug_edit(void);
EXPORT void lockgui(void);
EXPORT void unlockgui(void);
EXPORT void clearmsgs(void);
EXPORT void updatefps(void);
EXPORT int asl(STRPTR hail, STRPTR pattern, FLAG save, STRPTR startdir,
    STRPTR startfile, STRPTR passedfn);
EXPORT void breakparts(void);
EXPORT void invert(int i);
EXPORT void hostnamewindow(void);
EXPORT void clearkybd(void);
EXPORT void listen_open(void);
EXPORT void listen_loop(void);
EXPORT void listen_close(void);
EXPORT void infowindow(void);
#ifdef AMIGA
    EXPORT void detachsmlgads(void);
    EXPORT void attachsmlgads(void);
    EXPORT void writeicon(int kind, STRPTR passedfn);
    EXPORT void setmonitor(struct Gadget* gadgetptr, STRPTR contents);
    EXPORT FLAG lockscreen(FLAG panic);
    EXPORT void unlockscreen(void);
    EXPORT void rq(STRPTR text);
    EXPORT LONG easyrequest(FLAG handlesound);
#endif
#ifdef __GNUC__
    EXPORT int stcl_h(char* out, unsigned long lvalue);
#endif
#ifdef WIN32
    EXPORT int stcl_d(char* out, long lvalue);
    EXPORT int stcul_d(char* out, unsigned long lvalue);
    EXPORT int stcd_l(char* in, SLONG* lvalue);
    EXPORT int stch_l(char* in, SLONG* lvalue);
    EXPORT int stcl_h(char* out, unsigned long lvalue);
    EXPORT void reactivate(void);
    EXPORT void closewindow(void);
    EXPORT void openwindow(void);
    EXPORT void clearscreen(void);
    EXPORT void calcsize(FLAG autosize);
    EXPORT void setmonitor(HWND gadgetptr, STRPTR contents);
    EXPORT STRPTR GetCatalogStr(int ignored1, int ignored2, STRPTR thestring);
#endif

// cpu.c
EXPORT void cpu(void);
EXPORT void cpu_setup(void);

#ifdef AMIGA
    EXPORT void view_screen(void);
#endif

// uvi.c
EXPORT void uvi(void); // can't be __inlined because it's EXPORTed
EXPORT void uviwrite(UWORD address, UBYTE data);

// pvi.c
EXPORT void pvi(void); // can't be __inlined because it's EXPORTed
EXPORT UBYTE loadrand(void); // can't be __inlined because it's EXPORTed
EXPORT void saverand(UBYTE value); // can't be __inlined because it's EXPORTed
EXPORT void pviwrite(UWORD address, UBYTE data);

EXPORT void mousedown(int mousex, int mousey);
EXPORT void mouseup(void);
EXPORT void domouse(void);

// sound.c
#ifdef WIN32
    EXPORT void OpenChannels(void);
#endif
EXPORT void TonePlay(int guestchan, float hertz, int volume);
EXPORT void NoisePlay(int guestchan, float hertz, int volume);
EXPORT void SoundClose(int guestchan);
EXPORT void SoundStop(int guestchan);

// engine.c
EXPORT void change_read(STRPTR thestring);
EXPORT void change_write(STRPTR thestring, FLAG quiet);
EXPORT void engine_setup(void);
EXPORT FLAG engine_load(FLAG silent);
EXPORT void engine_save(int kind, FLAG updatetitle);
EXPORT void view_cram(void);
EXPORT void view_mram(void);
EXPORT void view_xvi(void);
EXPORT void view_udcs(void);
EXPORT void view_cpu(void);
EXPORT void view_psgs(void);
EXPORT void view_allmemory(void);
EXPORT void view_upperscreen(void);
EXPORT void view_lowerscreen(void);
EXPORT void view_bram(void);
EXPORT FLAG parse_bytes(FLAG loading, FLAG silent);
EXPORT void project_reset(FLAG keepalive);
EXPORT void project_open(void);
EXPORT void project_save(int kind);
EXPORT void runto(void);
EXPORT void debug_clearbp(void);
EXPORT void debug_clearwp(void);
EXPORT void updatemenu(int which);
EXPORT void updatemenus(void);
EXPORT void getfriendly(int number);
EXPORT int friendly_to_number(STRPTR thestring);
EXPORT void number_to_friendly(int number, STRPTR thestring, FLAG both);
EXPORT void docommand(int which);
EXPORT void fixextension(STRPTR extension, STRPTR passedfn, FLAG force);
EXPORT void macro_restartplayback(void);
EXPORT void macro_stop(void);
EXPORT void generate_crctable(void);
EXPORT void hex1(char* out, UBYTE value);
EXPORT void hex2(char* out, UWORD value);
EXPORT void decimal(char* out, unsigned long value);
EXPORT void calcrunningtime(void);
EXPORT void savescreenshot(int kind, FLAG wantasl);
EXPORT void debug_cli(void);
EXPORT void debug_edit(void);
EXPORT void view_imagery(void);
#ifdef PROFILE
    EXPORT void printprofile(void);
#endif
#ifdef COVERAGE
    EXPORT void printcoverage(void);
#endif
EXPORT void dopaddle(int source, int dest);
EXPORT void update_controls(void);
EXPORT void view_controls_engine(void);
EXPORT void updatemonitor(FLAG force);
EXPORT void initmonitor_engine(void);
EXPORT void closemonitors(FLAG both);
// network play
EXPORT FLAG NetExchange(char* In, const char* Out, int N);
EXPORT void NetPlay(int Switch);
EXPORT FLAG NetSendByte(UBYTE value);
EXPORT UBYTE NetReceiveByte(void);

#define PSU_S        128
#define PSU_F         64
#define PSU_II        32
// bits 4..3 are unused and unusable
#define PSU_SP         7 // $07 (%00000111)
#define PSU_WRITABLE 103 // $67 (%01100111)
#define PSU_READONLY 152 // $98 (%10011000)

#define ROMSIZE_MAX    (16 * KILOBYTE)
#define SNPSIZE_MAX    (ROMSIZE_MAX + 799 + (3 * KILOBYTE))

// these values are all in guest-pixels
#define A_BOXWIDTH       151 // (128 + 7) = 135 without margins
#define A_BOXHEIGHT      240 //             208 without margins
#define A_SPAREROWS        (BOXHEIGHT - 208) // currently 32 (16 each)
#define A_XMARGIN          ((BOXWIDTH - 135) / 2) // (currently 12 each)

#define IE_BOXWIDTH      180 // must be a multiple of 4!
#define IE_BOXHEIGHT     269
#define GRIDSTART_X       28 // as an offset from the left of the display box

#define BOXWIDTH         IE_BOXWIDTH
#define BOXHEIGHT        IE_BOXHEIGHT

// These are for Interton/Elektor, not Arcadia!
// Interton/Elektor colours are referred to in inverted RGB format!
#define BLACK            7 // 0,0,0 = 0
#define BLUE             6 // 0,0,1 = 1
#define GREEN            5 // 0,1,0 = 2 (different to Arcadia!)
#define CYAN             4 // 0,1,1 = 3 (different to Arcadia!)
#define RED              3 // 1,0,0 = 4 (different to Arcadia!)
#define PURPLE           2 // 1,0,1 = 5 (different to Arcadia!)
#define YELLOW           1 // 1,1,0 = 6
#define WHITE            0 // 1,1,1 = 7

#define GREY             8

// joysticks
#define JOYUP              1
#define JOYDOWN            2
#define JOYLEFT            4
#define JOYRIGHT           8
#define JOYFIRE1          16
#define JOYFIRE2          32
#define JOYFIRE3          64
#define JOYFIRE4         128
#define JOYSTART         256
#define JOYA             512
#define JOYB            1024

// CPU
#define PSU_SP   7 // %00000111

#define PSL_CC 192 // $11000000
#define PSL_IDC 32
#define PSL_RS  16
#define PSL_WC   8
#define PSL_OVF  4
#define PSL_COM  2
#define PSL_C    1

#define A_SPRITE0Y       6384 // $18F0 R/W
#define A_SPRITE0X       6385 // $18F1 R/W
#define A_SPRITE1Y       6386 // $18F2 R/W
#define A_SPRITE1X       6387 // $18F3 R/W
#define A_SPRITE2Y       6388 // $18F4 R/W
#define A_SPRITE2X       6389 // $18F5 R/W
#define A_SPRITE3Y       6390 // $18F6 R/W
#define A_SPRITE3X       6391 // $18F7 R/W
#define A_VSCROLL        6396 // $18FC -/W?
#define A_PITCH          6397 // $18FD -/W?
#define A_VOLUME         6398 // $18FE -/W?
#define A_CHARLINE       6399 // $18FF (dma)
#define A_P1LEFTKEYS     6400 // $1900
#define A_P1MIDDLEKEYS   6401 // $1901
#define A_P1RIGHTKEYS    6402 // $1902
#define A_P1PALLADIUM    6403 // $1903
#define A_P2LEFTKEYS     6404 // $1904
#define A_P2MIDDLEKEYS   6405 // $1905
#define A_P2RIGHTKEYS    6406 // $1906
#define A_P2PALLADIUM    6407 // $1907
#define A_CONSOLE        6408 // $1908
#define A_RESOLUTION     6648 // $19F8
#define A_BGCOLOUR       6649 // $19F9
#define A_SPRITES23CTRL  6650 // $19FA
#define A_SPRITES01CTRL  6651 // $19FB
#define A_BGCOLLIDE      6652 // $19FC
#define A_SPRITECOLLIDE  6653 // $19FD
#define A_P2PADDLE       6654 // $19FE
#define A_P1PADDLE       6655 // $19FF

#define I_NOISE          0x1E80
#define I_P1LEFTKEYS     0x1E88
#define I_P1MIDDLEKEYS   0x1E89
#define I_P1RIGHTKEYS    0x1E8A
#define I_CONSOLE        0x1E8B
#define I_P2LEFTKEYS     0x1E8C
#define I_P2MIDDLEKEYS   0x1E8D
#define I_P2RIGHTKEYS    0x1E8E
#define I_SPRITE0        0x1F00
#define I_SPRITE0AX      0x1F0A
#define I_SPRITE0BX      0x1F0B
#define I_SPRITE0AY      0x1F0C
#define I_SPRITE0BY      0x1F0D
#define I_SPRITE1        0x1F10
#define I_SPRITE1AX      0x1F1A
#define I_SPRITE1BX      0x1F1B
#define I_SPRITE1AY      0x1F1C
#define I_SPRITE1BY      0x1F1D
#define I_SPRITE2        0x1F20
#define I_SPRITE2AX      0x1F2A
#define I_SPRITE2BX      0x1F2B
#define I_SPRITE2AY      0x1F2C
#define I_SPRITE2BY      0x1F2D
#define I_SPRITE3        0x1F40
#define I_SPRITE3AX      0x1F4A
#define I_SPRITE3BX      0x1F4B
#define I_SPRITE3AY      0x1F4C
#define I_SPRITE3BY      0x1F4D
#define I_VERTGRID       0x1F80
#define I_HORIZ1         0x1FA8
#define I_HORIZ2         0x1FA9
#define I_HORIZ3         0x1FAA
#define I_HORIZ4         0x1FAB
#define I_HORIZ5         0x1FAC
#define I_SIZES          0x1FC0
#define I_SPR01COLOURS   0x1FC1
#define I_SPR23COLOURS   0x1FC2
#define I_SCORECTRL      0x1FC3
#define I_BGCOLOUR       0x1FC6
#define I_PITCH          0x1FC7
#define I_SCORELT        0x1FC8
#define I_SCORERT        0x1FC9
#define I_BGCOLLIDE      0x1FCA
#define I_SPRITECOLLIDE  0x1FCB
#define I_BGCOLLIDE2     0x1FDA
#define I_SPRITECOLLIDE2 0x1FDB
#define I_BGCOLLIDE3     0x1FEA
#define I_SPRITECOLLIDE3 0x1FEB
#define I_BGCOLLIDE4     0x1FFA
#define I_SPRITECOLLIDE4 0x1FFB
#define I_P1PADDLE       0x1FCC
#define I_P2PADDLE       0x1FCD

#define E_MONOB           0x800
#define E_MLINE           0x890
#define E_FSEQ            0x898
#define E_ENTM            0x899
#define E_MFUNC           0x89A
#define E_MONOB           0x800
#define E_MLINE           0x890
#define E_FSEQ            0x898
#define E_SILENC          0x898
#define E_ENTM            0x899
#define E_MFUNC           0x89A
#define E_MSCR            0x89B
#define E_RKBST           0x89D
#define E_LKBST           0x89E
#define E_MKBST           0x89F
#define E_FSCRM           0x8A0
#define E_REGM            0x8AC
#define E_BK1             0x8B5
#define E_BK2             0x8B7
#define E_INTADR          0x8B9
#define E_PC              0x8BE

#define E_PITCHA1_L      0x1D00
#define E_PITCHA1_H      0x1D01
#define E_PITCHB1_L      0x1D02
#define E_PITCHB1_H      0x1D03
#define E_PITCHC1_L      0x1D04
#define E_PITCHC1_H      0x1D05
#define E_PITCHD1        0x1D06
#define E_ENABLE1        0x1D07
#define E_AMPLITUDEA1    0x1D08
#define E_AMPLITUDEB1    0x1D09
#define E_AMPLITUDEC1    0x1D0A
#define E_PERIOD1_L      0x1D0B
#define E_PERIOD1_H      0x1D0C
#define E_SHAPE1         0x1D0D
#define E_PORTA1         0x1D0E
#define E_PORTB1         0x1D0F

#define E_PITCHA2_L      0x1D10
#define E_PITCHA2_H      0x1D11
#define E_PITCHB2_L      0x1D12
#define E_PITCHB2_H      0x1D13
#define E_PITCHC2_L      0x1D14
#define E_PITCHC2_H      0x1D15
#define E_PITCHD2        0x1D16
#define E_ENABLE2        0x1D17
#define E_AMPLITUDEA2    0x1D18
#define E_AMPLITUDEB2    0x1D19
#define E_AMPLITUDEC2    0x1D1A
#define E_PERIOD2_L      0x1D1B
#define E_PERIOD2_H      0x1D1C
#define E_SHAPE2         0x1D1D
#define E_PORTA2         0x1D1E
#define E_PORTB2         0x1D1F

#define E_RANDOM2        0x1D20
#define E_CASIN          0x1DBF
#define E_CASOUT         0x1DFF
#define E_RANDOM1G       0x1FAF
#define E_RANDOM1E       0x1FFF

#define MENUITEM_ANALOG           0
#define MENUITEM_AUTOFIRE         1
#define MENUITEM_AUTOSAVE         2
#define MENUITEM_CHANGE           3
#define MENUITEM_CLEARBP          4
#define MENUITEM_CLEARWP          5
#define MENUITEM_CLIENT           6
#define MENUITEM_COLLISIONS       7
#define MENUITEM_CONTROLS         8
#define MENUITEM_CRAM             9
#define MENUITEM_CREATEICONS     10
#define MENUITEM_DEMULTIPLEX     11
#define MENUITEM_FLAGLINE        12
#define MENUITEM_FULLSCREEN      13
#define MENUITEM_ABOUT           14
#define MENUITEM_JOY1            15
#define MENUITEM_LIMITREFRESH    16
#define MENUITEM_LOGINSTRUCTIONS 17
#define MENUITEM_LOGREADS        18
#define MENUITEM_LOGWRITES       19
#define MENUITEM_LOOP            20
#define MENUITEM_LOWERSCREEN     21
#define MENUITEM_MACHINE         22
#define MENUITEM_MACRO           23
#define MENUITEM_MANUAL          24
#define MENUITEM_MENUBAR         25 // never used
#define MENUITEM_MONITOR         26
#define MENUITEM_NARROW          27
#define MENUITEM_OPEN            28
#define MENUITEM_PAUSED          29
#define MENUITEM_POINTER         30
#define MENUITEM_PSGS            31
#define MENUITEM_QUICKLOAD       32
#define MENUITEM_QUICKSAVE       33
#define MENUITEM_REGION          34
#define MENUITEM_RESET           35
#define MENUITEM_RESETTOMONITOR  36
#define MENUITEM_REVERT          37
#define MENUITEM_RUNTODMA        38
#define MENUITEM_SAVESNP         39
#define MENUITEM_SAVEROM         40
#define MENUITEM_SAVETVC         41
#define MENUITEM_SENSEGAME       42
#define MENUITEM_SENSITIVITY     43
#define MENUITEM_SERVER          44
#define MENUITEM_SETBP           45
#define MENUITEM_SETWP           46
#define MENUITEM_SIZE            47
#define MENUITEM_SOUND           48
#define MENUITEM_STEP            49
#define MENUITEM_STRETCH         50
#define MENUITEM_SWAPPED         51
#define MENUITEM_TITLEBAR        52
#define MENUITEM_TOOLBAR         53
#define MENUITEM_TRACECPU        54
#define MENUITEM_USEHACKS        55
#define MENUITEM_WARP            56
#define MENUITEM_WATCHREADS      57
#define MENUITEM_CLI             58
#define MENUITEM_REQUIREBUTTON   59
#define MENUITEM_ABSOLUTE        60
#define MENUITEM_FILTER          61
#define MENUITEM_IMAGERY         62
#define MENUITEM_JOY0            63
#define MENUITEM_LOGMONITOR      64
#define MENUITEM_RUNTOINTERRUPT  65
#define MENUITEM_BRAM            66
#define MENUITEMS                67 // counting from 1

// these must not conflict with any MENUITEMs
#define COMMAND_ARCADIA          100
#define COMMAND_ELEKTOR          101
#define COMMAND_INTERTON         102
#define COMMAND_IGNORE           103

// small gadget positions, left to right, on GID_SB2
// controllers
#define GADPOS_ANALOG       0
#define GADPOS_AUTOFIRE     1
#define GADPOS_SWAPPED      2
// speed
#define GADPOS_PAUSED       3
#define GADPOS_WARP         4
// sprites
#define GADPOS_COLLISIONS   5
#define GADPOS_DEMULTIPLEX  6
// graphics
#define GADPOS_FLAGLINE     7
// misc
#define GADPOS_SOUND        8
// macro (not really part of toolbar in Amiga version)
#define GADPOS_MACRO        9

#define SENSEGAME_VERBOSE   0
#define SENSEGAME_QUIET     1
#define SENSEGAME_NO        2

#define NUMKEYS            24 // counting from 1 (last array index is NUMKEYS-1)
#define VERBOSELINES       12 // counting from 1 (last array index is VERBOSELINES-1)

#define KEYS               40 // counting from 1

#define KIND_BP             0
#define KIND_WP             1

#define MONITORGADS       178

#define ARCADIA  0
#define INTERTON 1
#define ELEKTOR  2
#define MACHINES 3 // counting from 1

#define MEMMAP_A 0
#define MEMMAP_B 1
#define MEMMAP_C 2
#define MEMMAP_D 3
#define MEMMAP_E 4
#define MEMMAP_F 5
#define MEMMAP_G 6

#define EXIT_WARNING  5
#define EXIT_ERROR   10

EXPORT struct KeyTableStruct
{   UWORD a_address;
    UBYTE a_mask;
    UWORD ie_address;
    UBYTE ie_mask;
    STRPTR ai_name;
    STRPTR e_name;
};

EXPORT struct KindStruct
{   const STRPTR extension,
                 pattern;
#ifdef WIN32
    const STRPTR hail;
#endif
#ifdef AMIGA
          STRPTR hail;
          STRPTR icon;
#endif
};
