/* ibm.c: Miscellaneous IBM-PC code

Known WinArcadia issues:
 (a) Some supposedly random noise samples sound more like tones, especially
     after playing a tone. Eg. SHOOTGAL.BIN, HUNTING.BIN demonstrate this.
     This could be caused by oddities in the host randomization, Windows
     multimedia subsystem, or sound hardware.
 (b) Very short noises are not really audible.
 (c) In DirectX mode, the main window, when it redraws its client area,
     does so regardless of the depth-arrangement (ie. it will redraw even
     areas that are supposed to be overlapped by other windows or the
     start bar). (Forcing the main window to always be frontmost would be
     a workaround for this.) (Also, DirectX mode seems slower than GDI
     mode, which is strange.)
 (d) Network play does not seem to work except when the server and client
     are both on the same local computer (timeout issues?). (This probably
     applies to AmiArcadia also.)
 (e) Sometimes moving the mouse when paused causes the emulation to run
     for a rastline. */

// always define one (only!) of these. Normally MODE_FPS.
#define MODE_FPS
// #define MODE_MHZ
// #define MODE_PERCENT

// always define one (only!) of these. Normally USEGDI.
#define USEGDI
// #define USEDIRECTX

/* The emulators currently run at these speeds...?
   For NTSC Arcadia:          0.879840MHz ( 879.840 KHz).
   For PAL  Arcadia:          0.733200MHz ( 733.200 KHz).
   For PAL  Interton/Elektor: 0.873600MHz ( 873.600 KHz).
   For NTSC Interton/Elektor: 1.048320MHz (1048.320 KHz).
   "Clock Frequency:          3.546895Mhz +/- 35Hz" - Fountain
                                                      Service Manual */

#include "ibm.h"

#include <commctrl.h>

#ifdef USEDIRECTX
    #include <ddraw.h>
#endif

#include "resource.h"

#include <stdio.h>
#include <fcntl.h> // for _O_TEXT
#include <io.h>    // for _open_osfhandle()

#include "aa.h"

const char g_szClassName[] = "WinArcadia";

#define IDC_TOOLBAR       102
#define IDC_FPS           103
#define IDC_STATIC        104

#define REGION_NTSC         0
#define REGION_PAL          1

#define DLGTYPE_BP          0
#define DLGTYPE_WP          1

#define A_TITLEBARTEXT      "WinArcadia 5.7"
#define I_TITLEBARTEXT     "WinInterton 5.7"
#define E_TITLEBARTEXT      "WinElektor 5.7"
#define TITLEBARTEXTLENGTH 23 // 23 is enough for "WinInterton x.xxx: "

#define CONFIGLENGTH      135 // in bytes, counting from 1 (data is 0..CONFIGLENGTH-1)
                              // (not including game path)

#define SMALLGADGETS_WIDE       21
#define SMALLGADGETS_NARROW     13
#define SMALLIMAGES_WIDE        15
#define SMALLIMAGES_NARROW      15

typedef struct
{   DWORD  vkCode;
    DWORD  scanCode;
    DWORD  flags;
    DWORD  time;
    ULONG* dwExtraInfo;
} KBDLLHOOKSTRUCT, *PKBDLLHOOKSTRUCT;

EXPORT FLAG              aborting;
EXPORT int               winleftx,
                         wintopy,
                         titleheight,
                         step                 = FALSE;
EXPORT ULONG*            display              = NULL;
EXPORT ULONG             oldframes,
                         newclock;
EXPORT UBYTE             screen[BOXWIDTH][BOXHEIGHT];
EXPORT HWND              ControlsWindowPtr    = NULL,
                         ListeningWindowPtr   = NULL,
                         MainWindowPtr        = NULL,
                         MonitorWindowPtr     = NULL,
                         hToolbar             = NULL;
EXPORT HMENU             MenuPtr              = NULL,
                         PopUpMenuPtr         = NULL;
EXPORT HBITMAP           MenuBitmapPtr[3]     = {NULL, NULL, NULL};
EXPORT TEXT              globalstring[13 + 1 + 1],
                         hostname[256 + 1]; // should hopefully be long enough
EXPORT UBYTE             joystickIDMapping[2] = {0, 1};
EXPORT HDC               ControlsRastPtr;
EXPORT int               CatalogPtr; // never used

EXPORT const struct KindStruct filekind[KINDS] =
{ { ".IFF",
    "IFF ILBM (*.IFF, *.ILBM, *.LBM)\0*.IFF;*.ILBM;*.LBM\0All Files (*.*)\0*.*\0",
    "Save Screenshot as IFF ILBM"
  },
  { ".IFF",
    "IFF ACBM (*.IFF, *.ACBM)\0*.IFF;*.ACBM\0All Files (*.*)\0*.*\0",
    "Save Screenshot as IFF ACBM"
  },
  { ".BMP",
    "BMP (*.BMP)\0*.BMP\0All Files (*.*)\0*.*\0",
    "Save Screenshot as BMP"
  },
  { ".BIN",
    "Arcadia/Interton ROM (*.BIN)\0*.BIN\0All Files (*.*)\0*.*\0",
    "Save Game ROM"
  },
  { ".PGM",
    "Elektor Program (*.PGM)\0*.PGM\0All Files (*.*)\0*.*\0",
    "Save Elektor Program (PGM)"
  },
  { ".TVC",
    "Elektor Program (*.TVC)\0*.TVC\0All Files (*.*)\0*.*\0",
    "Save Elektor Program (TVC)"
  },
  { ".EAS",
    "Arcadia Snapshot (*.EAS)\0*.EAS\0All Files (*.*)\0*.*\0",
    "Save Arcadia Snapshot"
  },
  { ".INS",
    "Interton Snapshot (*.INS)\0*.INS\0All Files (*.*)\0*.*\0",
    "Save Interton Snapshot"
  },
  { ".EAR",
    "Arcadia Recording (*.EAR)\0*.EAR\0All Files (*.*)\0*.*\0",
    "Start Arcadia Recording"
  },
  { ".INR",
    "Interton Recording (*.INR)\0*.INR\0All Files (*.*)\0*.*\0",
    "Start Interton Recording"
  },
  { ".TVS",
    "Elektor Snapshot (*.TVS)\0*.TVS\0All Files (*.*)\0*.*\0",
    "Save Elektor Snapshot"
  },
  { ".TVR",
    "Elektor Recording (*.TVR)\0*.TVR\0All Files (*.*)\0*.*\0",
    "Start Elektor Recording"
} };
                                          
#ifdef MODE_MHZ
    EXPORT ULONG oldelapsed;
#endif

EXPORT ULONG  pens[16] =
{   0x00FFFFFF, // white  (UVI #0, PVI #0)
    0x00FFFF00, // yellow (UVI #1, PVI #1)
    0x00FF00FF, // purple (UVI #4, PVI #2)
    0x00FF0000, // red    (UVI #5, PVI #3)
    0x0000FFFF, // cyan   (UVI #2, PVI #4)
    0x0000FF00, // green  (UVI #3, PVI #5)
    0x000000FF, // blue   (UVI #6, PVI #6)
    0x00000000, // black  (UVI #7, PVI #7)
    0x00BBBBBB, // dark white
    0x00CCCC00, // dark yellow
    0x00BB00BB, // dark purple
    0x00990000, // dark red
    0x0000AAAA, // dark cyan
    0x0000BB00, // dark green
    0x000000AA, // dark blue
    0x00000000  // dark black
};

MODULE UBYTE             KeyMatrix[64], // to allow keycodes up to 511
                         realsize,
                         realwide,
                         ConfigBuffer[CONFIGLENGTH];
MODULE TEXT              ProgDir[MAX_PATH + 1] = "",
                         fn_config[MAX_PATH + 1],
                         fpsstring[10 + 1],
                         titlebartext[MAX_PATH + 1 + TITLEBARTEXTLENGTH + 9]; // " (paused)" is 9 bytes
MODULE HWND              LabelPtr          = NULL,
                         FPSPtr            = NULL;
MODULE HGDIOBJ           OldObject         = NULL;
// MODULE HANDLE         AcceleratorPtr    = NULL;
MODULE HINSTANCE         InstancePtr       = NULL;
MODULE int               bpl,
                         clientheight,
                         DisplayWidth,
                         DisplayHeight,
                         dlgtype,
                         menuheight,
                         side,
                         toolbarheight,
                         whichkey,
                         xoffset,
                         yoffset;
MODULE ULONG             oldfps            = ~0,
                         newtime,
                         oldtime,
                         waittill          = 0;
MODULE SWORD             rastwidth,
                         rastheight;
MODULE RECT              therect;
MODULE DWORD             winwidth,
                         winheight;
MODULE UWORD             temp_console[4],
                         temp_keypads[2][NUMKEYS];
MODULE POINT             ThePoint;

MODULE struct
{   BITMAPINFOHEADER Header;
    DWORD            Colours[3];
} TheBitMapInfo;

#ifdef USEDIRECTX
    MODULE LPDIRECTDRAW        DD           = NULL;
    MODULE LPDIRECTDRAWSURFACE DDSurface    = NULL;
    MODULE LPDIRECTDRAWCLIPPER DDClipper    = NULL;
    MODULE DDSURFACEDESC       DDSD;
    MODULE HDC                 OurhDC       = NULL;
    MODULE HBITMAP             OurhBitMap   = NULL;
#endif

EXPORT HWND  // CPU chip...
             RGad[7],
             CCPtr,
             SensePtr,
             FlagPtr,
             InterruptPtr,
             IDCPtr,
             RSPtr,
             WCPtr,
             OVFPtr,
             COMPtr,
             CPtr,
             RASGad[8],
             SPGad,
             IARPtr,
             ClockPtr,
             FramePtr,
             RasterGad,
             TimePtr,
             monitorgad[MONITORGADS],
             SPRITEnAXYGad[4],
             SPRITEnBXYGad[4],
             PITCHAGad[2],
             PITCHBGad[2],
             PITCHCGad[2],
             PERIODGad[2];

// IMPORTED VARIABLES ----------------------------------------------------

// ReAction gadgets
IMPORT ULONG     analog,
                 autofire,
                 collisions,
                 demultiplex,
                 flagline,
                 fps,
                 paused,
                 sound,
                 swapped,
                 warp;
// other ULONGs here
IMPORT ULONG     elapsed,
                 frames,
                 frameskip,
                 joy0,
                 joy1,
                 region;
IMPORT UWORD     iar,
                 default_console[4],
                 default_keypads[2][NUMKEYS],
                 console[4],
                 keypads[2][NUMKEYS];
IMPORT UBYTE     memory[32768],
                 sx[2],
                 sy[2];
IMPORT       int    autosave,
                    bp,
                    connected,
                    consoleopen,
                    crippled,
                    fromnum,
                    fullscreen,
                    game,
                    limitrefresh,
                    loginstructions,
                    logmonitor,
                    logreads,
                    logwrites,
                    loop,
   				    machine,
   				    memmap,
                    rastn,
                    recmode,
                    requirebutton,
                    runtorastline,
                    runtodma,
                    runtoframe,
                    runtointerrupt,
                    sensegame,
//			        sensitivity,
                    showmenubar,
                    showpointer,
                    showtitlebar,
                    showtoolbar,
                    size,
                    stretch,
                    tonum,
                    tolimit,
                    trace,
                    usehacks,
                    watchreads,
                    wide,
                    wp;
IMPORT       TEXT   autotext[VERBOSELINES][60 + 1],
                    filepart[MAX_PATH + 1], // file only
                    friendly[21 + 1],
                    fromstr[5 + 1 + 1],
                    tostr[5 + 1 + 1],
                    lastchange[13 + 1 + 1],
                    pathpart[MAX_PATH + 1],
                    fn_game[MAX_PATH + 1], // the entire pathname (path and file)
                    fn_screenshot[MAX_PATH + 1]; // screen shot file name
IMPORT const STRPTR autoname[3];
IMPORT       FLAG   cli,
                    absolute;
IMPORT       RECT   keyrect[3][KEYS];

// MODULE ARRAYS ---------------------------------------------------------

MODULE int menucode[MENUITEMS] =
{   ID_OPTIONS_CONTROLLERS_ANALOG,
    ID_OPTIONS_CONTROLLERS_AUTOFIRE,
    ID_OPTIONS_AUTOSAVE,
    ID_DEBUG_CHANGE,
    ID_DEBUG_CLEARBP,
    ID_DEBUG_CLEARWP,
    ID_FILE_CLIENT,
    ID_OPTIONS_SPRITES_COLLISIONS,
    ID_VIEW_CONTROLS,
    ID_VIEW_CRAM,
    -1, // create icons
    ID_OPTIONS_SPRITES_DEMULTIPLEX,
    ID_OPTIONS_GRAPHICS_FLAGLINE,
    ID_OPTIONS_GRAPHICS_FULLSCREEN,
    -1, // about
    ID_CONTROLLERS_ENABLE2ND,
    ID_OPTIONS_SPEED_LIMITREFRESH,
    ID_DEBUG_LOGINSTRUCTIONS,
    ID_DEBUG_LOGREADS,
    ID_DEBUG_LOGWRITES,
    ID_MACRO_LOOP,
    ID_VIEW_LOWERSCREEN,
    -1, // machine
    -1, // macro
    -1, // manual
    ID_VIEW_MENU,
    ID_VIEW_MONITOR,
    ID_OPTIONS_GRAPHICS_NARROW,
    ID_FILE_OPEN,
    ID_OPTIONS_SPEED_PAUSED,
    ID_VIEW_POINTER,
    ID_VIEW_PSGS,
	ID_FILE_QUICKLOAD,
	ID_FILE_QUICKSAVE,
    -1, // region
    ID_FILE_RESET,
    ID_FILE_RESETTOMONITOR,
    ID_FILE_REVERT,
    ID_DEBUG_RUNTODMA,
    ID_FILE_SAVESNP,
    ID_FILE_SAVEROM,
    ID_FILE_SAVETVC,
    -1, // sensegame
    -1, // sensitivity
    ID_FILE_SERVER,
    ID_DEBUG_SETBP,
    ID_DEBUG_SETWP,
    -1, // size
    ID_OPTIONS_SOUND,
    ID_DEBUG_STEP,
    ID_OPTIONS_GRAPHICS_STRETCH,
    ID_OPTIONS_CONTROLLERS_SWAPPED,
    ID_VIEW_TITLE,
    ID_VIEW_TOOL,
    ID_DEBUG_TRACECPU,
    ID_MACHINE_USEHACKS,
    ID_OPTIONS_SPEED_WARP,
    ID_DEBUG_WATCHREADS,
	ID_DEBUG_CLI,
    ID_CONTROLLERS_REQUIREBUTTON,
    -1, // analog for host
    -1, // filter
    ID_VIEW_IMAGERY,
    ID_CONTROLLERS_ENABLE1ST,
    ID_DEBUG_LOGMONITOR,
    ID_DEBUG_RUNTOINTERRUPT,
    ID_VIEW_BRAM
};
MODULE int gadgetcode[] =
{   ID_OPTIONS_CONTROLLERS_ANALOG,   // 0
    ID_OPTIONS_CONTROLLERS_AUTOFIRE, // 1
    ID_OPTIONS_CONTROLLERS_SWAPPED,  // 2
    ID_OPTIONS_SPEED_PAUSED,         // 3
    ID_OPTIONS_SPEED_WARP,           // 4
    ID_OPTIONS_SPRITES_COLLISIONS,   // 5
    ID_OPTIONS_SPRITES_DEMULTIPLEX,  // 6
    ID_OPTIONS_GRAPHICS_FLAGLINE,    // 7
    ID_OPTIONS_SOUND,                // 8
    -1 // macro                      // 9
};
MODULE int gadgetimage[11][2] =
{ {  4, 15 },
  {  5, 16 },
  {  6, 17 },
  {  7, 18 },
  {  8, 19 },
  {  9, 20 },
  { 10, 21 },
  { 11, 22 },
  { 12, 23 },
  { 13, 24 },
  { 14, 25 }
};

/* Keys we don't hear: both Windows keys, both Alt keys, F10.
Also Power, Sleep, Wake, Fn.
Esc and PrtSc both use scancode 1, apparently. */
MODULE STRPTR keyname[512] =
{   "(0)",
    "PrtSc",
    "a1",
    "a2",
    "a3",
    "a4",
    "a5",
    "a6",
    "a7",
    "a8",
    "a9",
    "a0",
    "a-",
    "=",
    "BkSp",
    "Tab",
    "Q",
    "W",
    "E",
    "R",
    "T",
    "Y",
    "U",
    "I",
    "O",
    "P",
    "[",
    "]",
    "aE",
    "LCtrl",
    "A",
    "S",
    "D",
    "F",
    "G",
    "H",
    "J",
    "K",
    "L",
    ";",
    "'",
    "`",
    "LShift",
    "\\",
    "Z",
    "X",
    "C",
    "V",
    "B",
    "N",
    "M",
    ",",
    "a.",
    "a/",
    "RShift",
    "n*",
    "Alt",
    "Space",
    "CL",
    "F1",
    "F2",
    "F3",
    "F4",
    "F5",
    "F6",
    "F7",
    "F8",
    "F9",
    "F10",
    "Pause",
    "SL",
    "n7",
    "n8",
    "n9",
    "n-",
    "n4",
    "n5",
    "n6",
    "n+",
    "n1",
    "n2",
    "n3",
    "n0",
    "n.",
    "(84)",
    "(85)",
    "(86)",
    "F11",
    "F12",
    "(89)",
    "(90)",
    "(91)",
    "(92)",
    "(93)",
    "(94)",
    "(95)",
    "(96)",
    "(97)",
    "(98)",
    "(99)",
    "(100)",
    "(101)",
    "(102)",
    "(103)",
    "(104)",
    "(105)",
    "(106)",
    "(107)",
    "(108)",
    "(109)",
    "(110)",
    "(111)",
    "(112)",
    "(113)",
    "(114)",
    "(115)",
    "(116)",
    "(117)",
    "(118)",
    "(119)",
    "(120)",
    "(121)",
    "(122)",
    "(123)",
    "(124)",
    "(125)",
    "(126)",
    "(127)",
    "(128)",
    "(129)",
    "(130)",
    "(131)",
    "(132)",
    "(133)",
    "(134)",
    "(135)",
    "(136)",
    "(137)",
    "(138)",
    "(139)",
    "(140)",
    "(141)",
    "(142)",
    "(143)",
    "(144)",
    "(145)",
    "(146)",
    "(147)",
    "(148)",
    "(149)",
    "(150)",
    "(151)",
    "(152)",
    "(153)",
    "(154)",
    "(155)",
    "(156)",
    "(157)",
    "(158)",
    "(159)",
    "(160)",
    "(161)",
    "(162)",
    "(163)",
    "(164)",
    "(165)",
    "(166)",
    "(167)",
    "(168)",
    "(169)",
    "(170)",
    "(171)",
    "(172)",
    "(173)",
    "(174)",
    "(175)",
    "(176)",
    "(177)",
    "(178)",
    "(179)",
    "(180)",
    "(181)",
    "(182)",
    "(183)",
    "(184)",
    "(185)",
    "(186)",
    "(187)",
    "(188)",
    "(189)",
    "(190)",
    "(191)",
    "(192)",
    "(193)",
    "(194)",
    "(195)",
    "(196)",
    "(197)",
    "(198)",
    "(199)",
    "(200)",
    "(201)",
    "(202)",
    "(203)",
    "(204)",
    "(205)",
    "(206)",
    "(207)",
    "(208)",
    "(209)",
    "(210)",
    "(211)",
    "(212)",
    "(213)",
    "(214)",
    "(215)",
    "(216)",
    "(217)",
    "(218)",
    "(219)",
    "(220)",
    "(221)",
    "(222)",
    "(223)",
    "(224)",
    "(225)",
    "(226)",
    "(227)",
    "(228)",
    "(229)",
    "(230)",
    "(231)",
    "(232)",
    "(233)",
    "(234)",
    "(235)",
    "(236)",
    "(237)",
    "(238)",
    "(239)",
    "(240)",
    "(241)",
    "(242)",
    "(243)",
    "(244)",
    "(245)",
    "(246)",
    "(247)",
    "(248)",
    "(249)",
    "(250)",
    "(251)",
    "(252)",
    "(253)",
    "(254)",
    "(255)",
    "(256)",
    "(257)",
    "(258)",
    "(259)",
    "(260)",
    "(261)",
    "(262)",
    "(263)",
    "(264)",
    "(265)",
    "(266)",
    "(267)",
    "(268)",
    "(269)",
    "(270)",
    "(271)",
    "(272)",
    "(273)",
    "(274)",
    "(275)",
    "(276)",
    "(277)",
    "(278)",
    "(279)",
    "(280)",
    "(281)",
    "(282)",
    "(283)",
    "nE",
    "RCtrl",
    "(286)",
    "(287)",
    "(288)",
    "(289)",
    "(290)",
    "(291)",
    "(292)",
    "(293)",
    "(294)",
    "(295)",
    "(296)",
    "(297)",
    "(298)",
    "(299)",
    "(300)",
    "(301)",
    "(302)",
    "(303)",
    "(304)",
    "(305)",
    "(306)",
    "(307)",
    "(308)",
    "n/",
    "(310)",
    "(311)",
    "(312)",
    "(313)",
    "(314)",
    "(315)",
    "(316)",
    "(317)",
    "(318)",
    "(319)",
    "(320)",
    "(321)",
    "(322)",
    "(323)",
    "(324)",
    "NL",
    "(326)",
    "Home",
    "Up",
    "PgUp",
    "(330)",
    "Lt",
    "(332)",
    "Rt",
    "(334)",
    "End",
    "Dn",
    "PgDn",
    "Ins",
    "Del",
    "(340)",
    "(341)",
    "(342)",
    "(343)",
    "(344)",
    "(345)",
    "(346)",
    "(347)",
    "(348)",
    "Menu",
    "(350)",
    "(351)",
    "(352)",
    "(353)",
    "(354)",
    "(355)",
    "(356)",
    "(357)",
    "(358)",
    "(359)",
    "(360)",
    "(361)",
    "(362)",
    "(363)",
    "(364)",
    "(365)",
    "(366)",
    "(367)",
    "(368)",
    "(369)",
    "(370)",
    "(371)",
    "(372)",
    "(373)",
    "(374)",
    "(375)",
    "(376)",
    "(377)",
    "(378)",
    "(379)",
    "(380)",
    "(381)",
    "(382)",
    "(383)",
    "(384)",
    "(385)",
    "(386)",
    "(387)",
    "(388)",
    "(389)",
    "(390)",
    "(391)",
    "(392)",
    "(393)",
    "(394)",
    "(395)",
    "(396)",
    "(397)",
    "(398)",
    "(399)",
    "(400)",
    "(401)",
    "(402)",
    "(403)",
    "(404)",
    "(405)",
    "(406)",
    "(407)",
    "(408)",
    "(409)",
    "(410)",
    "(411)",
    "(412)",
    "(413)",
    "(414)",
    "(415)",
    "(416)",
    "(417)",
    "(418)",
    "(419)",
    "(420)",
    "(421)",
    "(422)",
    "(423)",
    "(424)",
    "(425)",
    "(426)",
    "(427)",
    "(428)",
    "(429)",
    "(430)",
    "(431)",
    "(432)",
    "(433)",
    "(434)",
    "(435)",
    "(436)",
    "(437)",
    "(438)",
    "(439)",
    "(440)",
    "(441)",
    "(442)",
    "(443)",
    "(444)",
    "(445)",
    "(446)",
    "(447)",
    "(448)",
    "(449)",
    "(450)",
    "(451)",
    "(452)",
    "(453)",
    "(454)",
    "(455)",
    "(456)",
    "(457)",
    "(458)",
    "(459)",
    "(460)",
    "(461)",
    "(462)",
    "(463)",
    "(464)",
    "(465)",
    "(466)",
    "(467)",
    "(468)",
    "(469)",
    "(470)",
    "(471)",
    "(472)",
    "(473)",
    "(474)",
    "(475)",
    "(476)",
    "(477)",
    "(478)",
    "(479)",
    "(480)",
    "(481)",
    "(482)",
    "(483)",
    "(484)",
    "(485)",
    "(486)",
    "(487)",
    "(488)",
    "(489)",
    "(490)",
    "(491)",
    "(492)",
    "(493)",
    "(494)",
    "(495)",
    "(496)",
    "(497)",
    "(498)",
    "(499)",
    "(500)",
    "(501)",
    "(502)",
    "(503)",
    "(504)",
    "(505)",
    "(506)",
    "(507)",
    "(508)",
    "(509)",
    "(510)",
    "(511)"
};

BOOL CALLBACK    ControlsDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK        HelpDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK     MonitorDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK         SetDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK       RastNDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK      ChangeDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK    RedefineDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK RedefineSubDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK    HostNameDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK   ListeningDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK   AutosenseDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam);
// BOOL CALLBACK SensitivityDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam);

MODULE __inline void thewait(void);
MODULE void freeall(void);
MODULE void resize(int newsize);
MODULE void maketoolbar(HWND hwnd);
MODULE FLAG ctrl(void);
MODULE FLAG shift(void);
MODULE void printusage(STRPTR progname);
MODULE void handlekybd(UINT code);
MODULE void refreshkeys(HWND hwnd);
MODULE void loadconfig(void);

LRESULT CALLBACK WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{   PAINTSTRUCT   ps;
 // int           iToolHeight;
    LPTOOLTIPTEXT lpttt;

    switch(msg)
    {   case WM_CREATE:
            maketoolbar(hwnd);
        break;
/*      case WM_SYSCOMMAND:
            if (wParam == SC_MAXIMIZE)
            {   if (fullscreen) fullscreen = FALSE; else fullscreen = TRUE;
                updatemenu(MENUITEM_FULLSCREEN);
                closewindow();
                openwindow();
                redrawscreen();
            } else
            {   return DefWindowProc(hwnd, msg, wParam, lParam);
            }
        break; */
        case WM_NCLBUTTONDBLCLK:
            if (fullscreen) fullscreen = FALSE; else fullscreen = TRUE;
            updatemenu(MENUITEM_FULLSCREEN);
            closewindow();
            openwindow();
            redrawscreen();
        break;
        case WM_DROPFILES:
            DragQueryFile((HDROP) wParam, 0, fn_game, MAX_PATH);
            DragFinish((HDROP) wParam);
            // runtime_// assert(strlen(fn_game) <= MAX_PATH);
            engine_load(FALSE);
        break;
        case WM_PAINT:
            BeginPaint(hwnd, &ps);
            redrawscreen();
            EndPaint(hwnd, &ps);
        break;
        case WM_CLOSE:
            cleanexit(EXIT_SUCCESS);
        break;
        case WM_DESTROY:
            PostQuitMessage(0);
        break;
        case WM_MOVE:
            if (fullscreen)
            {   SetWindowPos
                (   MainWindowPtr,
                    HWND_TOP,
                    0, 0,
                    winwidth, winheight,
                    SWP_NOCOPYBITS
                ); // move it back!
            } else
            {   winleftx = (int) LOWORD(lParam);
                wintopy  = (int) HIWORD(lParam);
            }
        break;
        case WM_COMMAND:
            if (!ListeningWindowPtr)
            {   switch(LOWORD(wParam))
                {
                case ID_FILE_RESET:
                    project_reset(FALSE);
                acase ID_FILE_RESETTOMONITOR:
                    docommand(MENUITEM_RESETTOMONITOR);
                acase ID_FILE_OPEN:
                    project_open();
                acase ID_FILE_REVERT:
                    docommand(MENUITEM_REVERT);
                acase ID_FILE_QUICKLOAD:
                    docommand(MENUITEM_QUICKLOAD);
                acase ID_FILE_SAVESNP:
                    if (machine == ARCADIA)
                    {   project_save(KIND_EAS);
                    } elif (machine == INTERTON)
                    {   project_save(KIND_INS);
                    } else
                    {   // assert(machine == ELEKTOR);
                        project_save(KIND_TVS);
                    }
                acase ID_FILE_SAVEROM:
                    docommand(MENUITEM_SAVEROM);
                acase ID_FILE_SAVETVC:
                    // assert(machine == ELEKTOR);
                    project_save(KIND_TVC);
                acase ID_FILE_SAVEILBM:
                    savescreenshot(KIND_ILBM, TRUE);
                acase ID_FILE_SAVEACBM:
                    savescreenshot(KIND_ACBM, TRUE);
                acase ID_FILE_SAVEBMP:
                    savescreenshot(KIND_BMP, TRUE);
                acase ID_FILE_QUICKSAVE:
                    docommand(MENUITEM_QUICKSAVE);
                acase ID_FILE_SERVER:
                    docommand(MENUITEM_SERVER);
                acase ID_FILE_CLIENT:
                    docommand(MENUITEM_CLIENT);
                acase ID_FILE_EXIT:
                    cleanexit(EXIT_SUCCESS);
                acase ID_OPTIONS_GRAPHICS_1XSIZE:
                    resize(1);
                acase ID_OPTIONS_GRAPHICS_2XSIZE:
                    resize(2);
                acase ID_OPTIONS_GRAPHICS_3XSIZE:
                    resize(3);
                acase ID_OPTIONS_SPEED_NTSC:
                    region = REGION_NTSC;
                    updatemenu(MENUITEM_REGION);
                acase ID_OPTIONS_SPEED_PAL:
                    region = REGION_PAL;
                    updatemenu(MENUITEM_REGION);
                acase ID_OPTIONS_SPEED_LIMITREFRESH:
                    if (limitrefresh) limitrefresh = FALSE; else limitrefresh = TRUE;
                    updatemenu(MENUITEM_LIMITREFRESH);
                acase ID_OPTIONS_GRAPHICS_FULLSCREEN:
                    if (fullscreen) fullscreen = FALSE; else fullscreen = TRUE;
                    updatemenu(MENUITEM_FULLSCREEN);
                    closewindow();
                    openwindow();
                    redrawscreen();
                acase ID_OPTIONS_GRAPHICS_NARROW:
                    // assert(MenuPtr);
                    if (wide == 1)
                    {   // it was checked, so now we uncheck it
                        wide = realwide = 2;
                    } else
                    {   // assert(wide == 2);
                        // it was unchecked, so now we check it
                        wide = realwide = 1;
                    }
                    updatemenu(MENUITEM_NARROW);
                    resize(size);
                acase ID_OPTIONS_GRAPHICS_STRETCH:
                    if (stretch) stretch = FALSE; else stretch = TRUE;
                    docommand(MENUITEM_STRETCH);
                acase ID_OPTIONS_GRAPHICS_FLAGLINE:
                    if (flagline) flagline = FALSE; else flagline = TRUE;
                    updatesmlgad(GADPOS_FLAGLINE, flagline, TRUE);
                    updatemenu(MENUITEM_FLAGLINE);
                    if (paused)
                    {   redrawscreen();
                    }
                acase ID_OPTIONS_CONTROLLERS_ANALOG:
                    if (analog) analog = FALSE; else analog = TRUE;
                    updatesmlgad(GADPOS_ANALOG, analog, TRUE);
                    updatemenu(MENUITEM_ANALOG);
                    updatemenu(MENUITEM_SENSITIVITY);
                acase ID_OPTIONS_CONTROLLERS_AUTOFIRE:
                    if (autofire) autofire = FALSE; else autofire = TRUE;
                    updatesmlgad(GADPOS_AUTOFIRE, autofire, TRUE);
                    updatemenu(MENUITEM_AUTOFIRE);
                acase ID_CONTROLLERS_REQUIREBUTTON:
                    if (requirebutton) requirebutton = FALSE; else requirebutton = TRUE;
                    updatemenu(MENUITEM_REQUIREBUTTON);
                acase ID_CONTROLLERS_ENABLE1ST:
                    if (joy0) joy0 = FALSE; else joy0 = TRUE;
                    updatemenu(MENUITEM_JOY0);
                acase ID_CONTROLLERS_ENABLE2ND:
                    if (joy1) joy1 = FALSE; else joy1 = TRUE;
                    updatemenu(MENUITEM_JOY1);
                acase ID_OPTIONS_CONTROLLERS_SWAPPED:
                    if (swapped) swapped = FALSE; else swapped = TRUE;
                    updatesmlgad(GADPOS_SWAPPED, swapped, TRUE);
                    updatemenu(MENUITEM_SWAPPED);
                acase ID_CONTROLLERS_REDEFINE:
                    sound_off(FALSE);
                    DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINE), MainWindowPtr, RedefineDlgProc);
                    sound_on();
                /* acase ID_CONTROLLERS_SENSITIVITY:
                    sound_off(FALSE);
                    DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_SENSITIVITY), MainWindowPtr, SensitivityDlgProc);
                    sound_on(); */
                acase ID_OPTIONS_SPEED_PAUSED:
                case ID_DEBUG_BREAKCONTINUE:
                    if (paused) unpause(); else pause();
                acase ID_OPTIONS_SPEED_WARP:
                    if (warp) warp = FALSE; else warp = TRUE;
                    updatesmlgad(GADPOS_WARP, warp, TRUE);
                    updatemenu(MENUITEM_WARP);
                acase ID_OPTIONS_SPRITES_COLLISIONS:
                    if (collisions) collisions = FALSE; else collisions = TRUE;
                    updatesmlgad(GADPOS_COLLISIONS, collisions, TRUE);
                    updatemenu(MENUITEM_COLLISIONS);
                acase ID_OPTIONS_SPRITES_DEMULTIPLEX:
                    if (demultiplex) demultiplex = FALSE; else demultiplex = TRUE;
                    docommand(MENUITEM_DEMULTIPLEX);
                acase ID_MACHINE_ARCADIA:
                    docommand(COMMAND_ARCADIA);
                acase ID_MACHINE_INTERTON:
                    docommand(COMMAND_INTERTON);
                acase ID_MACHINE_ELEKTOR:
                    docommand(COMMAND_ELEKTOR);
                acase ID_MACHINE_USEHACKS:
                    if (usehacks) usehacks = FALSE; else usehacks = TRUE;
                    updatemenu(MENUITEM_USEHACKS);
                acase ID_OPTIONS_AUTOSAVE:
                    // assert(MenuPtr);
                    if (autosave) autosave = FALSE; else autosave = TRUE;
                    updatemenu(MENUITEM_AUTOSAVE);
                acase ID_SENSEGAME_VERBOSE:
                    sensegame = SENSEGAME_VERBOSE;
                    updatemenu(MENUITEM_SENSEGAME);
                    updatemenu(MENUITEM_USEHACKS);
                acase ID_SENSEGAME_QUIET:
                    sensegame = SENSEGAME_QUIET;
                    updatemenu(MENUITEM_SENSEGAME);
                    updatemenu(MENUITEM_USEHACKS);
                acase ID_SENSEGAME_NO:
                    sensegame = SENSEGAME_NO;
                    updatemenu(MENUITEM_SENSEGAME);
                    updatemenu(MENUITEM_USEHACKS);
                acase ID_OPTIONS_SOUND:
                    if (sound) sound = FALSE; else sound = TRUE;
                    docommand(MENUITEM_SOUND);
                acase ID_DEBUG_WATCHREADS:
                    if (watchreads) watchreads = FALSE; else watchreads = TRUE;
                    updatemenu(MENUITEM_WATCHREADS);
                acase ID_VIEW_MENU:
                    if (showmenubar) showmenubar = FALSE; else showmenubar = TRUE;
                    docommand(MENUITEM_MENUBAR);
                acase ID_VIEW_POINTER:
                    if (showpointer) showpointer = FALSE; else showpointer = TRUE;
                    docommand(MENUITEM_POINTER);
                acase ID_VIEW_TITLE:
                    if (showtitlebar) showtitlebar = FALSE; else showtitlebar = TRUE;
                    docommand(MENUITEM_TITLEBAR);
                acase ID_VIEW_TOOL:
                    if (showtoolbar) showtoolbar = FALSE; else showtoolbar = TRUE;
                    docommand(MENUITEM_TOOLBAR);
                acase ID_VIEW_MONITOR:
                    view_monitor();
                acase ID_VIEW_PSGS:
                    // assert(machine == ELEKTOR);
                    view_psgs();
                acase ID_VIEW_ALLMEMORY:
                    view_allmemory();
                acase ID_VIEW_CRAM:
                    view_cram();
                acase ID_VIEW_MRAM:
                    view_mram();
                acase ID_VIEW_UPPERSCREEN:
                    view_upperscreen();
                acase ID_VIEW_LOWERSCREEN:
                    view_lowerscreen();
                acase ID_VIEW_UDCS:
                    view_udcs();
                acase ID_VIEW_XVI:
                    view_xvi();
                acase ID_VIEW_CPU:
                    view_cpu();
                acase ID_VIEW_IMAGERY:
                    view_imagery();
                acase ID_VIEW_BRAM:
                    view_bram();
                acase ID_MACRO_TOGGLERECORDING:
                    if (recmode == RECMODE_NORMAL)
                    {   if (machine == ARCADIA)
                        {   project_save(KIND_EAR);
                        } elif (machine == INTERTON)
                        {   project_save(KIND_INR);
                        } else
                        {   // assert(machine == ELEKTOR);
                            project_save(KIND_TVR);
                    }   }
                    elif (recmode == RECMODE_RECORD)
                    {   macro_stop();
                    } // else assert(recmode == RECMODE_PLAY);
                acase ID_MACRO_TOGGLEPLAYING:
                    if (recmode == RECMODE_PLAY)
                    {   macro_stop();
                    } else
                    {   ; // assert(recmode == RECMODE_NORMAL || recmode == RECMODE_RECORD);
                        // assert(0);
                    }
                acase ID_MACRO_STARTRECORDING:
                    if (machine == ARCADIA)
                    {   project_save(KIND_EAR);
                    } elif (machine == INTERTON)
                    {   project_save(KIND_INR);
                    } else
                    {   // assert(machine == ELEKTOR);
                        project_save(KIND_TVR);
                    }
                acase ID_MACRO_RESTARTPLAYBACK:
                    macro_restartplayback();
                acase ID_MACRO_STOP:
                    macro_stop();
                acase ID_MACRO_LOOP:
                    if (loop) loop = FALSE; else loop = TRUE;
			        updatemenu(MENUITEM_LOOP);
                acase ID_HELP_GG:
                    if (machine == ARCADIA)
                    {   ShellExecute(0, NULL, "http://amigan.classicgaming.gamespy.com/agg/", NULL, NULL, SW_SHOW);
                    } elif (machine == INTERTON)
                    {   ShellExecute(0, NULL, "http://amigan.classicgaming.gamespy.com/igg/", NULL, NULL, SW_SHOW);
                    } else
                    {   // assert(machine == ELEKTOR);
                        ShellExecute(0, NULL, "http://amigan.classicgaming.gamespy.com/egg/", NULL, NULL, SW_SHOW);
                    }
                acase ID_HELP_ABOUT:
                    sound_off(FALSE);
                    DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_ABOUT), MainWindowPtr, HelpDlgProc);
                    sound_on();
                acase ID_VIEW_CONTROLS:
                    view_controls();
                acase ID_DEBUG_CLI:
                    if (cli) cli = FALSE; else cli = TRUE;
					docommand(MENUITEM_CLI);
				acase ID_DEBUG_TRACECPU:
                    if (trace) trace = FALSE; else trace = TRUE;
                    updatemenu(MENUITEM_TRACECPU);
                acase ID_DEBUG_LOGINSTRUCTIONS:
                    if (loginstructions) loginstructions = FALSE; else loginstructions = TRUE;
                    updatemenu(MENUITEM_LOGINSTRUCTIONS);
                acase ID_DEBUG_LOGREADS:
                    if (logreads) logreads = FALSE; else logreads = TRUE;
                    updatemenu(MENUITEM_LOGREADS);
                acase ID_DEBUG_LOGWRITES:
                    if (logwrites) logwrites = FALSE; else logwrites = TRUE;
                    updatemenu(MENUITEM_LOGWRITES);
                acase ID_DEBUG_LOGMONITOR:
                    if (logmonitor) logmonitor = FALSE; else logmonitor = TRUE;
                    updatemenu(MENUITEM_LOGMONITOR);
                acase ID_DEBUG_SETBP:
                    debug_set(KIND_BP);
                acase ID_DEBUG_CLEARBP:
                    debug_clearbp();
                acase ID_DEBUG_SETWP:
                    debug_set(KIND_WP);
                acase ID_DEBUG_CLEARWP:
                    debug_clearwp();
                acase ID_DEBUG_STEP:
                    docommand(MENUITEM_STEP);
                acase ID_DEBUG_IGNORE:
                    docommand(COMMAND_IGNORE);
                acase ID_DEBUG_RUNTORASTLINE:
                    runtorastline = TRUE;
                    unpause();
                acase ID_DEBUG_RUNTORASTN:
                    debug_setrastn();
                    unpause();
                acase ID_DEBUG_RUNTODMA:
                    runtodma = TRUE;
                    unpause();
                acase ID_DEBUG_RUNTOINTERRUPT:
                    runtointerrupt = TRUE;
                    unpause();
                acase ID_DEBUG_RUNTOFRAME:
                    runtoframe = TRUE;
                    unpause();
                acase ID_DEBUG_EDIT:
                    debug_edit();
                adefault:
                break;
            }   }
        acase WM_NOTIFY:
            if ((((LPNMHDR) lParam)->code) == TTN_NEEDTEXT)
            {   lpttt = (LPTOOLTIPTEXT) lParam; // button info
                lpttt->hinst = InstancePtr; // used in our main dialog

                switch(lpttt->hdr.idFrom) // Check which button we hover on
                {
                case ID_FILE_RESET:
                    lpttt->lpszText = "Reset to game (F5)";
                acase ID_FILE_OPEN:
                    lpttt->lpszText = "Open (Ctrl+O)";
                acase ID_FILE_SAVESNP:
                    lpttt->lpszText = "Save snapshot (Ctrl+S)";
                acase ID_FILE_SAVEILBM:
                    lpttt->lpszText = "Save screenshot as IFF ILBM (Ctrl+_)";
                acase ID_OPTIONS_GRAPHICS_FLAGLINE:
                    lpttt->lpszText = "Flag line emulation? (Ctrl+P)";
                acase ID_OPTIONS_CONTROLLERS_ANALOG:
                    lpttt->lpszText = "Analog controllers for guest? (Ctrl+F6)";
                acase ID_OPTIONS_CONTROLLERS_AUTOFIRE:
                    lpttt->lpszText = "Autofire? (Ctrl+F7)";
                acase ID_OPTIONS_CONTROLLERS_SWAPPED:
                    lpttt->lpszText = "Controllers swapped? (F7)";
                acase ID_OPTIONS_SPEED_PAUSED:
                    lpttt->lpszText = "Paused? (Pause)";
                acase ID_OPTIONS_SPEED_WARP:
                    lpttt->lpszText = "Warp speed? (F8)";
                acase ID_OPTIONS_SPRITES_COLLISIONS:
                    lpttt->lpszText = "Collisions? (Ctrl+F8)";
                acase ID_OPTIONS_SPRITES_DEMULTIPLEX:
                    lpttt->lpszText = "Demultiplex sprites? (Ctrl+D)";
                acase ID_OPTIONS_SOUND:
                    lpttt->lpszText = "Sound? (Ctrl+U)";
                acase ID_MACRO_TOGGLEPLAYING:
                    lpttt->lpszText = "Play/stop (Ctrl+Plus (+)/Ctrl+T)";
                acase ID_MACRO_TOGGLERECORDING:
                    lpttt->lpszText = "Record/stop (Ctrl+R/Ctrl+T)";
                adefault:
                break;
            }   }
        adefault:
            return DefWindowProc(hwnd, msg, wParam, lParam);
    }
    return 0;
}

int PASCAL WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{   FLAG       settings       = FALSE;
    WNDCLASSEX wc;
    int        argc,
               i,
               joystickCount  = 0,
               joystickID     = 0,
               length,
               number,
               x, y;
    char**     argv;
    WSADATA    WSAInfo;
    JOYINFOEX  joyInfoEx;

    /* Start of program. */

    InitCommonControls();

    argc = __argc;
    argv = __argv;

    for (x = 0; x < BOXWIDTH; x++)
    {   for (y = 0; y < BOXHEIGHT; y++)
        {   screen[x][y] = 8;
    }   }

    // get joystick IDs
    joyInfoEx.dwSize  = sizeof(joyInfoEx);
	joyInfoEx.dwFlags = JOY_RETURNALL;

    for (joystickID = 0; joystickID < 15 && joystickCount < 2; joystickID++)
    {   if (joyGetPosEx(joystickID, &joyInfoEx) == JOYERR_NOERROR)
        {   joystickIDMapping[joystickCount] = joystickID;
			joystickCount++;
    }   }

    engine_setup();

    InstancePtr = GetModuleHandle(NULL);

    DisplayWidth  = GetSystemMetrics(SM_CXSCREEN);
    DisplayHeight = GetSystemMetrics(SM_CYSCREEN);

    DISCARD GetCurrentDirectory(MAX_PATH, ProgDir);
    filepart[0] =
    fn_game[0]  =
    hostname[0] = 0;
    strcpy(pathpart, ".\\Games\\");
    strcpy(fn_config, "WA.CFG");

    /* Initialize networking */
    WSAStartup(MAKEWORD(2, 2), &WSAInfo);

    loadconfig(); // load default configuration file

    if (argc >= 2)
    {   for (i = 1; i < argc; i++)
        {   if (!strcmp(argv[i], "?"))
            {   printusage(argv[0]);
            } elif (!strnicmp(argv[i], "ANALOG", 6))
            {   if (!stricmp(&argv[i][6], "=ON"))
                {   analog = TRUE;
                } elif (!stricmp(&argv[i][6], "=OFF"))
                {   analog = FALSE;
                } else
                {   printusage(argv[0]);
            }   }
            elif (!strnicmp(argv[i], "AUTOFIRE", 8))
            {   if (!stricmp(&argv[i][8], "=ON"))
                {   autofire = TRUE;
                } elif (!stricmp(&argv[i][8], "=OFF"))
                {   autofire = FALSE;
                } else
                {   printusage(argv[0]);
            }   }
            elif (!strnicmp(argv[i], "AUTOSAVE", 8))
            {   if (!stricmp(&argv[i][8], "=ON"))
                {   autosave = TRUE;
                } elif (!stricmp(&argv[i][8], "=OFF"))
                {   autosave = FALSE;
                } else
                {   printusage(argv[0]);
            }   }
            elif (!strnicmp(argv[i], "COLLISIONS", 10))
            {   if (!stricmp(&argv[i][10], "=ON"))
                {   collisions = TRUE;
                } elif (!stricmp(&argv[i][10], "=OFF"))
                {   collisions = FALSE;
                } else
                {   printusage(argv[0]);
            }   }
            elif (!strnicmp(argv[i], "DEMULTIPLEX", 11))
            {   if (!stricmp(&argv[i][11], "=ON"))
                {   demultiplex = TRUE;
                } elif (!stricmp(&argv[i][11], "=OFF"))
                {   demultiplex = FALSE;
                } else
                {   printusage(argv[0]);
            }   }
            elif (!strnicmp(argv[i], "FLAGLINE", 8))
            {   if (!stricmp(&argv[i][8], "=ON"))
                {   flagline = TRUE;
                } elif (!stricmp(&argv[i][8], "=OFF"))
                {   flagline = FALSE;
                } else
                {   printusage(argv[0]);
            }   }
            elif (!strnicmp(argv[i], "FULLSCREEN", 10))
            {   if (!stricmp(&argv[i][10], "=ON"))
                {   fullscreen = TRUE;
                } elif (!stricmp(&argv[i][10], "=OFF"))
                {   fullscreen = FALSE;
                } else
                {   printusage(argv[0]);
            }   }
            elif (!strnicmp(argv[i], "LIMITREFRESH", 12))
            {   if (!stricmp(&argv[i][12], "=ON"))
                {   limitrefresh = TRUE;
                } elif (!stricmp(&argv[i][12], "=OFF"))
                {   limitrefresh = FALSE;
                } else
                {   printusage(argv[0]);
            }   }
            elif (!strnicmp(argv[i], "MACHINE", 7))
            {   if (!stricmp(&argv[i][7], "=ARCADIA"))
                {   machine = ARCADIA;
                } elif (!stricmp(&argv[i][7], "=INTERTON"))
                {   machine = INTERTON;
                } elif (!stricmp(&argv[i][7], "=ELEKTOR"))
                {   machine = ELEKTOR;
                } else
                {   printusage(argv[0]);
            }   }
            elif (!strnicmp(argv[i], "MENUBAR", 7))
            {   if (!stricmp(&argv[i][7], "=ON"))
                {   showmenubar = TRUE;
                } elif (!stricmp(&argv[i][7], "=OFF"))
                {   showmenubar = FALSE;
                } else
                {   printusage(argv[0]);
            }   }
            elif (!strnicmp(argv[i], "NARROW", 6))
            {   if (!stricmp(&argv[i][6], "=ON"))
                {   wide = 1;
                } elif (!stricmp(&argv[i][6], "=OFF"))
                {   wide = 2;
                } else
                {   printusage(argv[0]);
            }   }
            elif (!strnicmp(argv[i], "PAUSED", 6))
            {   if (!stricmp(&argv[i][6], "=ON"))
                {   paused = TRUE;
                } elif (!stricmp(&argv[i][6], "=OFF"))
                {   paused = FALSE;
                } else
                {   printusage(argv[0]);
            }   }
            elif (!strnicmp(argv[i], "SETTINGS", 8))
            {   if (argv[i][8] != '=')
                {   printusage(argv[0]);
                }
                strcpy(fn_config, (STRPTR) &argv[i][9]);
                settings = TRUE;
                // we don't load it yet
            } elif (!strnicmp(argv[i], "SIZE", 4))
            {   if (!strcmp(&argv[i][4], "=1"))
                {   size = 1;
                } elif (!strcmp(&argv[i][4], "=2"))
                {   size = 2;
                } elif (!strcmp(&argv[i][4], "=3"))
                {   size = 3;
                } else
                {   printusage(argv[0]);
            }   }
            elif (!strnicmp(argv[i], "SOUND", 5))
            {   if (!stricmp(&argv[i][5], "=ON"))
                {   sound = TRUE;
                } elif (!stricmp(&argv[i][5], "=OFF"))
                {   sound = FALSE;
                } else
                {   printusage(argv[0]);
            }   }
            elif (!strnicmp(argv[i], "STRETCH", 7))
            {   if (!stricmp(&argv[i][7], "=ON"))
                {   stretch = TRUE;
                } elif (!stricmp(&argv[i][7], "=OFF"))
                {   stretch = FALSE;
                } else
                {   printusage(argv[0]);
            }   }
            elif (!strnicmp(argv[i], "SWAPPED", 7))
            {   if (!stricmp(&argv[i][7], "=ON"))
                {   swapped = TRUE;
                } elif (!stricmp(&argv[i][7], "=OFF"))
                {   swapped = FALSE;
                } else
                {   printusage(argv[0]);
            }   }
            elif (!strnicmp(argv[i], "TITLEBAR", 8))
            {   if (!stricmp(&argv[i][8], "=ON"))
                {   showtitlebar = TRUE;
                } elif (!stricmp(&argv[i][8], "=OFF"))
                {   showtitlebar = FALSE;
                } else
                {   printusage(argv[0]);
            }   }
            elif (!strnicmp(argv[i], "TOOLBAR", 7))
            {   if (!stricmp(&argv[i][7], "=ON"))
                {   showtoolbar = TRUE;
                } elif (!stricmp(&argv[i][7], "=OFF"))
                {   showtoolbar = FALSE;
                } else
                {   printusage(argv[0]);
            }   }
            elif (!strnicmp(argv[i], "WARP", 4))
            {   if (!stricmp(&argv[i][4], "=ON"))
                {   warp = TRUE;
                } elif (!stricmp(&argv[i][4], "=OFF"))
                {   warp = FALSE;
               } else
                {   printusage(argv[0]);
            }   }
            elif (!strnicmp(argv[i], "LOGINSTRUCTIONS", 15))
            {   if (!stricmp(&argv[i][15], "=ON"))
                {   loginstructions = TRUE;
                } elif (!stricmp(&argv[i][15], "=OFF"))
                {   loginstructions = FALSE;
                } else
                {   printusage(argv[0]);
            }   }
            elif (!strnicmp(argv[i], "LOGREADS", 8))
            {   if (!stricmp(&argv[i][8], "=ON"))
                {   logreads = TRUE;
                } elif (!stricmp(&argv[i][8], "=OFF"))
                {   logreads = FALSE;
                } else
                {   printusage(argv[0]);
            }   }
            elif (!strnicmp(argv[i], "LOGWRITES", 9))
            {   if (!stricmp(&argv[i][9], "=ON"))
                {   logwrites = TRUE;
                } elif (!stricmp(&argv[i][9], "=OFF"))
                {   logwrites = FALSE;
                } else
                {   printusage(argv[0]);
            }   }
            elif
            (    (!strnicmp(argv[i], "BREAKPOINT", 10))
              || (!strnicmp(argv[i], "BP", 2))
            )
            {   if (!strnicmp(argv[i], "BREAKPOINT", 10))
                {   length = 10;
                } else
                {   // assert(!strnicmp(argv[i], "BP", 2));
                    length = 2;
                }

                // sanity checks
                if (argv[i][length] != '=')
                {   printusage(argv[0]);
                }
                if (strlen(&argv[i][length + 1]) > 13)
                {   T1;
                    DISCARD printf("%s: BREAKPOINT argument must be <= 13 characters!\n", argv[0]);
                    cleanexit(EXIT_FAILURE);
                }

                number = friendly_to_number(&argv[i][length + 1]);
                if (number == OUTOFRANGE)
                {   T1;
                    DISCARD printf("%s: BREAKPOINT address must be $0000-$7FFF!\n", argv[0]);
                    cleanexit(EXIT_FAILURE);
                } else
                {   bp = number;
                    getfriendly(bp);
                    T1;
                    DISCARD printf("Set code breakpoint at %s!\n\n", friendly);
                    // no window to be reactivated yet
            }   }
            elif
            (   (!strnicmp(argv[i], "WATCHPOINT", 10))
             || (!strnicmp(argv[i], "WP", 2))
            )
            {   if (!strnicmp(argv[i], "WATCHPOINT", 10))
                {   length = 10;
                } else
                {   // assert(!strnicmp(argv[i], "WP", 2));
                    length = 2;
                }

                // sanity checks
                if (argv[i][length] != '=')
                {   printusage(argv[0]);
                }
                if (strlen(&argv[i][length + 1]) > 13)
                {   T1;
                    DISCARD printf("%s: WATCHPOINT argument must be <= 13 characters!\n", argv[0]);
                    cleanexit(EXIT_FAILURE);
                }

                number = friendly_to_number(&argv[i][length + 1]);
                if (number == OUTOFRANGE)
                {   T1;
                    DISCARD printf("%s: WATCHPOINT address must be $0000-$7FFF!\n", argv[0]);
                    cleanexit(EXIT_FAILURE);
                } else
                {   wp = number;
                    getfriendly(wp);
                    T1;
                    DISCARD printf("Set data watchpoint at %s!\n\n", friendly);
                    // no window to be reactivated yet
            }   }
            elif (!strnicmp(argv[i], "FILE", 4))
            {   if (argv[i][4] != '=')
                {   printusage(argv[0]);
                }
                strcpy(fn_game, (STRPTR) &argv[i][5]);
                // we don't load it yet
            } else
            {   strcpy(fn_game, argv[i]);
                // we don't load it yet
    }   }   }

    if (settings)
    {   loadconfig();
    }

    menuheight    = GetSystemMetrics(SM_CYMENU);    // normally 19
    titleheight   = GetSystemMetrics(SM_CYCAPTION); // normally 19
    toolbarheight = 30; // 12 + bitmap height
    // we get an exact value later

    OpenChannels(); // we keep the channels open permanently
    sound_on();

    // AcceleratorPtr = LoadAccelerators(InstancePtr, MAKEINTRESOURCE(IDR_ACCELERATOR)); // automatically freed at exit

    wc.cbSize        = sizeof(WNDCLASSEX);
    wc.style         = 0;
    wc.lpfnWndProc   = WndProc;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = 0;
    wc.hInstance     = InstancePtr;
    wc.hIcon         = LoadIcon(InstancePtr, MAKEINTRESOURCE(IDI_ICON));
    wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = (HBRUSH)(COLOR_GRAYTEXT+1); // was (HBRUSH)(COLOR_WINDOW+1);
    wc.lpszMenuName  = NULL;
    wc.lpszClassName = g_szClassName;
    wc.hIconSm       = LoadIcon(InstancePtr, MAKEINTRESOURCE(IDI_ICON));

    if(!RegisterClassEx(&wc))
    {
        MessageBox(NULL, "Window registration failed!", "WinArcadia: Error",
            MB_ICONEXCLAMATION | MB_OK);
        cleanexit(EXIT_FAILURE);
    }

    memset(&TheBitMapInfo, 0, sizeof(TheBitMapInfo));
    TheBitMapInfo.Header.biSize        = sizeof(BITMAPINFOHEADER);
    TheBitMapInfo.Header.biWidth       = BOXWIDTH;
    TheBitMapInfo.Header.biHeight      = -BOXHEIGHT;
    TheBitMapInfo.Header.biBitCount    = 32; // sizeof(ULONG) * 8;
    TheBitMapInfo.Header.biPlanes      = 1;
    TheBitMapInfo.Header.biCompression = BI_RGB; // BI_BITFIELDS seems to work just as well
    TheBitMapInfo.Colours[0]           = 0x00FF0000;
    TheBitMapInfo.Colours[1]           = 0x0000FF00;
    TheBitMapInfo.Colours[2]           = 0x000000FF;

    MenuBitmapPtr[0] = LoadBitmap(InstancePtr, MAKEINTRESOURCE(IDB_M_RESET   ));
    MenuBitmapPtr[1] = LoadBitmap(InstancePtr, MAKEINTRESOURCE(IDB_M_OPEN    ));
    MenuBitmapPtr[2] = LoadBitmap(InstancePtr, MAKEINTRESOURCE(IDB_M_SAVESNP ));
    MenuBitmapPtr[3] = LoadBitmap(InstancePtr, MAKEINTRESOURCE(IDB_M_SAVEILBM));

#ifdef USEGDI
    display = malloc(BOXWIDTH * BOXHEIGHT * sizeof(ULONG));
#endif
#ifdef USEDIRECTX
    if (!(OurhDC = CreateCompatibleDC(0)))
    {   MessageBox
        (   NULL,
            "CreateCompatibleDC() failed!",
            "WinArcadia: Error",
            MB_ICONEXCLAMATION | MB_OK
        );
        cleanexit(EXIT_FAILURE);
    }
    if (!(OurhBitMap = CreateDIBSection
    (   OurhDC,
        (BITMAPINFO *) &TheBitMapInfo,
        DIB_RGB_COLORS,
        &display,
        0,
        0
    )))
    {   MessageBox
        (   NULL,
            "CreateDIBSection() failed!",
            "WinArcadia: Error",
            MB_ICONEXCLAMATION | MB_OK
        );
        cleanexit(EXIT_FAILURE);
    }
    if (!SelectObject(OurhDC, OurhBitMap))
    {   MessageBox
        (   NULL,
            "SelectObject() failed!",
            "WinArcadia: Error",
            MB_ICONEXCLAMATION | MB_OK
        );
        cleanexit(EXIT_FAILURE);
    }
#endif

    // memset(display, 0, BOXWIDTH * BOXHEIGHT * sizeof(ULONG)); // probably not needed

    openwindow(); // calls redrawscreen()

    if (machine == ARCADIA)
    {   arcadia(TRUE);
    } elif (machine == INTERTON)
    {   interton(TRUE);
    } else
    {   // assert(machine == ELEKTOR);
        elektor(TRUE);
    }

    if (fn_game[0]) // filename from CLI
    {   if (!engine_load(FALSE))
        {   // don't beep, the user may have just wanted to set the directory.
            project_reset(FALSE);
    }   }
    else
    {   strcpy(fn_game, pathpart);
        if (autosave)
        {   changefilepart(autoname[machine]);
			if (!engine_load(TRUE))
            {   // don't beep
                project_reset(FALSE);
        }   }
        else
        {   project_reset(FALSE);
    }   }

#ifdef VERBOSE
    T1;
#ifdef USEGDI
    printf("Using GDI.\n\n");
#endif
#ifdef USEDIRECTX
    printf("Using DirectX.\n\n");
#endif
    T2;
#endif

    if (!paused) // must not be done until after loading the game of course
    {   sound_on();
    }
    if (!showpointer) ShowCursor(FALSE);

    for (;;)
    {   if (!paused)
        {   if (machine == ARCADIA)
            {   uvi();
            } else
            {   // assert(machine == INTERTON || machine == ELEKTOR);
                pvi();
            }
            updatemonitor(FALSE);
            checkinput();
            thewait();

            /* This uses a polling technique to determine whether a second has
               passed. Using timer requests would give better multi-tasking
               performance. */

            if (newtime >= oldtime + 1000)
            {   oldtime = newtime;

                if (!paused)
                {
#ifdef MODE_MHZ
                    fps = elapsed - oldelapsed; // overflow is possible but unlikely and not very harmful
                    oldelapsed = elapsed;
#else // MODE_FPS and MODE_PERCENT
                    fps = frames - oldframes; // overflow is possible but unlikely and not very harmful
                    oldframes = frames;
#endif

#ifdef MODE_PERCENT
                    fps *= 100;
                    if (region == REGION_NTSC)
                    {   fps /= 60;
                    } else
                    {   fps /= 50;
                    }
#endif

                    updatefps();
        }   }   }
        else
        {   updatemonitor(FALSE);

            DISCARD WaitMessage(); // multitask with other programs while paused

            checkinput();
    }   }

    // control can never reach here
    // "The return value is only really useful if your program is designed
    // to be called by another program and you want to return a specific
    // value."
    return 0;
}

EXPORT int stcl_d(char* out, long lvalue) // number -> decimal string
{   if (lvalue < 0)
    {   out[0] = '-';
        return(stcul_d(out + 1, (unsigned long int) abs((int) lvalue)) + 1);
    }
    return(stcul_d(out, (unsigned long int) lvalue));
}
EXPORT int stcul_d(char* out, unsigned long lvalue) // number -> decimal string
{   ULONG calc,
          i,
          where   = 0;
    FLAG  started = FALSE;

    for (i = ONE_BILLION; i >= 1; i /= 10)
    {   calc = lvalue / i;
        // assert(calc < 10);
        if (calc || started || i == 1)
        {   *(out + where) = (char) ('0' + calc);
            where++;
            started = TRUE;
        }
        lvalue %= i;
    }

    *(out + where) = 0;
    return((int) where);
}
EXPORT int stcl_h(char* out, unsigned long lvalue)
{   ULONG calc,
          i,
          where   = 0;
    FLAG  started = FALSE;

    for (i = 0x10000000; i >= 1; i /= 16)
    {   calc = lvalue / i;
        // assert(calc < 16);
        if (calc || started || i == 1)
        {   if (calc <= 9)
            {   *(out + where) = (char) ('0' + calc);
            } else
            {   *(out + where) = (char) ('A' + calc - 10);
            }
            where++;
            started = TRUE;
        }
        lvalue %= i;
    }

    *(out + where) = 0;
    return((int) where);
}
EXPORT int stcd_l(char* in, SLONG* lvalue) // string -> decimal number
{   *(lvalue) = atol(in);
    return 1; // not really
}
EXPORT int stch_l(char* in, SLONG* lvalue) // string -> hex number
{   *(lvalue) = strtol(in, NULL, 16);
    return 1; // not really
}

BOOL CALLBACK HelpDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam)
{   switch(Message)
    {
    case WM_INITDIALOG:
        return TRUE;
    acase WM_CLOSE:
        clearkybd();
        EndDialog(hwnd, 0);
        return TRUE;
    acase WM_COMMAND:
        if
        (   (LOWORD (wParam) == IDC_LINK1)
         || (LOWORD (wParam) == IDC_AMIGANSOFTWARE)
        )
        {   ShellExecute(0, NULL, "http://amigan.1emu.net/releases/", NULL, NULL, SW_SHOW);
        } elif
        (   (LOWORD (wParam) == IDC_LINK2)
         || (LOWORD (wParam) == IDC_EA2001CENTRAL)
        )
        {   ShellExecute(0, NULL, "http://amigan.classicgaming.gamespy.com/",   NULL, NULL, SW_SHOW);
        } elif (LOWORD (wParam) == IDOK)
        {   clearkybd();
            EndDialog(hwnd, 0);
        }
        return TRUE;
    adefault:
        return FALSE;
    break;
    }
    return TRUE;
}

BOOL CALLBACK RedefineDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam)
{   int  i, j;

    switch(Message)
    {
    case WM_INITDIALOG:
        for (i = 0; i < 2; i++)
        {   for (j = 0; j < NUMKEYS; j++)
            {   temp_keypads[i][j] = keypads[i][j];
        }   }
        for (i = 0; i < 4; i++)
        {   temp_console[i] = console[i];
        }

        if (machine != ARCADIA)
        {   // assert(machine == INTERTON || machine == ELEKTOR);
            EnableWindow(GetDlgItem(hwnd, IDC_P1_X1), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P1_X2), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P1_X3), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P1_X4), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P2_X1), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P2_X2), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P2_X3), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P2_X4), FALSE);
        }
        if (machine != ELEKTOR)
        {   EnableWindow(GetDlgItem(hwnd, IDC_KEY_RESET), FALSE);
            if (machine == INTERTON)
            {   EnableWindow(GetDlgItem(hwnd, IDC_KEY_B), FALSE);
        }   }
        refreshkeys(hwnd);

        return TRUE;
    acase WM_CLOSE:
        clearkybd();
        EndDialog(hwnd, 0);
        return TRUE;
    acase WM_COMMAND:
        switch(LOWORD(wParam))
        {
        case IDOK:
            goto DONE;
        acase IDCANCEL:
            clearkybd();
            EndDialog(hwnd, 0);
            return TRUE;
        acase IDC_RESETTODEFAULTS:
            for (i = 0; i < 2; i++)
            {   for (j = 0; j < NUMKEYS; j++)
                {   temp_keypads[i][j] = default_keypads[i][j];
            }   }
            for (i = 0; i < 4; i++)
            {   temp_console[i] = default_console[i];
            }
            refreshkeys(hwnd);
        acase IDC_P1_1ST:
            side = 0; whichkey = 0;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P1_2ND:
            side = 0; whichkey = 21;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P1_3RD:
            side = 0; whichkey = 22;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P1_4TH:
            side = 0; whichkey = 23;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P1_1:
            side = 0; whichkey = 1;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P1_2:
            side = 0; whichkey = 2;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P1_3:
            side = 0; whichkey = 3;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P1_4:
            side = 0; whichkey = 4;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P1_5:
            side = 0; whichkey = 5;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P1_6:
            side = 0; whichkey = 6;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P1_7:
            side = 0; whichkey = 7;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P1_8:
            side = 0; whichkey = 8;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P1_9:
            side = 0; whichkey = 9;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P1_C:
            side = 0; whichkey = 10;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P1_0:
            side = 0; whichkey = 11;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P1_E:
            side = 0; whichkey = 12;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P1_X1:
            side = 0; whichkey = 13;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P1_X2:
            side = 0; whichkey = 14;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P1_X3:
            side = 0; whichkey = 15;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P1_X4:
            side = 0; whichkey = 16;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P1_UP:
            side = 0; whichkey = 17;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P1_DOWN:
            side = 0; whichkey = 18;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P1_LEFT:
            side = 0; whichkey = 19;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P1_RIGHT:
            side = 0; whichkey = 20;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P2_1ST:
            side = 1; whichkey = 0;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P2_2ND:
            side = 1; whichkey = 21;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P2_3RD:
            side = 1; whichkey = 22;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P2_4TH:
            side = 1; whichkey = 23;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P2_1:
            side = 1; whichkey = 1;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P2_2:
            side = 1; whichkey = 2;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P2_3:
            side = 1; whichkey = 3;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P2_4:
            side = 1; whichkey = 4;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P2_5:
            side = 1; whichkey = 5;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P2_6:
            side = 1; whichkey = 6;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P2_7:
            side = 1; whichkey = 7;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P2_8:
            side = 1; whichkey = 8;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P2_9:
            side = 1; whichkey = 9;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P2_C:
            side = 1; whichkey = 10;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P2_0:
            side = 1; whichkey = 11;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P2_E:
            side = 1; whichkey = 12;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P2_X1:
            side = 1; whichkey = 13;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P2_X2:
            side = 1; whichkey = 14;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P2_X3:
            side = 1; whichkey = 15;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P2_X4:
            side = 1; whichkey = 16;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P2_UP:
            side = 1; whichkey = 17;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P2_DOWN:
            side = 1; whichkey = 18;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P2_LEFT:
            side = 1; whichkey = 19;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_P2_RIGHT:
            side = 1; whichkey = 20;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_KEY_START:
            side = 2; whichkey = 0;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_KEY_A:
            side = 2; whichkey = 1;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_KEY_B:
            side = 2; whichkey = 2;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        acase IDC_KEY_RESET:
            side = 2; whichkey = 3;
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
            refreshkeys(hwnd); // this is overkill
        adefault:
        break;
        }
        return FALSE;
    adefault:
        return FALSE;
    break;
    }

    return TRUE;

DONE:
    clearkybd();
    EndDialog(hwnd, 0);

    for (i = 0; i < 2; i++)
    {   for (j = 0; j < NUMKEYS; j++)
        {   keypads[i][j] = temp_keypads[i][j];
    }   }
    for (i = 0; i < 4; i++)
    {   console[i] = temp_console[i];
    }

    return TRUE;
}

MODULE void refreshkeys(HWND hwnd)
{   FLAG ok;
    int  i,
         j;

    SetDlgItemText(hwnd, IDC_KEY_START, keyname[temp_console[0]]);
    SetDlgItemText(hwnd, IDC_KEY_A    , keyname[temp_console[1]]);
    SetDlgItemText(hwnd, IDC_KEY_B    , keyname[temp_console[2]]);
    SetDlgItemText(hwnd, IDC_KEY_RESET, keyname[temp_console[3]]);

    SetDlgItemText(hwnd, IDC_P1_2     , keyname[temp_keypads[0][ 2]]);
    SetDlgItemText(hwnd, IDC_P1_X1    , keyname[temp_keypads[0][13]]);
    SetDlgItemText(hwnd, IDC_P1_X2    , keyname[temp_keypads[0][14]]);
    SetDlgItemText(hwnd, IDC_P1_X3    , keyname[temp_keypads[0][15]]);
    SetDlgItemText(hwnd, IDC_P1_1     , keyname[temp_keypads[0][ 1]]);
    SetDlgItemText(hwnd, IDC_P1_X4    , keyname[temp_keypads[0][16]]);
    SetDlgItemText(hwnd, IDC_P1_3     , keyname[temp_keypads[0][ 3]]);
    SetDlgItemText(hwnd, IDC_P1_4     , keyname[temp_keypads[0][ 4]]);
    SetDlgItemText(hwnd, IDC_P1_5     , keyname[temp_keypads[0][ 5]]);
    SetDlgItemText(hwnd, IDC_P1_6     , keyname[temp_keypads[0][ 6]]);
    SetDlgItemText(hwnd, IDC_P1_7     , keyname[temp_keypads[0][ 7]]);
    SetDlgItemText(hwnd, IDC_P1_8     , keyname[temp_keypads[0][ 8]]);
    SetDlgItemText(hwnd, IDC_P1_9     , keyname[temp_keypads[0][ 9]]);
    SetDlgItemText(hwnd, IDC_P1_C     , keyname[temp_keypads[0][10]]);
    SetDlgItemText(hwnd, IDC_P1_0     , keyname[temp_keypads[0][11]]);
    SetDlgItemText(hwnd, IDC_P1_E     , keyname[temp_keypads[0][12]]);
    SetDlgItemText(hwnd, IDC_P1_UP    , keyname[temp_keypads[0][17]]);
    SetDlgItemText(hwnd, IDC_P1_DOWN  , keyname[temp_keypads[0][18]]);
    SetDlgItemText(hwnd, IDC_P1_LEFT  , keyname[temp_keypads[0][19]]);
    SetDlgItemText(hwnd, IDC_P1_RIGHT , keyname[temp_keypads[0][20]]);
    SetDlgItemText(hwnd, IDC_P1_1ST   , keyname[temp_keypads[0][ 0]]);
    SetDlgItemText(hwnd, IDC_P1_2ND   , keyname[temp_keypads[0][21]]);
    SetDlgItemText(hwnd, IDC_P1_3RD   , keyname[temp_keypads[0][22]]);
    SetDlgItemText(hwnd, IDC_P1_4TH   , keyname[temp_keypads[0][23]]);

    SetDlgItemText(hwnd, IDC_P2_2     , keyname[temp_keypads[1][ 2]]);
    SetDlgItemText(hwnd, IDC_P2_X1    , keyname[temp_keypads[1][13]]);
    SetDlgItemText(hwnd, IDC_P2_X2    , keyname[temp_keypads[1][14]]);
    SetDlgItemText(hwnd, IDC_P2_X3    , keyname[temp_keypads[1][15]]);
    SetDlgItemText(hwnd, IDC_P2_1     , keyname[temp_keypads[1][ 1]]);
    SetDlgItemText(hwnd, IDC_P2_X4    , keyname[temp_keypads[1][16]]);
    SetDlgItemText(hwnd, IDC_P2_3     , keyname[temp_keypads[1][ 3]]);
    SetDlgItemText(hwnd, IDC_P2_4     , keyname[temp_keypads[1][ 4]]);
    SetDlgItemText(hwnd, IDC_P2_5     , keyname[temp_keypads[1][ 5]]);
    SetDlgItemText(hwnd, IDC_P2_6     , keyname[temp_keypads[1][ 6]]);
    SetDlgItemText(hwnd, IDC_P2_7     , keyname[temp_keypads[1][ 7]]);
    SetDlgItemText(hwnd, IDC_P2_8     , keyname[temp_keypads[1][ 8]]);
    SetDlgItemText(hwnd, IDC_P2_9     , keyname[temp_keypads[1][ 9]]);
    SetDlgItemText(hwnd, IDC_P2_C     , keyname[temp_keypads[1][10]]);
    SetDlgItemText(hwnd, IDC_P2_0     , keyname[temp_keypads[1][11]]);
    SetDlgItemText(hwnd, IDC_P2_E     , keyname[temp_keypads[1][12]]);
    SetDlgItemText(hwnd, IDC_P2_UP    , keyname[temp_keypads[1][17]]);
    SetDlgItemText(hwnd, IDC_P2_DOWN  , keyname[temp_keypads[1][18]]);
    SetDlgItemText(hwnd, IDC_P2_LEFT  , keyname[temp_keypads[1][19]]);
    SetDlgItemText(hwnd, IDC_P2_RIGHT , keyname[temp_keypads[1][20]]);
    SetDlgItemText(hwnd, IDC_P2_1ST   , keyname[temp_keypads[1][ 0]]);
    SetDlgItemText(hwnd, IDC_P2_2ND   , keyname[temp_keypads[1][21]]);
    SetDlgItemText(hwnd, IDC_P2_3RD   , keyname[temp_keypads[1][22]]);
    SetDlgItemText(hwnd, IDC_P2_4TH   , keyname[temp_keypads[1][23]]);

    ok = TRUE;
    for (i = 0; i < 2; i++)
    {   for (j = 0; j < NUMKEYS; j++)
        {   if (temp_keypads[i][j] != default_keypads[i][j])
            {   ok = FALSE;
                break; // for speed
    }   }   }
    if (ok)
    {   for (i = 0; i < 4; i++)
        {   if (temp_console[i] != default_console[i])
            {   ok = FALSE;
                break; // for speed
    }   }   }

    if (ok) // all are defaults
    {   EnableWindow(GetDlgItem(hwnd, IDC_RESETTODEFAULTS), FALSE);
    } else // not all are defaults
    {   EnableWindow(GetDlgItem(hwnd, IDC_RESETTODEFAULTS), TRUE );
}   }

BOOL CALLBACK RedefineSubDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam)
{   ULONG code;

    switch(Message)
    {
    case WM_INITDIALOG:
        clearkybd();
        return TRUE;
    acase WM_KEYUP: // WM_KEYDOWN causes problems with cursor down and with cursor right
        code = (lParam & 33488896) >> 16;

        if
        (    code != SCAN_PAUSE
      // &&  code != SCAN_SCRLK
      // &&  code != SCAN_RETURN
         && (code <  SCAN_F5 || code > SCAN_F10)
         &&  code != SCAN_F12
        )
        {   if (side == 2)
            {   temp_console[whichkey] = (UWORD) code;
            } else
            {   temp_keypads[side][whichkey] = (UWORD) code;
            }
            clearkybd();
            EndDialog(hwnd, 0);
        }

        return FALSE;
    acase WM_CLOSE:
        clearkybd();
        EndDialog(hwnd, 0);
        return TRUE;
    adefault:
        return FALSE;
    break;
    }

    return TRUE;
}

BOOL CALLBACK SetDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam)
{   int number;

    switch(Message)
    {
    case WM_INITDIALOG:
        if (dlgtype == DLGTYPE_BP)
        {   if (bp == OUTOFRANGE)
            {   SetDlgItemText(hwnd, IDC_POINT, "");
            } else
            {   number_to_friendly(bp, friendly, FALSE);
                SetDlgItemText(hwnd, IDC_POINT, friendly);
        }   }
        else
        {   // assert(dlgtype == DLGTYPE_WP);
            if (wp == OUTOFRANGE)
            {   SetDlgItemText(hwnd, IDC_POINT, "");
            } else
            {   number_to_friendly(wp, friendly, FALSE);
                SetDlgItemText(hwnd, IDC_POINT, friendly);
        }   }
        // DISCARD SetForegroundWindow(hwnd); doesn't work

        return TRUE;
    break;
    case WM_CLOSE:
        goto DONE;
    break;
    case WM_COMMAND:
        if (LOWORD (wParam) == IDOK)
        {   goto DONE;
        }

        return FALSE;
    break;
    default:
        return FALSE;
    break;
    }

    return TRUE;

DONE:
    GetWindowText(GetDlgItem(hwnd, IDC_POINT), friendly, sizeof(friendly));
	number = friendly_to_number(friendly);

    if (number == OUTOFRANGE)
    {   T1; printf("No effect (value is out of range).\n\n"); T2;
    } elif (dlgtype == DLGTYPE_BP)
    {   if (bp != number) // just to prevent needless messages
        {   bp = number;
            getfriendly(bp);
            T1;
            printf("Set code breakpoint to %s!\n\n", friendly);
            T2;
    }   }
    else
    {   // assert(dlgtype == DLGTYPE_WP);
        if (wp != number) // just to prevent needless messages
        {   wp = number;
			getfriendly(wp);
            T1;
            printf("Set data watchpoint to %s!\n\n", friendly);
            T2;
    }   }
    clearkybd();
    EndDialog(hwnd, 0);

    return TRUE;
}

BOOL CALLBACK RastNDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam)
{   switch(Message)
    {
    case WM_INITDIALOG:
        stcl_d(globalstring, rastn); // number -> decimal string
        SetDlgItemText(hwnd, IDC_POINT, globalstring);
        // DISCARD SetForegroundWindow(hwnd); doesn't work

        return TRUE;
    break;
    case WM_CLOSE:
        goto DONE;
    break;
    case WM_COMMAND:
        if (LOWORD (wParam) == IDOK)
        {   goto DONE;
        }
        if (LOWORD (wParam) == IDCANCEL)
        {   clearkybd();
            EndDialog(hwnd, 0);

            return TRUE;
        }

        return FALSE;
    break;
    default:
        return FALSE;
    break;
    }

    return TRUE;

DONE:
    GetWindowText(GetDlgItem(hwnd, IDC_POINT), globalstring, sizeof(globalstring));
    stcd_l(globalstring, &rastn); // string -> decimal number
    // we should really bounds check this number

    clearkybd();
    EndDialog(hwnd, 0);

    return TRUE;
}

BOOL CALLBACK MonitorDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam)
{   switch(Message)
    {
        case WM_INITDIALOG:
            return TRUE;
        break;
        case WM_CLOSE:
            DestroyWindow(hwnd);
            MonitorWindowPtr = NULL;
            unghost(MENUITEM_MONITOR);
            return TRUE;
        break;
        default:
            return FALSE;
    }
    return TRUE;
}

BOOL CALLBACK ControlsDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam)
{   switch(Message)
    {
        case WM_INITDIALOG:
            // this is so that paddle axes which are never read by the
            // game are reported to be centred.
            sx[0] =
            sy[0] =
            sx[1] =
            sy[1] = 128;

            return TRUE;
        acase WM_LBUTTONDBLCLK:
        case WM_LBUTTONDOWN:
            mousedown((int) (LOWORD(lParam)), (int) (HIWORD(lParam)));
        acase WM_LBUTTONUP:
            mouseup();
        acase WM_CLOSE:
            DestroyWindow(hwnd);
            ControlsWindowPtr = NULL;
            unghost(MENUITEM_CONTROLS);
            return TRUE;
        adefault:
            return FALSE;
    }
    return TRUE;
}

EXPORT ULONG getsize(STRPTR filename)
{   HANDLE hFile /* = NULL */ ;
    ULONG  size;

    hFile = CreateFile(filename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
    if (hFile == INVALID_HANDLE_VALUE)
    {   return(0);
    }
    size = GetFileSize(hFile, NULL);
    CloseHandle(hFile);
    // hFile = NULL;

    if (size == (ULONG) -1)
    {   return(0);
    }

    return(size);
}

EXPORT void checkinput(void)
{   MSG   Msg;
    ULONG scancode;

    if (paused && !cli)
    {   DISCARD WaitMessage(); // to give better multitasking
    }

    while(PeekMessage(&Msg, NULL, 0, 0, PM_REMOVE))
    {   switch(Msg.message)
        {
        case WM_KEYDOWN:
            scancode = (Msg.lParam & 33488896) >> 16;
            handlekybd(scancode);
        acase WM_KEYUP:
            scancode = (Msg.lParam & 33488896) >> 16;
            KeyMatrix[scancode / 8] &= (255 - (1 << (scancode % 8)));
        acase WM_RBUTTONUP:
            ThePoint.x = LOWORD(Msg.lParam);
            ThePoint.y = HIWORD(Msg.lParam);
            ClientToScreen(MainWindowPtr, &ThePoint);
            TrackPopupMenu
            (   PopUpMenuPtr,
                TPM_RIGHTBUTTON,
                ThePoint.x,
                ThePoint.y,
                0,
                MainWindowPtr,
                NULL
            );
            // TranslateMessage(&Msg);
            // DispatchMessage(&Msg);
        adefault:
            TranslateMessage(&Msg);
            DispatchMessage(&Msg);
        break;
    }   }

    return; // return Msg.wParam;
}

EXPORT UBYTE KeyDown(UWORD scancode) // scancode must be at least 16 bits wide!
{   if (KeyMatrix[scancode / 8] & (1 << (scancode % 8)))
    {   return(1);
    }
    return(0);
}

EXPORT void debug_set(int kind)
{   sound_off(FALSE);
    if (kind == KIND_BP)
    {   dlgtype = DLGTYPE_BP;
        DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_SETBP), MainWindowPtr, SetDlgProc);
    } else
    {   // assert(dlgtype = DLGTYPE_WP);
        dlgtype = DLGTYPE_WP;
        DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_SETWP), MainWindowPtr, SetDlgProc);
    }
    updatemenus();
    sound_on();
}
EXPORT void debug_setrastn(void)
{   sound_off(FALSE);
    DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_RUNTORASTN), MainWindowPtr, RastNDlgProc);
    sound_on();
}

EXPORT void redrawscreen(void)
{   int x, y;

    for (x = 0; x < BOXWIDTH; x++)
    {   for (y = 0; y < BOXHEIGHT; y++)
        {   display[x + (y * BOXWIDTH)] = pens[screen[x][y]];
    }   }

    updatescreen();
}

EXPORT void updatescreen(void)
{   PERSIST HDC OnScreenRastPort;

    if (!MainWindowPtr)
    {   return; // important!
    }

#ifdef USEGDI
    // caching the DC would be faster but we are not supposed to do that
    OnScreenRastPort = GetDC(MainWindowPtr);

    if (stretch && fullscreen)
    {   DISCARD StretchDIBits
        (   OnScreenRastPort,
            0,              // dest leftx
            yoffset,        // dest topy
            winwidth,       // dest width
            clientheight,   // dest height
            0,              // source leftx
            0,              // source topy
            BOXWIDTH,       // source width
            BOXHEIGHT,      // source height
            display,        // pointer to the bits
            (const struct tagBITMAPINFO *) &TheBitMapInfo, // pointer to BITMAPINFO structure
            DIB_RGB_COLORS, // format of data
            SRCCOPY         // blit mode
        );
    } else
    {   DISCARD StretchDIBits
        (   OnScreenRastPort,
            xoffset,        // dest leftx
            yoffset,        // dest topy
            rastwidth,      // dest width
            rastheight,     // dest height
            0,              // source leftx
            0,              // source topy
            BOXWIDTH,       // source width
            BOXHEIGHT,      // source height
            display,        // pointer to the bits
            (const struct tagBITMAPINFO *) &TheBitMapInfo, // pointer to BITMAPINFO structure
            DIB_RGB_COLORS, // format of data
            SRCCOPY         // blit mode
        );
    }

    DISCARD ReleaseDC(MainWindowPtr, OnScreenRastPort);
#endif
#ifdef USEDIRECTX
    if
    (   !DDSurface
     || (   IDirectDrawSurface_IsLost(DDSurface)  == DDERR_SURFACELOST
         && IDirectDrawSurface_Restore(DDSurface) != DD_OK
        )
     || IDirectDrawSurface_GetDC(DDSurface, &OnScreenRastPort) != DD_OK
    )
    {   // should at least call IDirectDrawSurface_ReleaseDC() if IDirectDrawSurface_GetDC() succeeded
        MessageBox
        (   NULL,
            "DirectX output failed!",
            "WinArcadia: Error",
            MB_ICONEXCLAMATION | MB_OK
        );
        cleanexit(EXIT_FAILURE);
    }

    // it would be better to call ClientToScreen() only when the window is opened/moved
    ThePoint.x =
    ThePoint.y = 0;
    ClientToScreen(MainWindowPtr, &ThePoint);

    if (stretch && fullscreen)
    {   DISCARD StretchBlt
        (   OnScreenRastPort,            // handle of dest device context
            ThePoint.x,                  // leftx  of dest rect
            ThePoint.y + yoffset,        // topy   of dest rect
            winwidth,                    // width  of dest rect
            clientheight,                // height of dest rect
            OurhDC,                      // handle of src  device context
            0,                           // leftx  of src  rect
            0,                           // topy   of src  rect
            BOXWIDTH,                    // width  of src  rect
            BOXHEIGHT,                   // height of src  rect
            SRCCOPY                      // raster operation code
        );
    } else
    {   DISCARD StretchBlt
        (   OnScreenRastPort,            // handle of dest device context
            ThePoint.x + xoffset,        // leftx  of dest rect
            ThePoint.y + yoffset,        // topy   of dest rect
            rastwidth,                   // width  of dest rect
            rastheight,                  // height of dest rect
            OurhDC,                      // handle of src  device context
            0,                           // leftx  of src  rect
            0,                           // topy   of src  rect
            BOXWIDTH,                    // width  of src  rect
            BOXHEIGHT,                   // height of src  rect
            SRCCOPY                      // raster operation code
        );
    }

    IDirectDrawSurface_ReleaseDC(DDSurface, OnScreenRastPort);
#endif

}

MODULE __inline void thewait(void)
{   PERSIST ULONG howlong = 0; // doesn't matter much what the starting value is, garbage would be fine too

    /* 1/3rd of the time we wait for 16 milliseconds.
       2/3rd of the time we wait for 17 milliseconds */

    newtime = timeGetTime();
    if (warp)
    {   if (limitrefresh && newtime >= waittill)
        {   updatescreen();
            waittill = newtime;

            if (region == REGION_NTSC)
            {   if ((howlong % 3) == 0)
                {   waittill += 16;
                } else
                {   waittill += 17;
                }
                howlong++;
            } else
            {   // assert(region == REGION_PAL);
                waittill += 20;
    }   }   }
    else
    {   // As soon as we have finished waiting, we calculate how much we
        // should wait next time. This way the time consumed in doing the
        // emulation is taken into account.

        // The way it is implemented, it will return immediately on the
        // first frame, but all subsequent frames are timed correctly.

        if (newtime < waittill)
        {   do
            {   newtime = timeGetTime();
            } while (newtime < waittill);
        } else
        {   waittill = newtime;
        }

        if (ctrl() && KeyDown(SCAN_APOSTROPHE))
        {   waittill += 1000;
        } elif (region == REGION_NTSC)
        {   if ((howlong % 3) == 0)
            {   if (ctrl() && KeyDown(SCAN_SEMICOLON))
                {   waittill += 66;
                } else
                {   waittill += 16;
            }   }
            else
            {   if (ctrl() && KeyDown(SCAN_SEMICOLON))
                {   waittill += 67;
                } else
                {   waittill += 17;
            }   }
            howlong++;
        } else
        {   // assert(region == REGION_PAL);
            if (ctrl() && KeyDown(SCAN_SEMICOLON))
            {   waittill += 80;
            } else
            {   waittill += 20;
}   }   }   }

EXPORT void freeall(void)
{   FILE*      LocalHandle /* = NULL */ ;
    int        i, j,
               localoffset;

    macro_stop();
    sound_off(FALSE);

    for (i = 0; i < 14; i++)
    {   SoundClose(i);
    }

    NetPlay(NET_OFF);

    if (ProgDir[0])
    {   DISCARD SetCurrentDirectory(ProgDir);
    }
    if (autosave && !crippled && game)
    {   changefilepart(autoname[machine]);
        // filepart will be wrong but we don't use it
        if (machine == ARCADIA)
        {   engine_save(KIND_EAS, FALSE);
        } elif (machine == INTERTON)
        {   engine_save(KIND_INS, FALSE);
        } else
        {   // assert(machine == ELEKTOR);
            engine_save(KIND_TVS, FALSE);
    }   }

#ifdef USEGDI
    if (display)
    {   free(display);
    }
#endif
#ifdef USEDIRECTX
    if (DD)
    {   if (DDSurface)
        {   DDSurface->lpVtbl->Release(DDSurface);
            DDSurface = NULL;
	    }
        DD->lpVtbl->Release(DD);
        DD = NULL;
	}
    if (OurhBitMap)
    {   DeleteObject(OurhBitMap);
        OurhBitMap = NULL;
    }
    if (OurhDC)
    {   DeleteDC(OurhDC);
        OurhDC = NULL;
    }
#endif

    if (MenuBitmapPtr[3])
    {   DeleteObject(MenuBitmapPtr[3]);
        MenuBitmapPtr[3] = NULL;
    }
    if (MenuBitmapPtr[2])
    {   DeleteObject(MenuBitmapPtr[2]);
        MenuBitmapPtr[2] = NULL;
    }
    if (MenuBitmapPtr[1])
    {   DeleteObject(MenuBitmapPtr[1]);
        MenuBitmapPtr[1] = NULL;
    }
    if (MenuBitmapPtr[0])
    {   DeleteObject(MenuBitmapPtr[0]);
        MenuBitmapPtr[0] = NULL;
    }

    if ((LocalHandle = fopen(fn_config, "wb")))
    {   ConfigBuffer[0]  = (UBYTE) size;
        ConfigBuffer[1]  = (UBYTE) wide;
        ConfigBuffer[2]  = (UBYTE) requirebutton;
        ConfigBuffer[3]  = (UBYTE) analog;
        ConfigBuffer[4]  = (UBYTE) autofire;
        ConfigBuffer[5]  = (UBYTE) swapped;
        ConfigBuffer[6]  = (UBYTE) paused;
        ConfigBuffer[7]  = (UBYTE) warp;
        ConfigBuffer[8]  = (UBYTE) collisions;
        ConfigBuffer[9]  = (UBYTE) sound;
        ConfigBuffer[10] = (UBYTE) demultiplex;
        ConfigBuffer[11] = (UBYTE) autosave;
        ConfigBuffer[12] = (UBYTE) fullscreen;
        ConfigBuffer[13] = (UBYTE) showmenubar;
        ConfigBuffer[14] = (UBYTE) showtitlebar;
        ConfigBuffer[15] = (UBYTE) showtoolbar;
        ConfigBuffer[16] = (UBYTE) flagline;
        ConfigBuffer[17] = (UBYTE) showpointer;
        ConfigBuffer[18] = (UBYTE) loop;
        ConfigBuffer[19] = (UBYTE) logreads;
        ConfigBuffer[20] = (UBYTE) logwrites;
        ConfigBuffer[21] = (UBYTE) 16; /* means V5.62+ */
        ConfigBuffer[22] = (UBYTE) limitrefresh;
        ConfigBuffer[23] = (UBYTE) machine;
        ConfigBuffer[24] = (UBYTE) loginstructions;
        ConfigBuffer[25] = (UBYTE) watchreads;
        ConfigBuffer[26] = (UBYTE) sensegame;
        ConfigBuffer[27] = (UBYTE) usehacks;

        localoffset = 28;
        for (i = 0; i < 2; i++)
        {   for (j = 0; j <= 23; j++)
            {   ConfigBuffer[localoffset    ] = keypads[i][j] / 256;
                ConfigBuffer[localoffset + 1] = keypads[i][j] % 256;
                localoffset += 2;
        }   }
        // assert(localoffset == 124);
        for (i = 0; i < 4; i++)
        {   ConfigBuffer[localoffset    ] = console[i] / 256;
            ConfigBuffer[localoffset + 1] = console[i] % 256;
            localoffset += 2;
        }

        // assert(localoffset == 132);
        ConfigBuffer[132] = (UBYTE) joy0;
        ConfigBuffer[133] = (UBYTE) joy1;
        ConfigBuffer[134] = (UBYTE) logmonitor;

        DISCARD fwrite(ConfigBuffer, CONFIGLENGTH,         1, LocalHandle);
        DISCARD fwrite(pathpart,     strlen(pathpart) + 1, 1, LocalHandle);
        DISCARD fclose(LocalHandle);
        // LocalHandle = NULL;
}   }

EXPORT void cleanexit(SBYTE rc)
{   if (consoleopen)
    {   printf("Exiting.\n");
    }

    freeall();

    // PostMessage(MainWindowPtr, WM_CLOSE, 0, 0);
    DestroyWindow(MainWindowPtr); // this will also destroy all child windows
    // MainWindowPtr = NULL;

    exit(rc);
}

EXPORT UWORD ReadJoystick(UWORD joynum)
{   JOYINFOEX joyInfoEx;
    MMRESULT  joyInfoExResult;
    UBYTE     rc = 0;
    UINT      id;

    // It would be better not to require separate reads of the buttons and paddles.

    // joynum will be '1' if they want the PRIMARY port
    // joynum will be '0' if they want the SECONDARY port
    if (joynum == 1)
    {   if (!joy0)
        {   return 0;
        }
        id = JOYSTICKID2; // changed for V5.3
    } else
    {   // assert(joynum == 0);
        if (!joy1)
        {   return 0;
        }
        id = JOYSTICKID1; // changed for V5.3
    }

    ZeroMemory(&joyInfoEx, sizeof(joyInfoEx));
    joyInfoEx.dwSize  = sizeof(joyInfoEx);
    joyInfoEx.dwFlags = JOY_RETURNALL;
    joyInfoExResult   = joyGetPosEx(joystickIDMapping[id], &joyInfoEx);

    if (joyInfoExResult == JOYERR_NOERROR)
    {   if (joyInfoEx.dwButtons & 1)
        {   rc |= JOYFIRE1;
        }
        if (joyInfoEx.dwButtons & 2)
        {   rc |= JOYFIRE2;
        }
        if (joyInfoEx.dwButtons & 4)
        {   rc |= JOYFIRE3;
        }
        if (joyInfoEx.dwButtons & 8)
        {   rc |= JOYFIRE4;
        }
        return rc;
    } else
    {   
#ifdef VERBOSE
        else { T1; printf("Failed to read buttons of joystick %ld!\n", 1 - joynum); T2; }
#endif
        return 0;
}   }

EXPORT void sound_off(FLAG force)
{   int i;

    if (!force && !sound)
    {   return; // important!
    }

    for (i = 0; i < 14; i++)
    {   SoundStop(i);
}   }

EXPORT void sound_on(void)
{   if (!sound)
    {   return; // important!
    }

    /* if (!waveOutGetNumDevs()) // if no sound output devices found
    {   MessageBox(MainWindowPtr, "waveOutGetNumDevs() failed!", "WinArcadia: Error", MB_OK | MB_ICONERROR);
    } */

    // we used to call OpenChannels() here

    if (!paused)
    {   playsound(TRUE);
}   }

EXPORT void calcsize(FLAG autosize)
{   int realsizex,
        realsizey;

    if (fullscreen || autosize)
    {   winwidth                 = DisplayWidth;
        winheight = clientheight = DisplayHeight;

        if (showtitlebar)
        {   clientheight -= titleheight;
        }
        if (showmenubar)
        {   clientheight -= menuheight;
        }
        if (showtoolbar)
        {   clientheight -= toolbarheight;
        }

        realwide = 2;
        realsizex = winwidth     / (BOXWIDTH * realwide);
        realsizey = clientheight / BOXHEIGHT;
        if (realsizey < realsizex)
        {   realsize = realsizey;
        } else
        {   realsize = realsizex;
        }
        rastwidth  = BOXWIDTH  * realsize * realwide;
        rastheight = BOXHEIGHT * realsize;
        xoffset = (winwidth     - rastwidth ) / 2;
    } else
    {   winwidth  = rastwidth  = (SWORD) BOXWIDTH  * size * wide;
        winheight = rastheight = (SWORD) BOXHEIGHT * size;

        if (showtitlebar)
        {   winheight += titleheight;                   // top border
            winheight += GetSystemMetrics(SM_CYBORDER); // bottom border (normally 1 pixel)
            xoffset   =  GetSystemMetrics(SM_CXBORDER);
            winwidth  += xoffset * 2;                   // left and right borders (normally 2 pixels each)
        } else
        {   xoffset   =  0;
        }
        if (showmenubar)
        {   winheight += menuheight;
            if (size == 1 && wide == 1)
            {   winheight += menuheight; // because menu will be two rows in height
        }   }
        if (showtoolbar)
        {   winheight += toolbarheight;
        }

        realwide = wide;
        realsize = size;
    }

    if (showtoolbar)
    {   yoffset = toolbarheight;
    } else
    {   yoffset = 0;
    }
    if (fullscreen && !stretch)
    {   yoffset += (clientheight - rastheight) / 2;
    }

    winleftx = (DisplayWidth  / 2) - (winwidth  / 2);
    wintopy  = (DisplayHeight / 2) - (winheight / 2);
    if (winleftx < 0)
    {   winleftx = 0;
    }
    if (wintopy < 0)
    {   wintopy = 0;
}   }

MODULE void resize(int newsize)
{   size = realsize = newsize;
    updatemenu(MENUITEM_SIZE);
    calcsize(FALSE);
    SetWindowPos
    (   MainWindowPtr,
        HWND_TOP,
        winleftx,
        wintopy,
        winwidth,
        winheight,
        SWP_NOCOPYBITS
    );
    maketoolbar(MainWindowPtr);
    redrawscreen();
}

EXPORT void beep(void)      { ; }
EXPORT void readkybd(void)  { ; }

MODULE void maketoolbar(HWND hwnd)
{   TBBUTTON tbb[SMALLGADGETS_WIDE];
    UINT     gadgets, images;

    // We can't use MainWindowPtr, as it is not set up yet!
    // (This function gets called when the window is 'half-open'.)

    if (hToolbar)
    {   DestroyWindow(hToolbar);
        hToolbar = NULL;
    }
    if (LabelPtr)
    {   DestroyWindow(LabelPtr);
        LabelPtr = NULL;
    }
    if (FPSPtr)
    {   DestroyWindow(FPSPtr);
        FPSPtr = NULL;
    }

    if (!showtoolbar)
    {   return;
    }

    ZeroMemory(tbb, sizeof(tbb));

    tbb[0].iBitmap = 0;
    tbb[0].fsState = 0;
    tbb[0].fsStyle = TBSTYLE_BUTTON;
    tbb[0].idCommand = ID_FILE_RESET;
    tbb[1].iBitmap = 1;
    tbb[1].fsState = TBSTATE_ENABLED;
    tbb[1].fsStyle = TBSTYLE_BUTTON;
    tbb[1].idCommand = ID_FILE_OPEN;
    tbb[2].iBitmap = 2;
    tbb[2].fsState = 0;
    tbb[2].fsStyle = TBSTYLE_BUTTON;
    tbb[2].idCommand = ID_FILE_SAVESNP;
    tbb[3].iBitmap = 3;
    tbb[3].fsState = TBSTATE_ENABLED;
    tbb[3].fsStyle = TBSTYLE_BUTTON;
    tbb[3].idCommand = ID_FILE_SAVEILBM;

    tbb[4].iBitmap = 0;
    tbb[4].fsState = 0;
    tbb[4].fsStyle = TBSTYLE_SEP;
    tbb[4].idCommand = 0;

    // Note that this assumes that there is enough width in fullscreen
    // mode to show the complete toolbar.
    if (fullscreen || (realwide >= 2 && realsize >= 2) || realsize >= 3)
    {   gadgets = SMALLGADGETS_WIDE;
        images  = SMALLIMAGES_WIDE;

        tbb[5].iBitmap = 4;
        tbb[5].fsState = TBSTATE_ENABLED;
        tbb[5].fsStyle = TBSTYLE_CHECK;
        tbb[5].idCommand = ID_OPTIONS_CONTROLLERS_ANALOG;
        tbb[6].iBitmap = 5;
        tbb[6].fsState = TBSTATE_ENABLED;
        tbb[6].fsStyle = TBSTYLE_CHECK;
        tbb[6].idCommand = ID_OPTIONS_CONTROLLERS_AUTOFIRE;
        tbb[7].iBitmap = 6;
        tbb[7].fsState = TBSTATE_ENABLED;
        tbb[7].fsStyle = TBSTYLE_CHECK;
        tbb[7].idCommand = ID_OPTIONS_CONTROLLERS_SWAPPED;

        tbb[8].iBitmap = 0;
        tbb[8].fsState = 0;
        tbb[8].fsStyle = TBSTYLE_SEP;
        tbb[8].idCommand = 0;

        tbb[9].iBitmap = 7;
        tbb[9].fsState = TBSTATE_ENABLED;
        tbb[9].fsStyle = TBSTYLE_CHECK;
        tbb[9].idCommand = ID_OPTIONS_SPEED_PAUSED;
        tbb[10].iBitmap = 8;
        tbb[10].fsState = TBSTATE_ENABLED;
        tbb[10].fsStyle = TBSTYLE_CHECK;
        tbb[10].idCommand = ID_OPTIONS_SPEED_WARP;

        tbb[11].iBitmap = 0;
        tbb[11].fsState = 0;
        tbb[11].fsStyle = TBSTYLE_SEP;
        tbb[11].idCommand = 0;

        tbb[12].iBitmap = 9;
        tbb[12].fsState = TBSTATE_ENABLED;
        tbb[12].fsStyle = TBSTYLE_CHECK;
        tbb[12].idCommand = ID_OPTIONS_SPRITES_COLLISIONS;
        tbb[13].iBitmap = 10;
        if (machine == ARCADIA)
        {   tbb[13].fsState = TBSTATE_ENABLED;
        } else
        {   // assert(machine == INTERTON || machine == ELEKTOR);
            tbb[13].fsState = 0;
        }
        tbb[13].fsStyle = TBSTYLE_CHECK;
        tbb[13].idCommand = ID_OPTIONS_SPRITES_DEMULTIPLEX;

        tbb[14].iBitmap = 0;
        tbb[14].fsState = 0;
        tbb[14].fsStyle = TBSTYLE_SEP;
        tbb[14].idCommand = 0;

        tbb[15].iBitmap = 11;
        tbb[15].fsStyle = TBSTYLE_BUTTON;
        tbb[15].idCommand = ID_OPTIONS_GRAPHICS_FLAGLINE;
        tbb[15].fsState = TBSTATE_ENABLED;

        tbb[16].iBitmap = 0;
        tbb[16].fsState = 0;
        tbb[16].fsStyle = TBSTYLE_SEP;
        tbb[16].idCommand = 0;

        tbb[17].iBitmap = 12;
        tbb[17].fsState = TBSTATE_ENABLED;
        tbb[17].fsStyle = TBSTYLE_CHECK;
        tbb[17].idCommand = ID_OPTIONS_SOUND;

        tbb[18].iBitmap = 0;
        tbb[18].fsState = 0;
        tbb[18].fsStyle = TBSTYLE_SEP;
        tbb[18].idCommand = 0;

        tbb[19].iBitmap = 13;
        tbb[19].fsState = 0;
        tbb[19].fsStyle = TBSTYLE_CHECK;
        tbb[19].idCommand = ID_MACRO_TOGGLEPLAYING;
        tbb[20].iBitmap = 14;
        tbb[20].fsState = 0;
        tbb[20].fsStyle = TBSTYLE_CHECK;
        tbb[20].idCommand = ID_MACRO_TOGGLERECORDING;
    } elif (realsize >= 2 || realwide >= 2)
    {   gadgets = SMALLGADGETS_NARROW;
        images  = SMALLIMAGES_NARROW;

        tbb[5].iBitmap = 5;
        tbb[5].fsState = TBSTATE_ENABLED;
        tbb[5].fsStyle = TBSTYLE_CHECK;
        tbb[5].idCommand = ID_OPTIONS_CONTROLLERS_AUTOFIRE;
        tbb[6].iBitmap = 7;
        tbb[6].fsState = TBSTATE_ENABLED;
        tbb[6].fsStyle = TBSTYLE_CHECK;
        tbb[6].idCommand = ID_OPTIONS_SPEED_PAUSED;
        tbb[7].iBitmap = 8;
        tbb[7].fsState = TBSTATE_ENABLED;
        tbb[7].fsStyle = TBSTYLE_CHECK;
        tbb[7].idCommand = ID_OPTIONS_SPEED_WARP;
        tbb[8].iBitmap = 9;
        tbb[8].fsState = TBSTATE_ENABLED;
        tbb[8].fsStyle = TBSTYLE_CHECK;
        tbb[8].idCommand = ID_OPTIONS_SPRITES_COLLISIONS;
        tbb[9].iBitmap = 12;
        tbb[9].fsState = TBSTATE_ENABLED;
        tbb[9].fsStyle = TBSTYLE_CHECK;
        tbb[9].idCommand = ID_OPTIONS_SOUND;

        tbb[10].iBitmap = 0;
        tbb[10].fsState = 0;
        tbb[10].fsStyle = TBSTYLE_SEP;
        tbb[10].idCommand = 0;

        tbb[11].iBitmap = 13;
        tbb[11].fsState = 0;
        tbb[11].fsStyle = TBSTYLE_CHECK;
        tbb[11].idCommand = ID_MACRO_TOGGLEPLAYING;
        tbb[12].iBitmap = 14;
        tbb[12].fsState = 0;
        tbb[12].fsStyle = TBSTYLE_CHECK;
        tbb[12].idCommand = ID_MACRO_TOGGLERECORDING;
    } else
    {   gadgets = 4;
        images = 4;
    }

/*  wide           narrow        very narrow
    (4x2,3x3,6x3)  (2x1, 2x2)    (1x1)
 0: reset          reset         reset
 1: open           open          open
 2: save snp       save snp      save snp
 3: save iff       save iff      save iff
 4: ---            ---
 5: analog         autofire      (we omit analog, autofire, swapped,
 6: autofire       paused        paused, warp, collisions, demultiplex,
 7: swapped        warp          flagline, sound, play, record)
 8: ---            collisions
 9: paused         sound
10: warp           ---
11: ---            play
12: collisions     record
13: demultiplex
14: ---            (we omit analog, swapped, demultiplex, flagline)
15: flagline
16: ---
17: sound
18: ---
19: play
20: record

1x1 (size = 1, narrow): very narrow toolbar, no FPS label
2x1 (size = 1, wide  ): narrow      toolbar, no FPS label
2x2 (size = 2, narrow): narrow      toolbar, no FPS label
4x2 (size = 2, wide  ): wide        toolbar,    FPS label
3x3 (size = 3, narrow): wide        toolbar, no FPS label
6x3 (size = 3, wide  ): wide        toolbar,    FPS label

Thus: for width of 1, very narrow toolbar and no FPS label
      for width of 2, narrow      toolbar and no FPS label
      for width of 3, wide        toolbar and no FPS label
      for width of 4, wide        toolbar and    FPS label
      for width of 6, wide        toolbar and    FPS label */

    // assert(hwnd);
    hToolbar = CreateToolbarEx
    (   hwnd,
        WS_CHILD | WS_VISIBLE | TBSTYLE_TOOLTIPS,
        IDC_TOOLBAR,
        images + 5,
        GetModuleHandle(NULL),
        (UINT) MAKEINTRESOURCE(IDB_TOOLBAR),
        tbb,
        gadgets,
        18,
        18,
        18,
        18,
        sizeof(TBBUTTON)
    );
    if (hToolbar == NULL)
    {   MessageBox(MainWindowPtr, "Can't create toolbar!", "WinArcadia: Error", MB_OK | MB_ICONERROR);
        cleanexit(EXIT_FAILURE);
    }

    // Send the TB_BUTTONSTRUCTSIZE message, which is required
    SendMessage(hToolbar, TB_BUTTONSTRUCTSIZE, (WPARAM) sizeof(TBBUTTON), 0);
    SendMessage(hToolbar, TB_AUTOSIZE, 0, 0);
    updatebiggads();
    updatesmlgads();

    if (realsize >= 2 && realwide >= 2)
    {   if (!(LabelPtr = CreateWindowEx
        (   0, "STATIC",

#ifdef MODE_FPS
            "FPS:",
#endif
#ifdef MODE_MHZ
            "MHz:",
#endif
#ifdef MODE_PERCENT
            "%:",
#endif

            WS_CHILD | WS_VISIBLE | SS_RIGHT,

#ifdef MODE_MHZ
            winwidth - 150, 8, 36, 18,
#else
            winwidth - 102, 8, 36, 18,
#endif

            hwnd, (HMENU) IDC_STATIC, GetModuleHandle(NULL), NULL
        )))
        {   MessageBox(MainWindowPtr, "Can't create static gadget!", "WinArcadia: Error", MB_OK | MB_ICONERROR);
            cleanexit(EXIT_FAILURE);
    }   }

    oldfps = ~0;

#ifdef MODE_MHZ
    if (realsize >= 2 || realwide >= 2)
    {   // WS_EX_CLIENTEDGE doesn't seem to work reliably
        if (!(FPSPtr = CreateWindowEx
        (   0, "STATIC", "-",
            WS_CHILD | WS_VISIBLE | SS_CENTER,
            winwidth - 110, 8, 100, 18,
            hwnd, (HMENU) IDC_FPS, GetModuleHandle(NULL), NULL
        )))
        {   MessageBox(MainWindowPtr, "Can't create static gadget!", "WinArcadia: Error", MB_OK | MB_ICONERROR);
            cleanexit(EXIT_FAILURE);
    }   }
#else // MODE_FPS and MODE_PERCENT
    if (!(FPSPtr = CreateWindowEx
    (   0, "STATIC", "-",
        WS_CHILD | WS_VISIBLE | SS_CENTER,
        winwidth - 62, 8, 52, 18,
        hwnd, (HMENU) IDC_FPS, GetModuleHandle(NULL), NULL
    )))
    {   MessageBox(MainWindowPtr, "Can't create static gadget!", "WinArcadia: Error", MB_OK | MB_ICONERROR);
        cleanexit(EXIT_FAILURE);
    }
#endif

    GetWindowRect(hToolbar, &therect);
    toolbarheight = therect.bottom - therect.top;

    /* If we ask for a large window on a small screen, the window
       is truncated to fit. This will have, for example, the
       effect of the FPS display not being visible. */
}

EXPORT void settitle(void)
{   STRPTR filepart;

    if (machine == ARCADIA)
    {   strcpy(titlebartext, A_TITLEBARTEXT);
    } elif (machine == INTERTON)
    {   strcpy(titlebartext, I_TITLEBARTEXT);
    } else
    {   // assert(machine == ELEKTOR);
        strcpy(titlebartext, E_TITLEBARTEXT);
    }
    if (game)
    {   if (machine == INTERTON)
        {   DISCARD GetFullPathName(fn_game, MAX_PATH, &titlebartext[TITLEBARTEXTLENGTH    ], &filepart);
        } else
        {   // assert(machine == ARCADIA || machine == ELEKTOR);
            DISCARD GetFullPathName(fn_game, MAX_PATH, &titlebartext[TITLEBARTEXTLENGTH - 1], &filepart);
        }
        if (filepart)
        {   strcat(titlebartext, ": ");
            strcat(titlebartext, filepart);
    }   }
    if (connected == NET_SERVER)
    {   strcat(titlebartext, " (server)");
    } elif (connected == NET_CLIENT)
    {   strcat(titlebartext, " (client)");
    }
    if (paused)
    {   strcat(titlebartext, " (paused)");
    }
    SetWindowText(MainWindowPtr, titlebartext);

    // printf("Setting title to %s!\n", titlebartext);
}

MODULE FLAG ctrl(void)
{   if
    (   (KeyMatrix[SCAN_LCTRL  / 8] & (1 << (SCAN_LCTRL  % 8)))
     || (KeyMatrix[SCAN_RCTRL  / 8] & (1 << (SCAN_RCTRL  % 8)))
    )
    {   return TRUE;
    } else
    {   return FALSE;
}   }
MODULE FLAG shift(void)
{   if
    (   (KeyMatrix[SCAN_LSHIFT / 8] & (1 << (SCAN_LSHIFT % 8)))
     || (KeyMatrix[SCAN_RSHIFT / 8] & (1 << (SCAN_RSHIFT % 8)))
    )
    {   return TRUE;
    } else
    {   return FALSE;
}   }

EXPORT void openwindow(void)
{   DWORD   style;

    calcsize(FALSE);

    if (fullscreen)
    {   style = WS_POPUP | WS_MINIMIZEBOX /* | WS_MAXIMIZEBOX */ ;
        // still seems to have a border anyway! :-(
    } else
    {   style = WS_POPUP | WS_MINIMIZEBOX /* | WS_MAXIMIZEBOX */ | WS_BORDER;
    }
    if (showmenubar)
    {   MenuPtr = LoadMenu(InstancePtr, MAKEINTRESOURCE(IDR_MAINMENU));
    } else
    {   MenuPtr = NULL;
    }
    if (showtitlebar)
    {   if (machine == ARCADIA)
        {   MainWindowPtr = CreateWindowEx(
                0,
                g_szClassName,
                A_TITLEBARTEXT,
                style | WS_CAPTION | WS_SYSMENU,
                winleftx, wintopy,
                winwidth, winheight,
                NULL,
                MenuPtr,
                InstancePtr, NULL);
        } elif (machine == INTERTON)
        {   MainWindowPtr = CreateWindowEx(
                0,
                g_szClassName,
                I_TITLEBARTEXT,
                style | WS_CAPTION | WS_SYSMENU,
                winleftx, wintopy,
                winwidth, winheight,
                NULL,
                MenuPtr,
                InstancePtr, NULL);
        } else
        {   // assert(machine == ELEKTOR);
            MainWindowPtr = CreateWindowEx(
                0,
                g_szClassName,
                E_TITLEBARTEXT,
                style | WS_CAPTION | WS_SYSMENU,
                winleftx, wintopy,
                winwidth, winheight,
                NULL,
                MenuPtr,
                InstancePtr, NULL);
    }   }
    else
    {   MainWindowPtr = CreateWindowEx(
            0,
            g_szClassName,
            NULL,
            style,
            winleftx, wintopy,
            winwidth, winheight,
            NULL,
            MenuPtr,
            InstancePtr, NULL);
    }

    if (!MainWindowPtr)
    {   MessageBox(NULL, "Window creation failed!", "WinArcadia: Error", MB_ICONEXCLAMATION | MB_OK);
        cleanexit(EXIT_FAILURE);
    }

    PopUpMenuPtr = LoadMenu(InstancePtr, MAKEINTRESOURCE(IDR_POPUPMENU));
    PopUpMenuPtr = GetSubMenu(PopUpMenuPtr, 0);

    if (MenuPtr)
    {   // assert(showmenubar);
        DISCARD SetMenuItemBitmaps
        (   MenuPtr,          // handle of menu 
            ID_FILE_RESET,    // menu item to receive new bitmaps 
            MF_BYCOMMAND,     // menu item flags 
            MenuBitmapPtr[0], // handle of unchecked bitmap 
            NULL              // handle of checked bitmap 
        );
        DISCARD SetMenuItemBitmaps
        (   MenuPtr,          // handle of menu 
            ID_FILE_OPEN,     // menu item to receive new bitmaps 
            MF_BYCOMMAND,     // menu item flags 
            MenuBitmapPtr[1], // handle of unchecked bitmap 
            NULL              // handle of checked bitmap 
        );
        DISCARD SetMenuItemBitmaps
        (   MenuPtr,          // handle of menu 
            ID_FILE_SAVESNP,  // menu item to receive new bitmaps 
            MF_BYCOMMAND,     // menu item flags 
            MenuBitmapPtr[2], // handle of unchecked bitmap 
            NULL              // handle of checked bitmap 
        );
        DISCARD SetMenuItemBitmaps
        (   MenuPtr,          // handle of menu 
            ID_FILE_SAVEILBM, // menu item to receive new bitmaps 
            MF_BYCOMMAND,     // menu item flags 
            MenuBitmapPtr[3], // handle of unchecked bitmap 
            NULL              // handle of checked bitmap 
        );
    }
    if (PopUpMenuPtr)
    {   DISCARD SetMenuItemBitmaps
        (   PopUpMenuPtr,     // handle of menu 
            ID_FILE_RESET,    // menu item to receive new bitmaps 
            MF_BYCOMMAND,     // menu item flags 
            MenuBitmapPtr[0], // handle of unchecked bitmap 
            NULL              // handle of checked bitmap 
        );
        DISCARD SetMenuItemBitmaps
        (   PopUpMenuPtr,     // handle of menu 
            ID_FILE_OPEN,     // menu item to receive new bitmaps 
            MF_BYCOMMAND,     // menu item flags 
            MenuBitmapPtr[1], // handle of unchecked bitmap 
            NULL              // handle of checked bitmap 
        );
        DISCARD SetMenuItemBitmaps
        (   PopUpMenuPtr,     // handle of menu 
            ID_FILE_SAVESNP,  // menu item to receive new bitmaps 
            MF_BYCOMMAND,     // menu item flags 
            MenuBitmapPtr[2], // handle of unchecked bitmap 
            NULL              // handle of checked bitmap 
        );
        DISCARD SetMenuItemBitmaps
        (   PopUpMenuPtr,     // handle of menu 
            ID_FILE_SAVEILBM, // menu item to receive new bitmaps 
            MF_BYCOMMAND,     // menu item flags 
            MenuBitmapPtr[3], // handle of unchecked bitmap 
            NULL              // handle of checked bitmap 
        );
    }

    DISCARD joySetCapture(MainWindowPtr, JOYSTICKID1, 0, FALSE);
    DISCARD joySetCapture(MainWindowPtr, JOYSTICKID2, 0, FALSE);
    // are these still necessary and relevant?
    // captures are automatically released when window is destroyed.

#ifdef USEDIRECTX
    // this must be done after CreateWindowEx() but before UpdateWindow()

    if
    (   DirectDrawCreate(NULL, &DD, NULL) != DD_OK
     || IDirectDraw_SetCooperativeLevel(DD, MainWindowPtr, DDSCL_NORMAL) != DD_OK
	)
	{   // there are no cleanup functions corresponding to the above two functions
        MessageBox
        (   NULL,
            "DirectX initialization failed!",
            "WinArcadia: Error",
            MB_ICONEXCLAMATION | MB_OK
        );
        cleanexit(EXIT_FAILURE);
	}
    memset(&DDSD, 0, sizeof(DDSD));
    DDSD.dwSize         = sizeof(DDSD);
    DDSD.dwFlags        = DDSD_CAPS;
    DDSD.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
    if
    (   IDirectDraw_CreateSurface(DD, &DDSD, &DDSurface, NULL)  != DD_OK
     || IDirectDraw_CreateClipper(DD, 0, &DDClipper, NULL)      != DD_OK
     || IDirectDrawClipper_SetHWnd(DDClipper, 0, MainWindowPtr) != DD_OK
     || IDirectDrawSurface_SetClipper(DDSurface, DDClipper)     != DD_OK
    )
    {   // there is probably cleanup (for the above 4 functions) which should be done here
        // (or does cleanexit() handle everything?)
        MessageBox
        (   NULL,
            "DirectX initialization failed!",
            "WinArcadia: Error",
            MB_ICONEXCLAMATION | MB_OK
        );
        cleanexit(EXIT_FAILURE);
    }
#endif

    DragAcceptFiles(MainWindowPtr, TRUE);
    ShowWindow(MainWindowPtr, SW_SHOW);
    UpdateWindow(MainWindowPtr);

    updatebiggads();
    updatesmlgads();
    updatemenus();
    redrawscreen();
    settitle();

    if (fps <= 9999)
    {   DISCARD stcul_d(fpsstring, fps);
    } else
    {   strcpy(fpsstring, "####");
    }
    DISCARD SetWindowText(FPSPtr, fpsstring);
}

EXPORT void closewindow(void)
{   SetMenu(MainWindowPtr, MenuPtr); // so it is automatically freed for us
    if (PopUpMenuPtr)
    {   DestroyMenu(PopUpMenuPtr);
        PopUpMenuPtr = NULL;
    }
    // PostMessage(MainWindowPtr, WM_CLOSE, 0, 0);
    DestroyWindow(MainWindowPtr);
    MainWindowPtr = NULL;
}

EXPORT void listen_close(void)
{   if (!ListeningWindowPtr) // important!
    {   return;
    }

    DestroyWindow(ListeningWindowPtr);
    ListeningWindowPtr = NULL;
}

/* MODULE FLAG getargv[i](LPSTR lpCmdLine)
{   AUTO    ULONG destwhere = 0;
    PERSIST ULONG srcwhere  = 0;

    while
    (   lpCmdLine[srcwhere] == ' '
     || lpCmdLine[srcwhere] == '\t'
    )
    {   srcwhere++;
    }
    if (lpCmdLine[srcwhere] == 0)
    {   return FALSE;
    }
    while
    (   lpCmdLine[srcwhere] != ' '
     && lpCmdLine[srcwhere] != '\t'
     && lpCmdLine[srcwhere] != 0
    )
    {   argv[i][destwhere++] = lpCmdLine[srcwhere++];
    }
    argv[i][destwhere] = 0;

    return TRUE;
} */

MODULE void printusage(STRPTR progname)
{   T1;
    printf(
                   "Usage: %s\n"
             "[[FILE]=<file>]\n"
             "[ANALOG=ON|OFF]\n"
           "[AUTOFIRE=ON|OFF]\n"
           "[AUTOSAVE=ON|OFF]\n"
      "[BREAKPOINT|BP=<address>]\n"
         "[COLLISIONS=ON|OFF]\n"
        "[DEMULTIPLEX=ON|OFF]\n"
           "[FLAGLINE=ON|OFF]\n"
         "[FULLSCREEN=ON|OFF]\n"
       "[LIMITREFRESH=ON|OFF]\n"
    "[LOGINSTRUCTIONS=ON|OFF]\n"
           "[LOGREADS=ON|OFF]\n"
          "[LOGWRITES=ON|OFF]\n"
            "[MACHINE=ARCADIA|INTERTON|ELEKTOR]\n"
            "[MENUBAR=ON|OFF]\n"
             "[NARROW=ON|OFF]\n"
             "[PAUSED=ON|OFF]\n"
           "[SETTINGS=<file>]\n"
               "[SIZE=1|2|3]\n"
              "[SOUND=ON|OFF]\n"
            "[STRETCH=ON|OFF]\n"
            "[SWAPPED=ON|OFF]\n"
           "[TITLEBAR=ON|OFF]\n"
            "[TOOLBAR=ON|OFF]\n"
               "[WARP=ON|OFF]\n"
      "[WATCHPOINT|WP=<address>]\n"
             , progname
    );

    cleanexit(EXIT_FAILURE);
}

EXPORT void debug_edit(void)
{   sound_off(FALSE);
    DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_CHANGE), MainWindowPtr, ChangeDlgProc);
    sound_on();
}

BOOL CALLBACK ChangeDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam)
{   PERSIST FLAG dontaddr,
                 dontto;

    // IAR and RAS0..RAS7 are allowed to be up to 32767.
    // All others are limited to 255.

    switch(Message)
    {
    case WM_INITDIALOG:
        tolimit = 255;
        SetDlgItemText(hwnd, IDC_CHANGE_ADDRESS, lastchange);
        dontto = dontaddr = FALSE;
        change_read(lastchange);
        SetDlgItemText(hwnd, IDC_CHANGE_FROM, fromstr);
        SetDlgItemText(hwnd, IDC_CHANGE_TO  ,   tostr);
        // DISCARD SetForegroundWindow(hwnd); doesn't work

        return TRUE;
    break;
    case WM_CLOSE:
        change_write(globalstring, FALSE);
        clearkybd();
        EndDialog(hwnd, 0);

        return TRUE;
    break;
    case WM_COMMAND:
        if (LOWORD (wParam) == IDOK)
        {   // assert(tonum >= 0 && tonum <= tolimit);
            change_write(globalstring, FALSE);
            clearkybd();
            EndDialog(hwnd, 0);

            return TRUE;
        }
        if (LOWORD (wParam) == IDCANCEL)
        {   T1; printf("Cancelled.\n\n"); T2;
            clearkybd();
            EndDialog(hwnd, 0);

            return TRUE;
        }
        if (HIWORD (wParam) == EN_UPDATE)
        {   if (LOWORD (wParam) == IDC_CHANGE_TO)
            {   if (dontto)
                {   dontto = FALSE;
                } else
                {   GetWindowText(GetDlgItem(hwnd, IDC_CHANGE_TO), tostr, sizeof(tostr));
                    if (strlen(tostr) > 5) // "$7FFF" or "32767"
                    {   tostr[5] = 0;
                        dontto = TRUE;
                        SetDlgItemText(hwnd, IDC_CHANGE_TO, tostr);
                    }
                    // maybe we should check for illegal chars?
                    if (tostr[0] == '$')
                    {   stch_l(&tostr[1], &tonum); // hexstring -> number
                    } else
                    {   stcd_l(tostr, &tonum); // decstring -> number
                    }
                    if (tonum < 0 || tonum > tolimit)
                    {   tonum = 0;
                        if (tolimit == 255)
                        {   strcpy(tostr, "$00");
                        } else
                        {   // assert(tolimit == 32767);
                            strcpy(tostr, "$0000");
                        }
                        dontto = TRUE;
                        SetDlgItemText(hwnd, IDC_CHANGE_TO, tostr);
            }   }   }
            elif (LOWORD (wParam) == IDC_CHANGE_ADDRESS)
            {   if (dontaddr)
                {   dontaddr = FALSE;
                } else
                {   GetWindowText(GetDlgItem(hwnd, IDC_CHANGE_ADDRESS), globalstring, sizeof(globalstring));
                    strupr(globalstring);
                    if (strlen(globalstring) > 13) // eg. "SPRITECOLLIDE"
                    {   globalstring[13] = 0;
                        dontaddr = TRUE;
                        SetDlgItemText(hwnd, IDC_CHANGE_ADDRESS, globalstring);
                    }

                    change_read(globalstring);
                    SetDlgItemText(hwnd, IDC_CHANGE_FROM, fromstr);
                    SetDlgItemText(hwnd, IDC_CHANGE_TO  ,   tostr);
        }   }   }

        return FALSE;
    break;
    default:
        return FALSE;
    }
    return TRUE;
}

EXPORT void openconsole(void)
{   int   hScrn, hKybd;
    FILE* hf;

    // assert(consoleopen);

    AllocConsole(); // might need freeing at program exit
    consoleopen = TRUE;
    hScrn = _open_osfhandle((long) GetStdHandle(STD_OUTPUT_HANDLE), _O_TEXT); // might need freeing at program exit
    hKybd = _open_osfhandle((long) GetStdHandle(STD_INPUT_HANDLE ), _O_TEXT); // might need freeing at program exit
    hf = _fdopen(hScrn, "w"); // might need freeing at program exit
    *stdout = *hf;
    hf = _fdopen(hKybd, "r"); // might need freeing at program exit
	*stdin = *hf;
    DISCARD setvbuf(stdout, NULL, _IONBF, 0);
	DISCARD setvbuf(stdin , NULL, _IONBF, 0);
}

EXPORT void breakparts(void)
{   signed int i,
               length,
               where;

    // Splits a complete pathname (in fn_game) into pathpart and filepart.

    strcpy(pathpart, fn_game);
    where = -1;
    length = strlen(fn_game);
    for (i = 0; i < length; i++)
    {   if
        (   fn_game[i] == ':'
         || fn_game[i] == '\\'
        )
        {   where = i;
            // don't break, we want the LAST component, not the FIRST
    }   }
    strcpy(filepart, &fn_game[where + 1]);
    pathpart[where + 1] = 0;
}

MODULE void handlekybd(UINT code)
{   if
    (   (!(ctrl()) && !(shift()))
     || code == SCAN_APOSTROPHE
     || code == SCAN_SEMICOLON
    )
    {   KeyMatrix[code / 8] |= (1 << (code % 8));
    }

    switch(code)
    {
    case SCAN_ESCAPE:
        if (!ctrl() && !shift())
        {   cleanexit(EXIT_SUCCESS);
        }
    acase SCAN_BACKSPACE:
        if (ctrl() && !shift())
        {   if (trace) trace = FALSE; else trace = TRUE;
	        updatemenu(MENUITEM_TRACECPU);
        }
/*  acase SCAN_RETURN:
        if (!ctrl() && !shift())
        {   docommand(MENUITEM_STEP);
        } */
    acase SCAN_BACKTICK:
        if
        (   !crippled
         && game
         && ctrl()
         && !shift()
        )
        {   docommand(MENUITEM_REVERT);
        }
    acase SCAN_A:
        if (ctrl() && !shift())
        {   docommand(MENUITEM_SAVEROM);
        }
    acase SCAN_B:
        if (ctrl() && !shift())
        {   if (showtitlebar) showtitlebar = FALSE; else showtitlebar = TRUE; // foo = ~foo; doesn't always work!
            docommand(MENUITEM_TITLEBAR);
        }
    acase SCAN_C:
        if (ctrl() && !shift())
        {   if (ControlsWindowPtr)
            {   DestroyWindow(ControlsWindowPtr);
                ControlsWindowPtr = NULL;
                unghost(MENUITEM_CONTROLS);
            } else
            {   view_controls();
        }   }
    acase SCAN_D:
        if (ctrl() && !shift())
        {   if (machine == ARCADIA)
            {   if (demultiplex) demultiplex = FALSE; else demultiplex = TRUE; // foo = ~foo; doesn't always work!
                docommand(MENUITEM_DEMULTIPLEX);
        }   }
    acase SCAN_E:
        if (ctrl() && !shift())
        {   debug_clearwp();
        }
    acase SCAN_F:
        if (ctrl() && !shift())
        {   if (fullscreen)
            {   if (stretch)
                {   stretch = FALSE;
                } else
                {   stretch = TRUE;
                }
                docommand(MENUITEM_STRETCH);
        }   }
    acase SCAN_G:
        if (ctrl() && !shift())
        {   debug_set(KIND_BP);
        }
    acase SCAN_H:
        if (ctrl() && !shift())
        {   debug_clearbp();
        }
    acase SCAN_I:
#ifdef PROFILE
        if (ctrl() && !shift())
        {   printprofile();
        }
#endif
#ifdef COVERAGE
        if (ctrl() && shift())
        {   printcoverage();
        }
#endif
        ;
    acase SCAN_K:
        if (ctrl() && !shift())
        {   debug_edit();
        }
    acase SCAN_L:
        if (ctrl() && !shift())
        {   if (showtoolbar) showtoolbar = FALSE; else showtoolbar = TRUE; // foo = ~foo; doesn't always work!
            docommand(MENUITEM_TOOLBAR);
        }
    acase SCAN_M:
        if (ctrl() && !shift())
        {   if (showmenubar) showmenubar = FALSE; else showmenubar = TRUE; // foo = ~foo; doesn't always work!
            docommand(MENUITEM_MENUBAR);
        }
    acase SCAN_N:
        if (ctrl() && !shift() && !fullscreen)
        {   if (wide == 1)
            {   wide = realwide = 2;
            } else
            {   // assert(wide == 2);
                wide = realwide = 1;
            }
            updatemenu(MENUITEM_NARROW);
            resize(size);
        }
    acase SCAN_O:
        if (!crippled && ctrl() && !shift())
        {   project_open();
        }
    acase SCAN_P:
        if (ctrl() && !shift())
        {   if (flagline) flagline = FALSE; else flagline = TRUE; // foo = ~foo; doesn't always work!
            updatesmlgad(GADPOS_FLAGLINE, flagline, TRUE);
            updatemenu(MENUITEM_FLAGLINE);
            if (paused)
            {   redrawscreen();
        }   }
    acase SCAN_R:
        if (ctrl() && !shift())
        {   if (machine == ARCADIA)
            {   project_save(KIND_EAR);
            } elif (machine == INTERTON)
            {   project_save(KIND_INR);
            } else
            {   // assert(machine == ELEKTOR);
                project_save(KIND_TVR);
        }   }
    acase SCAN_S:
        if (ctrl() && !shift())
        {   if (!crippled)
            {   if (machine == ARCADIA)
                {   project_save(KIND_EAS);
                } elif (machine == INTERTON)
                {   project_save(KIND_INS);
                } else
                {   // assert(machine == ELEKTOR);
                    project_save(KIND_TVS);
        }   }   }
    acase SCAN_T:
        if (ctrl() && !shift())
        {   macro_stop();
        }
    acase SCAN_U:
        if (ctrl() && !shift())
        {   if (sound) sound = FALSE; else sound = TRUE; // foo = ~foo; doesn't always work!
            docommand(MENUITEM_SOUND);
        }
    acase SCAN_W:
        if (ctrl() && !shift())
        {   debug_set(KIND_WP);
        }
    acase SCAN_X:
        if (ctrl() && !shift())
        {   if (showpointer) showpointer = FALSE; else showpointer = TRUE;
            docommand(MENUITEM_POINTER);
        }
    acase SCAN_Y:
        if (ctrl() && !shift())
        {   if (loop) loop = FALSE; else loop = TRUE; // foo = ~foo; doesn't always work!
	        updatemenu(MENUITEM_LOOP);
        }
    acase SCAN_Z:
        if (ctrl() && !shift())
        {   docommand(MENUITEM_STEP);
        }
    acase SCAN_F1:
        if (ctrl() && !shift() && !fullscreen && size != 1)
        {   resize(1);
        }
    acase SCAN_A1:
        if (ctrl())
        {   if (shift())
            {   runtorastline = TRUE;
                unpause();
            } else
            {   if (!fullscreen && size != 1)
                {   resize(1);
        }   }   }
    acase SCAN_F2:
        if (ctrl() && !shift() && !fullscreen && size != 2)
        {   resize(2);
        }
    acase SCAN_A2:
        if (ctrl())
        {   if (shift())
            {   if (machine == ARCADIA)
                {   runtodma = TRUE;
                    unpause();
            }   }
            else
            {   if (!fullscreen && size != 2)
                {   resize(2);
        }   }   }
    acase SCAN_F3:
        if (ctrl() && !shift() && !fullscreen && size != 3)
        {   resize(3);
        }
    acase SCAN_A3:
        if (ctrl())
        {   if (shift())
            {   runtoframe = TRUE;
                unpause();
            } else
            {   if (!fullscreen && size != 3)
                {   resize(3);
        }   }   }
    acase SCAN_A4:
		if (ctrl())
		{   if (shift())
            {   debug_setrastn();
                unpause();
			} else
			{	docommand(MENUITEM_QUICKLOAD);
	    }	}
    acase SCAN_F5:
        if (machine != ARCADIA && ctrl() && shift())
        {   runtointerrupt = TRUE;
            unpause();
        } elif (!crippled && !ctrl())
        {   if (shift())
            {   if (machine == ELEKTOR)
                {   docommand(MENUITEM_RESETTOMONITOR);
            }   }
            else
            {   project_reset(FALSE);
        }   }
    acase SCAN_A5:
        if (ctrl() && !shift())
		{	docommand(MENUITEM_QUICKSAVE);
		}
    acase SCAN_F6:
        if (!shift())
        {   if (ctrl())
            {   if (analog) analog = FALSE; else analog = TRUE; // foo = ~foo; doesn't always work!
                updatesmlgad(GADPOS_ANALOG, analog, TRUE);
                updatemenu(MENUITEM_ANALOG);
                updatemenu(MENUITEM_SENSITIVITY);
            } else
            {   if (MonitorWindowPtr)
                {   DestroyWindow(MonitorWindowPtr);
                    MonitorWindowPtr = NULL;
                    unghost(MENUITEM_MONITOR);
                } else
                {   view_monitor();
        }   }   }
    acase SCAN_A6:
        if (ctrl() && !shift())
        {   docommand(COMMAND_ARCADIA);
        }
    acase SCAN_F7:
        if (!shift())
        {   if (ctrl())
            {   if (autofire) autofire = FALSE; else autofire = TRUE; // foo = ~foo; doesn't always work!
                updatesmlgad(GADPOS_AUTOFIRE, autofire, TRUE);
                updatemenu(MENUITEM_AUTOFIRE);
            } else
            {   if (swapped) swapped = FALSE; else swapped = TRUE; // foo = ~foo; doesn't always work!
                updatesmlgad(GADPOS_SWAPPED, swapped, TRUE);
                updatemenu(MENUITEM_SWAPPED);
        }   }
    acase SCAN_A7:
        if (ctrl() && !shift())
        {   docommand(COMMAND_INTERTON);
        }
    acase SCAN_F8:
        if (!shift())
        {   if (ctrl())
            {   if (collisions) collisions = FALSE; else collisions = TRUE; // foo = ~foo; doesn't always work!
                updatesmlgad(GADPOS_COLLISIONS, collisions, TRUE);
                updatemenu(MENUITEM_COLLISIONS);
            } else
            {   if (warp)
                {   warp = FALSE;
                    waittill = timeGetTime(); // makes it effectively IMMEDIATELY
                } else warp = TRUE; // foo = ~foo; doesn't always work!
                updatesmlgad(GADPOS_WARP, warp, TRUE);
                updatemenu(MENUITEM_WARP);
        }   }
    acase SCAN_A8:
        if (ctrl())
        {   if (shift())
            {   sound_off(FALSE);
                DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINE), MainWindowPtr, RedefineDlgProc);
                sound_on();
            } else
            {   docommand(COMMAND_ELEKTOR);
        }   }
    acase SCAN_F9:
        if (!ctrl() && !shift())
        {   if (fullscreen) fullscreen = FALSE; else fullscreen = TRUE; // foo = ~foo; doesn't always work!
            updatemenu(MENUITEM_FULLSCREEN);
            closewindow();
            openwindow();
            redrawscreen();
        }
    acase SCAN_A9:
        if (ctrl())
        {   if (shift())
            {   if (joy0) joy0 = FALSE; else joy0 = TRUE;
                updatemenu(MENUITEM_JOY0);
            } else
            {   if (cli) cli = FALSE; else cli = TRUE;
                docommand(MENUITEM_CLI);
        }   }
    acase SCAN_A0:
        if (ctrl() && shift())
        {   if (joy1) joy1 = FALSE; else joy1 = TRUE;
            updatemenu(MENUITEM_JOY1);
        }
    acase SCAN_F12:
        if (!ctrl() && !shift())
        {   if (ControlsWindowPtr)
            {   DestroyWindow(ControlsWindowPtr);
                ControlsWindowPtr = NULL;
                unghost(MENUITEM_CONTROLS);
            } else
            {   view_controls();
        }   }
    acase SCAN_HYPHEN:
        if (ctrl() && shift())
        {   savescreenshot(KIND_ILBM, TRUE);
        }
    acase SCAN_EQUALS:
        if (ctrl() && shift())
        {   if (!crippled && recmode == RECMODE_PLAY)
            {   macro_restartplayback();
        }   }
    acase SCAN_BACKSLASH:
        if (ctrl() && !shift() && machine == ELEKTOR)
        {   project_save(KIND_TVC);
        }
    acase SCAN_OS:
        if (ctrl() && !shift())
        {   region = REGION_NTSC;
            updatemenu(MENUITEM_REGION);
        }
    acase SCAN_CS:
        if (ctrl() && !shift())
        {   region = REGION_PAL;
            updatemenu(MENUITEM_REGION);
        }
    acase SCAN_COMMA:
        if (ctrl() && !shift())
        {   if (machine == ELEKTOR && !crippled)
            {   docommand(MENUITEM_RESETTOMONITOR);
        }   }
    acase SCAN_SLASH:
        if (ctrl() && shift())
        {   sound_off(FALSE);
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_ABOUT), MainWindowPtr, HelpDlgProc);
            sound_on();
        }
    acase SCAN_PAUSE:
        if (!ctrl() && !shift())
        {   if (paused)
            {   unpause();
            } else
            {   pause();
        }   }
 /* acase SCAN_SCRLK:
        if (!ctrl() && !shift())
        {   if (frameskip == 0x7FFFFFFF)
            {   frameskip = 1;
            } else
            {   frameskip = 0x7FFFFFFF;
        }   } */
    acase SCAN_LCTRL:
    case SCAN_RCTRL:
    case SCAN_LSHIFT:
    case SCAN_RSHIFT:
        KeyMatrix[code / 8] |= (1 << (code % 8));
    adefault:
    break;
}   }

EXPORT void clearkybd(void)
{   KeyMatrix[SCAN_LCTRL  / 8] &= ~(1 << (SCAN_LCTRL  % 8));
    KeyMatrix[SCAN_RCTRL  / 8] &= ~(1 << (SCAN_RCTRL  % 8));
    KeyMatrix[SCAN_LSHIFT / 8] &= ~(1 << (SCAN_LSHIFT % 8));
    KeyMatrix[SCAN_RSHIFT / 8] &= ~(1 << (SCAN_RSHIFT % 8));
}

EXPORT void reactivate(void)
{   // for some reason, this doesn't work the first time, but does for subsequent times!
    SetFocus(MainWindowPtr);
    // DISCARD SetForegroundWindow(MainWindowPtr); doesn't seem to work, lame fucking bastards at Microfilth
}

EXPORT void updatebiggads(void)
{   if (!MainWindowPtr)
    {   return; // important!
    }

    if (crippled)
    {   SendMessage(hToolbar, TB_SETSTATE, ID_FILE_RESET,        0);
        SendMessage(hToolbar, TB_SETSTATE, ID_FILE_OPEN,         0);
    } else
    {   SendMessage(hToolbar, TB_SETSTATE, ID_FILE_RESET,        TBSTATE_ENABLED);
        SendMessage(hToolbar, TB_SETSTATE, ID_FILE_OPEN,         TBSTATE_ENABLED);
    }
    if (crippled || !game)
    {   SendMessage(hToolbar, TB_SETSTATE, ID_FILE_SAVESNP,      0);
    } else
    {   SendMessage(hToolbar, TB_SETSTATE, ID_FILE_SAVESNP,      TBSTATE_ENABLED);
    }

    DISCARD RedrawWindow(FPSPtr, NULL, NULL, RDW_INVALIDATE | RDW_UPDATENOW);
}

EXPORT void ghost(int which)
{   if (menucode[which] != -1)
    {   DISCARD EnableMenuItem(MenuPtr,      menucode[which], MF_BYCOMMAND | MF_GRAYED);
        DISCARD EnableMenuItem(PopUpMenuPtr, menucode[which], MF_BYCOMMAND | MF_GRAYED);
}   }
EXPORT void unghost(int which)
{   if (menucode[which] != -1)
    {   DISCARD EnableMenuItem(MenuPtr,      menucode[which], MF_BYCOMMAND | MF_ENABLED);
        DISCARD EnableMenuItem(PopUpMenuPtr, menucode[which], MF_BYCOMMAND | MF_ENABLED);
}   }
EXPORT void tick(int which)
{   if (menucode[which] != -1)
    {   DISCARD CheckMenuItem( MenuPtr,      menucode[which], MF_BYCOMMAND | MF_CHECKED);
        DISCARD CheckMenuItem( PopUpMenuPtr, menucode[which], MF_BYCOMMAND | MF_CHECKED);
}   }
EXPORT void untick(int which)
{   if (menucode[which] != -1)
    {   DISCARD CheckMenuItem( MenuPtr,      menucode[which], MF_BYCOMMAND | MF_UNCHECKED);
        DISCARD CheckMenuItem( PopUpMenuPtr, menucode[which], MF_BYCOMMAND | MF_UNCHECKED);
}   }
EXPORT void setbutton(int which, FLAG enabled, FLAG state)
{   if (gadgetcode[which] != -1)
    {   if (state)
        {   SendMessage(hToolbar, TB_CHANGEBITMAP, gadgetcode[which], gadgetimage[which][0]);
        } else
        {   SendMessage(hToolbar, TB_CHANGEBITMAP, gadgetcode[which], gadgetimage[which][1]);
        }
        if (!enabled)
        {   if (state)
            {   SendMessage(hToolbar, TB_SETSTATE, gadgetcode[which], TBSTATE_CHECKED);
            } else
            {   SendMessage(hToolbar, TB_SETSTATE, gadgetcode[which], 0);
        }   }
        else
        {   if (state)
            {   SendMessage(hToolbar, TB_SETSTATE, gadgetcode[which], TBSTATE_ENABLED | TBSTATE_CHECKED);
            } else
            {   SendMessage(hToolbar, TB_SETSTATE, gadgetcode[which], TBSTATE_ENABLED);
}   }   }   }

EXPORT void say(STRPTR text, FLAG handlesound)
{   if (handlesound)
    {   sound_off(FALSE);
    }
    if (machine == ARCADIA)
    {   MessageBox(MainWindowPtr, text,  "WinArcadia: Warning", MB_OK | MB_ICONEXCLAMATION);
    } elif (machine == INTERTON)
    {   MessageBox(MainWindowPtr, text, "WinInterton: Warning", MB_OK | MB_ICONEXCLAMATION);
    } else
    {   // assert(machine == ELEKTOR);
        MessageBox(MainWindowPtr, text,  "WinElektor: Warning", MB_OK | MB_ICONEXCLAMATION);
    }
    if (handlesound)
    {   sound_on();
}   }

EXPORT void infowindow(void)
{   sound_off(FALSE);
    DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_AUTOSENSE), MainWindowPtr, AutosenseDlgProc);
    sound_on();
}

BOOL CALLBACK ListeningDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam)
{   switch(Message)
    {
    case WM_INITDIALOG:
        return TRUE;
    acase WM_CLOSE:
    case WM_KEYDOWN:
        DestroyWindow(hwnd);
        ListeningWindowPtr = NULL;
        aborting = TRUE;
        return TRUE;
    adefault:
        return FALSE;
    break;
    }

    return TRUE;
}

BOOL CALLBACK AutosenseDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam)
{   HBITMAP     BitmapPtr;
    HDC         hDC,
                MemDCPtr;
    PAINTSTRUCT Ps;

    switch (Message)
    {
    case WM_INITDIALOG:
        SetWindowText(hwnd, "Game Information");

        SetDlgItemText(hwnd, IDC_AUTOSENSE1 , autotext[ 0]);
        SetDlgItemText(hwnd, IDC_AUTOSENSE2 , autotext[ 1]);
        SetDlgItemText(hwnd, IDC_AUTOSENSE3 , autotext[ 2]);
        SetDlgItemText(hwnd, IDC_AUTOSENSE4 , autotext[ 3]);
        SetDlgItemText(hwnd, IDC_AUTOSENSE5 , autotext[ 4]);
        SetDlgItemText(hwnd, IDC_AUTOSENSE6 , autotext[ 5]);
        SetDlgItemText(hwnd, IDC_AUTOSENSE7 , autotext[ 6]);
        SetDlgItemText(hwnd, IDC_AUTOSENSE8 , autotext[ 7]);
        SetDlgItemText(hwnd, IDC_AUTOSENSE9 , autotext[ 8]);
        SetDlgItemText(hwnd, IDC_AUTOSENSE10, autotext[ 9]);
        SetDlgItemText(hwnd, IDC_AUTOSENSE11, autotext[10]);
        SetDlgItemText(hwnd, IDC_AUTOSENSE12, autotext[11]);

        return TRUE;
    acase WM_CLOSE:
        clearkybd();
        EndDialog(hwnd, 0);

        return TRUE;
    acase WM_COMMAND:
        if
        (   LOWORD (wParam) == IDOK
         || LOWORD (wParam) == IDCANCEL
        )
        {   clearkybd();
            EndDialog(hwnd, 0);

            return TRUE;
        }

        return FALSE;
    acase WM_DESTROY:
        PostQuitMessage(WM_QUIT);
    acase WM_PAINT:
        hDC = BeginPaint(hwnd, &Ps);

        // Load the bitmap from the resource
        if (machine == ARCADIA)
        {   BitmapPtr = LoadBitmap(InstancePtr, MAKEINTRESOURCE(IDB_S_ARCADIA ));
        } elif (machine == INTERTON)
        {   BitmapPtr = LoadBitmap(InstancePtr, MAKEINTRESOURCE(IDB_S_INTERTON));
        } else
        {   // assert(machine == ELEKTOR);
            BitmapPtr = LoadBitmap(InstancePtr, MAKEINTRESOURCE(IDB_S_ELEKTOR ));
        }

        // Create a memory device compatible with the above DC variable
        MemDCPtr = CreateCompatibleDC(hDC);
        // Select the new bitmap
        SelectObject(MemDCPtr, BitmapPtr);

        // Copy the bits from the memory DC into the current dc
        BitBlt(hDC, 11, 11, 32, 32, MemDCPtr, 0, 0, SRCCOPY);

        // Restore the old bitmap
        DeleteDC(MemDCPtr);
        DeleteObject(BitmapPtr);
        EndPaint(hwnd, &Ps);
    adefault:
        return FALSE;
        // using return DefWindowProc(hwnd, Message, wParam, lParam); means that we would
        // have to manually reactivate the main window after closing this requester.
    break;
    }

    return FALSE;
}

EXPORT void updatefps(void)
{   if (!MainWindowPtr)
    {   return; // unimportant?
    }

    if (fps != oldfps)
    {   oldfps = fps;

#ifdef MODE_MHZ
        sprintf(fpsstring, "%f", (float) (fps / 1000000.0));
#else // MODE_FPS and MODE_PERCENT
        if (fps <= 9999)
        {   DISCARD stcul_d(fpsstring, fps);
        } else
        {   strcpy(fpsstring, "####");
        }
#endif

#ifdef MODE_PERCENT
        strcat(fpsstring, "%");
#endif

        DISCARD SetWindowText(FPSPtr, fpsstring);
}   }

EXPORT int asl
(   STRPTR hail,
    STRPTR pattern,
    FLAG   save,
    STRPTR startdir,
    STRPTR startfile,
    STRPTR passedfn
)
{   OPENFILENAME ofn;

    strcpy(passedfn, startfile);

    ZeroMemory(&ofn, sizeof(ofn));
    ofn.lStructSize     = sizeof(ofn);
    ofn.hwndOwner       = MainWindowPtr;
    ofn.lpstrTitle      = hail;
    ofn.lpstrFilter     = pattern;
    ofn.lpstrInitialDir = startdir;
    ofn.lpstrFile       = passedfn;
    ofn.nMaxFile        = MAX_PATH;
    if (save)
    {   ofn.Flags       = OFN_EXPLORER | OFN_PATHMUSTEXIST | OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT;
    } else
    {   ofn.Flags       = OFN_EXPLORER | OFN_PATHMUSTEXIST;
    }

    sound_off(FALSE);
    if (!showpointer)
    {   ShowCursor(TRUE);
    }
    clearkybd();
    if
    (   ( save && GetSaveFileName(&ofn))
     || (!save && GetOpenFileName(&ofn))
    )
    {   if (!showpointer)
        {   ShowCursor(FALSE);
        }
        sound_on();
        return EXIT_SUCCESS;
    } else
    {   if (!showpointer)
        {   ShowCursor(FALSE);
        }
        sound_on();
        return EXIT_WARNING;
}   }

EXPORT void view_monitor(void)
{   AUTO          int   i;
    PERSIST const UWORD ibmgad[MONITORGADS] = {
    ID_CONSOLE,       //   0
    ID_PITCH,         //   1
    ID_SIZES,         //   2
    ID_SCORECTRL,     //   3
    ID_SCORELT,       //   4
    ID_SCORERT,       //   5
    ID_BGCOLOUR,      //   6
    ID_BGCOLLIDE,     //   7
    ID_SPRITECOLLIDE, //   8
    ID_SPR01COLOURS,  //   9
    ID_SPR23COLOURS,  //  10
    ID_P1PADDLE,      //  11
    ID_P1LEFTKEYS,    //  12
    ID_P1MIDDLEKEYS,  //  13
    ID_P1RIGHTKEYS,   //  14
    ID_P2PADDLE,      //  15
    ID_P2LEFTKEYS,    //  16
    ID_P2MIDDLEKEYS,  //  17
    ID_P2RIGHTKEYS,   //  18
    ID_NOISE,         //  19
    ID_AMPLITUDEA1,   //  20
    ID_AMPLITUDEB1,   //  21
    ID_AMPLITUDEC1,   //  22
    ID_PITCHD1,       //  23
    ID_ENABLE1,       //  24
    ID_SHAPE1,        //  25
    ID_PORTA1,        //  26
    ID_PORTB1,        //  27
    ID_AMPLITUDEA2,   //  28
    ID_AMPLITUDEB2,   //  29
    ID_AMPLITUDEC2,   //  30
    ID_PITCHD2,       //  31
    ID_ENABLE2,       //  32
    ID_SHAPE2,        //  33
    ID_PORTA2,        //  34
    ID_PORTB2,        //  35
    ID_1F0E,          //  36
    ID_1F0F,          //  37
    ID_1F1E,          //  38
    ID_1F1F,          //  39
    ID_1F4E,          //  40
    ID_1F4F,          //  41
    ID_1F50,          //  42
    ID_1F51,          //  43
    ID_1F52,          //  44
    ID_1F53,          //  45
    ID_1F54,          //  46
    ID_1F55,          //  47
    ID_1F56,          //  48
    ID_1F57,          //  49
    ID_1F58,          //  50
    ID_1F59,          //  51
    ID_1F5A,          //  52
    ID_1F5B,          //  53
    ID_1F5C,          //  54
    ID_1F5D,          //  55
    ID_1F5E,          //  56
    ID_1F5F,          //  57
    ID_1F60,          //  58
    ID_1F61,          //  59
    ID_1F62,          //  60
    ID_1F63,          //  61
    ID_1F64,          //  62
    ID_1F65,          //  63
    ID_1F66,          //  64
    ID_1F67,          //  65
    ID_1F68,          //  66
    ID_1F69,          //  67
    ID_1F6A,          //  68
    ID_1F6B,          //  69
    ID_1F6C,          //  70
    ID_1F6D,          //  71
    ID_1FAD,          //  72
    ID_CONSOLE,       //  73
    ID_RESOLUTION,    //  74
    ID_BGCOLLIDE,     //  75
    ID_P1PADDLE,      //  76
    ID_P2PADDLE,      //  77
    ID_PITCH,         //  78
    ID_BGCOLOUR,      //  79
    ID_SPRITECOLLIDE, //  80
    ID_P1LEFTKEYS,    //  81
    ID_P2LEFTKEYS,    //  82
    ID_VOLUME,        //  83
    ID_VSCROLL,       //  84
    ID_SPRITES01CTRL, //  85
    ID_P1MIDDLEKEYS,  //  86
    ID_P2MIDDLEKEYS,  //  87
    ID_CHARLINE,      //  88
    ID_SPRITES23CTRL, //  89
    ID_P1RIGHTKEYS,   //  90
    ID_P2RIGHTKEYS,   //  91
    ID_P1PALLADIUM,   //  92
    ID_P2PALLADIUM,   //  93
    ID_18D0,          //  94
    ID_18D1,          //  95
    ID_18D2,          //  96
    ID_18D3,          //  97
    ID_18D4,          //  98
    ID_18D5,          //  99
    ID_18D6,          // 100
    ID_18D7,          // 101
    ID_18D8,          // 102
    ID_18D9,          // 103
    ID_18DA,          // 104
    ID_18DB,          // 105
    ID_18DC,          // 106
    ID_18DD,          // 107
    ID_18DE,          // 108
    ID_18DF,          // 109
    ID_18E0,          // 110
    ID_18E1,          // 111
    ID_18E2,          // 112
    ID_18E3,          // 113
    ID_18E4,          // 114
    ID_18E5,          // 115
    ID_18E6,          // 116
    ID_18E7,          // 117
    ID_18E8,          // 118
    ID_18E9,          // 119
    ID_18EA,          // 120
    ID_18EB,          // 121
    ID_18EC,          // 122
    ID_18ED,          // 123
    ID_18EE,          // 124
    ID_18EF,          // 125
    ID_18F8,          // 126
    ID_18F9,          // 127
    ID_18FA,          // 128
    ID_18FB,          // 129
    ID_1AD0,          // 130
    ID_1AD1,          // 131
    ID_1AD2,          // 132
    ID_1AD3,          // 133
    ID_1AD4,          // 134
    ID_1AD5,          // 135
    ID_1AD6,          // 136
    ID_1AD7,          // 137
    ID_1AD8,          // 138
    ID_1AD9,          // 139
    ID_1ADA,          // 140
    ID_1ADB,          // 141
    ID_1ADC,          // 142
    ID_1ADD,          // 143
    ID_1ADE,          // 144
    ID_1ADF,          // 145
    ID_1AE0,          // 146
    ID_1AE1,          // 147
    ID_1AE2,          // 148
    ID_1AE3,          // 149
    ID_1AE4,          // 150
    ID_1AE5,          // 151
    ID_1AE6,          // 152
    ID_1AE7,          // 153
    ID_1AE8,          // 154
    ID_1AE9,          // 155
    ID_1AEA,          // 156
    ID_1AEB,          // 157
    ID_1AEC,          // 158
    ID_1AED,          // 159
    ID_1AEE,          // 160
    ID_1AEF,          // 161
    ID_1AF0,          // 162
    ID_1AF1,          // 163
    ID_1AF2,          // 164
    ID_1AF3,          // 165
    ID_1AF4,          // 166
    ID_1AF5,          // 167
    ID_1AF6,          // 168
    ID_1AF7,          // 169
    ID_1AF8,          // 170
    ID_1AF9,          // 171
    ID_1AFA,          // 172
    ID_1AFB,          // 173
    ID_1AFC,          // 174
    ID_1AFD,          // 175
    ID_1AFE,          // 176
    ID_1AFF           // 177
    };

    if (machine == ARCADIA)
    {   MonitorWindowPtr = CreateDialog
        (   InstancePtr,                   // handle to application instance
            MAKEINTRESOURCE(IDD_AMONITOR), // identifies dialog box template name
            MainWindowPtr,
            MonitorDlgProc
        );
    } elif (machine == INTERTON)
    {   MonitorWindowPtr = CreateDialog
        (   InstancePtr,                   // handle to application instance
            MAKEINTRESOURCE(IDD_IMONITOR), // identifies dialog box template name
            MainWindowPtr,
            MonitorDlgProc
        );
    } else
    {   // assert(machine == ELEKTOR);
        if (memmap == MEMMAP_E)
        {   MonitorWindowPtr = CreateDialog
            (   InstancePtr,                    // handle to application instance
                MAKEINTRESOURCE(IDD_EMONITOR1), // identifies dialog box template name
                MainWindowPtr,
                MonitorDlgProc
            );
        } else
        {   // assert(memmap == MEMMAP_F);
            MonitorWindowPtr = CreateDialog
            (   InstancePtr,                    // handle to application instance
                MAKEINTRESOURCE(IDD_EMONITOR2), // identifies dialog box template name
                MainWindowPtr,
                MonitorDlgProc
            );
    }   }

    // assert(MonitorWindowPtr);

    RGad[0]      = GetDlgItem(MonitorWindowPtr, ID_R0);
    RGad[1]      = GetDlgItem(MonitorWindowPtr, ID_R1);
    RGad[2]      = GetDlgItem(MonitorWindowPtr, ID_R2);
    RGad[3]      = GetDlgItem(MonitorWindowPtr, ID_R3);
    RGad[4]      = GetDlgItem(MonitorWindowPtr, ID_R4);
    RGad[5]      = GetDlgItem(MonitorWindowPtr, ID_R5);
    RGad[6]      = GetDlgItem(MonitorWindowPtr, ID_R6);
    CCPtr        = GetDlgItem(MonitorWindowPtr, ID_CC);
    SensePtr     = GetDlgItem(MonitorWindowPtr, ID_SENSE);
    FlagPtr      = GetDlgItem(MonitorWindowPtr, ID_FLAG);
    InterruptPtr = GetDlgItem(MonitorWindowPtr, ID_INTERRUPT);
    IDCPtr       = GetDlgItem(MonitorWindowPtr, ID_IDC);
    RSPtr        = GetDlgItem(MonitorWindowPtr, ID_RS);
    WCPtr        = GetDlgItem(MonitorWindowPtr, ID_WC);
    OVFPtr       = GetDlgItem(MonitorWindowPtr, ID_OVF);
    COMPtr       = GetDlgItem(MonitorWindowPtr, ID_COM);
    CPtr         = GetDlgItem(MonitorWindowPtr, ID_C);
    SPGad        = GetDlgItem(MonitorWindowPtr, ID_SP);
    RASGad[0]    = GetDlgItem(MonitorWindowPtr, ID_RAS0);
    RASGad[1]    = GetDlgItem(MonitorWindowPtr, ID_RAS1);
    RASGad[2]    = GetDlgItem(MonitorWindowPtr, ID_RAS2);
    RASGad[3]    = GetDlgItem(MonitorWindowPtr, ID_RAS3);
    RASGad[4]    = GetDlgItem(MonitorWindowPtr, ID_RAS4);
    RASGad[5]    = GetDlgItem(MonitorWindowPtr, ID_RAS5);
    RASGad[6]    = GetDlgItem(MonitorWindowPtr, ID_RAS6);
    RASGad[7]    = GetDlgItem(MonitorWindowPtr, ID_RAS7);
    IARPtr       = GetDlgItem(MonitorWindowPtr, ID_IAR);

    FramePtr         = GetDlgItem(MonitorWindowPtr, ID_FRAME);
    ClockPtr         = GetDlgItem(MonitorWindowPtr, ID_CLOCK);
    RasterGad        = GetDlgItem(MonitorWindowPtr, ID_RASTER);
    TimePtr          = GetDlgItem(MonitorWindowPtr, ID_TIME);

    SPRITEnAXYGad[0]  = GetDlgItem(MonitorWindowPtr, ID_SPRITE0XY);
    SPRITEnAXYGad[1]  = GetDlgItem(MonitorWindowPtr, ID_SPRITE1XY);
    SPRITEnAXYGad[2]  = GetDlgItem(MonitorWindowPtr, ID_SPRITE2XY);
    SPRITEnAXYGad[3]  = GetDlgItem(MonitorWindowPtr, ID_SPRITE3XY);

    if (machine == ARCADIA)
    {   for (i = 73; i < MONITORGADS; i++)
        {   monitorgad[i] = GetDlgItem(MonitorWindowPtr, ibmgad[i]);
    }   }
    else
    {   // assert(machine == INTERTON || machine == ELEKTOR);
        SPRITEnBXYGad[0]  = GetDlgItem(MonitorWindowPtr, ID_SPRITE0XY2);
        SPRITEnBXYGad[1]  = GetDlgItem(MonitorWindowPtr, ID_SPRITE1XY2);
        SPRITEnBXYGad[2]  = GetDlgItem(MonitorWindowPtr, ID_SPRITE2XY2);
        SPRITEnBXYGad[3]  = GetDlgItem(MonitorWindowPtr, ID_SPRITE3XY2);

        if (machine == INTERTON)
        {   for (i = 0; i <= 19; i++)
            {   monitorgad[i] = GetDlgItem(MonitorWindowPtr, ibmgad[i]);
            }
            for (i = 36; i <= 72; i++)
            {   monitorgad[i] = GetDlgItem(MonitorWindowPtr, ibmgad[i]);
        }   }
        else
        {   // assert(machine == ELEKTOR);
            if (memmap == MEMMAP_E)
            {   for (i = 0; i <= 18; i++)
                {   monitorgad[i] = GetDlgItem(MonitorWindowPtr, ibmgad[i]);
            }   }
            else
            {   // assert(memmap == MEMMAP_F);
                for (i = 0; i <= 35; i++)
                {   monitorgad[i] = GetDlgItem(MonitorWindowPtr, ibmgad[i]);
                }
    
                PITCHAGad[0]     = GetDlgItem(MonitorWindowPtr, ID_PITCHA1);
                PITCHBGad[0]     = GetDlgItem(MonitorWindowPtr, ID_PITCHB1);
                PITCHCGad[0]     = GetDlgItem(MonitorWindowPtr, ID_PITCHC1);
                PERIODGad[0]     = GetDlgItem(MonitorWindowPtr, ID_PERIOD1);

                PITCHAGad[1]     = GetDlgItem(MonitorWindowPtr, ID_PITCHA2);
                PITCHBGad[1]     = GetDlgItem(MonitorWindowPtr, ID_PITCHB2);
                PITCHCGad[1]     = GetDlgItem(MonitorWindowPtr, ID_PITCHC2);
                PERIODGad[1]     = GetDlgItem(MonitorWindowPtr, ID_PERIOD2);
    }   }   }

    SetActiveWindow(MainWindowPtr);

    updatemenu(MENUITEM_MONITOR);
    updatemonitor(TRUE);
}

EXPORT void view_controls(void)
{   view_controls_engine();

    if (machine == ARCADIA)
    {   ControlsWindowPtr = CreateDialog
        (   InstancePtr,                    // handle to application instance
            MAKEINTRESOURCE(IDD_ACONTROLS), // identifies dialog box template name
            MainWindowPtr,
            ControlsDlgProc
        );
    } elif (machine == INTERTON)
    {   ControlsWindowPtr = CreateDialog
        (   InstancePtr,                    // handle to application instance
            MAKEINTRESOURCE(IDD_ICONTROLS), // identifies dialog box template name
            MainWindowPtr,
            ControlsDlgProc
        );
    } else
    {   // assert(machine == ELEKTOR);
        ControlsWindowPtr = CreateDialog
        (   InstancePtr,                    // handle to application instance
            MAKEINTRESOURCE(IDD_ECONTROLS), // identifies dialog box template name
            MainWindowPtr,
            ControlsDlgProc
        );
    }
    SetActiveWindow(MainWindowPtr);

    updatemenu(MENUITEM_CONTROLS);
}

/* BOOL CALLBACK SensitivityDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam)
{   switch (Message)
    {
    case WM_INITDIALOG:
        DISCARD SetScrollRange(GetDlgItem(hwnd, IDC_SENSITIVITY), SB_CTL, 1, SENSITIVITY_MAX, FALSE);
        DISCARD SetScrollPos(  GetDlgItem(hwnd, IDC_SENSITIVITY), SB_CTL, sensitivity,        FALSE);
		
        return TRUE;                                                                    
    acase WM_CLOSE:
        sensitivity = GetScrollPos(GetDlgItem(hwnd, IDC_SENSITIVITY), SB_CTL);
        clearkybd();
        EndDialog(hwnd, 0);

        return TRUE;
    acase WM_COMMAND:
        if (LOWORD (wParam) == IDOK)
        {   sensitivity = GetScrollPos(GetDlgItem(hwnd, IDC_SENSITIVITY), SB_CTL);
            clearkybd();
            EndDialog(hwnd, 0);

            return TRUE;
        }
        if (LOWORD (wParam) == IDCANCEL)
        {   clearkybd();
            EndDialog(hwnd, 0);

            return TRUE;
        }

        return FALSE;
    adefault:
        return FALSE;
    break;
    }

    return TRUE;
}
*/

EXPORT void changefilepart(STRPTR newfilepart)
{   signed int i,
               thelength,
               where      = -1;

    /* It doesn't actually do anything with filepart, you would need to
       call breakparts() afterwards if you wanted filepart to be
       accurate. */

    strcpy(pathpart, fn_game);
    where = -1;
    thelength = strlen(pathpart);
    for (i = 0; i < thelength; i++)
    {   if
        (   pathpart[i] == ':'
         || pathpart[i] == '\\'
        )
        {   where = i;
            // don't break, we want the LAST component, not the FIRST
    }   }
    pathpart[where + 1] = 0;
    strcpy(fn_game, pathpart);
    strcat(fn_game, newfilepart);
}

EXPORT void lockgui(void)
{   EnableMenuItem(MenuPtr,      0, MF_BYPOSITION | MF_GRAYED);
    EnableMenuItem(PopUpMenuPtr, 0, MF_BYPOSITION | MF_GRAYED);
    // this only does the File menu unfortunately!
}
EXPORT void unlockgui(void)
{   EnableMenuItem(MenuPtr,      0, MF_BYPOSITION | MF_ENABLED);
    EnableMenuItem(PopUpMenuPtr, 0, MF_BYPOSITION | MF_ENABLED);
    // this only does the File menu unfortunately!
}
EXPORT void clearmsgs(void)
{   MSG Msg;

    if (!MainWindowPtr)
    {   return;
    }

    while (PeekMessage(&Msg, MainWindowPtr, 0, 0, PM_REMOVE))
    {   TranslateMessage(&Msg);
        DispatchMessage(&Msg);
    }

    if (!ListeningWindowPtr)
    {   return;
    }

    while (PeekMessage(&Msg, ListeningWindowPtr, 0, 0, PM_REMOVE))
    {   TranslateMessage(&Msg);
        DispatchMessage(&Msg);
}   }

EXPORT void clearscreen(void)
{   HDC OnScreenRastPort;

    // assert(fullscreen);

    if (!stretch)
    {   if (showtoolbar)
        {   therect.top = toolbarheight;
        } else
        {   therect.top = 0;
        }
        therect.bottom = winheight;
        therect.left = 0;
        therect.right = winwidth;
        /* Note that there is slight flickering caused by this; ideally
        we would only fill the borders with grey, not the whole client area of the window. */
        OnScreenRastPort = GetDC(MainWindowPtr);
        FillRect(OnScreenRastPort, &therect, GetStockObject(GRAY_BRUSH));
        DISCARD ReleaseDC(MainWindowPtr, OnScreenRastPort);
}   }

MODULE void loadconfig(void)
{   FLAG  ok = FALSE;
    ULONG thesize;
    FILE* LocalHandle /* = NULL */ ;
    int   i, j,
          localoffset;

    thesize = getsize(fn_config);
    if (thesize <= CONFIGLENGTH + MAX_PATH)
    {   if ((LocalHandle = fopen(fn_config, "rb")))
        {   if (fread(ConfigBuffer, 134, 1, LocalHandle) == 1)
            {   if (ConfigBuffer[21] == 15 || ConfigBuffer[21] == 16)
                {   ok = TRUE;

                    // graphics
                    size            = (UBYTE) ConfigBuffer[0];
                    wide            = (UBYTE) ConfigBuffer[1];
                    requirebutton   = (int)   ConfigBuffer[2];
                    analog          = (ULONG) ConfigBuffer[3];
                    autofire        = (ULONG) ConfigBuffer[4];
                    swapped         = (ULONG) ConfigBuffer[5];
                    paused          = (ULONG) ConfigBuffer[6];
                    warp            = (ULONG) ConfigBuffer[7];
                    collisions      = (ULONG) ConfigBuffer[8];
                    sound           = (ULONG) ConfigBuffer[9];
                    demultiplex     = (ULONG) ConfigBuffer[10];
                    autosave        = (int)   ConfigBuffer[11];
                    fullscreen      = (int)   ConfigBuffer[12];
                    showmenubar     = (int)   ConfigBuffer[13];
                    showtitlebar    = (int)   ConfigBuffer[14];
                    showtoolbar     = (int)   ConfigBuffer[15];
                    flagline        = (int)   ConfigBuffer[16];
                    showpointer     = (int)   ConfigBuffer[17];
                    loop            = (int)   ConfigBuffer[18];
                    logreads        = (int)   ConfigBuffer[19];
                    logwrites       = (int)   ConfigBuffer[20];
                    // version is ConfigBuffer[21]
                    limitrefresh    = (int)   ConfigBuffer[22];
                    machine         = (int)   ConfigBuffer[23];
                    loginstructions = (int)   ConfigBuffer[24];
                    watchreads      = (int)   ConfigBuffer[25];
                    sensegame       = (int)   ConfigBuffer[26];
                    usehacks        = (int)   ConfigBuffer[27];
                    localoffset = 28;
                    for (i = 0; i < 2; i++)
                    {   for (j = 0; j <= 23; j++)
                        {   keypads[i][j] = (ConfigBuffer[localoffset] * 256)
                                          +  ConfigBuffer[localoffset + 1];
                            localoffset += 2;
                    }   }
                    for (i = 0; i < 4; i++)
                    {   console[i]    = (ConfigBuffer[localoffset] * 256)
                                      +  ConfigBuffer[localoffset + 1];
                        localoffset += 2;
                    }
                    // assert(localoffset == 132);

                    if (ConfigBuffer[21] == 15)
                    {   if (thesize > 134)
                        {   DISCARD fread(pathpart, thesize - 134, 1, LocalHandle);
                    }   }
                    else
                    {   // assert(ConfigBuffer[21] == 16);
                        DISCARD fread(&ConfigBuffer[134], 1, 1, LocalHandle);
                        logmonitor = (ULONG) ConfigBuffer[134];
                        if (thesize > 135)
                        {   DISCARD fread(pathpart, thesize - 135, 1, LocalHandle);
            }   }   }   }
            DISCARD fclose(LocalHandle);
            // LocalHandle = NULL;
    }   }
    if (!ok)
    {   calcsize(TRUE);
        size = realsize;
}   }

EXPORT void invert(int i)
{   InvertRect(ControlsRastPtr, &keyrect[machine][i]);
}

EXPORT void listen_open(void)
{   ListeningWindowPtr = CreateDialog
    (   InstancePtr,                    // handle to application instance
        MAKEINTRESOURCE(IDD_LISTENING), // identifies dialog box template name
        MainWindowPtr,
        ListeningDlgProc
    );
}
EXPORT void hostnamewindow(void)
{   DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_HOSTNAME), MainWindowPtr, HostNameDlgProc);
}

BOOL CALLBACK HostNameDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam)
{   switch (Message)
    {
    case WM_INITDIALOG:
        SetDlgItemText(hwnd, IDC_HOSTNAME, hostname);

        return TRUE;
    break;
    case WM_CLOSE:
        GetWindowText(GetDlgItem(hwnd, IDC_HOSTNAME), hostname, sizeof(hostname));
        clearkybd();
        EndDialog(hwnd, 0);

        return TRUE;
    break;
    case WM_COMMAND:
        if (LOWORD (wParam) == IDOK)
        {   GetWindowText(GetDlgItem(hwnd, IDC_HOSTNAME), hostname, sizeof(hostname));
            clearkybd();
            EndDialog(hwnd, 0);

            return TRUE;
        }
        if (LOWORD (wParam) == IDCANCEL)
        {   hostname[0] = 0;
            clearkybd();
            EndDialog(hwnd, 0);

            return TRUE;
        }

        return FALSE;
    break;
    default:
        return FALSE;
    break;
    }

    return TRUE;
}

EXPORT void close_monitor(void)
{   if (!MonitorWindowPtr)
    {   return;
    }

    DestroyWindow(MonitorWindowPtr);
    MonitorWindowPtr = NULL;
    updatemenu(MENUITEM_MONITOR);
}

EXPORT void close_controls(void)
{   if (!ControlsWindowPtr)
    {   return;
    }

    DestroyWindow(ControlsWindowPtr);
    ControlsWindowPtr = NULL;
    updatemenu(MENUITEM_CONTROLS);
}

EXPORT void setmonitor(struct Gadget* gadgetptr, STRPTR contents)
{   // assert(MonitorWindowPtr);

    DISCARD SetWindowText(gadgetptr, contents);
}

EXPORT STRPTR GetCatalogStr(int ignored1, int ignored2, STRPTR thestring)
{   return thestring;
}
