/* engine.c: Miscellaneous platform-independent code

INCLUDES--------------------------------------------------------------- */

#ifdef AMIGA
    #include "lint.h"
    #include "amiga.h"
    #include "ajoystick.h"
#endif
#ifdef WIN32
    #include "ibm.h"
    #define EXEC_TYPES_H
#endif

#include <stdio.h>             // for printf()
#include <stdlib.h>            // for malloc(), calloc(), free()
#include <string.h>            // for strcpy(), etc.
#include <time.h>              // for randomization

#include "aa.h"
#define CATCOMP_NUMBERS
#define CATCOMP_CODE
#define CATCOMP_BLOCK
#include "aa_strings.h"

#ifdef AMIGA
    #include <intuition/intuition.h>
    #define ALL_REACTION_CLASSES
    #define ALL_REACTION_MACROS
    #include <reaction/reaction.h>
    #include <proto/gadtools.h>
    #include <proto/intuition.h>
    #include <proto/locale.h>        // GetCatalogStr()

    #ifdef __SASC
        #define BASE_PAR_DECL

        struct hostent* gethostbyname(const UBYTE* name);
        LONG            recv(LONG s, UBYTE* buf, LONG len, LONG flags); /* V3 */
        LONG            closesocket(LONG d);
        LONG            connect(LONG s, const struct sockaddr* name, LONG namelen);
        LONG            send(LONG s, const UBYTE* msg, LONG len, LONG flags);
        LONG            socket(LONG domain, LONG type, LONG protocol);
        LONG            bind(LONG s, const struct sockaddr* name, LONG namelen);
        LONG            listen(LONG s, LONG backlog);
        LONG            accept(LONG s, struct sockaddr* addr, LONG* addrlen);
        LONG            ioctlsocket(BASE_PAR_DECL LONG d, ULONG request, char* argp);
        LONG            gethostname(STRPTR hostname, LONG size); /* V3 */

        #pragma libcall SocketBase socket         1E 21003
        #pragma libcall SocketBase bind           24 18003
        #pragma libcall SocketBase listen         2A 1002
        #pragma libcall SocketBase accept         30 98003
        #pragma libcall SocketBase connect        36 18003
        #pragma libcall SocketBase send           42 218004
        #pragma libcall SocketBase recv           4E 218004
        #pragma libcall SocketBase ioctlsocket    72 81003
        #pragma libcall SocketBase closesocket    78 001
        #pragma libcall SocketBase gethostbyname  D2 801
        #pragma libcall SocketBase gethostname   11A 0802

        #define htonl(x)          (x)
        #define htons(x)          (x)

        struct in_addr
        {   ULONG          s_addr;
        };
        struct sockaddr_in
        {   UBYTE          sin_len;
            UBYTE          sin_family;
            USHORT         sin_port;
            struct in_addr sin_addr;
            char           sin_zero[8];
        };
    #else
        #include <netinet/in.h>
        #define closesocket(d) CloseSocket(d)
        #define ioctlsocket(d, request, argp) IoctlSocket(d, request, argp)
    #endif

    struct hostent
    {   char*          h_name;
        char**         h_aliases;
        int            h_addrtype;
        int            h_length;
        char**         h_addr_list;
#define h_addr         h_addr_list[0] /* address, for backward compatiblity */
    };

    #define SOCK_STREAM               1 /* stream socket */
    #define AF_INET                   2 /* internetwork: UDP, TCP, etc. */
    #define _IOC(inout,group,num,len) (inout | ((len & IOCPARM_MASK) << 16) | ((group) << 8) | (num))
    #define _IOW(g,n,t)               _IOC(IOC_IN, (g), (n), sizeof(t))
    #define FIONBIO                   _IOW('f', 126, long) /* set/clear non-blocking i/o */
    #define EAGAIN                    35        /* Resource temporarily unavailable */
    #define EWOULDBLOCK               EAGAIN        /* Operation would block */
    #define WSAEWOULDBLOCK            EWOULDBLOCK
    #ifdef __SASC
        typedef unsigned long         u_long;
    #endif
    typedef u_long                    n_long; /* long as received from the net */
    #define INADDR_ANY                (u_long)0x00000000
    #define IOCPARM_MASK              0x1fff /* parameter length, at most 13 bits */
    #define IOC_IN                    0x80000000 /* copy in parameters */

#endif
#ifdef WIN32
    #include <commctrl.h>
    #include "resource.h"
#endif

// DEFINES----------------------------------------------------------------

#define FUDGE 271

// #define BASICMODE
// whether to prefer mode "E" for "E/F" games (otherwise uses mode "F").

// MODULE VARIABLES-------------------------------------------------------

MODULE       UBYTE  oldkeys[KEYS],
                    newkeys[KEYS];
MODULE       UBYTE  startaddr_h,
                    startaddr_l,
                    t;
MODULE       TEXT   astring[4 + 1],
                    bits[4][10][8 + 1],
                    bstring[3 + 1],
                    disp[16][4 + 1],
                    xvistring[3][10 + 1];
MODULE       int    filesize,
                    Socket           = -1,
                    whichgame        = -1;
MODULE       ULONG  crcTable[256];
MODULE       FLAG   golf             = FALSE; // may not require initialization
MODULE const STRPTR quickname[3] =
{   "QUICKSAV.EAS",
    "QUICKSAV.INS",
    "QUICKSAV.TVS"
};
MODULE const TEXT   arcadia_chars[64 + 1] = " /\\#bc_efghijklm0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ.,+$nopqrstu";

MODULE UWORD monitorras[8],
             monitoriar,
             // PSGs
             monitorPITCHA[2],
             monitorPITCHB[2],
             monitorPITCHC[2],
             monitorPERIOD[2];
MODULE UBYTE monitorr[7],
             monitorpsu,
             monitorpsl,
             monitorSPRITEnAX[4],
             monitorSPRITEnAY[4],
             monitorSPRITEnBX[4],
             monitorSPRITEnBY[4],
             monitorcontents[MONITORGADS];
MODULE ULONG // XVI...
             monitorelapsed,
             monitorframes;
MODULE int   monitorrast;

// EXPORTED VARIABLES-----------------------------------------------------

// variables which correspond to ReAction gadgets must be ULONGs
EXPORT ULONG  analog           = FALSE,
              autofire         = FALSE,
              collisions       = TRUE,
              swapped          = FALSE,
              demultiplex      = FALSE,
              flagline         = TRUE,
              fps              = 0,
              joy1             = FALSE,
              paused           = FALSE,
              region           = REGION_NTSC,
              sound            = TRUE,
              warp             = FALSE;
// other ULONGs go here
EXPORT ULONG  downframes       = 4,
              oldframes,
              frames,
              totalframes      = 16;
EXPORT TEXT   autotext[VERBOSELINES][60 + 1],
              friendly[21 + 1],
           // "$19FD (SPRITECOLLIDE)"
           //  0123456789012345678901
              timestring[11 + 1], // "hh:mm:ss.uu"
              fromstr[5 + 1 + 1],
              tostr[5 + 1 + 1],
              lastchange[13 + 1 + 1],
              filepart[MAX_PATH + 1],
              pathpart[MAX_PATH + 1],
              fn_game[MAX_PATH + 1] = "",  // complete pathname of the game
              fn_screenshot[MAX_PATH + 1]; // complete pathname of the screenshot
EXPORT FILE*  MacroHandle      = NULL;
EXPORT int    bp               = OUTOFRANGE,
              connected        = NET_OFF,
              cpl              = PVI_CPL,
              frameskip        = 1,
              fromnum,
              forceinterrupt   = FALSE,
              fudge            = FUDGE,
              machine          = ARCADIA,
              memmap           = MEMMAP_G,
              offset,
              oldmachine       = ARCADIA,
              oldmemmap,
              key1             = 2,
              key2             = 1,
              key3             = 3,
              key4             = 0,
              p1sprite         = 4,
              p2sprite         = 4,
              rast,
              rastn            = -1,
              recmode          = RECMODE_NORMAL,
              recsize,
              romsize,
              sensegame        = SENSEGAME_QUIET,
              sensitivity      = SENSITIVITY_DEFAULT,
              size             = 1,
              tonum,
              tolimit,
              wheremouse       = -1,
              wide             = 2,
              wp               = OUTOFRANGE,
              autosave         = TRUE,
              consoleopen      = FALSE,
              crippled         = FALSE,
              fullscreen       = FALSE,
              game             = FALSE,
              icons            = FALSE,
              inframe          = FALSE,
              limitrefresh     = FALSE,
              loginstructions  = FALSE,
              logmonitor       = FALSE,
              logreads         = FALSE,
              logwrites        = FALSE,
              loop             = TRUE,
              requirebutton    = FALSE,
              runtorastline    = FALSE,
              runtodma         = FALSE,
              runtoframe       = FALSE,
              runtointerrupt   = FALSE,
              showmenubar      = TRUE,
              showtitlebar     = TRUE,
              showtoolbar      = TRUE,
              showpointer      = TRUE,
              stretch          = TRUE,
              trace            = FALSE,
              usehacks         = TRUE,
              watchreads       = FALSE;
EXPORT SLONG  ax[2], // analog paddle X-coords
              ay[2]; // analog paddle Y-coords
EXPORT UBYTE* IOBuffer         = NULL;
EXPORT UBYTE  memflags[32768],
              sx[2], sy[2];
EXPORT UWORD  console[4],
              joystate,
              keypads[2][NUMKEYS],
              mirror_r[32768],
              mirror_w[32768];
EXPORT FLAG   guestplaying[14] = {FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE,
                                  FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE};

EXPORT const STRPTR autoname[3] =
{   "AUTOSAVE.EAS",
    "AUTOSAVE.INS",
    "AUTOSAVE.TVS"
};
EXPORT FLAG                cli        = FALSE;
#ifdef AMIGA
    EXPORT FLAG            absolute   = FALSE;
    EXPORT struct AJoyData AJoy;
#endif
#ifdef WIN32
    EXPORT FLAG            absolute   = TRUE;
    EXPORT ULONG           joy0       = FALSE;
#endif

EXPORT struct KeyTableStruct keytable[16] =
{   {   A_P1MIDDLEKEYS, 0x01, I_P1MIDDLEKEYS, 0x10, "0",  "+/2"    },
    {   A_P1LEFTKEYS,   0x08, I_P1LEFTKEYS,   0x80, "1",  "RCAS/D" },
    {   A_P1MIDDLEKEYS, 0x08, I_P1MIDDLEKEYS, 0x80, "2",  "WCAS/E" },
    {   A_P1RIGHTKEYS,  0x08, I_P1RIGHTKEYS,  0x80, "3",  "C/F"    },
    {   A_P1LEFTKEYS,   0x04, I_P1LEFTKEYS,   0x40, "4",  "BP/9"   },
    {   A_P1MIDDLEKEYS, 0x04, I_P1MIDDLEKEYS, 0x40, "5",  "REG/A"  },
    {   A_P1RIGHTKEYS,  0x04, I_P1RIGHTKEYS,  0x40, "6",  "8/B"    },
    {   A_P1LEFTKEYS,   0x02, I_P1LEFTKEYS,   0x20, "7",  "PC/5"   },
    {   A_P1MIDDLEKEYS, 0x02, I_P1MIDDLEKEYS, 0x20, "8",  "MEM/6"  },
    {   A_P1RIGHTKEYS,  0x02, I_P1RIGHTKEYS,  0x20, "9",  "4/7"    },
    {   A_P1LEFTKEYS,   0x01, I_P1LEFTKEYS,   0x10, "Cl", "-/1"    }, // 10
    {   A_P1RIGHTKEYS,  0x01, I_P1RIGHTKEYS,  0x10, "En", "0/3"    }, // 11
    {   A_P1PALLADIUM,  0x08, 0,                 0, "x1", "?"      }, // 12
    {   A_P1PALLADIUM,  0x04, 0,                 0, "x2", "?"      }, // 13
    {   A_P1PALLADIUM,  0x02, 0,                 0, "x3", "?"      }, // 14
    {   A_P1PALLADIUM,  0x01, 0,                 0, "x4", "?"      }  // 15
};

// IMPORTED VARIABLES-----------------------------------------------------

IMPORT       FLAG  aborting,
                   led;
IMPORT       ULONG cycles,
                   elapsed;
IMPORT       UBYTE psu,
                   psl,
                   r[7],
                   memory[32768],
                   screen[BOXWIDTH][BOXHEIGHT],
                   table_sz[256];
IMPORT       TEXT  hostname[256 + 1];
IMPORT       UWORD iar,
                   ie_spritedata[4],
                   page, // page select register
                   ras[8];
IMPORT       int   connected,
                   consoleopen,
                   interrupt,
                   overcalc,
                   step,
                   table_clockperiods[256],
                   voffset;
IMPORT       char  psubits[3 + 1],
                   pslbits[6 + 1];

IMPORT const char              ccstring[4][2 + 1];
IMPORT const int               from_a[8];
IMPORT const struct KindStruct filekind[KINDS];

#ifdef AMIGA
    IMPORT struct Gadget*  gadgets[GIDS + 1];
    IMPORT struct Menu*    MenuPtr;
    IMPORT struct Window  *MainWindowPtr,
                          *ControlsWindowPtr,
                          *MonitorWindowPtr;
    IMPORT UWORD           PointerData[6];
    IMPORT struct Screen*  ScreenPtr;
    IMPORT Object*         smallimage[SMALLIMAGES];
    IMPORT struct Library *AmigaGuideBase,
                          *SocketBase;
    IMPORT struct Catalog* CatalogPtr;
#endif
#ifdef WIN32
    IMPORT HMENU          MenuPtr,
                          PopUpMenuPtr;
    IMPORT HWND           ControlsWindowPtr,
                          MainWindowPtr,
                          MonitorWindowPtr,
                          hToolbar;
    IMPORT HDC            ControlsRastPtr;
    IMPORT int            titleheight,
                          winleftx;
    IMPORT UBYTE          joystickIDMapping[2];
    IMPORT int            CatalogPtr; // APTR doesn't work
#endif
#ifdef PROFILE
    IMPORT ULONG          profile[256];
#endif
#ifdef COVERAGE
    IMPORT UBYTE          coverage[32 * KILOBYTE];
#endif
#ifdef MODE_MHZ
    IMPORT ULONG          oldelapsed;
#endif

#ifdef AMIGA
    IMPORT struct Gadget
             // CPU chip...
             *RGad[7],
             *CCPtr,
             *SensePtr,
             *FlagPtr,
             *InterruptPtr,
             *IDCPtr,
             *RSPtr,
             *WCPtr,
             *OVFPtr,
             *COMPtr,
             *CPtr,
             *RASGad[8],
             *SPGad,
             *IARPtr,
             *ClockPtr,
             *FramePtr,
             *RasterGad,
             *TimePtr,
             // XVI chips...
             *SPRITEnAXYGad[4],
             *SPRITEnBXYGad[4],
             *monitorgad[MONITORGADS],
             // PSGs
             *PITCHAGad[2],
             *PITCHBGad[2],
             *PITCHCGad[2],
             *PERIODGad[2];
#endif
#ifdef WIN32
    IMPORT HWND
             // CPU chip...
             RGad[7],
             CCPtr,
             SensePtr,
             FlagPtr,
             InterruptPtr,
             IDCPtr,
             RSPtr,
             WCPtr,
             OVFPtr,
             COMPtr,
             CPtr,
             RASGad[8],
             SPGad,
             IARPtr,
             ClockPtr,
             FramePtr,
             RasterGad,
             TimePtr,
             // XVI chips...
             SPRITEnAXYGad[4],
             SPRITEnBXYGad[4],
             monitorgad[MONITORGADS],
             // PSGs
             PITCHAGad[2],
             PITCHBGad[2],
             PITCHCGad[2],
             PERIODGad[2];
#endif

// MODULE STRUCTURES------------------------------------------------------

MODULE const UWORD monitoraddr[MONITORGADS] =
{   I_CONSOLE,       //   0
    I_PITCH,         //   1
    I_SIZES,         //   2
    I_SCORECTRL,     //   3
    I_SCORELT,       //   4
    I_SCORERT,       //   5
    I_BGCOLOUR,      //   6
    I_BGCOLLIDE,     //   7
    I_SPRITECOLLIDE, //   8
    I_SPR01COLOURS,  //   9
    I_SPR23COLOURS,  //  10
    I_P1PADDLE,      //  11
    I_P1LEFTKEYS,    //  12
    I_P1MIDDLEKEYS,  //  13
    I_P1RIGHTKEYS,   //  14
    I_P2PADDLE,      //  15
    I_P2LEFTKEYS,    //  16
    I_P2MIDDLEKEYS,  //  17
    I_P2RIGHTKEYS,   //  18
    I_NOISE,         //  19
    E_AMPLITUDEA1,   //  20
    E_AMPLITUDEB1,   //  21
    E_AMPLITUDEC1,   //  22
    E_PITCHD1,       //  23
    E_ENABLE1,       //  24
    E_SHAPE1,        //  25
    E_PORTA1,        //  26
    E_PORTB1,        //  27
    E_AMPLITUDEA2,   //  28
    E_AMPLITUDEB2,   //  29
    E_AMPLITUDEC2,   //  30
    E_PITCHD2,       //  31
    E_ENABLE2,       //  32
    E_SHAPE2,        //  33
    E_PORTA2,        //  34
    E_PORTB2,        //  35
    0x1F0E,          //  36
    0x1F0F,          //  37
    0x1F1E,          //  38
    0x1F1F,          //  39
    0x1F4E,          //  40
    0x1F4F,          //  41
    0x1F50,          //  42
    0x1F51,          //  43
    0x1F52,          //  44
    0x1F53,          //  45
    0x1F54,          //  46
    0x1F55,          //  47
    0x1F56,          //  48
    0x1F57,          //  49
    0x1F58,          //  50
    0x1F59,          //  51
    0x1F5A,          //  52
    0x1F5B,          //  53
    0x1F5C,          //  54
    0x1F5D,          //  55
    0x1F5E,          //  56
    0x1F5F,          //  57
    0x1F60,          //  58
    0x1F61,          //  59
    0x1F62,          //  60
    0x1F63,          //  61
    0x1F64,          //  62
    0x1F65,          //  63
    0x1F66,          //  64
    0x1F67,          //  65
    0x1F68,          //  66
    0x1F69,          //  67
    0x1F6A,          //  68
    0x1F6B,          //  69
    0x1F6C,          //  70
    0x1F6D,          //  71
    0x1FAD,          //  72
    A_CONSOLE,       //  73
    A_RESOLUTION,    //  74
    A_BGCOLLIDE,     //  75
    A_P1PADDLE,      //  76
    A_P2PADDLE,      //  77
    A_PITCH,         //  78
    A_BGCOLOUR,      //  79
    A_SPRITECOLLIDE, //  80
    A_P1LEFTKEYS,    //  81
    A_P2LEFTKEYS,    //  82
    A_VOLUME,        //  83
    A_VSCROLL,       //  84
    A_SPRITES01CTRL, //  85
    A_P1MIDDLEKEYS,  //  86
    A_P2MIDDLEKEYS,  //  87
    A_CHARLINE,      //  88
    A_SPRITES23CTRL, //  89
    A_P1RIGHTKEYS,   //  90
    A_P2RIGHTKEYS,   //  91
    A_P1PALLADIUM,   //  92
    A_P2PALLADIUM,   //  93
    0x18D0,          //  94
    0x18D1,          //  95
    0x18D2,          //  96
    0x18D3,          //  97
    0x18D4,          //  98
    0x18D5,          //  99
    0x18D6,          // 100
    0x18D7,          // 101
    0x18D8,          // 102
    0x18D9,          // 103
    0x18DA,          // 104
    0x18DB,          // 105
    0x18DC,          // 106
    0x18DD,          // 107
    0x18DE,          // 108
    0x18DF,          // 109
    0x18E0,          // 110
    0x18E1,          // 111
    0x18E2,          // 112
    0x18E3,          // 113
    0x18E4,          // 114
    0x18E5,          // 115
    0x18E6,          // 116
    0x18E7,          // 117
    0x18E8,          // 118
    0x18E9,          // 119
    0x18EA,          // 120
    0x18EB,          // 121
    0x18EC,          // 122
    0x18ED,          // 123
    0x18EE,          // 124
    0x18EF,          // 125
    0x18F8,          // 126
    0x18F9,          // 127
    0x18FA,          // 128
    0x18FB,          // 129
    0x1AD0,          // 130
    0x1AD1,          // 131
    0x1AD2,          // 132
    0x1AD3,          // 133
    0x1AD4,          // 134
    0x1AD5,          // 135
    0x1AD6,          // 136
    0x1AD7,          // 137
    0x1AD8,          // 138
    0x1AD9,          // 139
    0x1ADA,          // 140
    0x1ADB,          // 141
    0x1ADC,          // 142
    0x1ADD,          // 143
    0x1ADE,          // 144
    0x1ADF,          // 145
    0x1AE0,          // 146
    0x1AE1,          // 147
    0x1AE2,          // 148
    0x1AE3,          // 149
    0x1AE4,          // 150
    0x1AE5,          // 151
    0x1AE6,          // 152
    0x1AE7,          // 153
    0x1AE8,          // 154
    0x1AE9,          // 155
    0x1AEA,          // 156
    0x1AEB,          // 157
    0x1AEC,          // 158
    0x1AED,          // 159
    0x1AEE,          // 160
    0x1AEF,          // 161
    0x1AF0,          // 162
    0x1AF1,          // 163
    0x1AF2,          // 164
    0x1AF3,          // 165
    0x1AF4,          // 166
    0x1AF5,          // 167
    0x1AF6,          // 168
    0x1AF7,          // 169
    0x1AF8,          // 170
    0x1AF9,          // 171
    0x1AFA,          // 172
    0x1AFB,          // 173
    0x1AFC,          // 174
    0x1AFD,          // 175
    0x1AFE,          // 176
    0x1AFF           // 177
};

#define A_EQUIVALENTS          29 // counting from 1
MODULE struct
{   char*  text;
    UWORD  number;
} a_equivalents[A_EQUIVALENTS] = {
{ "P1PADDLE"     , A_P1PADDLE      }, //  0
{ "P2PADDLE"     , A_P2PADDLE      },
{ "SPRITECOLLIDE", A_SPRITECOLLIDE },
{ "BGCOLLIDE"    , A_BGCOLLIDE     },
{ "SPRITES01CTRL", A_SPRITES01CTRL },
{ "SPRITES23CTRL", A_SPRITES23CTRL }, //  5
{ "BGCOLOUR"     , A_BGCOLOUR      },
{ "RESOLUTION"   , A_RESOLUTION    },
{ "CONSOLE"      , A_CONSOLE       },
{ "P2PALLADIUM"  , A_P2PALLADIUM   },
{ "P2RIGHTKEYS"  , A_P2RIGHTKEYS   }, // 10
{ "P2MIDDLEKEYS" , A_P2MIDDLEKEYS  },
{ "P2LEFTKEYS"   , A_P2LEFTKEYS    },
{ "P1PALLADIUM"  , A_P1PALLADIUM   },
{ "P1RIGHTKEYS"  , A_P1RIGHTKEYS   },
{ "P1MIDDLEKEYS" , A_P1MIDDLEKEYS  }, // 15
{ "P1LEFTKEYS"   , A_P1LEFTKEYS    },
{ "CHARLINE"     , A_CHARLINE      },
{ "VOLUME"       , A_VOLUME        },
{ "PITCH"        , A_PITCH         },
{ "VSCROLL"      , A_VSCROLL       }, // 20
{ "SPRITE3X"     , A_SPRITE3X      },
{ "SPRITE3Y"     , A_SPRITE3Y      },
{ "SPRITE2X"     , A_SPRITE2X      },
{ "SPRITE2Y"     , A_SPRITE2Y      },
{ "SPRITE1X"     , A_SPRITE1X      }, // 25
{ "SPRITE1Y"     , A_SPRITE1Y      },
{ "SPRITE0X"     , A_SPRITE0X      },
{ "SPRITE0Y"     , A_SPRITE0Y      }  // 28
};

#define I_EQUIVALENTS          41 // counting from 1
MODULE struct
{   char*  text;
    UWORD  number;
} i_equivalents[I_EQUIVALENTS] = {
{ "P1PADDLE"     , I_P1PADDLE      }, //  0
{ "P2PADDLE"     , I_P2PADDLE      },
{ "SPRITECOLLIDE", I_SPRITECOLLIDE },
{ "BGCOLLIDE"    , I_BGCOLLIDE     },
{ "SPR01COLOURS" , I_SPR01COLOURS  },
{ "SPR23COLOURS" , I_SPR23COLOURS  }, //  5
{ "BGCOLOUR"     , I_BGCOLOUR      },
{ "CONSOLE"      , I_CONSOLE       },
{ "P2RIGHTKEYS"  , I_P2RIGHTKEYS   },
{ "P2MIDDLEKEYS" , I_P2MIDDLEKEYS  },
{ "P2LEFTKEYS"   , I_P2LEFTKEYS    }, // 10
{ "P1RIGHTKEYS"  , I_P1RIGHTKEYS   },
{ "P1MIDDLEKEYS" , I_P1MIDDLEKEYS  },
{ "P1LEFTKEYS"   , I_P1LEFTKEYS    },
{ "PITCH"        , I_PITCH         },
{ "SPRITE3AX"    , I_SPRITE3AX     }, // 15
{ "SPRITE3AY"    , I_SPRITE3AY     },
{ "SPRITE2AX"    , I_SPRITE2AX     },
{ "SPRITE2AY"    , I_SPRITE2AY     },
{ "SPRITE1AX"    , I_SPRITE1AX     },
{ "SPRITE1AY"    , I_SPRITE1AY     }, // 20
{ "SPRITE0AX"    , I_SPRITE0AX     },
{ "SPRITE0AY"    , I_SPRITE0AY     },
{ "SPRITE3BX"    , I_SPRITE3BX     },
{ "SPRITE3BY"    , I_SPRITE3BY     },
{ "SPRITE2BX"    , I_SPRITE2BX     }, // 25
{ "SPRITE2BY"    , I_SPRITE2BY     },
{ "SPRITE1BX"    , I_SPRITE1BX     },
{ "SPRITE1BY"    , I_SPRITE1BY     },
{ "SPRITE0BX"    , I_SPRITE0BX     },
{ "SPRITE0BY"    , I_SPRITE0BY     }, // 30
{ "SCORECTRL"    , I_SCORECTRL     },
{ "SCORELT"      , I_SCORELT       },
{ "SCORERT"      , I_SCORERT       },
{ "HORIZ1"       , I_HORIZ1        },
{ "HORIZ2"       , I_HORIZ2        }, // 35
{ "HORIZ3"       , I_HORIZ3        },
{ "HORIZ4"       , I_HORIZ4        },
{ "HORIZ5"       , I_HORIZ5        },
{ "SIZES"        , I_SIZES         },
{ "NOISE"        , I_NOISE         }  // 40
// we don't support eg. VERT, as it is really just the start of a larger area.
};

#define E_EQUIVALENTS         255 // counting from 1 (221 monitor labels, 34 hardware registers)
MODULE struct
{   char*  text;
    UWORD  number;
} e_equivalents[E_EQUIVALENTS] = {
{ "RESET"        ,   0x0           },
{ "INTRR"        ,   0x3           },
{ "BVEC1"        ,   0x6           }, // ROM data?
{ "BVEC2"        ,   0x8           }, // ROM data?
{ "START"        ,   0xA           },
{ "INIT0"        ,  0x23           },
{ "INIT1"        ,  0x28           },
{ "START1"       ,  0x38           },
{ "MWAIT0"       ,  0x45           },
{ "MWAIT1"       ,  0x47           },
{ "MONINT"       ,  0x4B           },
{ "FREFR0"       ,  0x4E           },
{ "FREEF1"       ,  0x57           },
{ "FILELN"       ,  0x71           },
{ "FREFR1"       ,  0x77           },
{ "FREFR2"       ,  0x7A           }, // display last line only
{ "MPOBJ1"       ,  0xAD           }, // ROM data
{ "MPOBJ2"       ,  0xB1           }, // ROM data
{ "MPOBJ3"       ,  0xB5           }, // ROM data
{ "MPOBJ4"       ,  0xB9           }, // ROM data
{ "MKEY"         ,  0xBD           },
{ "MKEY1"        ,  0xCA           },
{ "MKEY2"        ,  0xCD           },
{ "MKEYR"        ,  0xD4           },
{ "MKEY3"        ,  0xDE           },
{ "MKEY5"        ,  0xEF           },
{ "MKEY6"        ,  0xFD           },
{ "FNOP"         , 0x10C           },
{ "FMEM"         , 0x10D           },
{ "FBK"          , 0x110           },
{ "FREG"         , 0x113           },
{ "FINIT"        , 0x116           },
{ "FPC"          , 0x119           },
{ "FRC"          , 0x11C           },
{ "FWC"          , 0x11F           },
{ "KBFTBL"       , 0x122           }, // ROM data
{ "SAVE"         , 0x13E           },
{ "SAVE1"        , 0x161           }, // initiate PVI
{ "CLROBJ"       , 0x16E           }, // clear objects
{ "PVIPAR"       , 0x177           }, // ROM data
{ "KBSCAN"       , 0x181           }, // keyboard scan
{ "KBSC0"        , 0x18A           },
{ "KBSC12"       , 0x193           },
{ "IA00"         , 0x195           },
{ "IA01"         , 0x198           },
{ "IA02"         , 0x19B           },
{ "IA03"         , 0x1A0           },
{ "IA04"         , 0x1A3           },
{ "KBSC5"        , 0x1B6           },
{ "KBSC6"        , 0x1B8           },
{ "KBSC7"        , 0x1BD           },
{ "KBSC1"        , 0x1C0           },
{ "KBSC2"        , 0x1C3           },
{ "KBSC3"        , 0x1CF           },
{ "KBSC8"        , 0x1D9           },
{ "KBSC11"       , 0x1DC           },
{ "KBSC10"       , 0x1DF           },
{ "KBSC4"        , 0x1E3           },
{ "KBSC9"        , 0x1E9           },
{ "KBSBR"        , 0x1EE           },
{ "KBSBR0"       , 0x1F4           },
{ "KBSBR4"       , 0x1FB           },
{ "KBSBR2"       , 0x1FD           },
{ "KBSBR1"       , 0x1FF           },
{ "LINE"         , 0x20E           }, // translate MLINE to display
{ "LINE2"        , 0x212           },
{ "LINE07"       , 0x21E           },
{ "LINE6"        , 0x234           },
{ "LINE7"        , 0x23A           },
{ "LINE8"        , 0x23C           },
{ "CHARPS"       , 0x24B           },
{ "LINE05"       , 0x255           },
{ "LINE09"       , 0x25B           },
{ "LINE06"       , 0x25F           },
{ "IMGTBL"       , 0x27B           }, // ROM data
{ "SCROLL"       , 0x2CF           }, // scroll, 8 spaces to MLINE
{ "SCROL1"       , 0x2D1           },
{ "SCROL3"       , 0x2D9           }, // load 8 spaces to MLINE
{ "SCROL4"       , 0x2DB           },
{ "NEWMES"       , 0x2E3           }, // scroll, message to MLINE
{ "MESSAG"       , 0x2EA           }, // output message to MLINE
{ "MESSL"        , 0x2F5           }, // ROM data
{ "ASMDAT"       , 0x31D           }, // assemble data
{ "ASSMC"        , 0x328           },
{ "ASSMC1"       , 0x332           },
{ "STRDAT"       , 0x339           },
{ "RESDAT"       , 0x340           },
{ "LOADAT"       , 0x346           },
{ "OUTADD"       , 0x34E           }, // output double byte to MLINE
{ "OUTADR"       , 0x350           }, // output double byte to MLINE
{ "OUTBYT"       , 0x354           }, // output single byte to MLINE
{ "NIBL"         , 0x35E           }, // split register
{ "OUTNAD"       , 0x368           },
{ "OUTADS"       , 0x36C           },
{ "SPACES"       , 0x377           },
{ "STDATA"       , 0x381           },
{ "STDAT1"       , 0x392           },
{ "STDAT2"       , 0x39B           },
{ "INCADR"       , 0x39F           }, // increment double byte
{ "REG01"        , 0x3B0           },
{ "REG1"         , 0x3B8           },
{ "REG7"         , 0x3D1           },
{ "REG2"         , 0x3D6           },
{ "IAREG"        , 0x3DB           },
{ "REG3"         , 0x3E0           },
{ "REG4"         , 0x3F4           },
{ "REG5"         , 0x400           },
{ "MEM0"         , 0x40C           },
{ "MEM1"         , 0x415           },
{ "MEM2"         , 0x41C           },
{ "MEM3"         , 0x42E           },
{ "MEM4"         , 0x43E           },
{ "MEM7"         , 0x44D           },
{ "MEM9"         , 0x452           },
{ "MEM8"         , 0x465           },
{ "MEM5"         , 0x467           },
{ "MEM6"         , 0x46A           },
{ "MEMFS3"       , 0x47D           },
{ "MEMBYT"       , 0x480           }, // output single byte to screen
{ "MEMLIN"       , 0x483           },
{ "MEMFS7"       , 0x486           },
{ "FMEM1"        , 0x495           },
{ "FMEM3"        , 0x499           },
{ "FMEM2"        , 0x4A2           },
{ "BKSC"         , 0x4A9           },
{ "FBK1"         , 0x4C3           },
{ "FBK2"         , 0x4CB           },
{ "FBK3"         , 0x4D1           },
{ "IABK"         , 0x4D6           },
{ "FBK4"         , 0x4DB           },
{ "FBK5"         , 0x4DD           },
{ "FBK20"        , 0x4F5           },
{ "FBK21"        , 0x50A           },
{ "GO0"          , 0x50E           },
{ "GO4"          , 0x515           },
{ "GO3"          , 0x529           }, // output double byte to screen
{ "GOLIN"        , 0x52C           },
{ "GO1"          , 0x52F           }, // assemble data on screen
{ "GO2"          , 0x532           },
{ "ZBRR1"        , 0x537           }, // ROM data?
{ "ZBRR2"        , 0x539           }, // ROM data?
{ "UMODE"        , 0x53B           }, // ROM data
{ "STRTUP"       , 0x540           },
{ "BREAK1"       , 0x594           },
{ "BREAKM"       , 0x5A4           },
{ "BREAK2"       , 0x5B2           },
{ "PCEQBK"       , 0x5C4           },
{ "BRKSBR"       , 0x5CD           },
{ "BRKSB1"       , 0x5E0           },
{ "WCAS0"        , 0x5E8           },
{ "WCAS1"        , 0x5F3           },
{ "WCAS2"        , 0x5F7           },
{ "WCAS3"        , 0x602           }, // scroll, message to screen
{ "WCALIN"       , 0x605           },
{ "WCAS4"        , 0x608           },
{ "WCAS6"        , 0x610           },
{ "WCAS2A"       , 0x626           },
{ "WCAS7"        , 0x628           },
{ "WCAS8"        , 0x63C           },
{ "WCASF1"       , 0x642           },
{ "WCAS9"        , 0x64A           },
{ "BLOCK"        , 0x655           },
{ "WCASB"        , 0x671           },
{ "WCAS10"       , 0x681           },
{ "IBOUT"        , 0x690           },
{ "WCAS11"       , 0x699           },
{ "WCAS12"       , 0x6A4           },
{ "WCAS13"       , 0x6B8           },
{ "BOUT"         , 0x6D0           },
{ "COUT"         , 0x6D3           },
{ "COUT1"        , 0x6DA           },
{ "COUT2"        , 0x6E1           },
{ "COUT3"        , 0x6E3           },
{ "ONE"          , 0x6EA           },
{ "ZERO"         , 0x6EE           },
{ "DELAY"        , 0x6F2           },
{ "PULSE"        , 0x6F7           }, // delay
{ "CHIN"         , 0x709           },
{ "CHI1"         , 0x70D           },
{ "CHI2"         , 0x710           },
{ "CHI3"         , 0x712           },
{ "CHI4"         , 0x71B           },
{ "CBCC"         , 0x726           },
{ "RBIT"         , 0x731           },
{ "RBIT1"        , 0x734           },
{ "RBIT2"        , 0x739           },
{ "RBIT3"        , 0x73B           },
{ "RBIT4"        , 0x73D           },
{ "RBIT5"        , 0x747           },
{ "RCAS0"        , 0x758           },
{ "RCAS7"        , 0x764           },
{ "CHLINE"       , 0x76A           },
{ "CHLIN1"       , 0x76D           },
{ "RCAS8"        , 0x774           },
{ "RCAS2"        , 0x778           },
{ "RCAS1"        , 0x788           },
{ "ICHIN1"       , 0x797           },
{ "IOBJA"        , 0x79E           },
{ "RCAS3"        , 0x7A8           },
{ "RCAS4"        , 0x7B0           },
{ "RCAS9"        , 0x7CE           },
{ "ICHIN"        , 0x7D1           },
{ "RCAS5"        , 0x7D9           },
{ "RCAS6"        , 0x7EE           },
{ "RCASER"       , 0x7F4           },
{ "MONOB"        , E_MONOB         }, // RAM data
{ "MLINE"        , E_MLINE         }, // RAM data
{ "FSEQ"         , E_FSEQ          }, // RAM data (aka SILENC)
{ "SILENC"       , E_SILENC        }, // RAM data (aka FSEQ)
{ "ENTM"         , E_ENTM          }, // RAM data
{ "MFUNC"        , E_MFUNC         }, // RAM data
{ "MSCR"         , E_MSCR          }, // RAM data
{ "RKBST"        , E_RKBST         }, // RAM data
{ "LKBST"        , E_LKBST         }, // RAM data
{ "MKBST"        , E_MKBST         }, // RAM data
{ "FSCRM"        , E_FSCRM         }, // RAM data
{ "REGM"         , E_REGM          }, // RAM data
{ "BK1"          , E_BK1           }, // RAM data
{ "BK2"          , E_BK2           }, // RAM data
{ "INTADR"       , E_INTADR        }, // RAM data
{ "PC"           , E_PC            }, // RAM data
{ "PITCHA1_L"    , E_PITCHA1_L     },
{ "PITCHA1_H"    , E_PITCHA1_H     },
{ "PITCHB1_L"    , E_PITCHB1_L     },
{ "PITCHB1_H"    , E_PITCHB1_H     },
{ "PITCHC1_L"    , E_PITCHC1_L     },
{ "PITCHC1_H"    , E_PITCHC1_H     },
{ "PITCHD1"      , E_PITCHD1       },
{ "ENABLE1"      , E_ENABLE1       },
{ "AMPLITUDEA1"  , E_AMPLITUDEA1   },
{ "AMPLITUDEB1"  , E_AMPLITUDEB1   },
{ "AMPLITUDEC1"  , E_AMPLITUDEC1   },
{ "PERIOD1_L"    , E_PERIOD1_L     },
{ "PERIOD1_H"    , E_PERIOD1_H     },
{ "SHAPE1"       , E_SHAPE1        },
{ "PORTA1"       , E_PORTA1        },
{ "PORTB1"       , E_PORTB1        },
{ "PITCHA2_L"    , E_PITCHA2_L     },
{ "PITCHA2_H"    , E_PITCHA2_H     },
{ "PITCHB2_L"    , E_PITCHB2_L     },
{ "PITCHB2_H"    , E_PITCHB2_H     },
{ "PITCHC2_L"    , E_PITCHC2_L     },
{ "PITCHC2_H"    , E_PITCHC2_H     },
{ "PITCHD2"      , E_PITCHD2       },
{ "ENABLE2"      , E_ENABLE2       },
{ "AMPLITUDEA2"  , E_AMPLITUDEA2   },
{ "AMPLITUDEB2"  , E_AMPLITUDEB2   },
{ "AMPLITUDEC2"  , E_AMPLITUDEC2   },
{ "PERIOD2_L"    , E_PERIOD2_L     },
{ "PERIOD2_H"    , E_PERIOD2_H     },
{ "SHAPE2"       , E_SHAPE2        },
{ "PORTA2"       , E_PORTA2        },
{ "PORTB2"       , E_PORTB2        },
{ "CASIN"        , E_CASIN         },
{ "CASOUT"       , E_CASOUT        }
};

/* These are listed as useful routines in the "TV Games Computer" book but
have no official name/label:
  {  0x55, "DISPLAY6LINES" }, // "display six lines"
  {  0x9E, "CLEARDUPS"     }, // "clear duplicates"
  {  0xA0, "LOADVODDUPS"   }, // "load VOD duplicates"
  { 0x16F, "LOADOBJECTS"   }, // "load objects"
  { 0x379, "LDSPCS2MLINE"  }, // "load spaces to MLINE"
  { 0x42B, "ADDRDTA2MLINE" }, // "address + data to MLINE"
  { 0x470, "OUT3BYTS2SCRN" }, // "output 3 bytes to screen"

We don't support eg. RANDOM2, what is the point?
*/

MODULE const STRPTR opcodename[256] = {
"-------", // $00
"LODZ r1",
"LODZ r2",
"LODZ r3",
"LODI,r0",
"LODI,r1",
"LODI,r2",
"LODI,r3",
"LODR,r0",
"LODR,r1",
"LODR,r2",
"LODR,r3",
"LODA,r0",
"LODA,r1",
"LODA,r2",
"LODA,r3",
"-------", // $10
"-------",
"SPSU",
"SPSL",
"RETC,eq",
"RETC,gt",
"RETC,lt",
"RETC,un",
"BCTR,eq",
"BCTR,gt",
"BCTR,lt",
"BCTR,un",
"BCTA,eq",
"BCTA,gt",
"BCTA,lt",
"BCTA,un",
"EORZ r0", // $20
"EORZ r1",
"EORZ r2",
"EORZ r3",
"EORI,r0",
"EORI,r1",
"EORI,r2",
"EORI,r3",
"EORR,r0",
"EORR,r1",
"EORR,r2",
"EORR,r3",
"EORA,r0",
"EORA,r1",
"EORA,r2",
"EORA,r3",
"REDC,r0", // $30
"REDC,r1",
"REDC,r2",
"REDC,r3",
"RETE,eq",
"RETE,gt",
"RETE,lt",
"RETE,un",
"BSTR,eq",
"BSTR,gt",
"BSTR,lt",
"BSTR,un",
"BSTA,eq",
"BSTA,gt",
"BSTA,lt",
"BSTA,un",
"HALT", // $40
"ANDZ r1",
"ANDZ r2",
"ANDZ r3",
"ANDI,r0",
"ANDI,r1",
"ANDI,r2",
"ANDI,r3",
"ANDR,r0",
"ANDR,r1",
"ANDR,r2",
"ANDR,r3",
"ANDA,r0",
"ANDA,r1",
"ANDA,r2",
"ANDA,r3",
"RRR,r0", // $50
"RRR,r1",
"RRR,r2",
"RRR,r3",
"REDE,r0",
"REDE,r1",
"REDE,r2",
"REDE,r3",
"BRNR,r0",
"BRNR,r1",
"BRNR,r2",
"BRNR,r3",
"BRNA,r0",
"BRNA,r1",
"BRNA,r2",
"BRNA,r3",
"IORZ r0", // $60
"IORZ r1",
"IORZ r2",
"IORZ r3",
"IORI,r0",
"IORI,r1",
"IORI,r2",
"IORI,r3",
"IORR,r0",
"IORR,r1",
"IORR,r2",
"IORR,r3",
"IORA,r0",
"IORA,r1",
"IORA,r2",
"IORA,r3",
"REDD,r0", // $70
"REDD,r1",
"REDD,r2",
"REDD,r3",
"CPSU",
"CPSL",
"PPSU",
"PPSL",
"BSNR,r0",
"BSNR,r1",
"BSNR,r2",
"BSNR,r3",
"BSNA,r0",
"BSNA,r1",
"BSNA,r2",
"BSNA,r3",
"ADDZ r0", // $80
"ADDZ r1",
"ADDZ r2",
"ADDZ r3",
"ADDI,r0",
"ADDI,r1",
"ADDI,r2",
"ADDI,r3",
"ADDR,r0",
"ADDR,r1",
"ADDR,r2",
"ADDR,r3",
"ADDA,r0",
"ADDA,r1",
"ADDA,r2",
"ADDA,r3",
"-------", // $90
"-------",
"LPSU",
"LPSL",
"DAR,r0",
"DAR,r1",
"DAR,r2",
"DAR,r3",
"BCFR,eq",
"BCFR,gt",
"BCFR,lt",
"ZBRR",
"BCFA,eq",
"BCFA,gt",
"BCFA,lt",
"BXA,r3",
"SUBZ r0", // $A0
"SUBZ r1",
"SUBZ r2",
"SUBZ r3",
"SUBI,r0",
"SUBI,r1",
"SUBI,r2",
"SUBI,r3",
"SUBR,r0",
"SUBR,r1",
"SUBR,r2",
"SUBR,r3",
"SUBA,r0",
"SUBA,r1",
"SUBA,r2",
"SUBA,r3",
"WRTC,r0", // $B0
"WRTC,r1",
"WRTC,r3",
"WRTC,r3",
"TPSU",
"TPSL",
"-------",
"-------",
"BSFR",
"BSFR",
"BSFR",
"ZBSR",
"BSFA",
"BSFA",
"BSFA",
"BSXA,r3",
"NOP", // $C0
"STRZ r1",
"STRZ r2",
"STRZ r3",
"-------",
"-------",
"-------",
"-------",
"STRR,r0",
"STRR,r1",
"STRR,r2",
"STRR,r3",
"STRA,r0",
"STRA,r1",
"STRA,r2",
"STRA,r3",
"RRL,r0", // $D0
"RRL,r1",
"RRL,r2",
"RRL,r3",
"WRTE,r0",
"WRTE,r1",
"WRTE,r2",
"WRTE,r3",
"BIRR,r0",
"BIRR,r1",
"BIRR,r2",
"BIRR,r3",
"BIRA,r0",
"BIRA,r1",
"BIRA,r2",
"BIRA,r3",
"COMZ,r0", // $E0
"COMZ,r1",
"COMZ,r2",
"COMZ,r3",
"COMI,r0",
"COMI,r1",
"COMI,r2",
"COMI,r3",
"COMR,r0",
"COMR,r1",
"COMR,r2",
"COMR,r3",
"COMA,r0",
"COMA,r1",
"COMA,r2",
"COMA,r3",
"WRTD,r0", // $F0
"WRTD,r1",
"WRTD,r2",
"WRTD,r3",
"TMI,r0",
"TMI,r1",
"TMI,r2",
"TMI,r3",
"BDRR,r0",
"BDRR,r1",
"BDRR,r2",
"BDRR,r3",
"BDRA,r0",
"BDRA,r1",
"BDRA,r2",
"BDRA,r3"  // $FF
};

EXPORT UWORD default_console[4] =
{   SCAN_F1,
    SCAN_F2,
    SCAN_F3,
    SCAN_F4
};

#ifdef AMIGA
EXPORT UWORD default_keypads[2][NUMKEYS] =
{ { SCAN_SPACEBAR,
    SCAN_A1, SCAN_A2, SCAN_A3,
    SCAN_Q,  SCAN_W,  SCAN_E,
    SCAN_A,  SCAN_S,  SCAN_D,
    SCAN_Z,  SCAN_X,  SCAN_C,
    SCAN_A4, SCAN_R,  SCAN_F,  SCAN_V,
    SCAN_I , SCAN_K,  SCAN_J,  SCAN_L,
    SCAN_A5, SCAN_A6, SCAN_A7
  },
  /* ( ) / *
     7 8 9 -
     4 5 6 +
     1 2 3 E
     -0- . n */
  { SCAN_N0,
    SCAN_OB, SCAN_CB, SCAN_SL,
    SCAN_N7, SCAN_N8, SCAN_N9,
    SCAN_N4, SCAN_N5, SCAN_N6,
    SCAN_N1, SCAN_N2, SCAN_N3,
    SCAN_AS, SCAN_MI, SCAN_PL, SCAN_NE,
    SCAN_UP, SCAN_DOWN, SCAN_LEFT, SCAN_RIGHT,
    SCAN_ND, SCAN_DEL,SCAN_HELP
} };
#endif
#ifdef WIN32
EXPORT UWORD default_keypads[2][NUMKEYS] =
{ { SCAN_SPACEBAR,                             //  0     (1st firebutton)
    SCAN_A1, SCAN_A2, SCAN_A3,                 //  1.. 3 (1st row)
    SCAN_Q,  SCAN_W,  SCAN_E,                  //  4.. 6 (2nd row)
    SCAN_A,  SCAN_S,  SCAN_D,                  //  7.. 9 (3rd row)
    SCAN_Z,  SCAN_X,  SCAN_C,                  // 10..12 (4th row)
    SCAN_A4, SCAN_R,  SCAN_F,  SCAN_V,         // 13..16 (Palladium keys)
    SCAN_I , SCAN_K,  SCAN_J,  SCAN_L,         // 17..20 (paddles)
    SCAN_A5, SCAN_A6, SCAN_A7                  // 21..23 (2nd..4th firebuttons)
  },
  /* N / * -
     7 8 9 +
     4 5 6 +
     1 2 3 E
     -0- . n */
  { SCAN_N0,                                   //  0     (1st firebutton)
    SCAN_NL, SCAN_SL, SCAN_AS,                 //  1.. 3 (1st row)
    SCAN_N7, SCAN_N8, SCAN_N9,                 //  4.. 6 (2nd row)
    SCAN_N4, SCAN_N5, SCAN_N6,                 //  7.. 9 (3rd row)
    SCAN_N1, SCAN_N2, SCAN_N3,                 // 10..12 (4th row)
    SCAN_MI, SCAN_PL, SCAN_NE, SCAN_ND,        // 13..16 (Palladium keys)
    SCAN_UP, SCAN_DOWN, SCAN_LEFT, SCAN_RIGHT, // 17..20 (paddles)
    SCAN_DEL,SCAN_END,SCAN_PGDN                // 21..23 (2nd..4th firebuttons)
} };
#endif

/* CD32 gamepads are like this:

Red    = Select  = 1st = bottom left     GY
Blue   = Stop    = 2nd = bottom right    RB
Yellow = Repeat  = 3rd = top    right
Green  = Shuffle = 4th = top    left

Arcadia games with "missing" buttons:
BOWLING:
    LBUTTON_7,      // KEY_LEFT  (hook left)
    LBUTTON_9,      // KEY_RIGHT (hook right)
    LBUTTON_1,      // KEY_L     (review score)
BLACKJAC:
    5, 6, 7, 8, 9, 0, Clear, Enter
BRAINQUI:
    5, 6, 7, 8, 9, 0, Clear
CAPTURE:
    5, 7, 9, 0
GOLF:
    Score, 3W, 3I, 5W, 5I, 7I, 9I.
HORSERAC:
    5, 6, 7, 8, 9, 0, Clear, Enter
JTRON:
    LBUTTON_5,      // KEY_UP
    LBUTTON_8,      // KEY_DOWN
    LBUTTON_7,      // KEY_LEFT
    LBUTTON_9,      // KEY_RIGHT
OCEANBAT:
    2 ships
SMISSION:
    Bridge, Launch Droneship, Steer Droneship
TANKSALO:
    Warp base
Also 2DSOCCER, BASEBALL, BASKETBA, BOXING, STARCHES.

Known:
 all dumped ROMs, good and bad, including overdumps
 enhanced versions of games
 non-obsolete compatibility-patched versions of games
 latest versions of homebrew games
 latest versions of homebrew examples
 available older versions of homebrews
 ESS software (games and utilities)
 TVGC book programs
 obsolete compatibility-patched versions of games
  (except unavailable older versions thereof)
not known:
 trimmed dumps
 unavailable older versions of homebrews
 trimmed/overdumped homebrews
*/

#define KNOWNGAMES    446 // counting from 1
#define GOLFPOS         0
MODULE struct
{   UBYTE  machine,
           memmap;
    ULONG  crc32;
    UWORD  size;
    UBYTE  key1,
           key2,
           key3,
           key4;
    FLAG   analog;
    UBYTE  sensitivity;
    FLAG   autofire,
           demultiplex,
           swapped,
           flagline;
    UBYTE  downframes,
           totalframes;
    SWORD  fudge;
    UBYTE  cpl;
    FLAG   forceinterrupt;
    UBYTE  p1sprite,
           p2sprite;
    STRPTR name;
// We could conceivably want to also do warp, etc. in this manner.
// Games marked with '*' require ascertainment of p1sprite, p2sprite.
} known[KNOWNGAMES] = {
// Emerson Arcadia 2001 games (58)
  { ARCADIA , MEMMAP_G, 0xA23C7A01,  6 * KILOBYTE,  2,  4,  0, 10, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Golf"                                           }, //*GOLF (must be at GOLFPOS!)
  { ARCADIA , MEMMAP_G, 0xB0B98A92,  4 * KILOBYTE,  2,  5,  8,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Soccer"                                         }, //*2DSOCCER
  { ARCADIA , MEMMAP_G, 0x4DA68DF8,  8 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 1, 0, "3D Soccer (Emerson version?)"                   }, // 3DSOCCEA
  { ARCADIA , MEMMAP_G, 0x1B5BE22A,  8 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 1, 0, "3D Soccer (Tele-Fever version)"                 }, // 3DSOCCEB
  { ARCADIA , MEMMAP_G, 0x1F65B21E,  4 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, TRUE , FALSE, FALSE, TRUE ,  1, 16,     0,       0, FALSE, 2, 4, "Alien Invaders"                                 }, // ALIENINV
  { ARCADIA , MEMMAP_G, 0xDDCF5B2C,  4 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, TRUE , TRUE , FALSE, TRUE ,  1,  2,     0,       0, FALSE, 1, 4, "Astro Invader"                                  }, // ASTROINV (1:1:1:1) - player bullets
  { ARCADIA , MEMMAP_G, 0x99AC48CA,  4 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 0, 1, "Auto Race"                                      }, // AUTORACE
  { ARCADIA , MEMMAP_G, 0x1CF08155,  4 * KILOBYTE, 10, 11,  0,  5, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 3, 3, "Baseball"                                       }, // BASEBALL
  { ARCADIA , MEMMAP_G, 0xE93938F6,  4 * KILOBYTE,  4,  1,  7,  5, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Basketball"                                     }, //*BASKETBA
  { ARCADIA , MEMMAP_G, 0x6818C8BD,  4 * KILOBYTE,  2,  1, 10,  2, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Battle"                                         }, //*BATTLE
  { ARCADIA , MEMMAP_G, 0x0930CE04,  4 * KILOBYTE,  1,  2,  3,  4, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Blackjack/Poker"                                }, //*BLACKJAC
  { ARCADIA , MEMMAP_G, 0xF3C08E19,  4 * KILOBYTE,  2,  8, 10, 11, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "3D Bowling"                                     }, // BOWLING
  { ARCADIA , MEMMAP_G, 0x9240891B,  4 * KILOBYTE,  2,  7,  9,  1, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Boxing"                                         }, //*BOXING
  { ARCADIA , MEMMAP_G, 0xCA58B3F8,  4 * KILOBYTE,  1,  2,  3,  4, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Brain Quiz"                                     }, //*BRAINQUI
  { ARCADIA , MEMMAP_G, 0x33C348C1,  4 * KILOBYTE,  2,  5,  2,  5, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Breakaway"                                      }, //*BREAKAWA
  { ARCADIA , MEMMAP_G, 0xAEB803E4,  2 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Capture"                                        }, //*CAPTURE
  { ARCADIA , MEMMAP_G, 0xFD2CF496,  4 * KILOBYTE,  2,  1,  3,  2, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Cat Trax"                                       }, //*CATTRAX
  { ARCADIA , MEMMAP_G, 0x086EBC8C,  4 * KILOBYTE,  2,  2,  2,  2, TRUE , SENSITIVITY_HIGH   , FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 2, 3, "Circus"                                         }, // CIRCUS
  { ARCADIA , MEMMAP_G, 0xA6BFFA9E,  4 * KILOBYTE,  2,  1, 10, 11, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 0, 1, "Combat"                                         }, // COMBAT
  { ARCADIA , MEMMAP_G, 0x76E773FA,  4 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, TRUE , TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Crazy Climber"                                  }, //*CRAZYCLI
  { ARCADIA , MEMMAP_G, 0xE84DF2EF,  2 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, TRUE , TRUE ,  4, 16,     0,       0, FALSE, 3, 4, "Crazy Gobbler"                                  }, // CRAZYGOB
  { ARCADIA , MEMMAP_G, 0x62C45881,  8 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Doraemon"                                       }, //*DORAEMON
  { ARCADIA , MEMMAP_G, 0x927C375A,  8 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, FALSE, TRUE , FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Dr. Slump"                                      }, //*DR-SLUMP (2:2) - enemies
  { ARCADIA , MEMMAP_G, 0xFC7EBA76,  4 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Escape"                                         }, //*ESCAPE
  { ARCADIA , MEMMAP_G, 0x77C19320,  8 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, TRUE , TRUE , TRUE , TRUE ,  4,  8,     0,       0, FALSE, 4, 4, "Funky Fish"                                     }, //*FUNKFISH (1:1) - player, enemies, player bullets
  { ARCADIA , MEMMAP_G, 0x22624414,  6 * KILOBYTE,  1,  7, 11, 10, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "American Football"                              }, // GRIDIRON
  { ARCADIA , MEMMAP_G, 0x1CEC4B21,  8 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, TRUE,  TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Hobo"                                           }, //*HOBO
  { ARCADIA , MEMMAP_G, 0x560DAA7F,  8 * KILOBYTE,  1,  2,  3,  4, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Horse Racing"                                   }, //*HORSERAC
  { ARCADIA , MEMMAP_G, 0x97060A54,  8 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, TRUE , FALSE, TRUE , TRUE ,  1,  2,     0,       0, FALSE, 4, 4, "Jump Bug (Emerson version)"                     }, //*JUMPBUGA
  { ARCADIA , MEMMAP_G, 0xDC0264B8,  8 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, TRUE , FALSE, TRUE , TRUE ,  1,  2,     0,       0, FALSE, 4, 4, "Jump Bug (Tele-Fever version)"                  }, //*JUMPBUGB
  { ARCADIA , MEMMAP_G, 0xDFB5DD55,  8 * KILOBYTE,  2,  1,  7,  2, FALSE, SENSITIVITY_DEFAULT, FALSE, TRUE , FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Jungler"                                        }, //*JUNGLER  (1:1) - player bullets
  { ARCADIA , MEMMAP_G, 0x4C885AF2,  8 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, TRUE , TRUE , FALSE, TRUE ,  2,  3,     0,       0, FALSE, 4, 4, "Super Dimension Fortress Macross"               }, //*MACROSS  (see below)
  { ARCADIA , MEMMAP_G, 0xA615F068,  4 * KILOBYTE,  1,  3,  3,  1, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Missile War"                                    }, //*MISSILEW (2:2:2:2 UDCs) - enemy bullets
  { ARCADIA , MEMMAP_G, 0x5E033F9C,  4 * KILOBYTE,  5,  0,  9,  7, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Monaco Grand Prix"                              }, //*MONACO
  { ARCADIA , MEMMAP_G, 0x7832F7AD,  4 * KILOBYTE,  2,  1,  3,  2, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Nibblemen"                                      }, //*NIBBLE
  { ARCADIA , MEMMAP_G, 0x4F538848,  4 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, TRUE , FALSE, FALSE, TRUE ,  1,  3,     0,       0, FALSE, 4, 4, "Parashooter"                                    }, //*PARASHOO
  { ARCADIA , MEMMAP_G, 0x449E80CB,  4 * KILOBYTE,  2, 10,  0, 11, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Ocean Battle"                                   }, //*OCEANBAT (this game probably supports analog during combat)
  { ARCADIA , MEMMAP_G, 0x4F19FD9B,  8 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, TRUE , FALSE, FALSE, TRUE ,  1,  2,     0,       0, FALSE, 4, 4, "Pleiades"                                       }, //*PLEIADES
  { ARCADIA , MEMMAP_G, 0xA0626E23,  4 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, FALSE, TRUE , TRUE , TRUE ,  1,  4,     0,       0, FALSE, 4, 4, "R2D Tank"                                       }, //*R2DTANK  (see below)
  { ARCADIA , MEMMAP_G, 0xA06F284B,  4 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, TRUE , TRUE , TRUE , TRUE ,  1,  2,     0,       0, FALSE, 4, 4, "Red Clash"                                      }, //*REDCLASH (1:1)
  { ARCADIA , MEMMAP_G, 0x4671B7F7,  4 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Robot Killer"                                   }, //*ROBOTKIL
  { ARCADIA , MEMMAP_G, 0x8169864E,  8 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Route 16"                                       }, //*ROUTE16
  { ARCADIA , MEMMAP_G, 0xE3794A2C,  4 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, TRUE , TRUE , TRUE , TRUE ,  1,  2,     0,       0, FALSE, 4, 4, "Space Attack (Emerson version)"                 }, //*SATTACKA (1:1)
  { ARCADIA , MEMMAP_G, 0x669632EC,  2 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, TRUE , TRUE , TRUE , TRUE ,  1,  2,     0,       0, FALSE, 4, 4, "Space Attack (unknown version)"                 }, //*SATTACKB (1:1)
  { ARCADIA , MEMMAP_G, 0xC91828CB,  2 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, TRUE , TRUE , TRUE , TRUE ,  1,  2,     0,       0, FALSE, 4, 4, "Space War (prototype)"                          }, //*SATTACKC (1:1)
  { ARCADIA , MEMMAP_G, 0x3FF224E5,  4 * KILOBYTE,  2,  7,  8,  9, FALSE, SENSITIVITY_DEFAULT, FALSE, TRUE , FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Space Mission"                                  }, //*SMISSION (?:?)
  { ARCADIA , MEMMAP_G, 0xBB88DAEA,  4 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, TRUE , TRUE , TRUE , TRUE ,  1,  2,     0,       0, FALSE, 4, 4, "Spiders"                                        }, //*SPIDERS  (?:?)
  { ARCADIA , MEMMAP_G, 0x88DEFACF,  4 * KILOBYTE,  2,  3,  2,  3, FALSE, SENSITIVITY_DEFAULT, TRUE , FALSE, FALSE, TRUE ,  1,  2,     0,       0, FALSE, 4, 4, "Space Raiders"                                  }, //*SRAIDERS (1:1 UDCs) - enemies
  { ARCADIA , MEMMAP_G, 0xCAA5F449,  4 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, TRUE , FALSE, FALSE, TRUE ,  1,  2,     0,       0, FALSE, 1, 4, "Space Squadron"                                 }, // SSQUADRO (1:1 UDCs) - enemies (rumble only affects his ship not his exhaust)
  { ARCADIA , MEMMAP_G, 0x83E6E40B,  4 * KILOBYTE,  4,  2, 10, 11, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Star Chess"                                     }, // STARCHES
  { ARCADIA , MEMMAP_G, 0x08D8B186,  4 * KILOBYTE,  2,  1,  3,  2, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Super Gobbler"                                  }, //*SUPERGOB
  { ARCADIA , MEMMAP_G, 0x7666A712,  6 * KILOBYTE,  2,  1,  2,  1, FALSE, SENSITIVITY_DEFAULT, TRUE , TRUE , FALSE, TRUE ,  1,  3,     0,       0, FALSE, 4, 4, "Space Vultures"                                 }, //*SVULTURE (?) - player bullets, enemy bullets
  { ARCADIA , MEMMAP_G, 0x0076422E,  4 * KILOBYTE,  2,  3,  1,  7, FALSE, SENSITIVITY_DEFAULT, FALSE, TRUE , FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Tanks a Lot"                                    }, //*TANKSALO (2:2)
  { ARCADIA , MEMMAP_G, 0x8BF2DFA9,  8 * KILOBYTE,  2,  3,  2,  3, FALSE, SENSITIVITY_DEFAULT, FALSE, TRUE , FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Grand Slam Tennis"                              }, //*TENNIS   (1:1)
  { ARCADIA , MEMMAP_G, 0xE66F362D,  4 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, TRUE , TRUE , TRUE , TRUE ,  1,  2,     0,       0, FALSE, 1, 4, "The End"                                        }, // THE-END  (1:1:1:1)
  { ARCADIA , MEMMAP_G, 0x306E39C1,  8 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, FALSE, TRUE , TRUE , TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Turtles/Turpin"                                 }, //*TURTLES  (1:1)
  { ARCADIA , MEMMAP_G, 0x6289F607, 12 * KILOBYTE, 12, 13, 14, 15, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Dictionary (English + German)"                  }, // VIDLEXEG
  { ARCADIA , MEMMAP_G, 0x0739504D, 12 * KILOBYTE, 12, 13, 14, 15, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Dictionary (German + English)"                  }, // VIDLEXGE
// Emerson Arcadia 2001 known bad dumps (1)
  { ARCADIA , MEMMAP_G, 0x2A3A5C91,  8 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 0, 4, "Mobile Soldier Gundam (bad dump)"               }, // GUNDAM- (rumble only affects his head not legs unfortunately)
// Emerson Arcadia 2001 homebrews (3)
  { ARCADIA , MEMMAP_G, 0xC1146B0B,  4 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Frogger 1.23 (obsolete)"                        }, //*FROGGER
  { ARCADIA , MEMMAP_G, 0x63513991,  2 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "JTron 1.14"                                     }, //*JTRON
  { ARCADIA , MEMMAP_G, 0x296C8465,  4 * KILOBYTE,  2,  0,  7,  9, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Tetris"                                         }, // TETRIS
// Emerson Arcadia 2001 enhanced games (8 here)
  { ARCADIA , MEMMAP_G, 0xC1FB9F3C,  6 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, TRUE , FALSE, FALSE, TRUE ,  1, 16,     0,       0, FALSE, 2, 4, "Alien Invaders (enhanced), original background" }, // ALIENIN1
  { ARCADIA , MEMMAP_G, 0x5F77087A,  6 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, TRUE , FALSE, FALSE, TRUE ,  1, 16,     0,       0, FALSE, 2, 4, "Alien Invaders (enhanced), background #2"       }, // ALIENIN2
  { ARCADIA , MEMMAP_G, 0x6F68DCD1,  6 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, TRUE , FALSE, FALSE, TRUE ,  1, 16,     0,       0, FALSE, 2, 4, "Alien Invaders (enhanced), background #3"       }, // ALIENIN3
  { ARCADIA , MEMMAP_G, 0x9F915C41,  6 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, TRUE , FALSE, FALSE, TRUE ,  1, 16,     0,       0, FALSE, 2, 4, "Alien Invaders (enhanced), background #4"       }, // ALIENIN4
  { ARCADIA , MEMMAP_G, 0x881E11AC,  6 * KILOBYTE,  2,  4,  0, 10, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Golf (patched)"                                 }, //*GOLF!
  { ARCADIA , MEMMAP_G, 0x0F34FFBB,  8 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 0, 4, "Mobile Soldier Gundam (semi-fixed)"             }, // GUNDAM! (rumble only affects his head not legs unfortunately)
  { ARCADIA , MEMMAP_G, 0xA6E9A90F,  6 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, TRUE , FALSE, FALSE, TRUE ,  1,  2,     0,       0, FALSE, 4, 4, "Space Squadron (enhanced)"                      }, // SSQUADR! (rumble only affects his ship not his exhaust)
  { ARCADIA , MEMMAP_G, 0xB46CACD8,  4 * KILOBYTE,  4,  2, 10, 11, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Star Chess (enhanced)"                          }, //*STARCHE!
// Interton VC 4000 games (50)
  { INTERTON, MEMMAP_A, 0xA0F5320A,  2 * KILOBYTE, 10, 11, 10, 11, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, TRUE , TRUE ,  4, 16, FUDGE, PVI_CPL, TRUE , 4, 4, "4 in a Row"                                     }, // 4INAROW  (  I) (variant of CHALLEN-)
  { INTERTON, MEMMAP_A, 0xA8503624,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Air/Sea Attack (suspect dump)"                  }, //*AIRSEAAT (F  ) (suspect dump) (variant of AIRSEABA)
  { INTERTON, MEMMAP_A, 0xF3D37699,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Air/Sea Battle"                                 }, //*AIRSEABA (  I) (variant of AIRSEAAT)
  { INTERTON, MEMMAP_C, 0x08CD8135,  4 * KILOBYTE,  2,  0, 10, 11, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE,      60, FALSE, 4, 4, "Backgammon"                                     }, //*BACKGAMM (  I)
  { INTERTON, MEMMAP_A, 0xEC2A91B3,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Blackjack"                                      }, //*BLACKJAC (F+I)
  { INTERTON, MEMMAP_A, 0xA47FE1BF,  2 * KILOBYTE,  2,  1,  3,  0, TRUE , SENSITIVITY_MEDIUM , FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Bowling"                                        }, //*BOWLING  (  I)
  { INTERTON, MEMMAP_B, 0x922C9F0D,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,   239, PVI_CPL, FALSE, 4, 4, "Boxing"                                         }, //*BOXING   (  I) (variant of PRIZEFI-)
  { INTERTON, MEMMAP_A, 0x619C07C9,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Capture"                                        }, //*CAPTURE  (F+I)
  { INTERTON, MEMMAP_A, 0x5C7F11E0,  2 * KILOBYTE,  2,  1,  3,  0, TRUE , SENSITIVITY_MEDIUM , FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Car Races"                                      }, // CARRACES (  I) (variant of GRNDPRIX)
  { INTERTON, MEMMAP_A, 0x50FD6A18,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,   251, PVI_CPL, FALSE, 4, 4, "Casino"                                         }, //*CASINO   (  I)
  { INTERTON, MEMMAP_C, 0xE9C53288,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,   252, PVI_CPL, FALSE, 4, 4, "Chess"                                          }, // CHESS1   (  I)
  { INTERTON, MEMMAP_D, 0x1B948AEB,  6 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,   252, PVI_CPL, FALSE, 4, 4, "Chess 2"                                        }, // CHESS2   (  I)
  { INTERTON, MEMMAP_A, 0x0AB80F3E,  2 * KILOBYTE,  2,  1,  3,  0, TRUE , SENSITIVITY_MEDIUM , FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 2, 3, "Circus"                                         }, // CIRCUS   (F+I)
  { INTERTON, MEMMAP_B, 0x11DD7F0D,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE,      51, FALSE, 4, 4, "Cockpit"                                        }, //*COCKPIT  (  I)
  { INTERTON, MEMMAP_A, 0x52B8C6E7,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Codebreaker"                                    }, //*CODEBREA (F  ) (variant of MASTERMI)
  { INTERTON, MEMMAP_A, 0x5E8EE224,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 0, 1, "Tank/Plane Battle"                              }, // COMBATA  (F  ) (variant of COMBATB)
  { INTERTON, MEMMAP_A, 0x468E4779,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 0, 1, "Tank Battle/Plane Battle"                       }, // COMBATB  (  I) (variant of COMBATA)
  { INTERTON, MEMMAP_C, 0x7B868473,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,   220, PVI_CPL, FALSE, 4, 4, "Draughts"                                       }, // DRAUGHTS (  I)
  { INTERTON, MEMMAP_B, 0xD399CE07,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE,      53, FALSE, 4, 4, "Golf"                                           }, //*GOLF     (  I) (variant or overdump of GOLF-)
  { INTERTON, MEMMAP_A, 0x2C770DC3,  2 * KILOBYTE,  2,  1,  3,  0, TRUE , SENSITIVITY_MEDIUM , FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Grand Prix"                                     }, // GRNDPRIX (F  ) (variant of CARRACES)
  { INTERTON, MEMMAP_A, 0xFF39521A,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Head On"                                        }, //*HEADON   (F  )
  { INTERTON, MEMMAP_A, 0xAAEA290F,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Hippodrome"                                     }, //*HIPPODRO (  I) (variant of HORSERAC)
  { INTERTON, MEMMAP_A, 0x4339E630,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Horse Racing (suspect dump)"                    }, //*HORSERAC (F  ) (suspect dump) (variant of HIPPODRO)
  { INTERTON, MEMMAP_A, 0x31978AD4,  2 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 2, 3, "Hunting"                                        }, // HUNTING  (  I) (variant of SHOOTGAL)
  { INTERTON, MEMMAP_A, 0xB3ED1129,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Hyperspace"                                     }, //*HYPERSPA (  I)
  { INTERTON, MEMMAP_A, 0x6AE47072,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, TRUE , FALSE, FALSE, TRUE ,  4, 16,   254, PVI_CPL, FALSE, 4, 4, "Invaders (Fountain version)"                    }, //*INVADERA (F  ) (variant of INVADERB)
  { INTERTON, MEMMAP_A, 0x9497204A,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, TRUE , FALSE, FALSE, TRUE ,  4, 16,   254, PVI_CPL, FALSE, 4, 4, "Invaders (Interton version)"                    }, //*INVADERB (  I) (variant of INVADERA)
  { INTERTON, MEMMAP_A, 0x7B2B2B61,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Labyrinth"                                      }, //*LABYRINT (  I) (variant of SUPERMA-)
  { INTERTON, MEMMAP_A, 0xC44FABB1,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 3, 4, "Laser Attack"                                   }, // LASERATT (F  )
  { INTERTON, MEMMAP_A, 0x52B8C6E7,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Mastermind"                                     }, // MASTERMI (  I) (variant of CODEBREA)
  { INTERTON, MEMMAP_A, 0x66C5975E,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Math 2"                                         }, // MATH2    (F+I)
  { INTERTON, MEMMAP_A, 0xE38E0C03,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Mathematics 1"                                  }, // MATHEMA1 (  I) (variant of MATH1-)
  { INTERTON, MEMMAP_B, 0xA640A330,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE,      54, FALSE, 4, 4, "Metropolis/Hangman"                             }, // METROPOL (  I)
  { INTERTON, MEMMAP_B, 0x27CCE96F,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 3, 4, "Monster Man"                                    }, // MONSTERM (  I) (variant of SPIDERS-)
  { INTERTON, MEMMAP_B, 0x55011F0A,  4 * KILOBYTE,  2,  1,  3,  0, TRUE , SENSITIVITY_MEDIUM , FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE,      53, TRUE , 4, 4, "Motocross"                                      }, //*MOTOCROS (  I)
  { INTERTON, MEMMAP_A, 0xD68DA80A,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Musical Games"                                  }, // MUSICALG (F+I)
  { INTERTON, MEMMAP_A, 0x682B876A,  2 * KILOBYTE,  2,  1,  3,  0, TRUE , SENSITIVITY_MEDIUM , FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 0, 1, "Olympics"                                       }, // OLYMPICS (F+I)
  { INTERTON, MEMMAP_A, 0xC6B2CD9A,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,   267, PVI_CPL, FALSE, 0, 4, "Electronic Pinball"                             }, // PINBALLA (F  ) (variant of PINBALLB)
  { INTERTON, MEMMAP_A, 0xADD99C6E,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,   267, PVI_CPL, FALSE, 0, 4, "Pinball"                                        }, // PINBALLB (  I) (variant of PINBALLA)
  { INTERTON, MEMMAP_A, 0x2AEB516E,  2 * KILOBYTE,  2,  1,  3,  0, TRUE , SENSITIVITY_MEDIUM , FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Planet Defender"                                }, //*PLANETDE (F  )
  { INTERTON, MEMMAP_A, 0x9C8E7DE6,  2 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 2, 3, "Shooting Gallery"                               }, // SHOOTGAL (F  ) (variant of HUNTING)
  { INTERTON, MEMMAP_A, 0xB589F093,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 0, 1, "Shoot Out"                                      }, // SHOOTOUT (F+I)
  { INTERTON, MEMMAP_A, 0x87977F92,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE,      57, FALSE, 4, 4, "Soccer (Fountain version)"                      }, //*SOCCERA  (F  ) (variant of SOCCERB)
  { INTERTON, MEMMAP_A, 0x05AF8228,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE,      57, FALSE, 4, 4, "Soccer (Interton version)"                      }, //*SOCCERB  (  I) (variant of SOCCERA)
  { INTERTON, MEMMAP_A, 0x08209D98,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Solitaire"                                      }, // SOLITAIR (  I)
  { INTERTON, MEMMAP_A, 0xB1C31F9A,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Space War"                                      }, //*SPACEWAR (F+I)
  { INTERTON, MEMMAP_A, 0xA46C2F9E,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Super Invaders"                                 }, //*SUPERINV (  I)
  { INTERTON, MEMMAP_B, 0x306E37BB,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Super-Space"                                    }, //*SUPERSPA (  I)
  { INTERTON, MEMMAP_A, 0xEA8F717F,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Treasure Hunt"                                  }, // TREASURE (F+I)
  { INTERTON, MEMMAP_B, 0xBDD652B7,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,   275, PVI_CPL, FALSE, 4, 4, "Winter Sports"                                  }, //*WINTERSP (  I)
// Interton VC 4000 known bad dumps (8)
  { INTERTON, MEMMAP_A, 0x8AB96827,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, TRUE , 4, 4, "Challenge (bad dump)"                           }, // CHALLEN- (F  ) (variant of 4INAROW)
  { INTERTON, MEMMAP_A, 0x3CF3BAF6,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE,      53, FALSE, 4, 4, "Golf (bad dump)"                                }, // GOLF-    (F  ) (variant or underdump of GOLF)
  { INTERTON, MEMMAP_A, 0x2737D62D,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Super Knockout (bad dump)"                      }, // KNOCKOU- (F  )
  { INTERTON, MEMMAP_A, 0x9E3AB142,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Maths 1 (bad dump)"                             }, // MATH1-   (F  ) (variant of MATHEMA1)
  { INTERTON, MEMMAP_A, 0x4D37502D,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Mathematics 1 (bad dump)"                       }, //   alternate bad dump of of MATHEMA1
  { INTERTON, MEMMAP_A, 0xA931407C,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,   239, PVI_CPL, FALSE, 4, 4, "Prizefight (bad dump)"                          }, // PRIZEFI- (F  ) (variant of BOXING)
  { INTERTON, MEMMAP_A, 0xE5E4C37B,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 3, 4, "Spider's Web (bad dump)"                        }, // SPIDERS- (F  ) (variant of MONSTERM)
  { INTERTON, MEMMAP_A, 0x4D3F9CFB,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Super Maze (bad dump)"                          }, // SUPERMA- (F  ) (variant of LABYRINT)
// Elektor TV Games Computer Elektor Software Service (ESS) games (56)
  { ELEKTOR , MEMMAP_E, 0x45494CBC,  7 * KILOBYTE, 10, 11, 10, 11, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-003-4: 4 in a Row (overdump)"               }, // E/-
  { ELEKTOR , MEMMAP_E, 0x18288A02,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE,      47, FALSE, 4, 4, "ESS-003-5: Surround (overdump)"                 }, //*E/-
  { ELEKTOR , MEMMAP_E, 0x2DC0DD9E,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-006-2: PVI Programming (overdump)"          }, // E/-
  { ELEKTOR , MEMMAP_E, 0x1D8C1EA5,  7 * KILOBYTE,  8,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-006-3: Rocket Shooting (overdump)"          }, //*E/-
  { ELEKTOR , MEMMAP_E, 0x7585F58E,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-1: Mastermind (overdump)"               }, // E/-
  { ELEKTOR , MEMMAP_E, 0x63F7D8FE,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-2: Code Breaker (overdump)"             }, // E/-
  { ELEKTOR , MEMMAP_E, 0x2DB9E572,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, TRUE , 4, 4, "ESS-007-3: Reversi (overdump)"                  }, // E/-
  { ELEKTOR , MEMMAP_E, 0xD4523830,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-4: Amazone (overdump)"                  }, //*E/-
  { ELEKTOR , MEMMAP_E, 0xE9403961,  7 * KILOBYTE,  8,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-5: Space Shoot-Out (overdump)"          }, //*E/-
  { ELEKTOR , MEMMAP_E, 0xDD59FC15,  7 * KILOBYTE, 10, 11, 10, 11, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-6: 4 in a Row (overdump)"               }, // E/-
  { ELEKTOR , MEMMAP_E, 0xF8D8218B,  7 * KILOBYTE, 10, 11, 10, 11, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-7: 4 in a Row (overdump)"               }, // E/-
  { ELEKTOR , MEMMAP_E, 0x80E543D2,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-8: Jackpot (overdump)"                  }, // E/-
  { ELEKTOR , MEMMAP_E, 0xE6262205,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE,      47, FALSE, 4, 4, "ESS-007-9: Surround (overdump)"                 }, //*E/-
  { ELEKTOR , MEMMAP_E, 0x309A17A5,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-009-1: Aggressor (overdump)"                }, //*E/-
  { ELEKTOR , MEMMAP_E, 0xFC32C62A,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-009-2: Offshore Fishing (overdump)"         }, //*E/-
  { ELEKTOR , MEMMAP_E, 0xE7A63D7F,  8 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-009-3: Circledrive"                         }, //*E/-
  { ELEKTOR , MEMMAP_E, 0xC782D585,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-009-4: Seawar (overdump)"                   }, //*E/-
#ifdef BASICMODE
  { ELEKTOR , MEMMAP_E, 0x7AE45C16,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,   265, PVI_CPL, FALSE, 4, 4, "ESS-009-5: Memory (overdump)"                   }, //*E/F
  { ELEKTOR , MEMMAP_E, 0x64BEF839,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,   257, PVI_CPL, FALSE, 3, 4, "ESS-009-6: Labyrinth"                           }, // E/F
#else
  { ELEKTOR , MEMMAP_F, 0x7AE45C16,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,   265, PVI_CPL, FALSE, 4, 4, "ESS-009-5: Memory (overdump)"                   }, //*E/F
  { ELEKTOR , MEMMAP_F, 0x64BEF839,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,   257, PVI_CPL, FALSE, 3, 4, "ESS-009-6: Labyrinth"                           }, // E/F
#endif
  { ELEKTOR , MEMMAP_E, 0xBD495BED,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-009-7: Destroyer (overdump)"                }, //*E/-
  { ELEKTOR , MEMMAP_E, 0x62B18147,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-009-8: Starship Enterprise (overdump)"      }, //*E/-
  { ELEKTOR , MEMMAP_E, 0x95A254AA,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-009-9: Blackjack (overdump)"                }, // E/-
  { ELEKTOR , MEMMAP_E, 0x231B2427,  7 * KILOBYTE, 10,  0, 11, 10, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-009-A: Card Trick (overdump)"               }, // E/-
  { ELEKTOR , MEMMAP_E, 0x90CADC60,  7 * KILOBYTE, 10,  0, 11,  9, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-009-B: Awari (overdump)"                    }, // E/-
  { ELEKTOR , MEMMAP_E, 0xE9F2688E,  7 * KILOBYTE,  3,  3,  3,  3, FALSE, SENSITIVITY_DEFAULT, FALSE, TRUE,  FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 1, 4, "ESS-009-C: UFO Shooting (overdump)"             }, // E/-
  { ELEKTOR , MEMMAP_E, 0xBDC3DA72,  7 * KILOBYTE,  8,  1, 11,  8, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-009-D: Raster (overdump)"                   }, //*E/-
  { ELEKTOR , MEMMAP_E, 0x83538875,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE,      57, FALSE, 4, 4, "ESS-009-E: Nim (overdump)"                      }, // E/-
  { ELEKTOR , MEMMAP_E, 0x21E8568B,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-009-F: Solitaire (overdump)"                }, // E/-
  { ELEKTOR , MEMMAP_E, 0xE02FDA29,  8 * KILOBYTE,  8,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-010-1: Invaders"                            }, //*E/-
  { ELEKTOR , MEMMAP_F, 0xC15FF62B,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-010-2: Pinball (overdump)"                  }, //*-/F
  { ELEKTOR , MEMMAP_E, 0x8EDE9085,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-010-3: Helicopter (overdump)"               }, //*E/-
  { ELEKTOR , MEMMAP_E, 0x7EC3B97F,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE,      55, FALSE, 4, 4, "ESS-010-4: Hard Nut (overdump)"                 }, // E/-
  { ELEKTOR , MEMMAP_E, 0xF193C235,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-010-5: Catapult (overdump)"                 }, //*E/-
  { ELEKTOR , MEMMAP_F, 0xDBD484BB,  7 * KILOBYTE,  2,  1,  3,  0, TRUE,  SENSITIVITY_MEDIUM , FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE,      58, FALSE, 4, 4, "ESS-010-6: Pilot (overdump)"                    }, // -/F
  { ELEKTOR , MEMMAP_E, 0x2066116A,  7 * KILOBYTE,  2,  1,  3,  0, TRUE,  SENSITIVITY_LOW    , FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-010-7: Attack from Space (overdump)"        }, //*E/-
  { ELEKTOR , MEMMAP_E, 0xC31C0B56,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-010-8: Reverse (overdump)"                  }, //*E/-
  { ELEKTOR , MEMMAP_E, 0xDA58E218,  8 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-010-9: Rocket Hunting"                      }, //*E/-
  { ELEKTOR , MEMMAP_F, 0x8B66AC75,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,   258, PVI_CPL, FALSE, 4, 4, "ESS-010-A: Basketball"                          }, //*-/F
  { ELEKTOR , MEMMAP_E, 0xE62573E0,  7 * KILOBYTE,  1,  2,  3,  0, TRUE,  SENSITIVITY_MEDIUM , FALSE, FALSE, FALSE, TRUE ,  4, 16,   220, PVI_CPL, FALSE, 4, 4, "ESS-010-B: Bursting Balloons (overdump)"        }, //*E/-
  { ELEKTOR , MEMMAP_F, 0x49E69462,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,   266,      54, FALSE, 4, 4, "ESS-010-E: Galgenspiel"                         }, // -/F
  { ELEKTOR , MEMMAP_F, 0x426B44D8,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,   266,      54, FALSE, 4, 4, "ESS-010-F: Hangman"                             }, // -/F
  { ELEKTOR , MEMMAP_E, 0x994BC0EA,  7 * KILOBYTE,  8,  1,  1,  8, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-1: Snakes and Ladders (overdump)"       }, // E/-
  { ELEKTOR , MEMMAP_E, 0xEC20152C,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE,      61, FALSE, 4, 4, "ESS-011-2: Molebasher (overdump)"               }, //*E/-
  { ELEKTOR , MEMMAP_E, 0x09EF8D41,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-3: Snap (overdump)"                     }, // E/-
#ifdef BASICMODE
  { ELEKTOR , MEMMAP_E, 0x5206E5CF,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-4: Mazes (overdump)"                    }, //*E/F
  { ELEKTOR , MEMMAP_E, 0x597615A6,  7 * KILOBYTE,  8,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, TRUE,  FALSE, FALSE, TRUE ,  1,  2, FUDGE, PVI_CPL, FALSE, 0, 4, "ESS-011-5: Asteroids (overdump)"                }, // E/F
#else
  { ELEKTOR , MEMMAP_F, 0x5206E5CF,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-4: Mazes (overdump)"                    }, //*E/F
  { ELEKTOR , MEMMAP_F, 0x597615A6,  7 * KILOBYTE,  8,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, TRUE,  FALSE, FALSE, TRUE ,  1,  2, FUDGE, PVI_CPL, FALSE, 0, 4, "ESS-011-5: Asteroids (overdump)"                }, // E/F
#endif
  { ELEKTOR , MEMMAP_F, 0x015E3686,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, TRUE , 4, 4, "ESS-011-6: Dragster"                            }, //*-/F
  { ELEKTOR , MEMMAP_E, 0xF5789409,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-7: Omega Landing (overdump)"            }, //*E/-
  { ELEKTOR , MEMMAP_E, 0x43E6DB7A,  7 * KILOBYTE,  1,  2,  3,  0, TRUE,  SENSITIVITY_MEDIUM , FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-8: Breakout (overdump)"                 }, //*E/-
  { ELEKTOR , MEMMAP_F, 0x743DFF28,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, FALSE,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-9: Tiny Tim (overdump)"                 }, //*-/F
  { ELEKTOR , MEMMAP_E, 0xD85303AD,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-A: Horse Race + Jackpot (overdump)"     }, //*E/-
  { ELEKTOR , MEMMAP_E, 0x77523CC2,  7 * KILOBYTE,  2,  1,  3,  0, TRUE,  SENSITIVITY_HIGH   , FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE,      55, FALSE, 4, 4, "ESS-011-B: Newton (overdump)"                   }, //*E/-
  { ELEKTOR , MEMMAP_F, 0xEBAB31FD,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-C: Horse Races (overdump)"              }, //*-/F
  { ELEKTOR , MEMMAP_E, 0xF55DD406,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-D: Painting (overdump)"                 }, //*E/-
  { ELEKTOR , MEMMAP_E, 0xED484047,  7 * KILOBYTE,  2,  1,  3,  0, TRUE , SENSITIVITY_MEDIUM , FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-E: Submarines + Racing (overdump)"      }, //*E/-
  { ELEKTOR , MEMMAP_F, 0x46D5A0EE,  7 * KILOBYTE,  1,  2,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  1, 50, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-F: Cosmic Adventure"                    }, //*-/F
// Elektor TV Games Computer Elektor Software Service (ESS) utilities (14)
  { ELEKTOR , MEMMAP_E, 0x0D7D868E,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-003-1: Play with the PVI (overdump)"        }, // E/-
  { ELEKTOR , MEMMAP_E, 0x10991986,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-003-2: Clock (overdump)"                    }, // E/-
  { ELEKTOR , MEMMAP_E, 0x51B61371,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,   228, PVI_CPL, FALSE, 4, 4, "ESS-003-3: Music Box (overdump)"                }, // E/-
  { ELEKTOR , MEMMAP_E, 0x94984EB9,  7 * KILOBYTE, 10, 10, 10, 10, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-003-6: Demonstration Program (overdump)"    }, // E/-
  { ELEKTOR , MEMMAP_E, 0x9CA4965D,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-006-1: Picture Pattern (overdump)"          }, // E/-
  { ELEKTOR , MEMMAP_E, 0x18A53913,  7 * KILOBYTE, 10, 10, 10, 10, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, FALSE,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-A: Shapes (overdump)"                   }, // E/-
  { ELEKTOR , MEMMAP_E, 0xA7B77712,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,   228, PVI_CPL, FALSE, 4, 4, "ESS-007-B: Piano (overdump)"                    }, // E/-
  { ELEKTOR , MEMMAP_E, 0xAEFB870B,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-C: PVI Progamming (overdump)"           }, // E/-
  { ELEKTOR , MEMMAP_E, 0xF06467B4,  8 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-D: Disassembler"                        }, // E/-
  { ELEKTOR , MEMMAP_E, 0x9CA4965D,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-E: Test Patterns (overdump)"            }, // E/-
  { ELEKTOR , MEMMAP_E, 0x5BEE68A3,  7 * KILOBYTE,  0, 10,  7, 11, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-F: Lotto (overdump)"                    }, // E/-
  { ELEKTOR , MEMMAP_F, 0xAD39EDC6,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-010-C: Fr Elise (overdump)"                }, // -/F
  { ELEKTOR , MEMMAP_F, 0x755114C2,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-010-D: Editor"                              }, // -/F
  { ELEKTOR , MEMMAP_F, 0x902A3B8C,  7 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE,      54, FALSE, 4, 4, "ESS-010-1x: Change Words for Hangman (overdump)"}, // -/F
// Elektor TV Games Computer book examples (18) (overdumps) (could perhaps show pages for these)
  { ELEKTOR , MEMMAP_E, 0x9653ABA1,  4 * KILOBYTE,  0,  0,  0,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Add 2 (overdump)"                               }, // E/-
  { ELEKTOR , MEMMAP_E, 0x22180C05,  4 * KILOBYTE,  0,  0,  0,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Add 4 (overdump)"                               }, // E/-
  { ELEKTOR , MEMMAP_E, 0x3E1B794C,  4 * KILOBYTE,  0,  0,  0,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Minus 2 (overdump)"                             }, // E/-
  { ELEKTOR , MEMMAP_E, 0x1E982CA3,  4 * KILOBYTE,  0,  0,  0,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Minus 4 (overdump)"                             }, // E/-
  { ELEKTOR , MEMMAP_F, 0xAAAC97CC,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Explosion (overdump)"                           }, // -/F
  { ELEKTOR , MEMMAP_F, 0xCDAE1CFD,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Gunshot (overdump)"                             }, // -/F
  { ELEKTOR , MEMMAP_E, 0xD5522589,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_VLOW   , FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Joystick (overdump)"                            }, // E/-
  { ELEKTOR , MEMMAP_E, 0x12EFAD0F,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Message (overdump)"                             }, // E/-
  { ELEKTOR , MEMMAP_E, 0xA1553056,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Practice (overdump)"                            }, // E/-
  { ELEKTOR , MEMMAP_F, 0x65B54C14,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Siren (overdump)"                               }, // -/F
  { ELEKTOR , MEMMAP_E, 0x4E917AD2,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Steam Engine (overdump)"                        }, // E/-
  { ELEKTOR , MEMMAP_E, 0xACDC1E77,  4 * KILOBYTE,  3,  3,  3,  3, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Table A (overdump)"                             }, //*E/-
  { ELEKTOR , MEMMAP_E, 0x740B53EA,  4 * KILOBYTE,  3,  3,  3,  3, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Table B (overdump)"                             }, //*E/-
  { ELEKTOR , MEMMAP_E, 0x686C7949,  4 * KILOBYTE,  3,  3,  3,  3, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Table C (overdump)"                             }, //*E/-
  { ELEKTOR , MEMMAP_E, 0xBCDD4049,  4 * KILOBYTE,  3,  3,  3,  3, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Table D (overdump)"                             }, //*E/-
  { ELEKTOR , MEMMAP_E, 0xF49B658B,  4 * KILOBYTE,  3,  3,  3,  3, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Table E (overdump)"                             }, //*E/-
  { ELEKTOR , MEMMAP_F, 0x06599141,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Wedding March (overdump)"                       }, // -/F
  { ELEKTOR , MEMMAP_F, 0x7DCCBA1A,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Wolf Whistle (overdump)"                        }, // -/F
// Emerson Arcadia 2001 overdumps (4)
  { ARCADIA , MEMMAP_G, 0x82D049D7,  8 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 0, 1, "Combat (overdump)"                              },
  { ARCADIA , MEMMAP_G, 0x7F674C85,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, TRUE , TRUE ,  4, 16,     0,       0, FALSE, 3, 4, "Crazy Gobbler (overdump)"                       },
  { ARCADIA , MEMMAP_G, 0x06A86F4A,  8 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, TRUE , TRUE , TRUE , TRUE ,  1,  2,     0,       0, FALSE, 4, 4, "Red Clash (overdump)"                           }, //*1:1
  { ARCADIA , MEMMAP_G, 0xF9D9EC5B,  8 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, TRUE , TRUE , TRUE , TRUE ,  1,  2,     0,       0, FALSE, 4, 4, "Spiders (overdump)"                             }, //*?:?
// Emerson Arcadia 2001 examples (24)
  { ARCADIA , MEMMAP_G, 0xA2757E4F,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "4 Sprites"                                      }, //*Charlie
  { ARCADIA , MEMMAP_G, 0x174EEBF4,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "4 UDCs"                                         }, // Charlie
  { ARCADIA , MEMMAP_G, 0xC8604449,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Animate"                                        }, //*Charlie
  { ARCADIA , MEMMAP_G, 0x995904CC,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Boot Initialization"                            }, // Charlie
  { ARCADIA , MEMMAP_G, 0xD7F47588,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Boxer"                                          }, //*Charlie
  { ARCADIA , MEMMAP_G, 0xA2BA446C,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Character Show"                                 }, // Charlie
  { ARCADIA , MEMMAP_G, 0x83633401,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Music"                                          }, // Charlie
  { ARCADIA , MEMMAP_G, 0x4753B462,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Read Buttons"                                   }, // Charlie
  { ARCADIA , MEMMAP_G, 0xD2C7E59C,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Read Paddles"                                   }, // Charlie
  { ARCADIA , MEMMAP_G, 0x13244D23,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Show Text"                                      }, // Charlie
  { ARCADIA , MEMMAP_G, 0xEC652918,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Character Test"                                 }, // Trionfo
  { ARCADIA , MEMMAP_G, 0x09B804ED,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Hello"                                          }, // Trionfo
  { ARCADIA , MEMMAP_G, 0xFD175D72,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Mothership"                                     }, // Trionfo
  { ARCADIA , MEMMAP_G, 0x365856B4,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Sprite"                                         }, //*Trionfo
  { ARCADIA , MEMMAP_G, 0x55B4B1F6,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, TRUE , FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "8 Sprites"                                      }, //*Turner (1:1)
  { ARCADIA , MEMMAP_G, 0x3EDD2C71,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Target 1"                                       }, //*Turner
  { ARCADIA , MEMMAP_G, 0xF446A67F,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Target 2"                                       }, //*Turner
  { ARCADIA , MEMMAP_G, 0xF848668B,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Colours 1.02"                                   }, // Jacobs
  { ARCADIA , MEMMAP_G, 0xE68C5C15,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Horizontal Scroll 1.01"                         }, //*Jacobs
  { ARCADIA , MEMMAP_G, 0xFC4D7DC7,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, TRUE , FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Multiplex 3.2"                                  }, //*Jacobs (1:1)
  { ARCADIA , MEMMAP_G, 0xC8E0C20B,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Sound 3.0"                                      }, // Jacobs
  { ARCADIA , MEMMAP_G, 0x2C402930,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Tester 1 1.02"                                  }, // Jacobs
  { ARCADIA , MEMMAP_G, 0x88078F4B,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Tester 2 1.01"                                  }, // Jacobs
  { ARCADIA , MEMMAP_G, 0xD4BE78B1,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Vertical Scroll 1.02"                           }, //*Jacobs
// miscellaneous
  { ELEKTOR,  MEMMAP_F, 0x407740F1,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Tester 1 1.01"                                  },
  { ARCADIA , MEMMAP_G, 0xE0A9DD9A,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Dictionary (main cart only)"                    }, // VIDLEX
  { ARCADIA , MEMMAP_G, 0x320B0A9D,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Dictionary (English subcart only)"              }, // VIDLEXE
  { ARCADIA , MEMMAP_G, 0xB2B9B5D8,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Dictionary (German subcart only)"               }, // VIDLEXG
  { ARCADIA , MEMMAP_G, 0xB10A08B5,  8 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "American Football (overdump)"                   },
  { ARCADIA , MEMMAP_G, 0x28AD5BCF,  8 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Auto Race (overdump?)"                          }, //*
  { ARCADIA , MEMMAP_G, 0xFF013B36,  4 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Frogger 1.03 (obsolete)"                        }, //*"Frogger v1.03 (200x)(Amigan Software).bin"
  { ARCADIA , MEMMAP_G, 0x8286BC61,  2 * KILOBYTE,  5,  7,  8,  9, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "JTron 1.0? (obsolete)"                          }, //*"JTron (2003)(Amigan Software)(PD).bin"
  { ARCADIA , MEMMAP_G, 0xAC32E456,  2 * KILOBYTE,  5,  7,  8,  9, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "JTron 1.1 (obsolete)"                           }, //*"JTron v1.1 (2003)(Amigan Software).bin"
  { INTERTON, MEMMAP_B, 0x78E94A1B,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Math 2 (patched) (obsolete)"                    }, // MATH2!   (F+I)
  { ELEKTOR,  MEMMAP_E, 0x6158CB0C,  8 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-009-3: Circledrive (patch 1.0) (obsolete)"  }, //*E/-
  { ELEKTOR,  MEMMAP_F, 0x816CBEEE,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Tester 2 1.0 (obsolete)"                        },
  { ELEKTOR , MEMMAP_F, 0xD458DC0E,  7 * KILOBYTE,  1,  2,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  1, 50, FUDGE, PVI_CPL, FALSE, 4, 4, "Cosmic Adventure (enhanced)"                    }, //*-/F
  { ARCADIA , MEMMAP_G, 0x5034F5A1,  4 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Frogger 1.24"                                   }, //*
  { ARCADIA , MEMMAP_G, 0x566C78A0,  6 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, TRUE , TRUE , TRUE , TRUE ,  1,  2,     0,       0, FALSE, 1, 4, "The End (enhanced)"                             }, // THE-END! (1:1:1:1)
  { ELEKTOR,  MEMMAP_F, 0x53277284,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Tester 2 1.01"                                  },
  { ELEKTOR , MEMMAP_E, 0x3B7977ED,  4 * KILOBYTE, 10, 11, 10, 11, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-003-4: 4 in a Row (patched)"                }, // E/-
  { ELEKTOR , MEMMAP_E, 0x632DAE6D,  4 * KILOBYTE, 10, 11, 10, 11, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-6: 4 in a Row (patched)"                }, // E/-
  { ELEKTOR , MEMMAP_E, 0xA6ED702F,  4 * KILOBYTE, 10, 11, 10, 11, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-7: 4 in a Row (patched)"                }, // E/-
  { INTERTON, MEMMAP_A, 0x35C47832,  2 * KILOBYTE, 10, 11, 10, 11, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, TRUE , TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "4 in a Row (patched)"                           }, // 4INAROW! (  I) (variant of CHALLEN-)
  { ARCADIA , MEMMAP_G, 0x617EEB43,  8 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, TRUE,  TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Hobo (enhanced)"                                }, //*HOBO!
  { ARCADIA , MEMMAP_G, 0x9AEADFB9,  4 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, TRUE , TRUE , TRUE , TRUE ,  1,  2,     0,       0, FALSE, 4, 4, "Space Attack (unknown) (MESS V2 patch 1.0)"     }, //*SATTACB! (1:1) (withdrawn but not obsolete)
  { ARCADIA , MEMMAP_G, 0x3C6F916C,  2 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, TRUE , TRUE , TRUE , TRUE ,  1,  2,     0,       0, FALSE, 4, 4, "Space War (prototype) (MESS V2 patch 1.1)"      }, //*SATTACC! (1:1) (withdrawn but not obsolete)
  { ARCADIA , MEMMAP_G, 0x58B80428,  6 * KILOBYTE,  2,  2,  2,  2, TRUE , SENSITIVITY_HIGH   , FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 2, 3, "Circus (MESS V1 patch 2.0) (obsolete)"          }, // CIRCUS!
  { ARCADIA , MEMMAP_G, 0x09A474AE,  8 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Doraemon (MESS V1 patch 1.0) (obsolete)"        }, //*DORAEMO!
  { ARCADIA , MEMMAP_G, 0xE462FA6E,  8 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, FALSE, TRUE , FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Dr. Slump (MESS V1 patch 1.1) (obsolete)"       }, //*DR-SLUM! (2:2) - enemies
  { ARCADIA , MEMMAP_G, 0x3859D736, 10 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, TRUE,  TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Hobo (MESS V1 patch 2.0) (obsolete)"            }, //*HOBO!
  { ARCADIA , MEMMAP_G, 0x1A3082F9,  8 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, TRUE , FALSE, TRUE , TRUE ,  1,  2,     0,       0, FALSE, 4, 4, "Jump Bug (Emerson) (MESS V1 patch 1.0) (obs.)"  }, //*JUMPBUG!
  { ARCADIA , MEMMAP_G, 0xFF4BC6FF,  4 * KILOBYTE,  2,  1,  3,  2, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Nibblemen (MESS V1 patch 1.0) (obsolete)"       }, //*NIBBLE!
  { ARCADIA , MEMMAP_G, 0xDBB340BF,  4 * KILOBYTE,  2, 10,  0, 11, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Ocean Battle (MESS V1 patch 1.0) (obsolete)"    }, //*OCEANBA!
  { ARCADIA , MEMMAP_G, 0x56C4E008,  4 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, TRUE , TRUE , TRUE , TRUE ,  1,  2,     0,       0, FALSE, 4, 4, "Red Clash (MESS V1 patch 1.0) (obsolete)"       }, //*REDCLAS! (1:1)
  { ARCADIA , MEMMAP_G, 0x8543544E,  4 * KILOBYTE,  2,  2,  2,  2, FALSE, SENSITIVITY_DEFAULT, TRUE , TRUE , TRUE , TRUE ,  1,  2,     0,       0, FALSE, 4, 4, "Spiders (MESS V1 patch 1.0) (obsolete)"         }, //*SPIDERS! (?:?)
  { ARCADIA , MEMMAP_G, 0xEACD58B4,  4 * KILOBYTE,  2,  1,  3,  2, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Super Gobbler (MESS V1 patch 1.0) (obsolete)"   }, //*SUPERGO!
  { ARCADIA , MEMMAP_G, 0xD1987FF5,  8 * KILOBYTE,  2,  1,  2,  1, FALSE, SENSITIVITY_DEFAULT, TRUE , TRUE , FALSE, TRUE ,  1,  3,     0,       0, FALSE, 4, 4, "Space Vultures (MESS V1 patch 1.2) (obsolete)"  }, //*SVULTUR! (?) - player bullets, enemy bullets
  { ARCADIA , MEMMAP_G, 0xEC69267D,  8 * KILOBYTE,  2,  3,  2,  3, FALSE, SENSITIVITY_DEFAULT, FALSE, TRUE , FALSE, TRUE ,  4, 16,     0,       0, FALSE, 4, 4, "Grand Slam Tennis (MESS V1 patch 1.0) (obs.)"   }, //*TENNIS!  (1:1)
// Elektor TV Games Computer Elektor Software Service (ESS) software (59)
  { ELEKTOR , MEMMAP_E, 0x337E7A3C,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-003-1: Play with the PVI"                   }, // E/-
  { ELEKTOR , MEMMAP_E, 0x4A75D8EB,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-003-2: Clock"                               }, // E/-
  { ELEKTOR , MEMMAP_E, 0x8A78E207,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,   228, PVI_CPL, FALSE, 4, 4, "ESS-003-3: Music Box"                           }, // E/-
  { ELEKTOR , MEMMAP_E, 0xA8FE170D,  4 * KILOBYTE, 10, 11, 10, 11, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-003-4: 4 in a Row"                          }, // E/-
  { ELEKTOR , MEMMAP_E, 0x6D15B13E,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE,      47, FALSE, 4, 4, "ESS-003-5: Surround"                            }, //*E/-
  { ELEKTOR , MEMMAP_E, 0xD97B72D0,  4 * KILOBYTE, 10, 10, 10, 10, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-003-6: Demonstration Program"               }, // E/-
  { ELEKTOR , MEMMAP_E, 0xDDB40355,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-006-1: Picture Pattern"                     }, // E/-
  { ELEKTOR , MEMMAP_E, 0xE0501EC9,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-006-2: PVI Programming"                     }, // E/-
  { ELEKTOR , MEMMAP_E, 0x6D4637D2,  4 * KILOBYTE,  8,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-006-3: Rocket Shooting"                     }, //*E/-
  { ELEKTOR , MEMMAP_E, 0xCDB52816,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-1: Mastermind"                          }, // E/-
  { ELEKTOR , MEMMAP_E, 0x8BC0C329,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-2: Code Breaker"                        }, // E/-
  { ELEKTOR , MEMMAP_E, 0x0BB4A564,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, TRUE , 4, 4, "ESS-007-3: Reversi"                             }, // E/-
  { ELEKTOR , MEMMAP_E, 0xF42B24A5,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-4: Amazone"                             }, //*E/-
  { ELEKTOR , MEMMAP_E, 0x608FE606,  4 * KILOBYTE,  8,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-5: Space Shoot-Out"                     }, //*E/-
  { ELEKTOR , MEMMAP_E, 0xF0AACE8D,  4 * KILOBYTE, 10, 11, 10, 11, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-6: 4 in a Row"                          }, // E/-
  { ELEKTOR , MEMMAP_E, 0x356A10CF,  4 * KILOBYTE, 10, 11, 10, 11, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-7: 4 in a Row"                          }, // E/-
  { ELEKTOR , MEMMAP_E, 0x6EE082D5,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-8: Jackpot"                             }, //*E/-
  { ELEKTOR , MEMMAP_E, 0xC36F5BDA,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE,      47, FALSE, 4, 4, "ESS-007-9: Surround"                            }, //*E/-
  { ELEKTOR , MEMMAP_E, 0xA37CE1E4,  3 * KILOBYTE, 10, 10, 10, 10, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, FALSE,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-A: Shapes"                              }, // E/-
  { ELEKTOR , MEMMAP_E, 0x59C2268B,  3 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,   228, PVI_CPL, FALSE, 4, 4, "ESS-007-B: Piano"                               }, // E/-
  { ELEKTOR , MEMMAP_E, 0xB3FC6C67,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-C: PVI Progamming"                      }, // E/-
  { ELEKTOR , MEMMAP_E, 0xDDB40355,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-E: Test Patterns"                       }, // E/-
  { ELEKTOR , MEMMAP_E, 0x2C0B67B9,  4 * KILOBYTE,  0, 10,  7, 11, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-F: Lotto"                               }, // E/-
  { ELEKTOR , MEMMAP_E, 0x900D9E22,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-009-1: Aggressor"                           }, //*E/-
  { ELEKTOR , MEMMAP_E, 0x5AFF5E53,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-009-2: Offshore Fishing"                    }, //*E/-
  { ELEKTOR , MEMMAP_E, 0x6C09F1E3,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-009-4: Seawar"                              }, // E/-
#ifdef BASICMODE
  { ELEKTOR , MEMMAP_E, 0x43B3295C,  5 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,   265, PVI_CPL, FALSE, 4, 4, "ESS-009-5: Memory"                              }, //*E/F
#else
  { ELEKTOR , MEMMAP_F, 0x43B3295C,  5 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,   265, PVI_CPL, FALSE, 4, 4, "ESS-009-5: Memory"                              }, //*E/F
#endif
  { ELEKTOR , MEMMAP_E, 0xA19AB7FB,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-009-7: Destroyer"                           }, //*E/-
  { ELEKTOR , MEMMAP_E, 0xEEA5525D,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-009-8: Starship Enterprise"                 }, //*E/-
  { ELEKTOR , MEMMAP_E, 0xAD905B98,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-009-9: Blackjack"                           }, // E/-
  { ELEKTOR , MEMMAP_E, 0x068165A0,  4 * KILOBYTE, 10,  0, 11, 10, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-009-A: Card Trick"                          }, // E/-
  { ELEKTOR , MEMMAP_E, 0x8C1579D0,  4 * KILOBYTE, 10,  0, 11,  9, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-009-B: Awari"                               }, // E/-
  { ELEKTOR , MEMMAP_E, 0x7BE35693,  4 * KILOBYTE,  3,  3,  3,  3, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, TRUE,  TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 1, 4, "ESS-009-C: UFO Shooting"                        }, // E/-
  { ELEKTOR , MEMMAP_E, 0x567D3224,  4 * KILOBYTE,  8,  1, 11,  8, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-009-D: Raster"                              }, //*E/-
  { ELEKTOR , MEMMAP_E, 0xFC329059,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE,      57, FALSE, 4, 4, "ESS-009-E: Nim"                                 }, // E/-
  { ELEKTOR , MEMMAP_E, 0xA3525ACC,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-009-F: Solitaire"                           }, // E/-
  { ELEKTOR , MEMMAP_F, 0xF4E4D797,  6 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-010-2: Pinball"                             }, //*-/F
  { ELEKTOR , MEMMAP_E, 0x13E56649,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-010-3: Helicopter"                          }, //*E/-
  { ELEKTOR , MEMMAP_E, 0x28828BCC,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE,      55, FALSE, 4, 4, "ESS-010-4: Hard Nut"                            }, // E/-
  { ELEKTOR , MEMMAP_E, 0x410D3424,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-010-5: Catapult"                            }, //*E/-
  { ELEKTOR , MEMMAP_F, 0x9C6B6EBB,  6 * KILOBYTE,  2,  1,  3,  0, TRUE,  SENSITIVITY_MEDIUM , FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE,      58, FALSE, 4, 4, "ESS-010-6: Pilot"                               }, // -/F
  { ELEKTOR , MEMMAP_E, 0x0048322B,  4 * KILOBYTE,  2,  1,  3,  0, TRUE,  SENSITIVITY_LOW    , FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-010-7: Attack from Space"                   }, //*E/-
  { ELEKTOR , MEMMAP_E, 0x15A90EB4,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-010-8: Reverse"                             }, //*E/-
  { ELEKTOR , MEMMAP_E, 0xE34220A0,  4 * KILOBYTE,  1,  2,  3,  0, TRUE,  SENSITIVITY_MEDIUM , FALSE, FALSE, FALSE, TRUE ,  4, 16,   220, PVI_CPL, FALSE, 4, 4, "ESS-010-B: Bursting Balloons"                   }, //*E/-
  { ELEKTOR , MEMMAP_F, 0x5D1C7D36,  6 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-010-C: Fr Elise"                           }, // -/F
  { ELEKTOR , MEMMAP_F, 0x09317D8A,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE,      54, FALSE, 4, 4, "ESS-010-1x: Change Words for Hangman"           }, // -/F
  { ELEKTOR , MEMMAP_E, 0x1527E074,  4 * KILOBYTE,  8,  1,  1,  8, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-1: Snakes and Ladders"                  }, // E/-
  { ELEKTOR , MEMMAP_E, 0x48EBA79F,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE,      61, FALSE, 4, 4, "ESS-011-2: Molebasher"                          }, //*E/-
  { ELEKTOR , MEMMAP_E, 0x46E1C06B,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-3: Snap"                                }, // E/-
#ifdef BASICMODE
  { ELEKTOR , MEMMAP_E, 0xB03A3EBD,  5 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-4: Mazes"                               }, //*E/F
  { ELEKTOR , MEMMAP_E, 0xD194B2D9,  6 * KILOBYTE,  8,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, TRUE,  FALSE, FALSE, TRUE ,  1,  2, FUDGE, PVI_CPL, FALSE, 0, 4, "ESS-011-5: Asteroids"                           }, //*E/F
#else
  { ELEKTOR , MEMMAP_F, 0xB03A3EBD,  5 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-4: Mazes"                               }, //*E/F
  { ELEKTOR , MEMMAP_F, 0xD194B2D9,  6 * KILOBYTE,  8,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, TRUE,  FALSE, FALSE, TRUE ,  1,  2, FUDGE, PVI_CPL, FALSE, 0, 4, "ESS-011-5: Asteroids"                           }, //*E/F
#endif
  { ELEKTOR , MEMMAP_E, 0x429D0A9B,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-7: Omega Landing"                       }, //*E/-
  { ELEKTOR , MEMMAP_E, 0xAB065283,  4 * KILOBYTE,  1,  2,  3,  0, TRUE,  SENSITIVITY_MEDIUM , FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-8: Breakout"                            }, //*E/-
  { ELEKTOR , MEMMAP_F, 0xB5A9FEDD,  5 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, FALSE,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-9: Tiny Tim"                            }, //*-/F
  { ELEKTOR , MEMMAP_E, 0x4EE3AE0F,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-A: Horse Race + Jackpot"                }, //*E/-
  { ELEKTOR , MEMMAP_E, 0xF544EEF0,  4 * KILOBYTE,  2,  1,  3,  0, TRUE,  SENSITIVITY_HIGH   , FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE,      55, FALSE, 4, 4, "ESS-011-B: Newton"                              }, //*E/-
  { ELEKTOR , MEMMAP_F, 0xB41E503E,  5 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-C: Horse Races"                         }, //*-/F
  { ELEKTOR , MEMMAP_E, 0x2F10BEE0,  4 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-D: Painting"                            }, //*E/-
  { ELEKTOR , MEMMAP_E, 0x347C584B,  4 * KILOBYTE,  2,  1,  3,  0, TRUE , SENSITIVITY_MEDIUM , FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-E: Submarines + Racing"                 }, //*E/-
// Elektor TV Games Computer book examples (18) (could perhaps show pages for these)
  { ELEKTOR , MEMMAP_E, 0x24DAED9C,  3 * KILOBYTE,  0,  0,  0,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Add 2"                                          }, // E/-
  { ELEKTOR , MEMMAP_E, 0xE1742A1C,  3 * KILOBYTE,  0,  0,  0,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Add 4"                                          }, // E/-
  { ELEKTOR , MEMMAP_E, 0x7256458B,  3 * KILOBYTE,  0,  0,  0,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Minus 2"                                        }, // E/-
  { ELEKTOR , MEMMAP_E, 0xAB977B2D,  3 * KILOBYTE,  0,  0,  0,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Minus 4"                                        }, // E/-
  { ELEKTOR , MEMMAP_F, 0x2AF63BB3,  3 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Explosion"                                      }, // -/F
  { ELEKTOR , MEMMAP_F, 0xC74E777E,  3 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Gunshot"                                        }, // -/F
  { ELEKTOR , MEMMAP_E, 0x18E5FE63,  3 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_VLOW   , FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Joystick"                                       }, // E/-
  { ELEKTOR , MEMMAP_E, 0x449032CB,  3 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Message"                                        }, // E/-
  { ELEKTOR , MEMMAP_E, 0x97B0FBEC,  3 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Practice"                                       }, // E/-
  { ELEKTOR , MEMMAP_F, 0xD849F992,  3 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Siren"                                          }, // -/F
  { ELEKTOR , MEMMAP_E, 0xD20C783D,  3 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Steam Engine"                                   }, // E/-
  { ELEKTOR , MEMMAP_E, 0x7503552D,  3 * KILOBYTE,  3,  3,  3,  3, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Table A"                                        }, //*E/-
  { ELEKTOR , MEMMAP_E, 0x03EEBDFE,  3 * KILOBYTE,  3,  3,  3,  3, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Table B"                                        }, //*E/-
  { ELEKTOR , MEMMAP_E, 0xA82DFDD4,  3 * KILOBYTE,  3,  3,  3,  3, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Table C"                                        }, //*E/-
  { ELEKTOR , MEMMAP_E, 0x5BC2E300,  3 * KILOBYTE,  3,  3,  3,  3, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Table D"                                        }, //*E/-
  { ELEKTOR , MEMMAP_E, 0xA7B6456D,  3 * KILOBYTE,  3,  3,  3,  3, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Table E"                                        }, //*E/-
  { ELEKTOR , MEMMAP_F, 0x8CCB4AFF,  3 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Wedding March"                                  }, // -/F
  { ELEKTOR , MEMMAP_F, 0xE68C74DB,  3 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Wolf Whistle"                                   }, // -/F
// TVC ESS programs
  { ELEKTOR , MEMMAP_E, 0x049B13BE,           774,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-003-1: Play with the PVI"                   }, // E/- $900/$900 spielen_mit_dem_pvi
  { ELEKTOR , MEMMAP_E, 0x3EB20A14,           774,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-003-2: Clock"                               }, // E/-           uhr
  { ELEKTOR , MEMMAP_E, 0x27936552,           774,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,   228, PVI_CPL, FALSE, 4, 4, "ESS-003-3: Music Box"                           }, // E/-           musikbox
  { ELEKTOR , MEMMAP_E, 0xDB0AC12C,          1797, 10, 11, 10, 11, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-003-4: 4 in a Row"                          }, // E/-           4_in_einer_reihe_ess034
  { ELEKTOR , MEMMAP_E, 0x1EE1671F,          1797,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE,      47, FALSE, 4, 4, "ESS-003-5: Surround"                            }, //*E/- $900/$900 surround
  { ELEKTOR , MEMMAP_E, 0x4071464B,           774, 10, 10, 10, 10, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-003-6: Demonstration Program"               }, // E/-           demonstrationsprogramm
  { ELEKTOR , MEMMAP_E, 0xF6EDB39A,          1861,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-006-1: Picture Pattern"                     }, // E/- $8C0/$8C0 bildmuster
  { ELEKTOR , MEMMAP_E, 0x93A4C8E8,          1797,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-006-2: PVI Programming"                     }, // E/- $900/$900 pvi_programmierung_ess062
  { ELEKTOR , MEMMAP_E, 0xF8B41252,          1077,  8,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-006-3: Rocket Shooting"                     }, //*E/- $8C0/$900 rakete_abschiessen
  { ELEKTOR , MEMMAP_E, 0xBC5443B5,          1747,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-1: Mastermind"                          }, // E/-
  { ELEKTOR , MEMMAP_E, 0x8FA9AAAF,          1541,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-2: Code Breaker"                        }, // E/- $900/$900 kode_knacker
  { ELEKTOR , MEMMAP_E, 0x78407345,          1797,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, TRUE , 4, 4, "ESS-007-3: Reversi"                             }, // E/- $900/$900 reversi
  { ELEKTOR , MEMMAP_E, 0xCCDF96C7,          1850,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-4: Amazone"                             }, //*E/-           amazone
  { ELEKTOR , MEMMAP_E, 0x6B1B3DF7,          1482,  8,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-5: Space Shoot-Out"                     }, //*E/- $8C0/$900 weltraumschlacht
  { ELEKTOR , MEMMAP_E, 0xFB5369A2,          1285, 10, 11, 10, 11, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-6: 4 in a Row"                          }, // E/-           4_in_einer_reihe_ess076
  { ELEKTOR , MEMMAP_E, 0x0F89B852,          1453, 10, 11, 10, 11, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-7: 4 in a Row"                          }, // E/-           4_in_einer_reihe
  { ELEKTOR , MEMMAP_E, 0x15F1D592,          1826,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-8: Jackpot"                             }, // E/-
  { ELEKTOR , MEMMAP_E, 0xB09B8DFB,          1797,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE,      47, FALSE, 4, 4, "ESS-007-9: Surround"                            }, //*E/- $900/$900 umzingelung
  { ELEKTOR , MEMMAP_E, 0x9853EB12,           511, 10, 10, 10, 10, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, FALSE,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-A: Shapes"                              }, // E/- $900/$900 figurenraten
  { ELEKTOR , MEMMAP_E, 0xCAC3F0CD,           719,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,   228, PVI_CPL, FALSE, 4, 4, "ESS-007-B: Piano"                               }, // E/-
  { ELEKTOR , MEMMAP_E, 0xC008BA46,          1797,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-C: PVI Progamming"                      }, // E/- $900/$900 pvi_programmierung
  { ELEKTOR , MEMMAP_E, 0x50BFE2F1,          5875,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-D: Disassembler"                        }, // E/-           disassembler
  { ELEKTOR , MEMMAP_E, 0xF6EDB39A,          1861,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-E: Test Patterns"                       }, // E/- $8C0/$8C0 testbilder
  { ELEKTOR , MEMMAP_E, 0x5FFFB198,          1797,  0, 10,  7, 11, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-007-F: Lotto"                               }, // E/-           lotto
  { ELEKTOR , MEMMAP_E, 0xA472ED35,          1827,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-009-1: Aggressor"                           }, //*E/-           angreifer
  { ELEKTOR , MEMMAP_E, 0x67C7D6D7,          1861,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-009-2: Offshore Fishing"                    }, //*E/- $8C0/$900 hochsee_angeln
  { ELEKTOR , MEMMAP_E, 0x00D8860D,          5874,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-009-3: Circledrive"                         }, //*E/-           rundfahrt
  { ELEKTOR , MEMMAP_E, 0x1FFD27C2,          1797,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-009-4: Seawar"                              }, //*E/-           seekrieg
#ifdef BASICMODE
  { ELEKTOR , MEMMAP_E, 0xD339CBB7,          2390,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,   265, PVI_CPL, FALSE, 4, 4, "ESS-009-5: Memory"                              }, //*E/F $8C0/$900 memory
  { ELEKTOR , MEMMAP_E, 0x3F21BDF6,          4348,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,   257, PVI_CPL, FALSE, 3, 4, "ESS-009-6: Labyrinth"                           }, // E/F $8C0/$900 labyrinth
#else
  { ELEKTOR , MEMMAP_F, 0xD339CBB7,          2390,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,   265, PVI_CPL, FALSE, 4, 4, "ESS-009-5: Memory"                              }, //*E/F $8C0/$900 memory
  { ELEKTOR , MEMMAP_F, 0x3F21BDF6,          4348,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,   257, PVI_CPL, FALSE, 3, 4, "ESS-009-6: Labyrinth"                           }, // E/F $8C0/$900 labyrinth
#endif
  { ELEKTOR , MEMMAP_E, 0xDCC5FC4F,          1784,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-009-7: Destroyer"                           }, //*E/-           zerstoerer
  { ELEKTOR , MEMMAP_E, 0xE9937B86,          1501,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-009-8: Starship Enterprise"                 }, //*E/-           raumschiff_enterprise
  { ELEKTOR , MEMMAP_E, 0x90ACD31C,          1861,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-009-9: Blackjack"                           }, // E/-           17_und_4
  { ELEKTOR , MEMMAP_E, 0x7FF32E0B,          1857, 10,  0, 11, 10, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-009-A: Card Trick"                          }, // E/-           kartentrick
  { ELEKTOR , MEMMAP_E, 0xB129F154,          1861, 10,  0, 11,  9, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-009-B: Awari"                               }, // E/-           awari
  { ELEKTOR , MEMMAP_E, 0x825CAECC,           985,  3,  3,  3,  3, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, TRUE,  TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 1, 4, "ESS-009-C: UFO Shooting"                        }, // E/- $8C0/$932 ufo_abschuss
  { ELEKTOR , MEMMAP_E, 0x84660731,          1332,  8,  1, 11,  8, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-009-D: Raster"                              }, //*E/- $8C0/$900 raster
  { ELEKTOR , MEMMAP_E, 0xC8466D70,          1813,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE,      57, FALSE, 4, 4, "ESS-009-E: Nim"                                 }, // E/-           nim
  { ELEKTOR , MEMMAP_E, 0x9E6ED248,          1861,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-009-F: Solitaire"                           }, // E/-           solitaer
  { ELEKTOR , MEMMAP_E, 0x39402B84,          5874,  8,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-010-1: Invaders"                            }, //*E/-           eindringlinge
  { ELEKTOR , MEMMAP_F, 0x5A2DA9C2,          3396,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-010-2: Pinball"                             }, //*-/F           flipper
  { ELEKTOR , MEMMAP_E, 0x2ED9EECD,          1861,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-010-3: Helicopter"                          }, //*E/- $8C0/$900 hubschrauber
  { ELEKTOR , MEMMAP_E, 0x5B765DED,          1797,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE,      55, FALSE, 4, 4, "ESS-010-4: Hard Nut"                            }, // E/-           harte_nuss
  { ELEKTOR , MEMMAP_E, 0xCEA44AF4,          1707,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-010-5: Catapult"                            }, //*E/-           katapult
  { ELEKTOR , MEMMAP_F, 0xA287992D,          3396,  2,  1,  3,  0, TRUE,  SENSITIVITY_MEDIUM , FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE,      58, FALSE, 4, 4, "ESS-010-6: Pilot"                               }, // -/F           pilot
  { ELEKTOR , MEMMAP_E, 0x3D74BAAF,          1861,  2,  1,  3,  0, TRUE,  SENSITIVITY_LOW    , FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-010-7: Attack from Space"                   }, //*E/-           angriff_aus_dem_all
  { ELEKTOR , MEMMAP_E, 0xDFDE0699,          1842,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-010-8: Reverse"                             }, //*E/- $8C0/$900 reverse
  { ELEKTOR , MEMMAP_E, 0x76F288A3,          5875,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-010-9: Rocket Hunting"                      }, //*E/- $8C0/$900
  { ELEKTOR , MEMMAP_E, 0xADB2A6AF,          5872,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-010-9: Rocket Hunting (bad dump)"           }, //*E/- $8C3/$900 raketenjagd
  { ELEKTOR , MEMMAP_F, 0x36541AE3,          4165,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,   258, PVI_CPL, FALSE, 4, 4, "ESS-010-A: Basketball"                          }, //*-/F           basketball
  { ELEKTOR , MEMMAP_E, 0xDE7EA824,          1861,  1,  2,  3,  0, TRUE,  SENSITIVITY_MEDIUM , FALSE, FALSE, FALSE, TRUE ,  4, 16,   220, PVI_CPL, FALSE, 4, 4, "ESS-010-B: Bursting Balloons"                   }, //*E/- $8C0/$900 platzende_ballons_ess10b
  { ELEKTOR , MEMMAP_F, 0x42B805C6,          3158,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-010-C: Fr Elise"                           }, // -/F           fuer_elise
  { ELEKTOR , MEMMAP_F, 0x9D1A5ED0,          1541,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-010-D: Editor"                              }, // -/F           editor
  { ELEKTOR , MEMMAP_F, 0x4CB754CE,          4101,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,   266,      54, FALSE, 4, 4, "ESS-010-E: Galgenspiel"                         }, // -/F           galgenspiel
  { ELEKTOR , MEMMAP_F, 0xFE5FCFB5,          4165,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16,   266,      54, FALSE, 4, 4, "ESS-010-F: Hangman"                             }, // -/F           hangman
  { ELEKTOR , MEMMAP_F, 0x3D4580A3,          1813,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE,      54, FALSE, 4, 4, "ESS-010-1x: Change Words for Hangman"           }, // -/F           wortaenderungsprogramm
  { ELEKTOR , MEMMAP_E, 0x281B68F0,          1861,  8,  1,  1,  8, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-1: Snakes and Ladders"                  }, // E/-           schlangen_und_leitern
  { ELEKTOR , MEMMAP_E, 0x39B9616B,          1851,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE,      61, FALSE, 4, 4, "ESS-011-2: Molebasher"                          }, //*E/- $8C0/$900 maulwurfschlaeger
  { ELEKTOR , MEMMAP_E, 0x7BDD48EF,          1861,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-3: Snap"                                }, // E/-           identa
#ifdef BASICMODE
  { ELEKTOR , MEMMAP_E, 0xFB3727C4,          2130,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-4: Mazes"                               }, //*E/F $8C0/$900 labyrinthe
  { ELEKTOR , MEMMAP_E, 0x6B20453B,          2954,  8,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, TRUE,  FALSE, FALSE, TRUE ,  1,  2, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-5: Asteroids"                           }, //*E/F
#else
  { ELEKTOR , MEMMAP_F, 0xFB3727C4,          2130,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-4: Mazes"                               }, //*E/F $8C0/$900 labyrinthe
  { ELEKTOR , MEMMAP_F, 0x6B20453B,          2954,  8,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, TRUE,  FALSE, FALSE, TRUE ,  1,  2, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-5: Asteroids"                           }, //*E/F
#endif
  { ELEKTOR , MEMMAP_F, 0x7F67C41A,          4101,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, TRUE , 4, 4, "ESS-011-6: Dragster"                            }, //*-/F           autorennen
  { ELEKTOR , MEMMAP_E, 0x7FA1821F,          1861,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-7: Omega Landing"                       }, //*E/-           landemanoever
  { ELEKTOR , MEMMAP_E, 0x963ADA07,          1861,  1,  2,  3,  0, TRUE,  SENSITIVITY_MEDIUM , FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-8: Breakout"                            }, //*E/- $8C0/$900 platzende_ballons
  { ELEKTOR , MEMMAP_F, 0x94417986,          2373,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, FALSE,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-9: Tiny Tim"                            }, //*-/F           der_kleine_tim
  { ELEKTOR , MEMMAP_E, 0x73DF268B,          1861,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-A: Horse Race + Jackpot"                }, //*E/- $8C0/$900 pferdewette_einarmiger_bandit
  { ELEKTOR , MEMMAP_E, 0x86B038D1,          1797,  2,  1,  3,  0, TRUE,  SENSITIVITY_HIGH   , FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE,      55, FALSE, 4, 4, "ESS-011-B: Newton"                              }, //*E/-           apfelfangen
  { ELEKTOR , MEMMAP_F, 0x4CDBB610,          2373,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-C: Horse Races"                         }, //*-/F $8C0/$900 springturnier
  { ELEKTOR , MEMMAP_E, 0x122C3664,          1861,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-D: Painting"                            }, //*E/- $8C0/$900 malkasten
  { ELEKTOR , MEMMAP_E, 0xF9F6942A,          1860,  2,  1,  3,  0, TRUE , SENSITIVITY_MEDIUM , FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-E: Submarines + Racing"                 }, //*E/- $8C0/$900 uboot_gegenverkehr
  { ELEKTOR , MEMMAP_F, 0xE059D788,          4573,  1,  2,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  1, 50, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-011-F: Cosmic Adventure"                    }, //*-/F $8C0/$900 galactische_eroberung
// TVC book examples
  { ELEKTOR , MEMMAP_E, 0x57A7FE9E,            45,  0,  0,  0,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Add 2"                                          }, // E/-
  { ELEKTOR , MEMMAP_E, 0xBC3D5AB2,            65,  0,  0,  0,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Add 4"                                          }, // E/-
  { ELEKTOR , MEMMAP_E, 0x6CD80D9E,            45,  0,  0,  0,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Minus 2"                                        }, // E/-
  { ELEKTOR , MEMMAP_E, 0x03F71EAE,            65,  0,  0,  0,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Minus 4"                                        }, // E/-
  { ELEKTOR , MEMMAP_E, 0x0474A87C,           225,  2,  1,  3,  0, FALSE,                   1, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Joystick"                                       }, // E/-
  { ELEKTOR , MEMMAP_E, 0xF150EFA1,           414,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Message"                                        }, // E/- $8B9/$9C7
  { ELEKTOR , MEMMAP_E, 0x1D48BE29,            98,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Practice"                                       }, // E/-
  { ELEKTOR , MEMMAP_E, 0x3ED1B240,           455,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Steam Engine"                                   }, // E/-
  { ELEKTOR , MEMMAP_E, 0x26F6F864,            60,  3,  3,  3,  3, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Table A"                                        }, //*E/- aka ship.tvc
  { ELEKTOR , MEMMAP_E, 0xD7B6FC2B,            94,  3,  3,  3,  3, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Table B"                                        }, //*E/-
  { ELEKTOR , MEMMAP_E, 0xDD2A5CAF,           175,  3,  3,  3,  3, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Table C"                                        }, //*E/-
  { ELEKTOR , MEMMAP_E, 0x1315D774,           237,  3,  3,  3,  3, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Table D (bad dump)"                             }, //*E/-
  { ELEKTOR , MEMMAP_E, 0x1B3D8BDA,           278,  3,  3,  3,  3, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Table E (bad dump)"                             }, //*E/-
// there are no available TVC versions of PSG examples (Explosion, Gunshot, Siren, Wedding March, Wolf Whistle)
  { ELEKTOR , MEMMAP_E, 0x67BED55D,  8 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "ESS-009-3: Circledrive (patched)"               }, // E/-
  { ARCADIA , MEMMAP_G, 0xC00DE933,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Scroller 1.0"                                   }, //*
  { ARCADIA , MEMMAP_G, 0x2AE9ABE1,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Arcadia stub 1.0"                               },
  { INTERTON, MEMMAP_A, 0x4CDC29A8,  2 * KILOBYTE,  2,  1,  3,  0, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Interton stub 1.0"                              },
  { ELEKTOR , MEMMAP_E, 0x272D5A21,           237,  3,  3,  3,  3, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Table D"                                        }, //*E/-
  { ELEKTOR , MEMMAP_E, 0xD38E95DB,           278,  3,  3,  3,  3, FALSE, SENSITIVITY_DEFAULT, FALSE, FALSE, FALSE, TRUE ,  4, 16, FUDGE, PVI_CPL, FALSE, 4, 4, "Table E"                                        }  //*E/-
};

/* Ratio numbers (eg. "1:1") indicate the demultiplexing pattern, as
   follows. After this is shown which game objects are multiplexed.

   "1:1" means a cycle of 2 frames, with each of the 2 sprite sets drawn
   for 1 frame. (Suitable for full demultiplexing.)

   "2:2" means a cycle of 4 frames, with each of the 2 sprite sets drawn
   for 2 frames.

   "1:1:1:1" means a cycle of 4 frames, with each of the 4 sprite sets
   drawn for 1 frame.

   MACROSS: enemies are multiplexed. Outside, a cycle of 4 frames, with
   each of the 2 sprite sets drawn for 2 frames. Inside, a cycle of 14
   frames: set 1 for 4 frames, set 2 for 4 frames, set 1 for 2 frames,
   set 2 for 4 frames. Inside, player bullets and enemy bullets are also
   multiplexed (not in quite the same multiplexing pattern though).

   R2DTANK: bullets are multiplexed. A cycle of 3 frames, with the bullet
   drawn for 1 frame, and undrawn for 2 frames. */

EXPORT RECT keyrect[3][KEYS] = { // they go left, top, right, bottom
// ARCADIA
{
{ 174,  73, 174 + 14,  73 + 9 }, // START
{ 193,  73, 193 + 14,  73 + 9 }, // A
{ 212,  73, 212 + 14,  73 + 9 }, // B
{ 168, 115, 168 + 36, 115 + 9 }, // RESET

{  37,  96,  37 + 14,  96 + 9 }, // '1'
{  38, 109,  38 + 14, 109 + 9 }, // '4'
{  39, 122,  39 + 14, 122 + 9 }, // '7'
{  40, 135,  40 + 14, 135 + 9 }, // 'Cl'
{  58,  95,  58 + 14,  95 + 9 }, // '2'
{  59, 108,  59 + 14, 108 + 9 }, // '5'
{  60, 121,  60 + 14, 121 + 9 }, // '8'
{  61, 134,  61 + 14, 134 + 9 }, // '0'
{  79,  94,  79 + 14,  94 + 9 }, // '3'
{  80, 107,  80 + 14, 107 + 9 }, // '6'
{  81, 120,  81 + 14, 120 + 9 }, // '9'
{  82, 133,  82 + 14, 133 + 9 }, // 'En'

{ 272,  90, 272 + 15,  90 + 9 }, // '1'
{ 272, 102, 272 + 15, 102 + 9 }, // '4'
{ 272, 115, 272 + 15, 115 + 9 }, // '7'
{ 272, 127, 272 + 15, 127 + 9 }, // 'Cl'
{ 293,  90, 293 + 15,  90 + 9 }, // '2'
{ 293, 102, 293 + 15, 102 + 9 }, // '5'
{ 293, 115, 293 + 15, 115 + 9 }, // '8'
{ 293, 127, 293 + 15, 127 + 9 }, // '0'
{ 314,  90, 314 + 15,  90 + 9 }, // '3'
{ 314, 102, 314 + 15, 102 + 9 }, // '6'
{ 314, 115, 314 + 15, 115 + 9 }, // '9'
{ 314, 127, 314 + 15, 127 + 9 }, // 'En'

{  57,  30,  57 +  8,  30 + 8 }, // 'I'
{  41,  45,  41 +  8,  45 + 8 }, // 'J'
{  57,  62,  57 +  8,  62 + 8 }, // 'K'
{  73,  45,  73 +  8,  45 + 8 }, // 'L'

{ 295,  29, 295 +  8,  29 + 8 }, // up
{ 278,  44, 278 +  8,  44 + 8 }, // left
{ 295,  57, 295 +  8,  57 + 8 }, // down
{ 310,  44, 310 +  8,  44 + 8 }, // right

{  36,  83,  36 + 14,  83 + 9 }, // p1 '2'
{  78,  81,  78 + 14,  81 + 9 }, // p1 '2'

{ 272,  78, 272 + 15,  78 + 9 }, // p2 '2'
{ 314,  78, 314 + 15,  78 + 9 }  // p2 '2'
},
// INTERTON
{
{ 199, 172, 199 + 29, 172 + 28 }, // START
{ 166, 172, 166 + 29, 172 + 28 }, // A
{   0,   0,        0,        0 }, // B (impossible)
{   0,   0,        0,        0 }, // RESET (impossible)

{  23, 134,  23 + 15, 134 +  9 }, // '1'
{  23, 151,  23 + 15, 151 +  9 }, // '4'
{  23, 168,  23 + 15, 168 +  9 }, // '7'
{  23, 185,  23 + 15, 185 +  9 }, // 'Cl'
{  42, 134,  42 + 15, 134 +  9 }, // '2'
{  42, 151,  42 + 15, 151 +  9 }, // '5'
{  42, 168,  42 + 15, 168 +  9 }, // '8'
{  42, 185,  42 + 15, 185 +  9 }, // '0'
{  61, 134,  61 + 15, 134 +  9 }, // '3'
{  61, 151,  61 + 15, 151 +  9 }, // '6'
{  61, 168,  61 + 15, 168 +  9 }, // '9'
{  61, 185,  61 + 15, 185 +  9 }, // 'En'

{ 249, 134, 249 + 15, 134 +  9 }, // '1'
{ 249, 150, 249 + 15, 150 +  9 }, // '4'
{ 249, 167, 249 + 15, 167 +  9 }, // '7'
{ 249, 183, 249 + 15, 183 +  9 }, // 'Cl'
{ 268, 134, 268 + 15, 134 +  9 }, // '2'
{ 268, 150, 268 + 15, 150 +  9 }, // '5'
{ 268, 167, 268 + 15, 167 +  9 }, // '8'
{ 268, 183, 268 + 15, 183 +  9 }, // '0'
{ 287, 134, 287 + 15, 134 +  9 }, // '3'
{ 287, 150, 287 + 15, 150 +  9 }, // '6'
{ 287, 167, 287 + 15, 167 +  9 }, // '9'
{ 287, 183, 287 + 15, 183 +  9 }, // 'En'

{  49,  55,  49 +  8,  55 +  8 }, // 'I'
{  30,  71,  30 +  8,  71 +  8 }, // 'J'
{  49,  87,  49 +  8,  87 +  8 }, // 'K'
{  66,  71,  66 +  8,  71 +  8 }, // 'L'

{ 268,  56, 268 +  8,  56 +  8 }, // up
{ 251,  72, 251 +  8,  72 +  8 }, // left
{ 268,  88, 268 +  8,  88 +  8 }, // down
{ 286,  72, 286 +  8,  72 +  8 }, // right

{  23, 117,  23 + 15, 117 +  9 }, // p1 '2'
{  61, 117,  61 + 15, 117 +  9 }, // p1 '2'

{ 249, 117, 249 + 15, 117 +  9 }, // p2 '2'
{ 287, 117, 287 + 15, 117 +  9 }  // p2 '2'
},
// ELEKTOR
{
{  21, 100,  21 + 29, 100 + 25 }, // START
{  21,  62,  21 + 29,  62 + 25 }, // UC
{  21, 138,  21 + 29, 138 + 26 }, // LC
{  21, 177,  21 + 29, 177 + 25 }, // RESET

{  77,  62,  77 + 28,  62 + 25 }, // RCAS
{  77, 100,  77 + 28, 100 + 25 }, // BP1/2
{  77, 138,  77 + 28, 138 + 26 }, // PC
{  77, 177,  77 + 28, 177 + 25 }, // -
{ 104,  62, 104 + 29,  62 + 25 }, // WCAS
{ 104, 100, 104 + 29, 100 + 25 }, // REG
{ 104, 138, 104 + 29, 138 + 26 }, // MEM
{ 104, 177, 104 + 29, 177 + 25 }, // +
{ 132,  62, 132 + 28,  62 + 25 }, // C
{ 132, 100, 132 + 28, 100 + 25 }, // 8
{ 132, 138, 132 + 28, 138 + 26 }, // 4
{ 132, 177, 132 + 28, 177 + 25 }, // 0

{ 203,  62, 203 + 28,  62 + 25 }, // D
{ 203, 100, 203 + 28, 100 + 25 }, // 9
{ 203, 138, 203 + 28, 138 + 26 }, // 5
{ 203, 177, 203 + 28, 177 + 25 }, // 1
{ 230,  62, 230 + 29,  62 + 25 }, // E
{ 230, 100, 230 + 29, 100 + 25 }, // A
{ 230, 138, 230 + 29, 138 + 26 }, // 6
{ 230, 177, 230 + 29, 177 + 25 }, // 2
{ 258,  62, 258 + 28,  62 + 25 }, // F
{ 258, 100, 258 + 28, 100 + 25 }, // B
{ 258, 138, 258 + 28, 138 + 26 }, // 7
{ 258, 177, 258 + 28, 277 + 25 }, // 3

{ 115,  21, 115 +  8,  21 +  8 }, // 'I'
{ 109,  27, 109 +  8,  27 +  8 }, // 'J'
{ 115,  33, 115 +  8,  33 +  8 }, // 'K'
{ 121,  27, 121 +  8,  27 +  8 }, // 'L'

{ 240,  21, 240 +  8,  21 +  8 }, // up
{ 234,  27, 234 +  8,  27 +  8 }, // left
{ 240,  33, 240 +  8,  33 +  8 }, // down
{ 246,  27, 246 +  8,  27 +  8 }  // right
/*
{   0,   0,   0,   0 }, // p1 '2'
{   0,   0,   0,   0 }, // p1 '2'

{   0,   0,   0,   0 }, // p2 '2'
{   0,   0,   0,   0 }  // p2 '2'
*/
} };

MODULE struct
{   const UWORD address;
    const UBYTE mask;
} a_keyemul[KEYS] =
{   { A_CONSOLE,      0x01 }, // START
    { A_CONSOLE,      0x02 }, // A
    { A_CONSOLE,      0x04 }, // B
    { 0,                 0 }, // RESET
    { A_P1LEFTKEYS,   0x08 }, // p1 '1'
    { A_P1LEFTKEYS,   0x04 }, // p1 '4'
    { A_P1LEFTKEYS,   0x02 }, // p1 '7'
    { A_P1LEFTKEYS,   0x01 }, // p1 'Cl'
    { A_P1MIDDLEKEYS, 0x08 }, // p1 '2'
    { A_P1MIDDLEKEYS, 0x04 }, // p1 '5'
    { A_P1MIDDLEKEYS, 0x02 }, // p1 '8'
    { A_P1MIDDLEKEYS, 0x01 }, // p1 '0'
    { A_P1RIGHTKEYS,  0x08 }, // p1 '3'
    { A_P1RIGHTKEYS,  0x04 }, // p1 '6'
    { A_P1RIGHTKEYS,  0x02 }, // p1 '9'
    { A_P1RIGHTKEYS,  0x01 }, // p1 'En'
    { A_P2LEFTKEYS,   0x08 }, // p2 '1'
    { A_P2LEFTKEYS,   0x04 }, // p2 '4'
    { A_P2LEFTKEYS,   0x02 }, // p2 '7'
    { A_P2LEFTKEYS,   0x01 }, // p2 'Cl'
    { A_P2MIDDLEKEYS, 0x08 }, // p2 '2'
    { A_P2MIDDLEKEYS, 0x04 }, // p2 '5'
    { A_P2MIDDLEKEYS, 0x02 }, // p2 '8'
    { A_P2MIDDLEKEYS, 0x01 }, // p2 '0'
    { A_P2RIGHTKEYS,  0x08 }, // p2 '3'
    { A_P2RIGHTKEYS,  0x04 }, // p2 '6'
    { A_P2RIGHTKEYS,  0x02 }, // p2 '9'
    { A_P2RIGHTKEYS,  0x01 }, // p2 'En'
    { 0,                 0 }, // p1 up
    { 0,                 0 }, // p1 left
    { 0,                 0 }, // p1 down
    { 0,                 0 }, // p1 right
    { 0,                 0 }, // p2 up
    { 0,                 0 }, // p2 left
    { 0,                 0 }, // p2 down
    { 0,                 0 }, // p2 right
    { A_P1MIDDLEKEYS, 0x08 }, // p1 '2'
    { A_P1MIDDLEKEYS, 0x08 }, // p1 '2'
    { A_P2MIDDLEKEYS, 0x08 }, // p1 '2'
    { A_P2MIDDLEKEYS, 0x08 }  // p1 '2'
}, i_keyemul[KEYS] =
{   { I_CONSOLE,      0x40 }, // START
    { I_CONSOLE,      0x80 }, // A
    { 0,                 0 }, // B
    { 0,                 0 }, // RESET
    { I_P1LEFTKEYS,   0x80 }, // p1 '1'
    { I_P1LEFTKEYS,   0x40 }, // p1 '4'
    { I_P1LEFTKEYS,   0x20 }, // p1 '7'
    { I_P1LEFTKEYS,   0x10 }, // p1 'Cl'
    { I_P1MIDDLEKEYS, 0x80 }, // p1 '2'
    { I_P1MIDDLEKEYS, 0x40 }, // p1 '5'
    { I_P1MIDDLEKEYS, 0x20 }, // p1 '8'
    { I_P1MIDDLEKEYS, 0x10 }, // p1 '0'
    { I_P1RIGHTKEYS,  0x80 }, // p1 '3'
    { I_P1RIGHTKEYS,  0x40 }, // p1 '6'
    { I_P1RIGHTKEYS,  0x20 }, // p1 '9'
    { I_P1RIGHTKEYS,  0x10 }, // p1 'En'
    { I_P2LEFTKEYS,   0x80 }, // p2 '1'
    { I_P2LEFTKEYS,   0x40 }, // p2 '4'
    { I_P2LEFTKEYS,   0x20 }, // p2 '7'
    { I_P2LEFTKEYS,   0x10 }, // p2 'Cl'
    { I_P2MIDDLEKEYS, 0x80 }, // p2 '2'
    { I_P2MIDDLEKEYS, 0x40 }, // p2 '5'
    { I_P2MIDDLEKEYS, 0x20 }, // p2 '8'
    { I_P2MIDDLEKEYS, 0x10 }, // p2 '0'
    { I_P2RIGHTKEYS,  0x80 }, // p2 '3'
    { I_P2RIGHTKEYS,  0x40 }, // p2 '6'
    { I_P2RIGHTKEYS,  0x20 }, // p2 '9'
    { I_P2RIGHTKEYS,  0x10 }, // p2 'En'
    { 0,                 0 }, // p1 up
    { 0,                 0 }, // p1 left
    { 0,                 0 }, // p1 down
    { 0,                 0 }, // p1 right
    { 0,                 0 }, // p2 up
    { 0,                 0 }, // p2 left
    { 0,                 0 }, // p2 down
    { 0,                 0 }, // p2 right
    { I_P1MIDDLEKEYS, 0x80 }, // p1 '2'
    { I_P1MIDDLEKEYS, 0x80 }, // p1 '2'
    { I_P2MIDDLEKEYS, 0x80 }, // p1 '2'
    { I_P2MIDDLEKEYS, 0x80 }  // p1 '2'
}, e_keyemul[36] =
{   { I_CONSOLE,      0x40 }, // START
    { I_CONSOLE,      0x80 }, // UC
    { I_CONSOLE,      0x20 }, // LC
    { I_CONSOLE,      0x10 }, // RESET
    { I_P1LEFTKEYS,   0x80 }, // p1
    { I_P1LEFTKEYS,   0x40 }, // p1
    { I_P1LEFTKEYS,   0x20 }, // p1
    { I_P1LEFTKEYS,   0x10 }, // p1
    { I_P1MIDDLEKEYS, 0x80 }, // p1
    { I_P1MIDDLEKEYS, 0x40 }, // p1
    { I_P1MIDDLEKEYS, 0x20 }, // p1
    { I_P1MIDDLEKEYS, 0x10 }, // p1
    { I_P1RIGHTKEYS,  0x80 }, // p1
    { I_P1RIGHTKEYS,  0x40 }, // p1
    { I_P1RIGHTKEYS,  0x20 }, // p1
    { I_P1RIGHTKEYS,  0x10 }, // p1
    { I_P2LEFTKEYS,   0x80 }, // p2
    { I_P2LEFTKEYS,   0x40 }, // p2
    { I_P2LEFTKEYS,   0x20 }, // p2
    { I_P2LEFTKEYS,   0x10 }, // p2
    { I_P2MIDDLEKEYS, 0x80 }, // p2
    { I_P2MIDDLEKEYS, 0x40 }, // p2
    { I_P2MIDDLEKEYS, 0x20 }, // p2
    { I_P2MIDDLEKEYS, 0x10 }, // p2
    { I_P2RIGHTKEYS,  0x80 }, // p2
    { I_P2RIGHTKEYS,  0x40 }, // p2
    { I_P2RIGHTKEYS,  0x20 }, // p2
    { I_P2RIGHTKEYS,  0x10 }, // p2
    { 0,                 0 }, // p1 up
    { 0,                 0 }, // p1 left
    { 0,                 0 }, // p1 down
    { 0,                 0 }, // p1 right
    { 0,                 0 }, // p2 up
    { 0,                 0 }, // p2 left
    { 0,                 0 }, // p2 down
    { 0,                 0 }  // p2 right
};

// MODULE ARRAYS----------------------------------------------------------

MODULE UBYTE a_bios[2048] = {
    0x20,0x1B,0x01,0x17,0x92,0x93,0x76,0x20,
    0x77,0x02,0xCC,0x18,0xFC,0xB4,0x80,0x98,
    0x7C,0xB4,0x80,0x18,0x7C,0xC1,0xCD,0x77,
    0xFF,0xCD,0x78,0xFF,0xCD,0x79,0xFF,0xF9,
    0x75,0x04,0xE0,0xCC,0x18,0xFC,0x04,0x7F,
    0xCC,0x19,0xF9,0x04,0x36,0xCC,0x19,0xFB,
    0xCC,0x19,0xFA,0x05,0x50,0x0D,0x60,0x80,
    0xCD,0x78,0x6F,0xF9,0x78,0x05,0x10,0x0D,
    0x60,0x70,0xCD,0x79,0x7F,0xCD,0x79,0x8F,
    0xF9,0x75,0x04,0x71,0xCC,0x18,0xF0,0xCC,
    0x18,0xF2,0x04,0x61,0xCC,0x18,0xF4,0xCC,
    0x18,0xF6,0x04,0x7B,0xCC,0x18,0xF1,0x04,
    0x83,0xCC,0x18,0xF3,0x04,0x73,0xCC,0x18,
    0xF5,0x04,0x7B,0xCC,0x18,0xF7,0x1B,0x7E,
    0x00,0x01,0x03,0x07,0x0F,0x1F,0x3F,0x7F,
    0xFF,0xFF,0xFE,0xFC,0xF8,0xF0,0xE0,0xC0,
    0x80,0x00,0x00,0x00,0x00,0x00,0x00,0x8C,
    0x8E,0x4C,0x4E,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x8C,0x8E,
    0x4C,0x4E,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x1E,0x26,0x1E,0x2B,0x2C,0x28,0x27,
    0x00,0x1A,0x2B,0x1C,0x1A,0x1D,0x22,0x1A,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x12,
    0x10,0x10,0x11,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
}, i_bios[2048] = {
    0x1F,0x00,0x23,0x0F,0x1F,0xCA,0xF7,0x08,
    0x18,0x01,0x37,0x0F,0x1F,0x65,0xE7,0x01,
    0x18,0x10,0x07,0xFE,0xCF,0x1F,0x0D,0x07,
    0x01,0xCF,0x1F,0x65,0x07,0x1A,0xCF,0x1F,
    0xC1,0x37,0x37,0x20,0x93,0x92,0x76,0x20,
    0x77,0x02,0xC1,0xCD,0x7F,0x00,0xF9,0x7B,
    0x04,0x03,0xCC,0x1F,0xC3,0x04,0x40,0xCC,
    0x1F,0xC8,0x20,0xCC,0x1F,0xC9,0x04,0x1A,
    0xCC,0x1F,0xC6,0x04,0x10,0xCC,0x1F,0x0A,
    0xCC,0x1F,0x0B,0x04,0x30,0xCC,0x1F,0x1A,
    0x04,0x50,0xCC,0x1F,0x2A,0x04,0x70,0xCC,
    0x1F,0x4A,0x04,0x94,0xCC,0x1F,0x0C,0xCC,
    0x1F,0x1C,0xCC,0x1F,0x2C,0xCC,0x1F,0x4C,
    0x04,0xAA,0xCC,0x1F,0xC0,0x04,0x0A,0xCC,
    0x1F,0x0D,0x04,0xFE,0xCC,0x1F,0x1D,0xCC,
    0x1F,0x2D,0xCC,0x1F,0x4D,0x04,0x12,0xCC,
    0x1F,0xC1,0xCC,0x1F,0xC2,0x20,0xCC,0x1F,
    0x65,0x05,0x0A,0x0D,0x60,0xF4,0xCD,0x7E,
    0xFF,0xF9,0x78,0x05,0x0A,0x0D,0x60,0xFE,
    0xCD,0x7F,0x0F,0xF9,0x78,0x05,0x0A,0x0D,
    0x61,0x08,0xCD,0x7F,0x1F,0xF9,0x78,0x05,
    0x0A,0x0D,0x61,0x12,0xCD,0x7F,0x3F,0xF9,
    0x78,0x74,0x20,0x04,0x12,0xCC,0x1F,0xC1,
    0x04,0x0A,0xCC,0x1F,0x0D,0x05,0x0A,0x0D,
    0x60,0xF4,0xCD,0x7E,0xFF,0xF9,0x78,0x04,
    0x94,0xCC,0x1F,0x0C,0x0C,0x1F,0x65,0x18,
    0x7B,0x05,0x0A,0x0D,0x60,0xEA,0xCD,0x7E,
    0xFF,0xF9,0x78,0xB4,0x80,0x98,0x7C,0xB4,
    0x80,0x18,0x7C,0x20,0xCC,0x1F,0x65,0x1F,
    0x00,0xB3,0x00,0xAE,0xA8,0xA8,0x48,0x4E,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0xEC,0x4A,0x4A,0x4A,0xEA,0x00,
    0x00,0x00,0x00,0x00,0xEE,0x48,0x4E,0x48,
    0x4E,0x00,0x00,0x00,0x00,0x00,0xCE,0xA4,
    0xC4,0xA4,0xA4,0x00,0x00,0x00,0x00,0x00,
    0x4C,0xAA,0xAA,0xAA,0x4A,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
}, e_bios[2048] = {
// This is the (fixed) Elektor monitor ROM ($0..$7FF).
    0x1F,0x00,0x0A,0x1F,0x88,0xB9,0x05,0x94,
    0x05,0xB2,0x76,0x20,0x0C,0x08,0xB9,0xE4,
    0x09,0x3C,0x05,0xCD,0x20,0xCC,0x08,0x9A,
    0x05,0x18,0x6D,0x48,0x78,0x59,0x7B,0x44,
    0x11,0x18,0x15,0x76,0x60,0x20,0x05,0xC0,
    0xCD,0x5F,0x00,0xCD,0x68,0x00,0x59,0x78,
    0x06,0x28,0x3F,0x06,0x02,0x3F,0x01,0x61,
    0x04,0x4B,0xCC,0x08,0xBA,0x04,0x00,0xCC,
    0x08,0xB9,0xCC,0x08,0x9F,0x77,0x02,0x74,
    0x27,0x1B,0x7C,0xB4,0x80,0x16,0x3F,0x01,
    0x81,0x01,0x1E,0x00,0xBD,0x05,0x04,0x0D,
    0x40,0xAD,0xCD,0x7F,0x0A,0x0D,0x60,0xB1,
    0xCD,0x7F,0x1A,0x0D,0x60,0xB5,0xCD,0x7F,
    0x2A,0x0D,0x60,0xB9,0xCD,0x7F,0x4A,0x59,
    0x66,0xB4,0x80,0x18,0x7C,0x04,0x5C,0xF8,
    0x7E,0xC2,0x0D,0x68,0x00,0xCE,0x7F,0x00,
    0x0D,0x68,0x06,0xCE,0x7F,0x10,0x0D,0x68,
    0x0C,0xCE,0x7F,0x20,0x0D,0x28,0x11,0xCE,
    0x3F,0x3F,0xE6,0x06,0x98,0x64,0x04,0xA0,
    0x85,0x12,0xE5,0x8F,0x1A,0x59,0x04,0xFE,
    0xCC,0x1F,0x0D,0xCC,0x1F,0x1D,0xCC,0x1F,
    0x2D,0xCC,0x1F,0x4D,0x17,0x1F,0x1F,0x10,
    0xFF,0x3F,0x3F,0x10,0xFF,0x5F,0x5F,0x10,
    0xFF,0x7F,0x7F,0x10,0xFF,0x0F,0x08,0x9A,
    0x45,0x1F,0x0D,0x61,0x22,0x1A,0x17,0xCC,
    0x08,0xA1,0x3F,0x01,0x6E,0x0E,0x08,0xA0,
    0x60,0xBF,0x01,0x0C,0x04,0x7F,0x4C,0x08,
    0x9F,0xC8,0xFC,0x1F,0x00,0x45,0xF4,0x40,
    0x18,0x1B,0x44,0x1F,0xC8,0xD8,0x05,0xFF,
    0xE3,0x98,0x04,0x09,0x83,0x25,0x7F,0xCD,
    0x08,0x98,0xC3,0x20,0x06,0x09,0xCE,0x48,
    0xA0,0x5A,0x7B,0x1B,0x4D,0xC1,0xCC,0x08,
    0x99,0xE7,0x0D,0x3E,0x01,0x6E,0x01,0x05,
    0x00,0x1F,0x00,0xCD,0x17,0x1F,0x04,0x0C,
    0x1F,0x04,0xA9,0x1F,0x03,0xB0,0x1F,0x00,
    0x23,0x1F,0x05,0x0E,0x1F,0x07,0x58,0x1F,
    0x05,0xE8,0xC0,0x8D,0x84,0x90,0xE0,0x81,
    0x87,0x93,0x00,0x04,0x08,0x0C,0x80,0x80,
    0x8A,0x80,0x01,0x05,0x09,0x0D,0x02,0x06,
    0x0A,0x0E,0x03,0x07,0x0B,0x0F,0xCC,0x08,
    0xAC,0x13,0xCC,0x08,0xB3,0x12,0x76,0x20,
    0xCC,0x08,0xB4,0x75,0x10,0xCD,0x08,0xAD,
    0xCE,0x08,0xAE,0xCF,0x08,0xAF,0x77,0x10,
    0xCD,0x08,0xB0,0xCE,0x08,0xB1,0xCF,0x08,
    0xB2,0x05,0x0A,0x0D,0x41,0x77,0xCD,0x7F,
    0xC0,0x59,0x78,0xCD,0x1E,0x80,0x20,0x06,
    0x50,0xCE,0x5F,0x00,0x5A,0x7B,0x17,0xAA,
    0x09,0x09,0x00,0x00,0x00,0x19,0x00,0xAA,
    0xAA,0x75,0x19,0x20,0xC8,0x93,0xC8,0x9C,
    0x06,0x07,0x05,0x03,0x0E,0x5E,0x88,0x44,
    0xF0,0x98,0x30,0x5A,0x75,0x0C,0x08,0x9D,
    0x0D,0x08,0x9B,0x3F,0x01,0xEE,0xC9,0xF6,
    0x0C,0x08,0x9E,0x0D,0x08,0x9C,0x3B,0xF4,
    0xC9,0xF7,0x09,0xED,0x18,0x08,0x08,0xF4,
    0x18,0x06,0x05,0xFF,0x1B,0x02,0x09,0xEC,
    0x0C,0x08,0x9F,0x3B,0xDF,0xC9,0xFA,0x17,
    0xF9,0x00,0xD0,0x9A,0x7B,0x44,0x7F,0x98,
    0x16,0x02,0xD0,0xD0,0x81,0x64,0x60,0xE6,
    0x03,0x19,0x10,0x09,0x85,0x18,0x02,0x04,
    0xFF,0xCC,0x08,0x9C,0x1F,0x01,0x93,0x04,
    0xFF,0x1B,0x6C,0x09,0x85,0x18,0x02,0x04,
    0xFF,0xCC,0x08,0x9B,0x1B,0xEF,0x59,0x04,
    0x44,0x9F,0xC1,0x17,0xE5,0xFF,0x18,0x03,
    0x60,0x9A,0x04,0x64,0x20,0xC1,0x17,0xF4,
    0x20,0x16,0xF4,0x40,0x98,0x77,0xE1,0x98,
    0x72,0x65,0x80,0x45,0xBF,0x17,0x05,0x08,
    0x75,0x08,0xF5,0x01,0x18,0x08,0x06,0x06,
    0x20,0xCE,0x48,0x78,0x5A,0x7B,0x0D,0x48,
    0x90,0xCD,0x08,0x9B,0x06,0x00,0xF4,0x01,
    0x98,0x0A,0x86,0x02,0xF5,0x01,0x18,0x02,
    0x86,0x01,0x1B,0x06,0xF5,0x01,0x98,0x02,
    0x86,0x01,0x3B,0x0F,0x0D,0x42,0x7B,0xE6,
    0x02,0x1A,0x12,0x44,0x0E,0xE6,0x02,0x18,
    0x16,0x1B,0x10,0x07,0x06,0x44,0xFE,0xC1,
    0xD0,0x81,0x83,0xC1,0x17,0x44,0xE0,0xE6,
    0x00,0x18,0x04,0xD0,0xD0,0xD0,0xD0,0x6F,
    0x48,0x78,0xCF,0x68,0x78,0x5B,0x55,0x0C,
    0x08,0x9B,0x14,0x3B,0x5E,0x0F,0x48,0x78,
    0xCD,0x48,0x78,0x5B,0x78,0x0D,0x08,0x9B,
    0x1F,0x02,0x12,0xE2,0xA6,0xA2,0xA2,0xA2,
    0xE2,0xEE,0x22,0xEE,0x82,0x82,0xEE,0xAE,
    0xA8,0xAE,0xE2,0x22,0x2E,0xEE,0x82,0x82,
    0xE2,0xA2,0xE2,0xEE,0xAA,0xEE,0xA2,0xA2,
    0xEE,0xE8,0xA8,0xEE,0xAA,0xAA,0xAE,0xE2,
    0x82,0x8E,0x8A,0x8A,0xEE,0xEE,0x88,0xEE,
    0x88,0x88,0xE8,0xE8,0x88,0x88,0x88,0xA8,
    0xEE,0xE0,0x40,0x4E,0x4A,0x4A,0xEA,0xE0,
    0xA0,0xEE,0x88,0x88,0x88,0x00,0x00,0x00,
    0xE0,0x00,0xE0,0x40,0x40,0xEE,0x40,0x40,
    0x00,0x00,0x4A,0x04,0x4A,0x00,0x00,0x05,
    0x88,0x0D,0x67,0x90,0xCD,0x67,0x78,0xD9,
    0x78,0x06,0x08,0x04,0x17,0xCE,0x48,0x90,
    0x5A,0x7B,0x17,0xCE,0x08,0x9B,0x3B,0x67,
    0x0A,0xFA,0x07,0x04,0x0E,0x42,0xF5,0xCF,
    0x48,0x90,0x5B,0x78,0x17,0x14,0x0C,0x17,
    0x16,0x0A,0x0D,0x17,0x16,0x15,0x17,0x17,
    0x16,0x0B,0x14,0x01,0x16,0x0B,0x14,0x02,
    0x16,0x0B,0x0E,0x10,0x16,0x0E,0x13,0x0D,
    0x16,0x05,0x0A,0x0D,0x16,0x0F,0x12,0x11,
    0x16,0x12,0x12,0x12,0x12,0x0C,0x08,0xA7,
    0x38,0x1E,0x04,0x01,0xC8,0xF8,0x3B,0x1E,
    0x77,0x0A,0x75,0x01,0xD0,0xD0,0xD0,0xD0,
    0x07,0x04,0xD0,0xD1,0xD2,0xFB,0x7B,0x75,
    0x08,0xCD,0x08,0xA3,0xCE,0x08,0xA2,0x17,
    0x20,0xC8,0xF7,0xC8,0xF8,0x17,0x0C,0x08,
    0xA1,0x09,0xEF,0x0A,0xF0,0x17,0x07,0x08,
    0x3B,0x02,0x02,0xC1,0x3B,0x08,0xCF,0x48,
    0x8F,0x01,0xCF,0x48,0x91,0x17,0x01,0x50,
    0x50,0x50,0x50,0x44,0x0F,0x45,0x0F,0x17,
    0x07,0x08,0x3B,0x5A,0x3B,0x62,0x0C,0x08,
    0xA6,0x14,0xC1,0xF8,0x00,0xC8,0xF8,0x03,
    0xC2,0x04,0x17,0xCE,0x28,0x8F,0xF9,0x7B,
    0x17,0x0C,0x08,0xA3,0x0D,0x08,0xA4,0x0E,
    0x08,0xA5,0xE5,0x08,0x98,0x04,0xE6,0xC0,
    0x1A,0x09,0xCC,0x88,0xA4,0xE5,0x18,0x1A,
    0x02,0xE0,0x17,0xEC,0x88,0xA4,0x17,0x77,
    0x0B,0x05,0x02,0x0D,0x48,0xA4,0x84,0x00,
    0xCD,0x68,0xA4,0x59,0x76,0x75,0x09,0x17,
    0x1A,0x2E,0x01,0x1A,0x03,0x02,0x98,0x1E,
    0x06,0x0C,0x3F,0x02,0xE3,0x04,0x01,0xCC,
    0x08,0xA0,0x0A,0x9E,0xCE,0x08,0x91,0x0E,
    0x68,0xAC,0xC1,0xC8,0x98,0x20,0xCC,0x08,
    0xA7,0x07,0x06,0x1F,0x04,0x80,0x3F,0x03,
    0x1D,0x04,0x17,0xCC,0x08,0x93,0x1B,0x71,
    0xC3,0x0E,0x08,0xA4,0x0C,0x08,0xA3,0xCE,
    0x28,0xAB,0xF7,0x20,0x98,0x12,0xE6,0x09,
    0x1A,0x02,0x06,0x00,0xCA,0xEC,0x04,0x16,
    0xC8,0xE2,0x3F,0x02,0x0E,0x1F,0x03,0xB8,
    0x77,0x0B,0xA6,0x02,0x9A,0x02,0x06,0x08,
    0x75,0x09,0x1B,0x68,0x1A,0x0E,0x01,0x1A,
    0x04,0x02,0x9C,0x04,0x65,0x06,0x08,0x04,
    0x01,0x1F,0x05,0xF7,0x0C,0x08,0xA0,0x99,
    0x1D,0x05,0x02,0x0D,0x48,0xA2,0xCD,0x68,
    0xA4,0x59,0x78,0x3F,0x02,0xD9,0x04,0xFD,
    0xC8,0xB9,0x0C,0x88,0xA4,0xCC,0x08,0xA3,
    0x20,0xCC,0x08,0xA7,0x1B,0x2C,0x3F,0x03,
    0x81,0x98,0x0A,0x0C,0x08,0x99,0xE4,0xC0,
    0x18,0x08,0x3F,0x03,0x9F,0x3F,0x02,0xCF,
    0x1B,0x5C,0x04,0x0A,0x93,0x05,0x02,0x0D,
    0x48,0xA4,0xA4,0x00,0xCD,0x68,0xA4,0x59,
    0x76,0x75,0x09,0x1B,0x68,0x1A,0x1F,0x3F,
    0x03,0x1D,0x0C,0x08,0xA0,0x1D,0x05,0x32,
    0x07,0x04,0x0D,0x08,0xA5,0x0E,0x08,0xA4,
    0x3F,0x03,0x50,0x07,0x08,0x0D,0x08,0xA3,
    0x3F,0x03,0x54,0x1F,0x02,0x0E,0x0C,0x08,
    0xA7,0x98,0x17,0xE6,0xFD,0x18,0x06,0x3F,
    0x02,0xCF,0x3F,0x03,0x9F,0x04,0xFF,0xC8,
    0xD2,0x3B,0x4C,0x3F,0x03,0x81,0x9C,0x04,
    0x15,0x17,0x3B,0x75,0x20,0xC8,0xE0,0x1B,
    0xDB,0x0D,0x08,0x98,0xF5,0x20,0x9C,0x04,
    0xF5,0x60,0x1A,0x27,0x02,0x19,0x1A,0x06,
    0x02,0x0E,0x48,0xB5,0xCE,0x68,0xA2,0x5A,
    0x78,0x06,0x10,0x3F,0x02,0xE3,0x04,0x01,
    0xCC,0x08,0xA0,0x3F,0x03,0x46,0x1F,0x05,
    0x29,0x3F,0x03,0x1D,0x04,0x17,0xCC,0x08,
    0x93,0x1B,0x70,0x06,0x02,0x05,0x16,0xC9,
    0xF6,0xF4,0x20,0xBC,0x03,0x40,0x05,0x02,
    0x0D,0x48,0xA2,0xCE,0x48,0xB5,0x59,0x78,
    0xCD,0x08,0x9A,0x1B,0x56,0x60,0x1A,0x12,
    0x02,0x98,0x56,0x05,0x02,0x0D,0x48,0xB7,
    0xCD,0x68,0xA2,0x59,0x78,0x06,0x14,0x1F,
    0x04,0xC3,0x06,0x04,0x1B,0x4F,0x1A,0x30,
    0x02,0x98,0x1C,0x06,0x04,0x3F,0x02,0xE3,
    0x0D,0x08,0xBF,0x0E,0x08,0xBE,0x04,0x03,
    0xCC,0x08,0xA6,0xCC,0x08,0xA0,0x3F,0x03,
    0x39,0x3F,0x03,0x4E,0x1F,0x02,0x0E,0x3F,
    0x03,0x1D,0x3F,0x03,0x68,0x1B,0xF6,0x9B,
    0x86,0x9B,0x88,0x09,0x03,0x77,0x00,0x1F,
    0x05,0x02,0x0D,0x48,0xA2,0xCD,0x68,0xBE,
    0x0D,0xE8,0xB5,0xCD,0x68,0xA2,0x0D,0xE8,
    0xB7,0xCD,0x68,0xA4,0x0D,0x65,0x37,0xCD,
    0xE8,0xB5,0x0D,0x65,0x39,0xCD,0xE8,0xB7,
    0x59,0x60,0x05,0x05,0x0D,0x45,0x3B,0xCD,
    0x68,0xB9,0x59,0x78,0x0C,0x08,0xB3,0xCC,
    0x08,0xBC,0x75,0x10,0x0D,0x08,0xAD,0x0E,
    0x08,0xAE,0x0F,0x08,0xAF,0x77,0x10,0x0D,
    0x08,0xB0,0x0E,0x08,0xB1,0x0F,0x08,0xB2,
    0x0C,0x08,0xB4,0x92,0x0C,0x08,0xAC,0x75,
    0xFF,0x1F,0x08,0xBB,0x3B,0x37,0x0D,0x08,
    0xB6,0x0E,0x08,0xB5,0xC8,0xF9,0xC8,0xFA,
    0x3B,0x22,0x06,0x10,0x3F,0x02,0xEA,0x04,
    0x15,0xCC,0x08,0x91,0x3F,0x02,0x0E,0x1F,
    0x00,0x38,0x3B,0x19,0x0D,0x08,0xB8,0x0E,
    0x08,0xB7,0xC8,0xF9,0xC8,0xFA,0x3B,0x04,
    0x06,0x14,0x1B,0x60,0xCD,0x08,0xBF,0xCE,
    0x08,0xBE,0x1F,0x03,0x4E,0x3F,0x01,0x3E,
    0x05,0x02,0x0D,0x48,0xA2,0xCD,0xE8,0xB5,
    0x0D,0x68,0xA4,0xCD,0xE8,0xB7,0x59,0x72,
    0x3F,0x02,0xCF,0x20,0xCC,0x08,0x9A,0x17,
    0x1A,0x26,0x01,0x1A,0x06,0x02,0x1D,0x05,
    0x2F,0x1A,0x15,0x04,0x01,0x06,0x18,0xCC,
    0x08,0xA0,0x04,0x03,0xCC,0x08,0xA6,0x3F,
    0x03,0x40,0x3F,0x02,0xE3,0x1F,0x02,0x0E,
    0x0C,0x08,0xA1,0xCC,0x08,0x95,0x1B,0xF6,
    0x02,0x1E,0x06,0x4A,0x06,0x02,0xE4,0x02,
    0x9A,0x0E,0x0E,0x48,0xA2,0xCE,0x68,0xA4,
    0x5A,0x78,0x04,0x02,0x06,0x1C,0x1B,0x4F,
    0x98,0x12,0x77,0x09,0x0E,0x48,0xA2,0x84,
    0x00,0xCE,0x68,0xA8,0x5A,0x76,0x04,0x03,
    0x06,0x20,0x1B,0x6A,0x04,0xFF,0xCC,0x08,
    0xA0,0x20,0xCC,0x08,0xA1,0x06,0x24,0x1F,
    0x06,0x02,0x20,0x07,0x1E,0x3F,0x06,0xF7,
    0xF8,0x79,0x3F,0x06,0xF2,0x07,0x09,0x3F,
    0x06,0xF0,0x20,0xCC,0x08,0xAA,0x04,0x4C,
    0x3B,0xAF,0x08,0xDF,0x3B,0xAB,0x77,0x0B,
    0x0D,0x08,0xA9,0x0C,0x08,0xA8,0xAD,0x08,
    0xA5,0xAC,0x08,0xA4,0x1E,0x05,0xF3,0x19,
    0x08,0x01,0x1C,0x06,0xBB,0xE4,0x10,0x1A,
    0x02,0x04,0x10,0xCC,0x08,0xAB,0x08,0xEA,
    0x3B,0x87,0x08,0xE3,0x3B,0x83,0x08,0xF4,
    0x3F,0x06,0xD0,0x08,0x90,0x3B,0xFA,0x07,
    0x00,0x0F,0xE8,0xA4,0xEB,0xE6,0x18,0x04,
    0x3B,0xEF,0xDB,0x75,0x0C,0x08,0xAA,0x3B,
    0xE8,0x04,0x0A,0x93,0x05,0x02,0x03,0x8D,
    0x48,0xA4,0xCD,0x68,0xA4,0x20,0x59,0x77,
    0x1F,0x06,0x55,0x0C,0x08,0xA2,0x3F,0x06,
    0xD0,0x0C,0x08,0xA3,0x3B,0xF9,0x20,0xCC,
    0x08,0x9A,0x3B,0xF3,0x08,0xD7,0x3B,0xEF,
    0x3F,0x07,0x26,0x77,0x10,0x75,0x08,0x06,
    0x08,0x50,0xD0,0x1A,0x04,0x3B,0x0F,0x1B,
    0x02,0x3B,0x07,0xFA,0x75,0x3B,0x0B,0x75,
    0x19,0x17,0x07,0x06,0x1B,0x02,0x07,0x03,
    0x3B,0x05,0x05,0x32,0xF9,0x7E,0x17,0x05,
    0xFF,0xCD,0x1D,0xFF,0x05,0x08,0xF9,0x7E,
    0xC9,0xF8,0x05,0x07,0xF9,0x7E,0xFB,0x6F,
    0x17,0x77,0x1A,0x06,0x08,0xCE,0x08,0x9B,
    0x3B,0x1F,0xE6,0x86,0x1A,0x05,0xDA,0x78,
    0x1F,0x07,0x88,0x0C,0x08,0x9C,0xD2,0xD0,
    0xC8,0xFA,0x0A,0xEA,0xFA,0x67,0x75,0x18,
    0xC1,0x2C,0x08,0xAA,0xD0,0xC8,0xFB,0x01,
    0x17,0x09,0x8B,0x20,0xCC,0x08,0x98,0x06,
    0x77,0x07,0x03,0xC0,0x01,0x0D,0x1D,0xBF,
    0x21,0x9A,0x04,0xDA,0x74,0xFA,0x72,0xE6,
    0x79,0xFB,0x70,0x15,0x08,0xE7,0xF8,0x64,
    0x0C,0x08,0xA8,0x18,0x5C,0x1F,0x07,0x88,
    0x1A,0x1E,0x0C,0x08,0xA1,0x19,0x05,0x06,
    0x24,0x1F,0x06,0x02,0xCC,0x08,0x95,0x1F,
    0x02,0x0E,0xCE,0x08,0x97,0x3B,0xF9,0x05,
    0x78,0x1F,0x00,0x71,0x3B,0x74,0x1B,0x10,
    0x44,0x20,0xCC,0x08,0xA8,0x05,0x18,0x58,
    0x02,0x05,0x19,0xCD,0x08,0x93,0x3B,0x65,
    0x3F,0x07,0x31,0xE6,0x86,0x1A,0x79,0xE6,
    0x8C,0x19,0x75,0x20,0xCC,0x08,0xAA,0x3F,
    0x07,0x09,0xE4,0x4C,0x98,0x6A,0x0C,0x1F,
    0x15,0x24,0x0A,0xC8,0xFA,0x3B,0xF1,0xC2,
    0x0C,0x08,0xA1,0x18,0x03,0xE2,0x98,0x44,
    0x07,0xFE,0xCF,0x1F,0x4A,0x3B,0x9B,0xCF,
    0x67,0xA6,0xCF,0x67,0xC0,0xDB,0x76,0x3B,
    0x91,0x60,0x98,0x0A,0x06,0x17,0x3F,0x07,
    0x6A,0x05,0xE1,0x1F,0x00,0xBD,0xCC,0x08,
    0xAB,0x3F,0x07,0x09,0x08,0x99,0x98,0x1C,
    0xC3,0x3B,0xF7,0xEB,0xF2,0x18,0x0F,0x0D,
    0x08,0xA8,0x18,0x03,0xCF,0xE8,0xA4,0xEF,
    0xE8,0xA4,0x98,0x08,0xDB,0x6B,0x0C,0x08,
    0xAA,0x1C,0x07,0x88,0x3F,0x05,0xE0,0x06,
    0x08,0x3F,0x05,0x15,0x1F,0x00,0xD4,0x82
};

// MODULE FUNCTIONS-------------------------------------------------------

MODULE UWORD ReadWord(void);
MODULE UBYTE ReadByte(void);
MODULE UWORD ReadWordAt(int where);
MODULE UBYTE ReadByteAt(int where);
MODULE void WriteWord(UWORD what);
MODULE void WriteByte(UBYTE what);
MODULE void view_pviscreen(void);
MODULE void inputmirrors(int address);
MODULE void a_setmemmap(void);
MODULE void i_setmemmap(void);
MODULE void e_setmemmap(void);
MODULE void alloc_iobuffer(ULONG thesize);
MODULE void free_iobuffer(void);
MODULE FLAG autosense(int thesize, FLAG silent);
MODULE void unautosense(void);
MODULE ULONG getcrc32(UBYTE* address, int thesize);
MODULE void configure(FLAG silent);
MODULE void a_readrom(void);
MODULE void i_readrom(void);
MODULE void engine_reset(FLAG user);
MODULE void help_commands(void);
// network play
MODULE void NetConnect(const char* Server, unsigned short Port);
MODULE void NetClose(void);
MODULE const char* NetMyName(char* Buffer, int MaxChars);

// function pointers
MODULE UBYTE a_cpuread(int address);
MODULE UBYTE i_cpuread(int address);
MODULE UBYTE e_cpuread(int address);

// function pointers
// (these are really variables, so they are defined here instead of aa.h)
EXPORT UBYTE (* cpuread)(int address);

// CODE-------------------------------------------------------------------

EXPORT void change_read(STRPTR thestring)
{   int number;

    /* thestring: contents of the "address" gadget

       This looks at thestring and sets fromnum, tonum, fromstr, tostr,
       tolimit. It is used when the "address" field in the "change"
       requester is modified (and also used by the debugger CLI
       interface). */

    if (!stricmp(thestring, "PSU"))
    {   fromnum = (int) psu;
        tolimit = 255;
    } elif (!stricmp(thestring, "PSL"))
    {   fromnum = (int) psl;
        tolimit = 255;
    } elif (!stricmp(thestring, "IAR"))
    {   fromnum = (int) (page + iar);
        tolimit = 32767;
    } elif
    (   strlen(thestring) == 4
     && !strnicmp(thestring, "RAS", 3)
     && thestring[3] >= '0'
     && thestring[3] <= '7'
    )
    {   fromnum = (int) ras[thestring[3] - '0'];
        tolimit = 32767;
    } elif
    (   strlen(thestring) == 2
     && (thestring[0] == 'R' || thestring[0] == 'r')
     && thestring[1] >= '0'
     && thestring[1] <= '6'
    )
    {   fromnum = (int) r[thestring[1] - '0'];
        tolimit = 255;
    } else
    {   tolimit = 255;

        number = friendly_to_number(thestring);
        if (number < OUTOFRANGE)
        {   fromnum = (int) memory[number];
        } else
        {   fromnum = OUTOFRANGE;
            fromstr[0] = 0;
    }   }

    if (fromnum < OUTOFRANGE)
    {   fromstr[0] = '$';
        stcl_h(&fromstr[1], fromnum);
    }

    tonum = fromnum;
    strcpy(tostr, fromstr);
}

EXPORT void change_write(STRPTR thestring, FLAG quiet)
{   /* thestring: contents of the "address" gadget

       This looks at thestring and tonum and actually changes the value of
       the register that the user has specified. It is used at the end of
       use of the "change" requester, and also by the CLI-based debugger
       interface. */

    FLAG found = TRUE;
    int  number;

    // assert(strlen(lastchange) <= 14);
    strupr(thestring);
    strcpy(lastchange, thestring); // for next time that we reopen this requester

    if (tonum > tolimit)
    {   if (!quiet)
        {   T1;
            printf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_OUTOFBOUNDS,
                    "Value is out of bounds (limit is %d)!\n\n"
                ),
                tolimit
            );
            T2;
        }
        return;
    }

    if (!strcmp(thestring, "PSU"))
    {   psu = (UBYTE) tonum;
    } elif (!strcmp(thestring, "PSL"))
    {   psl = (UBYTE) tonum;
    } elif (!strcmp(thestring, "IAR"))
    {   page = (UWORD) tonum & PAGE;
        iar =  (UWORD) tonum & ~(PAGE);
    } elif
    (   strlen(thestring) == 4
     && !strncmp(thestring, "RAS", 3)
     && thestring[3] >= '0'
     && thestring[3] <= '7'
    )
    {   ras[thestring[3] - '0'] = (UWORD) tonum;
    } elif
    (   strlen(thestring) == 2
     && (thestring[0] == 'R' || thestring[0] == 'r')
     && thestring[1] >= '0'
     && thestring[1] <= '6'
    )
    {   r[thestring[1] - '0'] = (UBYTE) tonum;
    } else
    {   found = FALSE;

        number = friendly_to_number(thestring);
        if (number < OUTOFRANGE)
        {   memory[number] = (UBYTE) tonum;
            found = TRUE;
            number_to_friendly(number, thestring, TRUE);
    }   }

    if (found && fromnum != tonum)
    {   if (!quiet)
        {   T1;
            printf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_CHANGEDCONTENTS,
                    "Changed contents of %s from $%lX to $%lX!\n\n"
                ),
                thestring,
                (ULONG) fromnum,
                (ULONG) tonum
            );
            T2;
        }
        updatemonitor(FALSE);
    } else
    {   if (!quiet)
        {   T1;
            printf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_NOCHANGE,
                    "No effect (no change to contents of address).\n\n"
            )   );
            T2;
}   }   }

EXPORT void getfriendly(int number)
{   number_to_friendly(number, friendly, TRUE);
}

EXPORT void number_to_friendly(int number, STRPTR thestring, FLAG both)
{   int i;

/*  number:    input number
    thestring: pointer to output string buffer
    both:      do we want friendly and hex names (TRUE) (eg. "BGCOLOUR
               ($19FD)") or just friendly names (eg. "BGCOLOUR")? */

    if (machine == ARCADIA)
    {   for (i = 0; i < A_EQUIVALENTS; i++)
        {   // T1;
            // printf("Comparing %s against %s...\n", thestring, a_equivalents[i].text);
            // T2;
            if (number == (int) a_equivalents[i].number)
            {   strcpy(thestring, a_equivalents[i].text);
                if (both)
                {   strcat(thestring, " ($");
                    stcl_h(&thestring[strlen(thestring)], number); // integer -> hexstring
                    strcat(thestring, ")");
                }
                return;
    }   }   }
    else
    {   // assert(machine == INTERTON || machine == ELEKTOR);

        for (i = 0; i < I_EQUIVALENTS; i++)
        {   // T1;
            // printf("Comparing %s against %s...\n", thestring, i_equivalents[i].text);
            // T2;
            if (number == (int) i_equivalents[i].number)
            {   strcpy(thestring, i_equivalents[i].text);
                if (both)
                {   strcat(thestring, " ($");
                    stcl_h(&thestring[strlen(thestring)], number); // integer -> hexstring
                    strcat(thestring, ")");
                }
                return;
        }   }

        if (machine == ELEKTOR)
        {   for (i = 0; i < E_EQUIVALENTS; i++)
            {   // T1;
                // printf("Comparing %s against %s...\n", thestring, e_equivalents[i].text);
                // T2;
                if (number == (int) e_equivalents[i].number)
                {   strcpy(thestring, e_equivalents[i].text);
                    if (both)
                    {   strcat(thestring, " ($");
                        stcl_h(&thestring[strlen(thestring)], number); // integer -> hexstring
                        strcat(thestring, ")");
                    }
                    return;
    }   }   }   }

    strcpy(thestring, "$");
    stcl_h(&thestring[1], number); // integer -> hexstring
}

EXPORT int friendly_to_number(STRPTR thestring)
{   /* This never returns a negative number. */

    int i,
        number;

    // this assumes the longest friendly string is 13 characters (plus
    // null terminator) (eg. "SPRITECOLLIDE")
    if (strlen(thestring) < 1 || strlen(thestring) > 13)
    {   thestring[0] = 0;
        return(OUTOFRANGE);
    }

    if (machine == ARCADIA)
    {   for (i = 0; i < A_EQUIVALENTS; i++)
        {   // T1;
            // printf("Comparing %s against %s...\n", thestring, a_equivalents[i].text);
            // T2;
            if (!stricmp(thestring, a_equivalents[i].text))
            {   return((int) a_equivalents[i].number);
    }   }   }
    else
    {   // assert(machine == INTERTON || machine == ELEKTOR);

        for (i = 0; i < I_EQUIVALENTS; i++)
        {   // T1;
            // printf("Comparing %s against %s...\n", thestring, i_equivalents[i].text);
            // T2;
            if (!stricmp(thestring, i_equivalents[i].text))
            {   return((int) i_equivalents[i].number);
        }   }

        if (machine == ELEKTOR)
        {   for (i = 0; i < E_EQUIVALENTS; i++)
            {   // T1;
                // printf("Comparing %s against %s...\n", thestring, e_equivalents[i].text);
                // T2;
                if (!stricmp(thestring, e_equivalents[i].text))
                {   return((int) e_equivalents[i].number);
    }   }   }   }

    // still no match, must be a numerical address
    if (thestring[0] == '$')
    {   stch_l(&thestring[1], (long *) &number); // hexstring -> integer
    } elif (thestring[0] >= '0' && thestring[0] <= '9')
    {   stcd_l(thestring, (long *) &number);     // decstring -> integer
        thestring[0] = '$';
        stcl_h(&thestring[1], number); // integer   -> hexstring
    } else
    {   number = OUTOFRANGE;
    }

    if (number < 0 || number >= OUTOFRANGE)
    {   thestring[0] = 0;
        return(OUTOFRANGE);
    }

    return(number);
}

EXPORT void project_reset(FLAG keepalive)
{   if (!keepalive)
    {   NetPlay(NET_OFF);
        updatemenu(MENUITEM_SERVER);
        updatemenu(MENUITEM_CLIENT);
    }

    macro_stop();
    engine_reset(TRUE);
}

EXPORT void macro_stop(void)
{   if (recmode == RECMODE_PLAY)
    {   free_iobuffer();
    } elif (recmode == RECMODE_RECORD)
    {   // assert(MacroHandle);
        DISCARD fclose(MacroHandle);
        MacroHandle = NULL;
    }

    if (recmode != RECMODE_NORMAL) // avoids useless refreshing
    {   recmode = RECMODE_NORMAL;
        updatesmlgad(GADPOS_MACRO, (ULONG) recmode, TRUE);
        updatemenus();
}   }

EXPORT void macro_restartplayback(void)
{   offset = 0;
    parse_bytes(FALSE, FALSE); // this sets recmode = RECMODE_PLAY for us
}

EXPORT void arcadia(FLAG user)
{   if (user)
    {   memmap = MEMMAP_G; // Arcadia memory map
        unautosense();
    }

    machine = ARCADIA;
    if (oldmachine != ARCADIA)
    {   // changing from Interton or Elektor to Arcadia
        closemonitors(TRUE);
        oldmachine = ARCADIA;
    }

    macro_stop();
    sound_off(FALSE);
    a_setmemmap();
    sound_on();
    region = REGION_NTSC;

    if (!game || user)
    {   engine_reset(FALSE);
}   }

EXPORT void interton(FLAG user)
{   if (user)
    {   memmap = MEMMAP_C; // in case of !game, or unsensed game
        unautosense();
    }

    machine = INTERTON;
    if (oldmachine != INTERTON)
    {   // changing from Arcadia or Elektor to Interton
        closemonitors(TRUE);
        oldmachine = INTERTON;
    }

    macro_stop();
    sound_off(FALSE);
    i_setmemmap();
    sound_on();
    region = REGION_PAL;

    if (!game || user)
    {   engine_reset(FALSE);
}   }

EXPORT void elektor(FLAG user)
{   if (user)
    {   memmap = MEMMAP_F; // in case of !game, or unsensed game
        unautosense();
    }

    machine = ELEKTOR;
    if (oldmachine != ELEKTOR)
    {   closemonitors(TRUE);
        oldmachine = ELEKTOR;
    } elif
    (   (oldmemmap == MEMMAP_E && memmap == MEMMAP_F)
     || (oldmemmap == MEMMAP_F && memmap == MEMMAP_E)
    )
    {   closemonitors(FALSE);
    }

    macro_stop();
    sound_off(FALSE);
    e_setmemmap();
    sound_on();
    region = REGION_PAL;

    if (!game || user)
    {   engine_reset(FALSE);
}   }

EXPORT void debug_clearbp(void)
{   // harmless if it's already cleared
    if (bp < OUTOFRANGE)
    {   bp = OUTOFRANGE;
        updatemenu(MENUITEM_CLEARBP);
        T1;
        printf
        (   GetCatalogStr(CatalogPtr,
                MSG_ENGINE_CLEAREDBP,
                "Cleared code breakpoint.\n\n"
        )   );
        T2;
}   }
EXPORT void debug_clearwp(void)
{   // harmless if it's already cleared
    if (wp < OUTOFRANGE)
    {   wp = OUTOFRANGE;
        updatemenu(MENUITEM_CLEARWP);
        T1;
        printf
        (   GetCatalogStr(CatalogPtr,
                MSG_ENGINE_CLEAREDWP,
                "Cleared data watchpoint.\n\n"
        )   );
        T2;
}   }

EXPORT void pause(void)
{   if (!paused)
    {   paused = TRUE;
        sound_off(FALSE);
        updatesmlgad(GADPOS_PAUSED, paused, TRUE);
        updatemenu(MENUITEM_PAUSED);
        fps = 0;
        updatefps();
        settitle();
        redrawscreen(); // helpful in frame skipping mode (ensures most
        // recent executed frame is shown, not a frame from a while ago).
    }
    debug_cli();
}

EXPORT void unpause(void)
{   if (paused) // important!
    {   paused = FALSE;
        sound_on();
        updatesmlgad(GADPOS_PAUSED, paused, TRUE);
        updatemenu(MENUITEM_PAUSED);
        settitle();
}   }

EXPORT void cripple(void)
{   crippled = FALSE;
    updatemenus();
    updatebiggads();
    updatesmlgads();
}
EXPORT void uncripple(void)
{   crippled = FALSE;
    updatemenus();
    updatebiggads();
    updatesmlgads();
}

EXPORT void updatemenu(int which)
{   if (!MainWindowPtr)
    {   return; // for speed
    }

#ifdef AMIGA
    if (!MenuPtr)
    {   return; // for speed
    }
#endif

    switch(which)
    {
    case MENUITEM_ANALOG:
        if (analog)                     tick(which);
        else                            untick(which);
    acase MENUITEM_AUTOFIRE:
        if (autofire)                   tick(which);
        else                            untick(which);
    acase MENUITEM_AUTOSAVE:
        if (autosave)                   tick(which);
        else                            untick(which);
    acase MENUITEM_BRAM:
        if (machine != ELEKTOR)         ghost(which);
        else                            unghost(which);
    acase MENUITEM_CHANGE:
#ifdef AMIGA
        if (fullscreen)                 ghost(which);
        else                            unghost(which);
#endif
#ifdef WIN32
        unghost(which);
#endif
    acase MENUITEM_CLEARBP:
        if (bp == OUTOFRANGE)           ghost(which);
        else                            unghost(which);
    acase MENUITEM_CLEARWP:
        if (wp == OUTOFRANGE)           ghost(which);
        else                            unghost(which);
    acase MENUITEM_CLI:
        if (cli)                        tick(which);
        else                            untick(which);
    acase MENUITEM_CLIENT:
#ifdef WIN32
        if (crippled)                   ghost(which);
#endif
#ifdef AMIGA
        if (crippled || !SocketBase)    ghost(which);
#endif
        else                            unghost(which);
        if (connected == NET_CLIENT)    tick(which);
        else                            untick(which);
    acase MENUITEM_COLLISIONS:
        if (collisions)                 tick(which);
        else                            untick(which);
    acase MENUITEM_CONTROLS:
#ifdef AMIGA
        if (fullscreen || ControlsWindowPtr)
#endif
#ifdef WIN32
        if (ControlsWindowPtr)
#endif
        {                               ghost(which);
        } else                          unghost(which);
    acase MENUITEM_CRAM:
        if (machine != INTERTON || memmap == MEMMAP_A || memmap == MEMMAP_B)
        {                               ghost(which);
        } else                          unghost(which);
    acase MENUITEM_CREATEICONS:
        if (icons)                      tick(which);
        else                            untick(which);
    acase MENUITEM_DEMULTIPLEX:
        if (machine != ARCADIA)         ghost(which);
        else                            unghost(which);
        if (demultiplex)                tick(which);
        else                            untick(which);
    acase MENUITEM_FLAGLINE:
        if (flagline)                   tick(which);
        else                            untick(which);
    acase MENUITEM_FULLSCREEN:
        if (fullscreen)                 tick(which);
        else                            untick(which);
    acase MENUITEM_IMAGERY:
        if (!game)                      ghost(which);
        else                            unghost(which);
    acase MENUITEM_JOY1:
        if (joy1)                       tick(which);
        else                            untick(which);
    acase MENUITEM_LIMITREFRESH:
        if (limitrefresh)               tick(which);
        else                            untick(which);
    acase MENUITEM_LOGINSTRUCTIONS:
        if (loginstructions)            tick(which);
        else                            untick(which);
    acase MENUITEM_LOGMONITOR:
        if (machine != ELEKTOR)         ghost(which);
        else                            unghost(which);
        if (logmonitor)                 tick(which);
        else                            untick(which);
    acase MENUITEM_LOGREADS:
        if (logreads)                   tick(which);
        else                            untick(which);
    acase MENUITEM_LOGWRITES:
        if (logwrites)                  tick(which);
        else                            untick(which);
    acase MENUITEM_LOOP:
        if (loop)                       tick(which);
        else                            untick(which);
    acase MENUITEM_LOWERSCREEN:
        if (machine != ARCADIA)         ghost(which);
        else                            unghost(which);
    acase MENUITEM_MACHINE:
#ifdef AMIGA
        if (crippled)
        {   DISCARD OffMenu(MainWindowPtr, FULLMENUNUM(MN_SETTINGS, IN_MACHINE, SN_ARCADIA ));
            DISCARD OffMenu(MainWindowPtr, FULLMENUNUM(MN_SETTINGS, IN_MACHINE, SN_INTERTON));
            DISCARD OffMenu(MainWindowPtr, FULLMENUNUM(MN_SETTINGS, IN_MACHINE, SN_ELEKTOR ));
        } else
        {   DISCARD  OnMenu(MainWindowPtr, FULLMENUNUM(MN_SETTINGS, IN_MACHINE, SN_ARCADIA ));
            DISCARD  OnMenu(MainWindowPtr, FULLMENUNUM(MN_SETTINGS, IN_MACHINE, SN_INTERTON));
            DISCARD  OnMenu(MainWindowPtr, FULLMENUNUM(MN_SETTINGS, IN_MACHINE, SN_ELEKTOR ));
        }
        if (machine == ARCADIA)
        {   ItemAddress(MenuPtr, FULLMENUNUM(MN_SETTINGS, IN_MACHINE, SN_ARCADIA ))->Flags |= CHECKED;
        } else
        {   ItemAddress(MenuPtr, FULLMENUNUM(MN_SETTINGS, IN_MACHINE, SN_ARCADIA ))->Flags &= ~CHECKED;
        }
        if (machine == INTERTON)
        {   ItemAddress(MenuPtr, FULLMENUNUM(MN_SETTINGS, IN_MACHINE, SN_INTERTON))->Flags |= CHECKED;
        } else
        {   ItemAddress(MenuPtr, FULLMENUNUM(MN_SETTINGS, IN_MACHINE, SN_INTERTON))->Flags &= ~CHECKED;
        }
        if (machine == ELEKTOR)
        {   ItemAddress(MenuPtr, FULLMENUNUM(MN_SETTINGS, IN_MACHINE, SN_ELEKTOR ))->Flags |= CHECKED;
        } else
        {   ItemAddress(MenuPtr, FULLMENUNUM(MN_SETTINGS, IN_MACHINE, SN_ELEKTOR ))->Flags &= ~CHECKED;
        }
#endif
#ifdef WIN32
        if (crippled)
        {   DISCARD EnableMenuItem(MenuPtr,      ID_MACHINE_ARCADIA,  MF_BYCOMMAND | MF_GRAYED );
            DISCARD EnableMenuItem(MenuPtr,      ID_MACHINE_INTERTON, MF_BYCOMMAND | MF_GRAYED );
            DISCARD EnableMenuItem(MenuPtr,      ID_MACHINE_ELEKTOR,  MF_BYCOMMAND | MF_GRAYED );
            DISCARD EnableMenuItem(PopUpMenuPtr, ID_MACHINE_ARCADIA,  MF_BYCOMMAND | MF_GRAYED );
            DISCARD EnableMenuItem(PopUpMenuPtr, ID_MACHINE_INTERTON, MF_BYCOMMAND | MF_GRAYED );
            DISCARD EnableMenuItem(PopUpMenuPtr, ID_MACHINE_ELEKTOR,  MF_BYCOMMAND | MF_GRAYED );
        } else
        {   DISCARD EnableMenuItem(MenuPtr,      ID_MACHINE_ARCADIA,  MF_BYCOMMAND | MF_ENABLED);
            DISCARD EnableMenuItem(MenuPtr,      ID_MACHINE_INTERTON, MF_BYCOMMAND | MF_ENABLED);
            DISCARD EnableMenuItem(MenuPtr,      ID_MACHINE_ELEKTOR,  MF_BYCOMMAND | MF_ENABLED);
            DISCARD EnableMenuItem(PopUpMenuPtr, ID_MACHINE_ARCADIA,  MF_BYCOMMAND | MF_ENABLED);
            DISCARD EnableMenuItem(PopUpMenuPtr, ID_MACHINE_INTERTON, MF_BYCOMMAND | MF_ENABLED);
            DISCARD EnableMenuItem(PopUpMenuPtr, ID_MACHINE_ELEKTOR,  MF_BYCOMMAND | MF_ENABLED);
        }
        if (machine == ARCADIA )
        {   DISCARD CheckMenuRadioItem(MenuPtr,      ID_MACHINE_ARCADIA, ID_MACHINE_ELEKTOR, ID_MACHINE_ARCADIA , MF_BYCOMMAND);
            DISCARD CheckMenuRadioItem(PopUpMenuPtr, ID_MACHINE_ARCADIA, ID_MACHINE_ELEKTOR, ID_MACHINE_ARCADIA , MF_BYCOMMAND);
        } elif (machine == INTERTON)
        {   DISCARD CheckMenuRadioItem(MenuPtr,      ID_MACHINE_ARCADIA, ID_MACHINE_ELEKTOR, ID_MACHINE_INTERTON, MF_BYCOMMAND);
            DISCARD CheckMenuRadioItem(PopUpMenuPtr, ID_MACHINE_ARCADIA, ID_MACHINE_ELEKTOR, ID_MACHINE_INTERTON, MF_BYCOMMAND);
        } elif (machine == ELEKTOR )
        {   DISCARD CheckMenuRadioItem(MenuPtr,      ID_MACHINE_ARCADIA, ID_MACHINE_ELEKTOR, ID_MACHINE_ELEKTOR , MF_BYCOMMAND);
            DISCARD CheckMenuRadioItem(PopUpMenuPtr, ID_MACHINE_ARCADIA, ID_MACHINE_ELEKTOR, ID_MACHINE_ELEKTOR , MF_BYCOMMAND);
        } // else assert(0);
#endif
    acase MENUITEM_MACRO:
#ifdef AMIGA
        if (crippled || !game)
        {   DISCARD OffMenu(MainWindowPtr, FULLMENUNUM(MN_MACRO, IN_STARTRECORDING,  NOSUB));
            DISCARD OffMenu(MainWindowPtr, FULLMENUNUM(MN_MACRO, IN_RESTARTPLAYBACK, NOSUB));
            DISCARD OffMenu(MainWindowPtr, FULLMENUNUM(MN_MACRO, IN_STOP,            NOSUB));
        } elif (recmode == RECMODE_PLAY)
        {   DISCARD OffMenu(MainWindowPtr, FULLMENUNUM(MN_MACRO, IN_STARTRECORDING,  NOSUB));
            DISCARD  OnMenu(MainWindowPtr, FULLMENUNUM(MN_MACRO, IN_RESTARTPLAYBACK, NOSUB));
            DISCARD  OnMenu(MainWindowPtr, FULLMENUNUM(MN_MACRO, IN_STOP,            NOSUB));
        } elif (recmode == RECMODE_RECORD)
        {   DISCARD OffMenu(MainWindowPtr, FULLMENUNUM(MN_MACRO, IN_STARTRECORDING,  NOSUB));
            DISCARD OffMenu(MainWindowPtr, FULLMENUNUM(MN_MACRO, IN_RESTARTPLAYBACK, NOSUB));
            DISCARD  OnMenu(MainWindowPtr, FULLMENUNUM(MN_MACRO, IN_STOP,            NOSUB));
        } elif (recmode == RECMODE_NORMAL)
        {   DISCARD  OnMenu(MainWindowPtr, FULLMENUNUM(MN_MACRO, IN_STARTRECORDING,  NOSUB));
            DISCARD OffMenu(MainWindowPtr, FULLMENUNUM(MN_MACRO, IN_RESTARTPLAYBACK, NOSUB));
            DISCARD OffMenu(MainWindowPtr, FULLMENUNUM(MN_MACRO, IN_STOP,            NOSUB));
        } // else assert(0);
#endif
#ifdef WIN32
        if (crippled || !game)
        {   DISCARD EnableMenuItem(MenuPtr, ID_MACRO_STARTRECORDING,  MF_BYCOMMAND | MF_GRAYED );
            DISCARD EnableMenuItem(MenuPtr, ID_MACRO_RESTARTPLAYBACK, MF_BYCOMMAND | MF_GRAYED );
            DISCARD EnableMenuItem(MenuPtr, ID_MACRO_STOP,            MF_BYCOMMAND | MF_GRAYED );
            DISCARD EnableMenuItem(PopUpMenuPtr, ID_MACRO_STARTRECORDING,  MF_BYCOMMAND | MF_GRAYED );
            DISCARD EnableMenuItem(PopUpMenuPtr, ID_MACRO_RESTARTPLAYBACK, MF_BYCOMMAND | MF_GRAYED );
            DISCARD EnableMenuItem(PopUpMenuPtr, ID_MACRO_STOP,            MF_BYCOMMAND | MF_GRAYED );
        } elif (recmode == RECMODE_PLAY)
        {   DISCARD EnableMenuItem(MenuPtr, ID_MACRO_STARTRECORDING,  MF_BYCOMMAND | MF_GRAYED );
            DISCARD EnableMenuItem(MenuPtr, ID_MACRO_RESTARTPLAYBACK, MF_BYCOMMAND | MF_ENABLED);
            DISCARD EnableMenuItem(MenuPtr, ID_MACRO_STOP,            MF_BYCOMMAND | MF_ENABLED);
            DISCARD EnableMenuItem(PopUpMenuPtr, ID_MACRO_STARTRECORDING,  MF_BYCOMMAND | MF_GRAYED );
            DISCARD EnableMenuItem(PopUpMenuPtr, ID_MACRO_RESTARTPLAYBACK, MF_BYCOMMAND | MF_ENABLED);
            DISCARD EnableMenuItem(PopUpMenuPtr, ID_MACRO_STOP,            MF_BYCOMMAND | MF_ENABLED);
        } elif (recmode == RECMODE_RECORD)
        {   DISCARD EnableMenuItem(MenuPtr, ID_MACRO_STARTRECORDING,  MF_BYCOMMAND | MF_GRAYED );
            DISCARD EnableMenuItem(MenuPtr, ID_MACRO_RESTARTPLAYBACK, MF_BYCOMMAND | MF_GRAYED );
            DISCARD EnableMenuItem(MenuPtr, ID_MACRO_STOP,            MF_BYCOMMAND | MF_ENABLED);
            DISCARD EnableMenuItem(PopUpMenuPtr, ID_MACRO_STARTRECORDING,  MF_BYCOMMAND | MF_GRAYED );
            DISCARD EnableMenuItem(PopUpMenuPtr, ID_MACRO_RESTARTPLAYBACK, MF_BYCOMMAND | MF_GRAYED );
            DISCARD EnableMenuItem(PopUpMenuPtr, ID_MACRO_STOP,            MF_BYCOMMAND | MF_ENABLED);
        } elif (recmode == RECMODE_NORMAL)
        {   DISCARD EnableMenuItem(MenuPtr, ID_MACRO_STARTRECORDING,  MF_BYCOMMAND | MF_ENABLED);
            DISCARD EnableMenuItem(MenuPtr, ID_MACRO_RESTARTPLAYBACK, MF_BYCOMMAND | MF_GRAYED );
            DISCARD EnableMenuItem(MenuPtr, ID_MACRO_STOP,            MF_BYCOMMAND | MF_GRAYED );
            DISCARD EnableMenuItem(PopUpMenuPtr, ID_MACRO_STARTRECORDING,  MF_BYCOMMAND | MF_ENABLED);
            DISCARD EnableMenuItem(PopUpMenuPtr, ID_MACRO_RESTARTPLAYBACK, MF_BYCOMMAND | MF_GRAYED );
            DISCARD EnableMenuItem(PopUpMenuPtr, ID_MACRO_STOP,            MF_BYCOMMAND | MF_GRAYED );
        } // else assert(0);
#endif
    acase MENUITEM_MONITOR:
#ifdef AMIGA
        if (fullscreen || MonitorWindowPtr)
#endif
#ifdef WIN32
        if (MonitorWindowPtr)
#endif
        {                               ghost(which);
        } else                          unghost(which);
    acase MENUITEM_NARROW:
        if (fullscreen)                 ghost(which);
        else                            unghost(which);
        if   (wide == 1)                tick(which);
        elif (wide == 2)                untick(which);
        // else assert(0);
    acase MENUITEM_OPEN:
    case MENUITEM_QUICKLOAD:
        if (crippled)                   ghost(which);
        else                            unghost(which);
    acase MENUITEM_PAUSED:
        if (paused)                     tick(which);
        else                            untick(which);
    acase MENUITEM_POINTER:
        if (showpointer)                tick(which);
        else                            untick(which);
    acase MENUITEM_PSGS:
        if (machine != ELEKTOR || memmap == MEMMAP_E)
        {                               ghost(which);
        } else                          unghost(which);
    acase MENUITEM_REGION:
#ifdef AMIGA
        if (region == REGION_NTSC)
        {    ItemAddress(MenuPtr, FULLMENUNUM(MN_SETTINGS, IN_SPEED, SN_NTSC))->Flags |= CHECKED;
             ItemAddress(MenuPtr, FULLMENUNUM(MN_SETTINGS, IN_SPEED, SN_PAL ))->Flags &= ~CHECKED;
        } else
        {    // assert(region == REGION_PAL);
             ItemAddress(MenuPtr, FULLMENUNUM(MN_SETTINGS, IN_SPEED, SN_NTSC))->Flags &= ~CHECKED;
             ItemAddress(MenuPtr, FULLMENUNUM(MN_SETTINGS, IN_SPEED, SN_PAL ))->Flags |= CHECKED;
        }
#endif
#ifdef WIN32
        if (region == REGION_NTSC)
        {   DISCARD CheckMenuRadioItem(MenuPtr,      ID_OPTIONS_SPEED_NTSC, ID_OPTIONS_SPEED_PAL, ID_OPTIONS_SPEED_NTSC, MF_BYCOMMAND);
            DISCARD CheckMenuRadioItem(PopUpMenuPtr, ID_OPTIONS_SPEED_NTSC, ID_OPTIONS_SPEED_PAL, ID_OPTIONS_SPEED_NTSC, MF_BYCOMMAND);
        } elif (region == REGION_PAL)
        {   DISCARD CheckMenuRadioItem(MenuPtr,      ID_OPTIONS_SPEED_NTSC, ID_OPTIONS_SPEED_PAL, ID_OPTIONS_SPEED_PAL , MF_BYCOMMAND);
            DISCARD CheckMenuRadioItem(PopUpMenuPtr, ID_OPTIONS_SPEED_NTSC, ID_OPTIONS_SPEED_PAL, ID_OPTIONS_SPEED_PAL , MF_BYCOMMAND);
        } // else assert(0);
#endif
    acase MENUITEM_REQUIREBUTTON:
        if (requirebutton)              tick(which);
        else                            untick(which);
    acase MENUITEM_RESET:
        if (crippled)                   ghost(which);
        else                            unghost(which);
    acase MENUITEM_RESETTOMONITOR:
        if (crippled || machine != ELEKTOR) // actually it might be safe when crippled?
        {                               ghost(which);
        } else                          unghost(which);
    acase MENUITEM_RUNTODMA:
        if (machine != ARCADIA)         ghost(which);
        else                            unghost(which);
    acase MENUITEM_RUNTOINTERRUPT:
        if (machine == ARCADIA)         ghost(which);
        else                            unghost(which);
    acase MENUITEM_QUICKSAVE:
    case MENUITEM_REVERT:
    case MENUITEM_SAVEROM:
    case MENUITEM_SAVESNP:
        if (crippled || !game)          ghost(which);
        else                            unghost(which);
    acase MENUITEM_SAVETVC:
        if (crippled || !game || machine != ELEKTOR)
        {                               ghost(which);
        } else                          unghost(which);
    acase MENUITEM_SENSEGAME:
#ifdef AMIGA
// Other options are automatically cleared if the user selected the menu item (or its keyboard shortcut)
// Manual clearing of other options is only needed if the setting has been changed programmatically
// This never occurs for sensegame
        if   (sensegame == SENSEGAME_VERBOSE) ItemAddress(MenuPtr, FULLMENUNUM(MN_SETTINGS, IN_SENSEGAME, SN_VERBOSE))->Flags |= CHECKED;
        elif (sensegame == SENSEGAME_QUIET  ) ItemAddress(MenuPtr, FULLMENUNUM(MN_SETTINGS, IN_SENSEGAME, SN_QUIET  ))->Flags |= CHECKED;
        elif (sensegame == SENSEGAME_NO     ) ItemAddress(MenuPtr, FULLMENUNUM(MN_SETTINGS, IN_SENSEGAME, SN_NO     ))->Flags |= CHECKED;
#endif
#ifdef WIN32
        if (sensegame == SENSEGAME_VERBOSE)
        {   CheckMenuRadioItem(MenuPtr,      ID_SENSEGAME_VERBOSE, ID_SENSEGAME_NO, ID_SENSEGAME_VERBOSE, MF_BYCOMMAND);
            CheckMenuRadioItem(PopUpMenuPtr, ID_SENSEGAME_VERBOSE, ID_SENSEGAME_NO, ID_SENSEGAME_VERBOSE, MF_BYCOMMAND);
        } elif (sensegame == SENSEGAME_QUIET)
        {   CheckMenuRadioItem(MenuPtr,      ID_SENSEGAME_VERBOSE, ID_SENSEGAME_NO, ID_SENSEGAME_QUIET  , MF_BYCOMMAND);
            CheckMenuRadioItem(PopUpMenuPtr, ID_SENSEGAME_VERBOSE, ID_SENSEGAME_NO, ID_SENSEGAME_QUIET  , MF_BYCOMMAND);
        } elif (sensegame == SENSEGAME_NO)
        {   CheckMenuRadioItem(MenuPtr,      ID_SENSEGAME_VERBOSE, ID_SENSEGAME_NO, ID_SENSEGAME_NO     , MF_BYCOMMAND);
            CheckMenuRadioItem(PopUpMenuPtr, ID_SENSEGAME_VERBOSE, ID_SENSEGAME_NO, ID_SENSEGAME_NO     , MF_BYCOMMAND);
        } // else assert(0);
#endif
    acase MENUITEM_SERVER:
#ifdef WIN32
        if (crippled)                   ghost(which);
#endif
#ifdef AMIGA
        if (crippled || !SocketBase)    ghost(which);
#endif
        else                            unghost(which);
        if (connected == NET_SERVER)    tick(which);
        else                            untick(which);
    acase MENUITEM_SIZE:
#ifdef AMIGA
        if (fullscreen)
        {                     DISCARD OffMenu(MainWindowPtr, FULLMENUNUM(MN_SETTINGS, IN_GRAPHICS, SN_1XSIZE));
                              DISCARD OffMenu(MainWindowPtr, FULLMENUNUM(MN_SETTINGS, IN_GRAPHICS, SN_2XSIZE));
                              DISCARD OffMenu(MainWindowPtr, FULLMENUNUM(MN_SETTINGS, IN_GRAPHICS, SN_3XSIZE));
        } else
        {                      DISCARD OnMenu(MainWindowPtr, FULLMENUNUM(MN_SETTINGS, IN_GRAPHICS, SN_1XSIZE));
                               DISCARD OnMenu(MainWindowPtr, FULLMENUNUM(MN_SETTINGS, IN_GRAPHICS, SN_2XSIZE));
                               DISCARD OnMenu(MainWindowPtr, FULLMENUNUM(MN_SETTINGS, IN_GRAPHICS, SN_3XSIZE));
        }
// Other options are automatically cleared if the user selected the menu item (or its keyboard shortcut)
// Manual clearing of other options is only needed if the setting has been changed programmatically
// This never occurs for sizes
        if   (size == 1)                ItemAddress(MenuPtr, FULLMENUNUM(MN_SETTINGS, IN_GRAPHICS, SN_1XSIZE))->Flags |= CHECKED;
        elif (size == 2)                ItemAddress(MenuPtr, FULLMENUNUM(MN_SETTINGS, IN_GRAPHICS, SN_2XSIZE))->Flags |= CHECKED;
        elif (size == 3)                ItemAddress(MenuPtr, FULLMENUNUM(MN_SETTINGS, IN_GRAPHICS, SN_3XSIZE))->Flags |= CHECKED;
#endif
#ifdef WIN32
        if (fullscreen)
        {   DISCARD EnableMenuItem(MenuPtr,      ID_OPTIONS_GRAPHICS_1XSIZE,  MF_BYCOMMAND | MF_GRAYED);
            DISCARD EnableMenuItem(MenuPtr,      ID_OPTIONS_GRAPHICS_2XSIZE,  MF_BYCOMMAND | MF_GRAYED);
            DISCARD EnableMenuItem(MenuPtr,      ID_OPTIONS_GRAPHICS_3XSIZE,  MF_BYCOMMAND | MF_GRAYED);
            DISCARD EnableMenuItem(PopUpMenuPtr, ID_OPTIONS_GRAPHICS_1XSIZE,  MF_BYCOMMAND | MF_GRAYED);
            DISCARD EnableMenuItem(PopUpMenuPtr, ID_OPTIONS_GRAPHICS_2XSIZE,  MF_BYCOMMAND | MF_GRAYED);
            DISCARD EnableMenuItem(PopUpMenuPtr, ID_OPTIONS_GRAPHICS_3XSIZE,  MF_BYCOMMAND | MF_GRAYED);
        } else
        {   DISCARD EnableMenuItem(MenuPtr,      ID_OPTIONS_GRAPHICS_1XSIZE,  MF_BYCOMMAND | MF_ENABLED);
            DISCARD EnableMenuItem(MenuPtr,      ID_OPTIONS_GRAPHICS_2XSIZE,  MF_BYCOMMAND | MF_ENABLED);
            DISCARD EnableMenuItem(MenuPtr,      ID_OPTIONS_GRAPHICS_3XSIZE,  MF_BYCOMMAND | MF_ENABLED);
            DISCARD EnableMenuItem(PopUpMenuPtr, ID_OPTIONS_GRAPHICS_1XSIZE,  MF_BYCOMMAND | MF_ENABLED);
            DISCARD EnableMenuItem(PopUpMenuPtr, ID_OPTIONS_GRAPHICS_2XSIZE,  MF_BYCOMMAND | MF_ENABLED);
            DISCARD EnableMenuItem(PopUpMenuPtr, ID_OPTIONS_GRAPHICS_3XSIZE,  MF_BYCOMMAND | MF_ENABLED);
        }
        if (size == 1)
        {   DISCARD CheckMenuRadioItem(MenuPtr,      ID_OPTIONS_GRAPHICS_1XSIZE, ID_OPTIONS_GRAPHICS_3XSIZE, ID_OPTIONS_GRAPHICS_1XSIZE, MF_BYCOMMAND);
            DISCARD CheckMenuRadioItem(PopUpMenuPtr, ID_OPTIONS_GRAPHICS_1XSIZE, ID_OPTIONS_GRAPHICS_3XSIZE, ID_OPTIONS_GRAPHICS_1XSIZE, MF_BYCOMMAND);
        } elif (size == 2)
        {   DISCARD CheckMenuRadioItem(MenuPtr,      ID_OPTIONS_GRAPHICS_1XSIZE, ID_OPTIONS_GRAPHICS_3XSIZE, ID_OPTIONS_GRAPHICS_2XSIZE, MF_BYCOMMAND);
            DISCARD CheckMenuRadioItem(PopUpMenuPtr, ID_OPTIONS_GRAPHICS_1XSIZE, ID_OPTIONS_GRAPHICS_3XSIZE, ID_OPTIONS_GRAPHICS_2XSIZE, MF_BYCOMMAND);
        } elif (size == 3)
        {   DISCARD CheckMenuRadioItem(MenuPtr,      ID_OPTIONS_GRAPHICS_1XSIZE, ID_OPTIONS_GRAPHICS_3XSIZE, ID_OPTIONS_GRAPHICS_3XSIZE, MF_BYCOMMAND);
            DISCARD CheckMenuRadioItem(PopUpMenuPtr, ID_OPTIONS_GRAPHICS_1XSIZE, ID_OPTIONS_GRAPHICS_3XSIZE, ID_OPTIONS_GRAPHICS_3XSIZE, MF_BYCOMMAND);
        } // else assert(0);
#endif
    acase MENUITEM_SOUND:
        if (sound)                      tick(which);
        else                            untick(which);
    acase MENUITEM_STRETCH:
        if (!fullscreen)                ghost(which);
        else                            unghost(which);
        if (stretch)                    tick(which);
        else                            untick(which);
    acase MENUITEM_SWAPPED:
        if (swapped)                    tick(which);
        else                            untick(which);
    acase MENUITEM_TITLEBAR:
#ifdef AMIGA
        if (!fullscreen)                ghost(which);
        else                            unghost(which);
#endif
        if (showtitlebar)               tick(which);
        else                            untick(which);
    acase MENUITEM_TRACECPU:
        if (trace)                      tick(which);
        else                            untick(which);
    acase MENUITEM_USEHACKS:
        if (machine == ARCADIA || sensegame == SENSEGAME_NO)
        {                               ghost(which);
        } else                          unghost(which);
        if (usehacks)                   tick(which);
        else                            untick(which);
    acase MENUITEM_WARP:
        if (warp)                       tick(which);
        else                            untick(which);
    acase MENUITEM_WATCHREADS:
        if (watchreads)                 tick(which);
        else                            untick(which);

#ifdef AMIGA
    acase MENUITEM_ABOUT:
        if (fullscreen)                 ghost(which);
        else                            unghost(which);
    acase MENUITEM_ABSOLUTE:
        if (absolute)                   tick(which);
        else                            untick(which);
    acase MENUITEM_FILTER:
        if (led)                        tick(which);
        else                            untick(which);
    acase MENUITEM_MANUAL:
        if (!AmigaGuideBase)            ghost(which);
        else                            unghost(which);
    acase MENUITEM_SENSITIVITY:
        if (!analog)                    ghost(which);
        else                            unghost(which);
    acase MENUITEM_SETBP:
        if (fullscreen)                 ghost(which);
        else                            unghost(which);
    acase MENUITEM_SETWP:
        if (fullscreen)                 ghost(which);
        else                            unghost(which);
#endif
#ifdef WIN32
    acase MENUITEM_JOY0:
        if (joy0)                       tick(which);
        else                            untick(which);
    acase MENUITEM_MENUBAR:
        if (showmenubar)                tick(which);
        else                            untick(which);
    acase MENUITEM_TOOLBAR:
        if (showtoolbar)                tick(which);
        else                            untick(which);
#endif

    adefault:
    break;
}   }

EXPORT void updatesmlgad(int which, ULONG state, FLAG now)
{   if (!MainWindowPtr)
    {   return; // important!
    }

#ifdef AMIGA
    if (fullscreen)
    {   return; // for safety
    }

    if (now)
    {   detachsmlgads();
    }
#endif
#ifdef WIN32
    if (!showtoolbar)
    {   return; // for speed
    }
#endif

    switch(which)
    {
    case GADPOS_FLAGLINE:
    case GADPOS_AUTOFIRE:
    case GADPOS_ANALOG:
    case GADPOS_SWAPPED:
    case GADPOS_PAUSED:
    case GADPOS_COLLISIONS:
    case GADPOS_SOUND:
    case GADPOS_WARP:
        setbutton(which, TRUE, (FLAG) state);
    break;
    case GADPOS_DEMULTIPLEX:
        if (machine == ARCADIA)
        {   setbutton(which, TRUE, (FLAG) state);
        } else
        {   // assert(machine == INTERTON || machine == ELEKTOR);
            setbutton(which, FALSE, (FLAG) state);
        }
    break;
    case GADPOS_MACRO:
#ifdef AMIGA
        DISCARD SetGadgetAttrs
        (   gadgets[GID_BU1], MainWindowPtr, NULL,
            GA_Image,    smallimage[recmode],
            GA_Disabled, crippled || !game,
        TAG_DONE);
        // this autorefreshes
#endif
#ifdef WIN32
        if (crippled || !game)
        {   SendMessage(hToolbar, TB_SETSTATE, ID_MACRO_TOGGLEPLAYING,   0);
            SendMessage(hToolbar, TB_SETSTATE, ID_MACRO_TOGGLERECORDING, 0);
        } elif (recmode == RECMODE_PLAY)
        {   SendMessage(hToolbar, TB_SETSTATE, ID_MACRO_TOGGLEPLAYING,   TBSTATE_ENABLED | TBSTATE_CHECKED);
            SendMessage(hToolbar, TB_SETSTATE, ID_MACRO_TOGGLERECORDING, 0);
        } elif (recmode == RECMODE_RECORD)
        {   SendMessage(hToolbar, TB_SETSTATE, ID_MACRO_TOGGLEPLAYING,   0);
            SendMessage(hToolbar, TB_SETSTATE, ID_MACRO_TOGGLERECORDING, TBSTATE_ENABLED | TBSTATE_CHECKED);
        } else
        {   // assert(recmode == RECMODE_NORMAL);
            SendMessage(hToolbar, TB_SETSTATE, ID_MACRO_TOGGLEPLAYING,   0);
            SendMessage(hToolbar, TB_SETSTATE, ID_MACRO_TOGGLERECORDING, TBSTATE_ENABLED);
        }
#endif
    break;
    default:
    break;
    }

#ifdef AMIGA
    if (now)
    {   attachsmlgads();
    }
#endif
}

EXPORT void updatesmlgads(void)
{   if (!MainWindowPtr)
    {   return; // important!
    }

#ifdef AMIGA
    if (fullscreen)
    {   return;
    }

    detachsmlgads();
#endif

    // graphics
    updatesmlgad(GADPOS_FLAGLINE   , flagline   , FALSE);
    // controllers
    updatesmlgad(GADPOS_AUTOFIRE   , autofire   , FALSE);
    updatesmlgad(GADPOS_ANALOG     , analog     , FALSE);
    updatesmlgad(GADPOS_SWAPPED    , swapped    , FALSE);
    // speed
    updatesmlgad(GADPOS_PAUSED     , paused     , FALSE);
    updatesmlgad(GADPOS_WARP       , warp       , FALSE);
    // sprites
    updatesmlgad(GADPOS_COLLISIONS , collisions , FALSE);
    updatesmlgad(GADPOS_DEMULTIPLEX, demultiplex, FALSE);
    // misc
    updatesmlgad(GADPOS_SOUND      , sound      , FALSE);

#ifdef AMIGA
    attachsmlgads();
#endif

    updatesmlgad(GADPOS_MACRO      , (ULONG) recmode, TRUE );
}

EXPORT void updatemenus(void)
{   int i;

    if (!MainWindowPtr)
    {   return; // important!
    }

#ifdef AMIGA
    if (!MenuPtr)
    {   return;
    }
#endif

    for (i = 0; i < MENUITEMS; i++)
    {   updatemenu(i);
}   }

MODULE void alloc_iobuffer(ULONG thesize)
{   // If, for example, we were playing an EAR and then a BIN was dragged
    // over, we will need to free the old IOBuffer
    if (IOBuffer)
    {   free_iobuffer();
    }
    IOBuffer = (UBYTE *) malloc(thesize);
#ifdef VERBOSE
    T1; printf("Allocated I/O buffer!\n"); T2;
#endif
}
MODULE void free_iobuffer(void)
{   // assert(IOBuffer);
    free(IOBuffer);
    IOBuffer = NULL;
#ifdef VERBOSE
    T1; printf("Freed I/O buffer!\n"); T2;
#endif
}

MODULE void a_setmemmap(void)
{   int i, address;

    cpuread = a_cpuread;
    game = FALSE;

    for (i = 0; i <= 0x7FF; i++)
    {   memory[i] = a_bios[i];
    }
    for (i = 0x800; i <= 0x7FFF; i++)
    {   memory[i] = 0;
    }
    for (i = 0x1909; i <= 0x197F; i++)
    {   memory[i] = 0xC0; // some of them (randomly determined?) are $40 though
    }
    for (i = 0x19C0; i <= 0x19F7; i++)
    {   memory[i] = 0xFF;
    }

    for (address = 0; address < 32 * KILOBYTE; address++)
    {   memflags[address] = 0;
        mirror_r[address] =
        mirror_w[address] = (UWORD) address;

        if
        (    address <= 0x0FFF                       // ROM
    // not occurring:    $1000..$17FF                   mirror
    // legal:            $1800..$18CF                   upper screen
    // legal:            $18D0..$18EF                   user RAM
    // legal:            $18F0..$18F7                   R/W hardware registers
    // legal:            $18F8..$18FB                   user RAM
    // legal:            $18FC                          R/W hardware register
    // legal:            $18FD                          -/W hardware register
    // legal:            $18FE                          R/W hardware register
         || (address >= 0x18FF && address <= 0x197F) // CHARLINE, PnXKEYS/PALLADIUM, CONSOLE, unmapped
    // legal:            $1980..$19BF                   R/W hardware registers
         || (address >= 0x19C0 && address <= 0x19F7) // unmapped
    // legal:            $19F8..$19FB                   -/W hardware registers
         || (address >= 0x19FC && address <= 0x19FF) // BGCOLLIDE, SPRITECOLLIDE, PnPADDLE
    // legal:            $1A00..$1ACF                   lower screen
    // legal:            $1AD0..$1AFF                   user RAM
         ||  address >= 0x1B00                       // not occurring, ROM
        )
        {   memflags[address] |= NOWRITE;
        }

        if
        (
        // legal:                $0000              ..$0FFF     ROM
        // not occurring:        $1000              ..$17FF     mirrored
        // legal:                $1800              ..$18CF     upper screen
        // legal:                $18D0              ..$18EF     user RAM
        // legal:                $18F0              ..$18F7     R/W hardware regs
        // legal:                $18F8              ..$18FB     user RAM
        // legal:                $18FC              ..$18FE     R/W hardware regs
        // legal:                $18FF              ..$1908     R/- hardware regs
                    (address >= 0x1909 && address <= 0x197F) // unmapped (apparently reads as $40 or $C0)
        // legal:                $1980              ..$19BF     R/W hardware regs
                 || (address >= 0x19C0 && address <= 0x19FB) // unmapped (apparently reads as $FF), -/W hardware regs
        // legal:                $19FC              ..$19FF     read-once hardware regs
        // legal:                $1A00              ..$1ACF     lower screen
        // legal:                $1AD0              ..$1AFF     user RAM
        // not occurring:        $1B00              ..$1FFF     mirrored
        // legal:                $2000              ..$2FFF     ROM
        // not occurring:        $3000              ..$3FFF     mirrored
        // legal:                $4000              ..$4FFF     ROM
        // not occurring:        $5000              ..$5FFF     mirrored
        // legal:                $6000              ..$6FFF     ROM
        // not occurring:        $7000              ..$7FFF     mirrored
        )
        {   memflags[address] |= NOREAD;
        }

        if (address >= 0x19FC && address <= 0x19FF) // read-once: $19FC..$19FF: PnPADDLE, SPRITECOLLIDE, BGCOLLIDE
        {   memflags[address] |= READONCE;
        }

        if (address == A_PITCH || address == A_VOLUME)
        {   memflags[address] |= AUDIBLE;
        }

        if ((address & 0x1000) && (address <= 0x17FF || address >= 0x1B00))
        // $1000..$17FF
        // $1B00..$1FFF
        // $3000..$3FFF
        // $5000..$5FFF
        // $7000..$7FFF
        {   memflags[address] |= MIRRORED_R | MIRRORED_W;

            // primary mirroring
            mirror_r[address] &= 0x13FF; // -> $1000..$13FF (mask of $0001 0011 1111 1111)
            mirror_r[address] |= 0x0800; // -> $1800..$1BFF (set  of $0000 1000 0000 0000)

            // secondary mirroring
            if (mirror_r[address] >= 0x1B00 && mirror_r[address] <= 0x1BFF)
            {   mirror_r[address] -=  0x200; // $1B00..$1BFF -> $1900..$19FF
            }

            mirror_w[address] = mirror_r[address];
}   }   }

MODULE void i_setmemmap(void)
{   int i, address;

    // assert(!crippled);
    // assert(!inframe);
    cpuread = i_cpuread;
    game = FALSE;

    for (i = 0; i <= 0x7FF; i++)
    {   memory[i] = i_bios[i];
    }
    for (i = 0x800; i <= 0x7FFF; i++)
    {   memory[i] = 0;
    }

    // only MIRRORED_R and/or MIRRORED_W need be set for mirrored regions.

    for (address = 0; address < 32 * KILOBYTE; address++)
    {   memflags[address] = 0;
        mirror_r[address] =
        mirror_w[address] = (UWORD) address;

        // this is only responsible for $0000..$15FF and $1800..$1DFF
        // regions
        switch(memmap)
        {
        case MEMMAP_A: // 2K ROM + 0K RAM
            if
            (    address <= 0x15FF
             || (address >= 0x1800 && address <= 0x1DFF)
            )
            {   memflags[address] |= NOWRITE;
            }
            if
            (   (address >=  0x800 && address <= 0x15FF)
             || (address >= 0x1800 && address <= 0x1DFF)
            )
            {   memflags[address] |= NOREAD;
            }
            // $0000..$07FF: ROM
            // $0800..$15FF: unused
            // $1600..$17FF: mirror of $1E00..$1FFF, as usual
        acase MEMMAP_B: // 4K ROM + 0K RAM
            if
            (    address <= 0x15FF
             || (address >= 0x1800 && address <= 0x1DFF)
            )
            {   memflags[address] |= NOWRITE;
            }
            if
            (   (address >= 0x1000 && address <= 0x15FF)
             || (address >= 0x1800 && address <= 0x1DFF)
            )
            {   memflags[address] |= NOREAD;
            }
            // $0000..$0FFF: ROM
            // $1000..$15FF: unused
            // $1600..$17FF: mirror of $1E00..$1FFF, as usual
        acase MEMMAP_C: // 4K ROM + 1K RAM
            if (address <= 0x0FFF)
            {   memflags[address] |= NOWRITE;
            }
            // we are allowed to read everything (ROM, cartridge RAM,
            // mirror of noise/input/PVI, mirror of cartridge RAM)
            // $0000..$0FFF: ROM
            // $1000..$15FF: cartridge RAM, and mirror of cartridge RAM
            // $1600..$17FF: mirror of $1E00..$1FFF, as usual
            // $1800..$1DFF: mirror of cartridge RAM
        acase MEMMAP_D: // 6K ROM + 1K RAM
            if (address <= 0x15FF)
            {   memflags[address] |= NOWRITE; // $0000..$15FF: 5.5K of ROM
            }
            // we are allowed to read everything (ROM, cartridge RAM,
            // mirror of noise/input/PVI, mirror of cartridge RAM)
            // $0000..$15FF: ROM
            // $1600..$17FF: mirror of $1E00..$1FFF, as usual
            // $1800..$1DFF: cartridge RAM, and mirror of cartridge RAM
        adefault:
            // assert(0);
        break;
        }

        if
        (
            (address >= 0x1E00 && address <= 0x1E7F) // unmapped
        // legal:        $1E80                          hardware register
         || (address >= 0x1E81 && address <= 0x1EFF) // unknown, R/- hardware registers, unknown
        // legal:        $1F00..$1F2D                   hardware registers, RAM, h/w regs, RAM, h/w regs
         || (address >= 0x1F2E && address <= 0x1F3F) // unmapped
        // legal:        $1F40..$1F6D                   hardware registers, RAM
         || (address >= 0x1F6E && address <= 0x1F7F) // unmapped
        // legal:        $1F80..$1FAD                   grid, RAM
         || (address >= 0x1FAE && address <= 0x1FBF) // unmapped
        // legal:        $1FC0..$1FC3                   hardware registers
         || (address >= 0x1FC4 && address <= 0x1FC5) // unused
        // legal:        $1FC6..$1FC9                   hardware registers
         ||  address >= 0x1FCA                       // R/- hardware registers, unused, mirrors (not occurring), semi-mirrors
        )
        {   memflags[address] |= NOWRITE;
        }

        if
        (   (address >= 0x1E00 && address <= 0x1E7F) // unmapped
    // legal:            $1E80                          hardware register
         || (address >= 0x1E81 && address <= 0x1E87) // unknown
    // legal:            $1E88              ..$1E8E     R/- hardware registers
         || (address >= 0x1E8F && address <= 0x1EFF) // unknown
    // legal:            $1F00              ..$1F2D     hardware registers, RAM, h/w regs, RAM, h/w regs
         || (address >= 0x1F2E && address <= 0x1F3F) // unmapped
    // legal:            $1F40              ..$1F6D     hardware registers, RAM
         || (address >= 0x1F6E && address <= 0x1F7F) // unmapped
    // legal:            $1F80              ..$1FAD     grid, RAM
         || (address >= 0x1FAE && address <= 0x1FC9) // unmapped, -/W (write-only) and -/- (unusable) hardware registers
    // legal:            $1FCA              ..$1FCD     R/- (read-only) hardware registers
         || (address >= 0x1FCE && address <= 0x1FD9) // unused, mirrors (not occurring)
    // legal:            $1FDA              ..$1FDB     semi-mirrors
         || (address >= 0x1FDC && address <= 0x1FE9) // mirrors (not occurring)
    // legal:            $1FEA              ..$1FEB     semi-mirrors
         || (address >= 0x1FEC && address <= 0x1FF9) // mirrors (not occurring)
    // legal:            $1FFA              ..$1FFB     semi-mirrors
         ||  address >= 0x1FFC                       // mirrors (not occurring)
        )
        {   memflags[address] |= NOREAD;
        }

        if
        (   (address >= 0x1FCA && address <= 0x1FCB)
         || (address >= 0x1FDA && address <= 0x1FDB)
         || (address >= 0x1FEA && address <= 0x1FEB)
         || (address >= 0x1FFA && address <= 0x1FFB)
        )
        {   memflags[address] |= READONCE;
        }
        if (address == I_PITCH || address == I_NOISE)
        {   memflags[address] |= AUDIBLE;
        }

        /* An address such as $77FF will be a 3-level mirror!
        $77FF -> $17FF after primary   mirroring
        $17FF -> $1FFF after secondary mirroring
        $1FFF -> $1FCF after tertiary  mirroring */

        // primary write mirroring
        if (mirror_r[address] >= 0x2000) // could use address instead of memmap[address]
        // $2000..$3FFF: mirror of $0000..$1FFF
        // $4000..$5FFF: mirror of $0000..$1FFF
        // $6000..$7FFF: mirror of $0000..$1FFF
        {   memflags[address] |= MIRRORED_R;
            mirror_r[address] &= 0x1FFF;
        }
        // secondary read mirroring
        if (mirror_r[address] >= 0x1600 && mirror_r[address] <= 0x17FF) // mirror of $1E00..$1FFF
        {   memflags[address] |= MIRRORED_R;
            mirror_r[address] += 0x800;
        }
        // tertiary read mirroring
        // mirrors of $1C00..$1C09 and $1C0C..$1C0F.
        // $1C0A..$1C0B are semi-mirrored.
        if
        (   (mirror_r[address] >= 0x1FD0 && mirror_r[address] <= 0x1FD9)
         || (mirror_r[address] >= 0x1FDC && mirror_r[address] <= 0x1FE9)
         || (mirror_r[address] >= 0x1FEC && mirror_r[address] <= 0x1FF9)
         || (mirror_r[address] >= 0x1FFC && mirror_r[address] <= 0x1FFE) // $1FFF is a randomizer
        )
        {   memflags[address] |= MIRRORED_R;
            mirror_r[address] &= 0xFFCF; // apply mask of %1111111111001111
        }

        // primary write mirroring
        if (mirror_w[address] >= 0x2000) // could use address instead of memmap[address]
        // $2000..$3FFF: mirror of $0000..$1FFF
        // $4000..$5FFF: mirror of $0000..$1FFF
        // $6000..$7FFF: mirror of $0000..$1FFF
        {   memflags[address] |= MIRRORED_W;
            mirror_w[address] &= 0x1FFF;
        }
        // secondary write mirroring
        if (mirror_w[address] >= 0x1600 && mirror_w[address] <= 0x17FF) // mirror of $1E00..$1FFF
        {   memflags[address] |= MIRRORED_W;
            mirror_w[address] += 0x800;
        }
        // tertiary write mirroring
        // mirrors of $1C00..$1C09 and $1C0C..$1C0F.
        // $1C0A..$1C0B are semi-mirrored.
        if
        (   (mirror_r[address] >= 0x1FD0 && mirror_r[address] <= 0x1FD9)
         || (mirror_r[address] >= 0x1FDC && mirror_r[address] <= 0x1FE9)
         || (mirror_r[address] >= 0x1FEC && mirror_r[address] <= 0x1FF9)
         || (mirror_r[address] >= 0x1FFC && mirror_r[address] <= 0x1FFE) // $1FFF is a randomizer
        )
        {   memflags[address] |= MIRRORED_W;
            mirror_r[address] &= 0xFFCF; // apply mask of %1111111111001111
        }

        inputmirrors(address);

        if (memmap == MEMMAP_C)
        {   if (mirror_r[address] >= 0x1400 && mirror_r[address] <= 0x15FF) // mirror of $1000..$11FF
            {   memflags[address] |= MIRRORED_R;
                mirror_r[address] -= 0x400;
            } elif (mirror_r[address] >= 0x1800 && mirror_r[address] <= 0x1BFF) // mirror of $1000..$13FF
            {   memflags[address] |= MIRRORED_R;
                mirror_r[address] -= 0x800;
            } elif (mirror_r[address] >= 0x1C00 && mirror_r[address] <= 0x1DFF) // mirror of $1400..$15FF (which is itself a mirror of $1000..$11FF)
            {   memflags[address] |= MIRRORED_R;
                mirror_r[address] -= 0xC00; // 2-level mirror, resolved completely here
            }

            if (mirror_w[address] >= 0x1400 && mirror_w[address] <= 0x15FF) // mirror of $1000..$11FF
            {   memflags[address] |= MIRRORED_W;
                mirror_w[address] -= 0x400;
            } elif (mirror_w[address] >= 0x1800 && mirror_w[address] <= 0x1BFF) // mirror of $1000..$13FF
            {   memflags[address] |= MIRRORED_W;
                mirror_w[address] -= 0x800;
            } elif (mirror_w[address] >= 0x1C00 && mirror_w[address] <= 0x1DFF) // mirror of $1400..$15FF (which is itself a mirror of $1000..$11FF)
            {   memflags[address] |= MIRRORED_W;
                mirror_w[address] -= 0xC00; // 2-level mirror, resolved completely here
        }   }
        elif (memmap == MEMMAP_D)
        {   if (mirror_r[address] >= 0x1C00 && mirror_r[address] <= 0x1DFF) // mirror of $1800..$19FF
            {   mirror_r[address] -= 0x400;
            }

            if (mirror_w[address] >= 0x1C00 && mirror_w[address] <= 0x1DFF) // mirror of $1800..$19FF
            {   mirror_w[address] -= 0x400;
}   }   }   }

MODULE void e_setmemmap(void)
{   int i, address;

    // assert(!crippled);
    // assert(!inframe);
    cpuread = e_cpuread;
    game = FALSE;

    for (i = 0; i <= 0x7FF; i++)
    {   memory[i] = e_bios[i];
    }
    for (i = 0x800; i <= 0xFFF; i++)
    {   memory[i] = 0;
    }
    if (memmap == MEMMAP_E)
    {   for (i = 0x1000; i <= 0x15FF; i++)
        {   memory[i] = 0xFF; // unimportant, illegal reads don't actually look at the stored value
            /* "If extended memory is not available, the value
                retrieved from address $1000 will be $FF." - "TV Games
                Computer" book, p. 185. */
    }   }
    else
    {   // assert(memmap == MEMMAP_F);
        for (i = 0x1000; i <= 0x15FF; i++)
        {   memory[i] = 0;
    }   }
    for (i = 0x1600; i <= 0x7FFF; i++)
    {   memory[i] = 0;
    }

    // only MIRRORED_R and/or MIRRORED_W need be set for mirrored regions.

    for (address = 0; address < 32 * KILOBYTE; address++)
    {   memflags[address] = 0;
        mirror_r[address] =
        mirror_w[address] = (UWORD) address;

        if (memmap == MEMMAP_E)
        {   if
            (    address <=  0x7FF                       // ROM
        // legal:             $800..               $FFF     monitor RAM, user RAM
             || (address >= 0x1000 && address <= 0x1EFF) // unused, mirrors, I/O, input and output lines (and mirrors thereof), unmapped, R/- hardware regs, unmapped, mirrors
        // legal:            $1F00..              $1F2D     R/W hardware regs, user RAM
             || (address >= 0x1F2E && address <= 0x1F3F) // unmapped
        // legal:            $1F40..              $1F6D     R/W hardware regs, user RAM
             || (address >= 0x1F6E && address <= 0x1F7F) // unmapped
        // legal:            $1F80..              $1FAD     R/W hardware regs, user RAM
             || (address >= 0x1FAE && address <= 0x1FBF) // unmapped
        // legal:            $1FC0..              $1FC3     -/W hardware regs
             || (address >= 0x1FC4 && address <= 0x1FC5) // unused
        // legal:            $1FC6..              $1FC9     -/W hardware regs
             ||  address >= 0x1FCA                       // R/- hardware registers, unused, semi-mirrors, unused
            )
            {   memflags[address] |= NOWRITE;
            }

            if
            (
        // legal:               $0..               $FFF     ROM, monitor RAM, user RAM
                (address >= 0x1000 && address <= 0x1E87) // unused, mirrors, I/O, input and output lines (and mirrors thereof), unused, unmapped
        // legal:            $1E88..              $1E8E     R/- hardware registers
             || (address >= 0x1E8F && address <= 0x1EFF) // unmapped, mirrors
        // legal:            $1F00..              $1F2D     R/W hardware registers, user RAM
             || (address >= 0x1F2E && address <= 0x1F3F) // unmapped
        // legal:            $1F40..              $1F6D     R/W hardware registers, user RAM
             || (address >= 0x1F6E && address <= 0x1F7F) // unmapped
        // legal:            $1F80..              $1FAD     R/W hardware registers, user RAM
             || (address >= 0x1FAE && address <= 0x1FC9) // unmapped, -/W hardware registers
        // legal:            $1FCA              ..$1FCD     R/- hardware registers
             || (address >= 0x1FCE && address <= 0x1FD9) // unused, mirrors (not occurring)
        // legal:            $1FDA              ..$1FDB     semi-mirrors
             || (address >= 0x1FDC && address <= 0x1FE9) // mirrors (not occurring)
        // legal:            $1FEA              ..$1FEB     semi-mirrors
             || (address >= 0x1FEC && address <= 0x1FF9) // mirrors (not occurring)
        // legal:            $1FFA              ..$1FFB     semi-mirrors
             ||  address >= 0x1FFC                       // mirrors (not occurring)
            )
            {   memflags[address] |= NOREAD;
        }   }
        else
        {   // assert(memmap == MEMMAP_F);
            if
            (    address <=  0x7FF                       // ROM
        // legal:             $800..              $1BFF     monitor RAM, user RAM
             || (address >= 0x1C00 && address <= 0x1CFF) // unused, mirrors
        // legal:            $1D00..              $1D0E     PSG #0
             ||  address == 0x1D0F                       // PORTB1
        // legal:            $1D10..              $1D1D     PSG #1
             || (address == 0x1D1E && address <= 0x1E7F) // PORTA2, PORTB2, randomizer #2, I/O, input and output lines (and mirrors thereof), unmapped
        // legal:            $1E80                          noise register
             || (address >= 0x1E81 && address <= 0x1EFF) // unmapped, R/- hardware regs, unmapped, mirrors
        // legal:            $1F00..              $1F2D     R/W hardware regs, user RAM
             || (address >= 0x1F2E && address <= 0x1F3F) // unmapped
        // legal:            $1F40..              $1F6D     R/W hardware regs, user RAM
             || (address >= 0x1F6E && address <= 0x1F7F) // unmapped
        // legal:            $1F80..              $1FAD     R/W hardware regs, user RAM
             || (address >= 0x1FAE && address <= 0x1FBF) // unmapped
        // legal:            $1FC0..              $1FC3     -/W hardware regs
             || (address >= 0x1FC4 && address <= 0x1FC5) // unused
        // legal:            $1FC6..              $1FC9     -/W hardware regs
             ||  address >= 0x1FCA                       // R/- hardware registers, unused, randomizer, semi-mirrors, unused
            )
            {   memflags[address] |= NOWRITE;
            }

            /* Writes to PORTB1, PORTA2, PORTB2 are considered illegal,
               even though these areas are write-only. This is because
               no game would write there. */

            if
            (
        // legal:               $0..              $1BFF     ROM, monitor RAM, user RAM
                (address >= 0x1C00 && address <= 0x1D1F) // unused, mirrors, PSGs (-/W)
        // legal:            $1D20                          randomizer
             || (address >= 0x1D20 && address <= 0x1E87) // I/O, input and output lines (and mirrors thereof), unused, noise register (assuming -/W), unmapped
        // legal:            $1E88..              $1E8E     R/- hardware registers
             || (address >= 0x1E8F && address <= 0x1EFF) // unmapped, mirrors
        // legal:            $1F00..              $1F2D     R/W hardware registers, user RAM
             || (address >= 0x1F2E && address <= 0x1F3F) // unmapped
        // legal:            $1F40..              $1F6D     R/W hardware registers, user RAM
             || (address >= 0x1F6E && address <= 0x1F7F) // unmapped
        // legal:            $1F80..              $1FAD     R/W hardware registers, user RAM
             ||  address == 0x1FAE                       // unmapped
        // legal:            $1FAF                          1st German randomizer
             || (address >= 0x1FB0 && address <= 0x1FC9) // unmapped, -/W hardware registers
        // legal:            $1FCA              ..$1FCD     R/- hardware registers
             || (address >= 0x1FCE && address <= 0x1FD9) // unused, mirrors (not occurring)
        // legal:            $1FDA              ..$1FDB     semi-mirrors
             || (address >= 0x1FDC && address <= 0x1FE9) // mirrors (not occurring)
        // legal:            $1FEA              ..$1FEB     semi-mirrors
             || (address >= 0x1FEC && address <= 0x1FF9) // mirrors (not occurring)
        // legal:            $1FFA              ..$1FFB     semi-mirrors
             || (address >= 0x1FFC && address <= 0x1FFE) // mirrors (not occurring)
        // legal:            $1FFF                          1st English randomizer
             ||  address >= 0x2000                       // unused
            )
            {   memflags[address] |= NOREAD;
        }   }

        if
        (   (address >= 0x1FCA && address <= 0x1FCB)
         || (address >= 0x1FDA && address <= 0x1FDB)
         || (address >= 0x1FEA && address <= 0x1FEB)
         || (address >= 0x1FFA && address <= 0x1FFB)
        )
        {   memflags[address] |= READONCE;
        }

        if
        (   address == I_PITCH
         || (   memmap == MEMMAP_F
             && (   address == I_NOISE
                 || (address >= 0x1D00 && address <= 0x1D1F) // not really every byte in this range
        )   )   )
        {   memflags[address] |= AUDIBLE;
        }

        if (memmap == MEMMAP_E)
        {   // primary mirroring, must go first
            if (mirror_r[address] >= 0x2000)
            {   memflags[address] |= MIRRORED_R;
                mirror_r[address] &= 0x1FFF;
            }
            if (mirror_w[address] >= 0x2000)
            {   memflags[address] |= MIRRORED_W;
                mirror_w[address] &= 0x1FFF;
            }

            // secondary mirroring
            if (mirror_r[address] >= 0x1600 && mirror_r[address] <= 0x17FF) // mirror of $1E00..$1FFF
            {   memflags[address] |= MIRRORED_R;
                mirror_r[address] += 0x800;
            } elif (mirror_r[address] >= 0x1800 && mirror_r[address] <= 0x1CFF)
            {   memflags[address] |= MIRRORED_R;
                mirror_r[address] &= 0xF8FF;
                mirror_r[address] |= 0x0500;
            }
            if (mirror_w[address] >= 0x1600 && mirror_w[address] <= 0x17FF) // mirror of $1E00..$1FFF
            {   memflags[address] |= MIRRORED_W;
                mirror_w[address] += 0x800;
            } elif (mirror_w[address] >= 0x1800 && mirror_w[address] <= 0x1CFF)
            {   memflags[address] |= MIRRORED_W;
                mirror_w[address] &= 0xF8FF;
                mirror_w[address] |= 0x0500;
        }   }

        // tertiary mirroring

        // mirrors of $1C00..$1C09 and $1C0C..$1C0F.
        // $1C0A..$1C0B are semi-mirrored.
        if
        (   (mirror_r[address] >= 0x1FD0 && mirror_r[address] <= 0x1FD9)
         || (mirror_r[address] >= 0x1FDC && mirror_r[address] <= 0x1FE9)
         || (mirror_r[address] >= 0x1FEC && mirror_r[address] <= 0x1FF9)
         || (mirror_r[address] >= 0x1FFC && mirror_r[address] <= 0x1FFE) // $1FFF is a randomizer
        )
        {   memflags[address] |= MIRRORED_R;
            mirror_r[address] &= 0xFFCF; // apply mask of %1111111111001111
        }

        if (mirror_r[address] >= 0x1C20 && mirror_r[address] <= 0x1CFF) // mirror of $1D20..$1DFF
        {   memflags[address] |= MIRRORED_R;
            mirror_r[address] += 0x100;
        }

        // mirrors of $1C00..$1C09 and $1C0C..$1C0F.
        // $1C0A..$1C0B are semi-mirrored.
        if
        (   (mirror_w[address] >= 0x1FD0 && mirror_w[address] <= 0x1FD9)
         || (mirror_w[address] >= 0x1FDC && mirror_w[address] <= 0x1FE9)
         || (mirror_w[address] >= 0x1FEC && mirror_w[address] <= 0x1FF9)
         || (mirror_w[address] >= 0x1FFC && mirror_w[address] <= 0x1FFE) // $1FFF is a randomizer
        )
        {   memflags[address] |= MIRRORED_W;
            mirror_w[address] &= 0xFFCF; // apply mask of %1111111111001111
        }
        if (mirror_w[address] >= 0x1C20 && mirror_w[address] <= 0x1CFF) // mirror of $1D20..$1DFF
        {   memflags[address] |= MIRRORED_W;
            mirror_w[address] += 0x100;
        }

        inputmirrors(address);
    }

/* Unresolved issues:
 how should writes to $1FFF be handled?
  a) mirror them: they will resolve to $1FCF; or...
  b) handle them as (illegal) writes to randomizer #1
  we're handling them as (b) currently.
 we're assuming $1D20 is mirrored at $1C20, but it might not be.
*/

    if (memory[0x73E] == 0x19) // buggy version of monitor
    {   memflags[0x199C] |= MIRRORED_R | MIRRORED_W;
        memflags[0x199D] |= MIRRORED_R | MIRRORED_W;
        memflags[0x199E] |= MIRRORED_R | MIRRORED_W;
        memflags[0x199F] |= MIRRORED_R | MIRRORED_W;
        memflags[0x19BC] |= MIRRORED_R | MIRRORED_W;
        memflags[0x19BD] |= MIRRORED_R | MIRRORED_W;
        memflags[0x19BE] |= MIRRORED_R | MIRRORED_W;
        memflags[0x19BF] |= MIRRORED_R | MIRRORED_W;
        mirror_r[0x199C] = mirror_w[0x199C] = 0x1D9C;
        mirror_r[0x199D] = mirror_w[0x199D] = 0x1D9D;
        mirror_r[0x199E] = mirror_w[0x199E] = 0x1D9E;
        mirror_r[0x199F] = mirror_w[0x199F] = 0x1D9F;
        mirror_r[0x19BC] = mirror_w[0x19BC] = 0x1DBC;
        mirror_r[0x19BD] = mirror_w[0x19BD] = 0x1DBD;
        mirror_r[0x19BE] = mirror_w[0x19BE] = 0x1DBE;
        mirror_r[0x19BF] = mirror_w[0x19BF] = 0x1DBF;
    } // elif (memory[0x73E] != 0x1D) assert(0); // not fixed version either, so panic!
}

MODULE void inputmirrors(int address)
{   // write mirrors
    if (    mirror_w[address] >= 0x1E81 && mirror_w[address] <= 0x1EFF)
    {   memflags[address] |= MIRRORED_W;
        mirror_w[address] = 0x1E80;
    }
    if (             address  == 0x1E80) // only applies to real $1E80, not mirrors thereof!
    {   memflags[address] |= NOREAD;
    }

    // read mirrors
    if (    mirror_r[address] >= 0x1E81 && mirror_r[address] <= 0x1E97)
    {   memflags[address] |= NOWRITE;
    } elif (mirror_r[address] >= 0x1E98 && mirror_r[address] <= 0x1E9B)
    {   memflags[address] |= MIRRORED_R;
        mirror_r[address] -= 0x10;
    } elif (mirror_r[address] >= 0x1E9C && mirror_r[address] <= 0x1EA7)
    {   memflags[address] |= NOWRITE;
    } elif (mirror_r[address] >= 0x1EA8 && mirror_r[address] <= 0x1EAE)
    {   memflags[address] |= MIRRORED_R;
        mirror_r[address] -= 0x20;
    } elif (mirror_r[address] >= 0x1EAF && mirror_r[address] <= 0x1EB7)
    {   memflags[address] |= NOWRITE;
    } elif (mirror_r[address] >= 0x1EB8 && mirror_r[address] <= 0x1EBB)
    {   memflags[address] |= MIRRORED_R;
        mirror_r[address] -= 0x30;
    } elif (mirror_r[address] >= 0x1EBC && mirror_r[address] <= 0x1EC7)
    {   memflags[address] |= NOWRITE;
    } elif (mirror_r[address] >= 0x1EC8 && mirror_r[address] <= 0x1ECE)
    {   memflags[address] |= MIRRORED_R;
        mirror_r[address] -= 0x40;
    } elif (mirror_r[address] >= 0x1ECF && mirror_r[address] <= 0x1ED7)
    {   memflags[address] |= NOWRITE;
    } elif (mirror_r[address] >= 0x1ED8 && mirror_r[address] <= 0x1EDB)
    {   memflags[address] |= MIRRORED_R;
        mirror_r[address] -= 0x50;
    } elif (mirror_r[address] >= 0x1EDC && mirror_r[address] <= 0x1EE7)
    {   memflags[address] |= NOWRITE;
    } elif (mirror_r[address] >= 0x1EE8 && mirror_r[address] <= 0x1EEE)
    {   memflags[address] |= MIRRORED_R;
        mirror_r[address] -= 0x60;
    } elif (mirror_r[address] >= 0x1EEF && mirror_r[address] <= 0x1EF7)
    {   memflags[address] |= NOWRITE;
    } elif (mirror_r[address] >= 0x1EF8 && mirror_r[address] <= 0x1EFB)
    {   memflags[address] |= MIRRORED_R;
        mirror_r[address] -= 0x70;
    } elif (mirror_r[address] >= 0x1EFC && mirror_r[address] <= 0x1EFF)
    {   memflags[address] |= NOWRITE;
}   }

MODULE void engine_reset(FLAG user)
{   int i;

    // We probably don't need to do these, as Arcadia/Interton games
    // (at least) have no right to assume they are in any particular
    // state.
    if (machine != ELEKTOR)
    {   for (i = 0; i <= 6; i++)
        {   r[i] = 0;
        }
        for (i = 0; i <= 7; i++)
        {   ras[i] = 0;
        }
        psu =
        psl = 0;
        iar = 0;
    } else
    {   // assert(machine == ELEKTOR);

        // $8A2/$8A3          = start address of game
        // $8B9/$8BA (INTADR) = address of game interrupt vector
        // $8BE/$8BF (PC)     = start address of game

        /* There are 3 feasible possibilities for starting the monitor:
        iar =     0; // gives black background with white "00 00" at top
        iar =  0x23; // gives blue  background with yellow "IIII" at bottom
        iar = 0x540; // gives blue  background
        */
        if (game)
        {   iar = 0x540;
            memory[0x8A2] = startaddr_h;
            memory[0x8A3] = startaddr_l;
        } else
        {   iar =  0x23; // gives blue  background with yellow "IIII" at bottom
        }

        for (i = 0x1D00; i <= 0x1D1F; i++)
        {   memory[i] = 0; // PSGs are cleared at reset
        }

        playsound(TRUE);
    }

    interrupt = FALSE;
    page      = 0;
    elapsed   = 0;
    frames    = 0;
    oldframes = 0;
    overcalc  = 0;

#ifdef PROFILE
    for (i = 0; i <= 255; i++)
    {   profile[i] = 0;
    }
#endif
#ifdef COVERAGE
    for (i = 0; i < 32 * KILOBYTE; i++)
    {   coverage[i] = 0;
    }
#endif
#ifdef MODE_MHZ
    oldelapsed = 0;
#endif

    /* We should really clear all memory (important for netplay, to */
    /* keep states at both sides consistent). */
}

EXPORT FLAG engine_load(FLAG silent)
{   FILE*  LocalHandle /* = NULL */ ;

    filesize = (int) getsize(fn_game);

    alloc_iobuffer((ULONG) filesize);
    if (!(LocalHandle = fopen((char *) fn_game, "rb"))) // just cast for lint
    {
#ifdef VERBOSE
        T1; printf("Can't open %s for reading!\n", fn_game); T2;
#endif
        return FALSE;
    }
    if (fread(IOBuffer, (size_t) filesize, 1, LocalHandle) != 1)
    {   fclose(LocalHandle);
        // LocalHandle = NULL;
        return FALSE;
    }
    fclose(LocalHandle);
    // LocalHandle = NULL;

    breakparts();

    if (!parse_bytes(TRUE, silent))
    {   return FALSE;
    }

    // assert(MainWindowPtr);

    updatemenus();
    updatesmlgads();
    updatebiggads();
    settitle();
    redrawscreen();

    return TRUE;
}

EXPORT void engine_save(int kind, FLAG updatetitle)
{   FILE* LocalHandle /* = NULL */ ;
    int i;
    UWORD loadaddress,
          endaddress;

    if (kind == KIND_BIN || kind == KIND_PGM || kind == KIND_TVC)
    {   alloc_iobuffer((ULONG) ROMSIZE_MAX);
    } else
    {   /* assert(kind == KIND_EAS || kind == KIND_EAR
               || kind == KIND_INS || kind == KIND_INR
               || kind == KIND_TVS || kind == KIND_TVR
                 ); */
        alloc_iobuffer((ULONG) SNPSIZE_MAX);
    }

    offset = 0;

    // magic byte (1 byte)
    switch(kind)
    {
    case KIND_PGM:
        WriteByte( 0x0); // indeterminate opcode (LODZ,r0)
    acase KIND_TVC:
        WriteByte( 0x2); // valid opcode (LODZ,r2) :-(
                         // (although useless as a first instruction)
    acase KIND_EAS:
        WriteByte(0x10); // unused opcode
    acase KIND_EAR:
        WriteByte(0x11); // unused opcode
    acase KIND_INS:
        WriteByte(0xB6); // unused opcode
    acase KIND_INR:
        WriteByte(0xB7); // unused opcode
    acase KIND_TVS:
        WriteByte(0x90); // unused opcode
    acase KIND_TVR:
        WriteByte(0x91); // unused opcode
    adefault:
        // assert(kind == KIND_BIN);
    break;
    }

    if (kind != KIND_BIN && kind != KIND_PGM && kind != KIND_TVC)
    {   /* assert(kind == KIND_EAS || kind == KIND_EAR
               || kind == KIND_INS || kind == KIND_INR
               || kind == KIND_TVS || kind == KIND_TVR
                 ); */

        // state
        WriteWord((UWORD) romsize);
        WriteWord((UWORD) iar);
        WriteByte((UBYTE) page);
        WriteByte(psu);
        WriteByte(psl);
        for (i = 0; i < 8; i++)
        {   WriteWord(ras[i]);
        }
        for (i = 0; i < 7; i++)
        {   WriteByte(r[i]);
        }
        WriteByte(6); // version ID (6 means V4.62+)
        WriteByte((UBYTE) overcalc);
        WriteByte((UBYTE) interrupt);
        WriteByte((UBYTE) ax[0]);
        WriteByte((UBYTE) ay[0]);
        WriteByte((UBYTE) ax[1]);
        WriteByte((UBYTE) ay[1]);

        /* Supported file versions are:
           BIN/PGM/TVC:  any
           EAS:           6 (V4.62+)
           EAR:           6 (V4.62+)
           INS:           6 (V4.62+)
           INR:           6 (V4.62+)
           TVS:           6 (V4.62+)
           TVR:           6 (V4.62+)
           CFG (Win):    11 (V4.71+)
           config (Ami):  8 (V4.3 -4.81), 12 (V4.82+) */

        if (machine == ARCADIA)
        {   // assert(kind == KIND_EAS || kind == KIND_EAR);

            // UVI/RAM dump
            for (i = 0x1800; i <= 0x1908; i++) /* 265 bytes */ WriteByte(memory[i]);
            for (i = 0x1980; i <= 0x19BF; i++) /*  64 bytes */ WriteByte(memory[i]);
            for (i = 0x19F8; i <= 0x1AFF; i++) /* 264 bytes */ WriteByte(memory[i]);
                                             /* = 593 bytes */
        } elif (machine == INTERTON)
        {   // assert(kind == KIND_INS || kind == KIND_INR);

            for (i = 0x1000; i <= 0x15FF; i++) // 1.5K
            {   WriteByte(memory[i]);
            }
            for (i = 0x1800; i <= 0x1DFF; i++) // 1.5K
            {   WriteByte(memory[i]);
            }
            for (i = 0x1E88; i <= 0x1E8E; i++) // 7 bytes
            {   WriteByte(memory[i]);
            }
            for (i = 0x1F00; i <= 0x1F2D; i++) // 46 bytes
            {   WriteByte(memory[i]);
            }
            for (i = 0x1F40; i <= 0x1F6D; i++) // 46 bytes
            {   WriteByte(memory[i]);
            }
            for (i = 0x1F80; i <= 0x1FAD; i++) // 46 bytes
            {   WriteByte(memory[i]);
            }
            for (i = 0x1FC0; i <= 0x1FC3; i++) // 4 bytes
            {   WriteByte(memory[i]);
            }
            for (i = 0x1FC6; i <= 0x1FCD; i++) // 8 bytes
            {   WriteByte(memory[i]);
        }   }
        else
        {   // assert(machine == ELEKTOR);
            // assert(kind == KIND_TVS || kind == KIND_TVR);

            WriteWord((UWORD) whichgame);

            for (i = 0x800; i <= 0x1FFF; i++)
            {   WriteByte(memory[i]);
    }   }   }

    // assert(romsize <= ROMSIZE_MAX);

    if (machine == ELEKTOR)
    {   if (kind == KIND_PGM)
        {   // assert(offset == 1);
            WriteByte(memory[0x8BE]);
            WriteByte(memory[0x8BF]);
            for (i = 3; i <= 0x7FF; i++)
            {   WriteByte(0);
            }
            for (i = 0x800; i < romsize; i++)
            {   WriteByte(memory[i]);
        }   }
        elif (kind == KIND_TVC)
        {   // find the first nonzero byte
            loadaddress = // fallback for the rather theoretical case where it is completely empty from $8C0 onwards
            endaddress  = (UWORD) 0x8C0;
            offset      = 0x8C0;
            for (i = 0x8C0; i < romsize; i++)
            {   if (ReadByte() != 0)
                {   loadaddress = (UWORD) i;
                    break;
            }   }
            // find the last nonzero byte
            for (i = (int) loadaddress; i < romsize; i++)
            {   if (ReadByte() != 0)
                {   endaddress = (UWORD) i;
            }   }
            // write out the header
            offset = 1;
            WriteWord(loadaddress);   // bytes 1 and 2
            WriteByte(memory[0x8BE]); // byte  3
            WriteByte(memory[0x8BF]); // byte  4
            // write out the body
            for (i = (int) loadaddress; i <= (int) endaddress; i++)
            {   WriteByte(memory[i]);
            }
#ifdef VERBOSE
        T1;
        printf("Load  address is $%lX.\n", loadaddress);
        printf("Start address is $%lX.\n", (memory[0x8BE] * 256) + memory[0x8BF]);
        T2;
#endif
    }   }
    elif (machine == INTERTON || romsize <= 4096)
    {   for (i = 0; i < romsize; i++)
        {   WriteByte(memory[i]);              // $0000..$0FFF -> $0000..$0FFF
    }   }
    elif (romsize <= 8192)
    {   if (machine == ARCADIA)
        {   for (i = 0; i < 4096; i++)
            {   WriteByte(memory[i]);          // $0000..$0FFF -> $0000..$0FFF
            }
            for (i = 0; i < romsize - 4096; i++)
            {   WriteByte(memory[8192 + i]);   // $2000..$2FFF -> $1000..$1FFF
        }   }
        else
        {   // assert(machine == INTERTON);
            for (i = 0; i < 0x15FF; i++)
            {   WriteByte(memory[i]);          // $0000..$15FF -> $0000..$15FF
    }   }   }
    elif (romsize <= 12 * KILOBYTE)
    {   for (i = 0; i < 4096; i++)
        {   WriteByte(memory[i]);              // $0000..$0FFF -> $0000..$0FFF
        }
        for (i = 0; i < 4096; i++)
        {   WriteByte(memory[0x2000 + i]);     // $2000..$2FFF -> $1000..$1FFF
        }
        for (i = 0; i < romsize - 8192; i++)
        {   WriteByte(memory[0x4000 + i]); // $4000..$4FFF -> $2000..$2FFF
    }   }
    else
    {   // assert(romsize <= 16 * KILOBYTE);
        for (i = 0; i < 4096; i++)
        {   WriteByte(memory[i]);          // $0000..$0FFF -> $0000..$0FFF
        }
        for (i = 0; i < 4096; i++)
        {   WriteByte(memory[0x2000 + i]); // $2000..$2FFF -> $1000..$1FFF
        }
        for (i = 0; i < 4096; i++)
        {   WriteByte(memory[0x4000 + i]); // $4000..$4FFF -> $2000..$2FFF
        }
        for (i = 0; i < romsize - (12 * KILOBYTE); i++)
        {   WriteByte(memory[0x6000 + i]); // $6000..$6FFF -> $3000..$3FFF
    }   }

    if ((LocalHandle = fopen(fn_game, "wb")))
    {   DISCARD fwrite(IOBuffer, (size_t) offset, 1, LocalHandle);
        DISCARD fclose(LocalHandle);
        // LocalHandle = NULL;
    } else
    {   beep();
        free_iobuffer();
        return;
    }
    free_iobuffer();

#ifdef AMIGA
    writeicon(kind, fn_game);
#endif

    if
    (   kind != KIND_EAR
     && kind != KIND_INR
     && kind != KIND_TVR
    )
    {   if (updatetitle)
        {   settitle();
        }
        return;
    }

    if (!(MacroHandle = fopen(fn_game, "a+b")))
    {   beep();
        return;
    }

    // assert(updatetitle == TRUE);
    settitle();
    recmode = RECMODE_RECORD;

    if (MainWindowPtr)
    {   updatesmlgad(GADPOS_MACRO, (ULONG) recmode, TRUE);
}   }

MODULE UWORD ReadWord(void)
{   UWORD result;

    // big endian 16-bit word
    result = (UWORD) (256 * IOBuffer[offset])
                          + IOBuffer[offset + 1];
    offset += 2;
    return(result);
}
MODULE UBYTE ReadByte(void)
{   UBYTE result;

    result = IOBuffer[offset++];
    return(result);
}
MODULE UWORD ReadWordAt(int where)
{   UWORD result;

    offset = where;
    // big endian 16-bit word
    result = (UWORD) (256 * IOBuffer[offset])
                          + IOBuffer[offset + 1];
    offset += 2;
    return(result);
}
MODULE UBYTE ReadByteAt(int where)
{   UBYTE result;

    offset = where;
    result = IOBuffer[offset++];
    return(result);
}
MODULE void WriteWord(UWORD what)
{   // big endian 16-bit word
    IOBuffer[offset]     = what / 256;
    IOBuffer[offset + 1] = what % 256;
    offset += 2;
}
MODULE void WriteByte(UBYTE what)
{   IOBuffer[offset++]   = what;
}

MODULE UBYTE a_cpuread(int address)
{   // assert(address <= 32767);
    // assert(machine == ARCADIA);

    /* Reads to memory from the UVI do not go through this routine.
    Strictly speaking, we should use this routine for ALL CPU read
    accesses, eg. reading opcodes/operands, relative addresses, etc.
    Currently, we only use it for absolute accesses: LODA, STRA, etc.

    Testing on an authentic Emerson Arcadia 2001 console has been
    undertaken to make the read/write property emulation as accurate as
    possible. However, this has not occurred for the Interton VC 4000.

    Reads of read-once registers, even if it is the second or later read,
    are not logged as illegal reads.

    Reads of unused ROM areas (eg. $0800..$0FFF for 2K ROMs) are not
    logged as illegal reads. */

    // handle mirroring

    if (memflags[address] & MIRRORED_R)
    {   if (logreads)
        {   getfriendly((int) mirror_r[address]);
            T1;
            printf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_READMIRROR,
                    "Instruction at $%X (raster %ld) read from mirrored address $%X (resolves to %s)!\n\n"
                ),
                (unsigned int) (page + iar),
                (long int)     rast,
                (unsigned int) address,
                               friendly
            );
            T2;
        }
        address = (int) mirror_r[address];
    }

#ifdef COVERAGE
    coverage[address] |= COVERAGE_READ;
#endif

    if (memflags[address] & NOREAD)
    {   t = 0xFF; // some really return eg. $40, $C0 instead of $FF
    } else
    {   t = memory[address];
    }
    if (memflags[address] & READONCE)
    {   memory[address] = 0xFF;
        // we really should use uviwrite() for this, but any watchpoint will be triggered below
    }

    // handle illegal reads

    if (logreads && (memflags[address] & NOREAD))
    {   getfriendly(address);
        T1;
        printf
        (   GetCatalogStr(CatalogPtr,
                MSG_ENGINE_ILLEGALREAD,
                "Instruction at $%X (raster %ld) is attempting to read from unreadable address %s!\n\n"
            ),
            (unsigned int) (page + iar),
            (long int)     rast,
                           friendly
        );
        T2;
    }

    if (address == wp && (watchreads || memflags[address] & READONCE))
    {   getfriendly(wp);
        T1;
        if (watchreads)
        {   printf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_HITWP,
                    "Instruction at $%X (raster %ld) is reading $%X from %s!\n" // yes, just one \n
                ),
                (unsigned int) (page + iar),
                (long int)     rast,
                               t,
                               friendly
            );
        }
        if (memflags[address] & READONCE)
        {   printf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_READONCE,
                    "Read-once register %s has been reset to $%X!\n" // yes, just one \n
                ),
                               friendly,
                (unsigned int) memory[address]
            );
        }
        printf("\n");
        T2;
        runto();
    }

    return(t);
}

MODULE UBYTE i_cpuread(int address)
{   // assert(address <= 32767);
    // assert(machine == INTERTON);

    /* Reads of read-once registers, even if it is the second or later
    read, are not logged as illegal reads.

    Reads of unused ROM areas (eg. $0800..$0FFF for 2K ROMs) are not
    logged as illegal reads. */

    // handle mirroring

    if (memflags[address] & MIRRORED_R)
    {   if (logreads)
        {   getfriendly((int) mirror_r[address]);
            T1;
            printf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_READMIRROR,
                    "Instruction at $%X (raster %ld) read from mirrored address $%X (resolves to %s)!\n\n"
                ),
                (unsigned int) (page + iar),
                (long int)     rast,
                (unsigned int) address,
                               friendly
            );
            T2;
        }
        address = (int) mirror_r[address];
    }

#ifdef COVERAGE
    coverage[address] |= COVERAGE_READ;
#endif

    if (memflags[address] & NOREAD)
    {   if (address == I_NOISE)
        {   t = 0x0C; // assuming same behaviour as Elektor
        } elif
        (   // It would probably be better to handle these in setmemmap().
            (address >= 0x1F30 && address <= 0x1F3F)
         || (address >= 0x1F6E && address <= 0x1F7F)
         || (address >= 0x1FAE && address <= 0x1FBF) // this region is $00 for Elektor, so perhaps also for Interton?
        )
        {   t = 0xFF; // same as how MESS V2 initializes these regions
        } else
        {   t = 0x00;
            /* this is assumed rather than known behaviour, but:
            Interton BLACKJAC expects the high nybble (at least) of $1E8F to be $0! */
    }   }
    else
    {   t = memory[address];
    }

    if (memflags[address] & READONCE)
    {   memory[address] = 0;
        // we deliberately don't zero the semi-mirrors of this, that is
        // why they are only semi-mirrors.
        // we really should use pviwrite() for this, but if there's a
        // watchpoint set, that would mean breaking in the middle of
        // executing a CPU instruction.
    }

    if (logreads && (memflags[address] & NOREAD))
    {   getfriendly(address);
        T1;
        printf
        (   GetCatalogStr(CatalogPtr,
                MSG_ENGINE_ILLEGALREAD,
                "Instruction at $%X (raster %ld) is attempting to read from unreadable address %s!\n\n"
            ),
            (unsigned int) (page + iar),
            (long int)     rast,
                           friendly
        );
        T2;
    }

    if (address == wp && (watchreads || memflags[address] & READONCE))
    {   getfriendly(wp);
        T1;
        if (watchreads)
        {   printf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_HITWP,
                    "Instruction at $%X (raster %ld) is reading $%X from %s!\n" // yes, just one \n
                ),
                (unsigned int) (page + iar),
                (long int)     rast,
                               t,
                               friendly
            );
        }
        if (memflags[address] & READONCE)
        {   printf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_READONCE,
                    "Read-once register %s has been reset to $%lX!\n" // yes, just one \n
                ),
                                    friendly,
                (long unsigned int) memory[address]
            );
        }
        printf("\n");
        T2;
        runto();
    }

    return(t);
}

MODULE UBYTE e_cpuread(int address)
{   // assert(address <= 32767);
    // assert(machine == INTERTON);

    /* Reads of read-once registers, even if it is the second or later
    read, are not logged as illegal reads. */

    // handle mirroring

    if (memflags[address] & MIRRORED_R)
    {   if (logreads)
        {   getfriendly((int) mirror_r[address]);
            T1;
            printf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_READMIRROR,
                    "Instruction at $%X (raster %ld) read from mirrored address $%X (resolves to %s)!\n\n"
                ),
                (unsigned int) (page + iar),
                (long int)     rast,
                (unsigned int) address,
                               friendly
            );
            T2;
        }
        address = (int) mirror_r[address];
    }

#ifdef COVERAGE
    coverage[address] |= COVERAGE_READ;
#endif

    if
    (   memmap == MEMMAP_F
     && (   address == E_RANDOM1E
         || address == E_RANDOM1G
         || address == E_RANDOM2
    )   )
    {   if (recmode == RECMODE_PLAY)
        {   t = loadrand();
        } else
        {   if (connected == NET_SERVER)
            {   t = rand() % 256;
                DISCARD NetSendByte(t); // should check return code!
            } elif (connected == NET_CLIENT)
            {   t = NetReceiveByte(); // should check return code!
            } else
            {   t = rand() % 256;
            }
            if (recmode == RECMODE_RECORD)
            {   saverand(t);
    }   }   }
    elif (memflags[address] & NOREAD)
    /* It would be better to make a memread[32768] array, containing what
       values each read-only address returns. (Setting memory[] as
       appropriate wouldn't work, because of write-only registers.) */
    {   if (address >= 0x1000 && address <= 0x15FF)
        {   // assert(memmap == MEMMAP_E);
            t = 0xFF; // games which attempt to sense whether in mode "E"/"F" expect unavailable memory to return $FF
        } elif (address >= 0x1D80 && address <= 0x1DBF)
        {   t = 0x80;
            // TESTER2.PGM proves that $1D80 reads as $80.
        } elif (address >= 0x1DC0 && address <= 0x1DFF)
        {   t = 0xC0;
            // TESTER2.PGM proves that $1DC0 reads as $C0.
        } elif (address == I_NOISE)
        {   t = 0x0C;
            // TESTER2.PGM proves that NOISE reads as $0C.
        } elif
        (   (address >= 0x1E8F && address <= 0x1E97)
         || (address >= 0x1E9C && address <= 0x1EA7)
         || (address >= 0x1EAF && address <= 0x1EB7)
         || (address >= 0x1EBC && address <= 0x1EC7)
         || (address >= 0x1ECF && address <= 0x1ED7)
         || (address >= 0x1EDC && address <= 0x1EE7)
         || (address >= 0x1EEF && address <= 0x1EF7)
         || (address >= 0x1EFC && address <= 0x1EFF)
        )
        {   t = 0x0F;
            // TESTER2.PGM proves that $1E93 reads as $0F.
        } elif (address >= 0x1FAE && address <= 0x1FBF)
        {   // Elektor Labyrinth reads $1FB3..$1FB5 and expects them to be $00
            // Also, TESTER2.PGM proves that $1FB8 reads as $00.
            t = 0;
    }   }
    else
    {   t = memory[address];
    }

    // handle read-once

    if (memflags[address] & READONCE)
    {   memory[address] = 0;
        // we deliberately don't zero the semi-mirrors of this, that is
        // why they are only semi-mirrors.
        // we really should use pviwrite() for this?
    }

    // handle illegal reads
    // GRNDPRIX attempts to read from $1F6E..$1F70!
    // not sure whether NOISE ($1E80) is R/W or -/W! (assuming R/W)

    if (logreads && (memflags[address] & NOREAD))
    {   getfriendly(address);
        T1;
        printf
        (   GetCatalogStr(CatalogPtr,
                MSG_ENGINE_ILLEGALREAD,
                "Instruction at $%X (raster %ld) is attempting to read from unreadable address %s!\n\n"
            ),
            (unsigned int) (page + iar),
            (long int)     rast,
                           friendly
        );
        T2;
    }

    if (address == wp && (watchreads || memflags[address] & READONCE))
    {   getfriendly(wp);
        T1;
        if (watchreads)
        {   printf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_HITWP,
                    "Instruction at $%X (raster %ld) is reading $%X from %s!\n" // yes, just one \n
                ),
                (unsigned int) (page + iar),
                (long int)     rast,
                               t,
                               friendly
            );
        }
        if (memflags[address] & READONCE)
        {   printf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_READONCE,
                    "Read-once register %s has been reset to $%lX!\n" // yes, just one \n
                ),
                                    friendly,
                (long unsigned int) memory[address]
            );
        }
        printf("\n");
        T2;
        runto();
    }

    return(t);
}

MODULE void view_pviscreen(void)
{   // assert(machine == INTERTON || machine == ELEKTOR);

    UWORD where = 0x1F80;
    UBYTE data;
    int   i, j;

    T1;

    for (j = 0; j < 20; j++)
    {   hex2(astring, where);
        hex1(disp[0], memory[where]);
        hex1(disp[1], memory[where + 1]);
        printf("$%s: %s %s ", astring, disp[0], disp[1]);

        data = memory[where++];
        for (i = 0; i < 8; i++)
        {   if (data & (128 >> i))
            {   printf("#");
            } else
            {   printf(".");
        }   }
        printf(" ");
        data = memory[where++];
        for (i = 0; i < 8; i++)
        {   if (data & (128 >> i))
            {   printf("#");
            } else
            {   printf(".");
        }   }
        printf("\n");
    }

    for (i = 0; i < 5; i++)
    {   hex1(disp[i], memory[0x1FA8 + i]);
    }
    printf("$1FA8: %s %s %s %s %s\n\n", disp[0], disp[1], disp[2], disp[3], disp[4]);

    T2;
}

EXPORT void view_upperscreen(void)
{   SLONG x, y;
    TEXT  rowchars[16 + 1];

    if (machine == INTERTON || machine == ELEKTOR)
    {   view_pviscreen();
        return;
    }

    rowchars[16] = 0;

    T1;
    for (y = 0; y <= 12; y++)
    {   hex2(astring, (UWORD) (A_OFFSET_UPPERSCREEN + (y * 16)));
        for (x = 0; x <= 15; x++)
        {   hex1(disp[x], memory[A_OFFSET_UPPERSCREEN + x + (y * 16)]);
            rowchars[x] = arcadia_chars[memory[A_OFFSET_UPPERSCREEN + x + (y * 16)] % 64];
        }
        printf
        (   "$%s: %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s\n",
            astring,
            disp[0], disp[1],  disp[2],  disp[3],  disp[4],  disp[5],  disp[6],  disp[7],
            disp[8], disp[9], disp[10], disp[11], disp[12], disp[13], disp[14], disp[15],
            rowchars
        );
    }
    printf("\n");
    T2;
}
EXPORT void view_lowerscreen(void)
{   SLONG x, y;
    TEXT  rowchars[16 + 1];

    if (machine == INTERTON || machine == ELEKTOR)
    {   view_pviscreen();
        return;
    }

    rowchars[16] = 0;

    T1;
    for (y = 0; y <= 12; y++)
    {   hex2(astring, (UWORD) (A_OFFSET_LOWERSCREEN + (y * 16)));
        for (x = 0; x <= 15; x++)
        {   hex1(disp[x], memory[A_OFFSET_LOWERSCREEN + x + (y * 16)]);
            rowchars[x] = arcadia_chars[memory[A_OFFSET_LOWERSCREEN + x + (y * 16)] % 64];
        }
        printf
        (   "$%s: %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s\n",
            astring,
            disp[0], disp[1],  disp[2],  disp[3],  disp[4],  disp[5],  disp[6],  disp[7],
            disp[8], disp[9], disp[10], disp[11], disp[12], disp[13], disp[14], disp[15],
            rowchars
        );
    }
    printf("\n");
    T2;
}
EXPORT void view_screen(void)
{   if (machine == ARCADIA)
    {   view_upperscreen();
        view_lowerscreen();
    } else
    {   // assert(machine == INTERTON || machine == ELEKTOR);
        view_pviscreen();
}   }

EXPORT void view_cram(void)
{   int i, j,
        where;

    if
    (   machine != INTERTON
     || (memmap != MEMMAP_C && memmap != MEMMAP_D)
    )
    {   return;
    }

    T1;

    if (memmap == MEMMAP_C)
    {   where = 0x1000; // 1K ($1000..$13FF)
    } else
    {   // assert(memmap == MEMMAP_D);
        where = 0x1800; // 1K ($1800..$1BFF)
    }
    for (i = where; i <= where + 1023; i += 16)
    {   hex2(astring, (UWORD) i);
        for (j = 0; j <= 15; j++)
        {   hex1(disp[j], memory[i + j]);
        }
        printf("$%s: %s %s %s %s %s %s %s %s  %s %s %s %s %s %s %s %s\n",
                astring,
                disp[0], disp[1],  disp[2],  disp[3],  disp[4],  disp[5],  disp[6],  disp[7],
                disp[8], disp[9], disp[10], disp[11], disp[12], disp[13], disp[14], disp[15]
              );
    }
    printf("\n");

    T2;
}

EXPORT void view_allmemory(void)
{   int i, j,
        max;

    if (machine == ARCADIA)
    {   max = 0x7FFF;
        // perhaps we should suppress showing $3000..$3FFF, $5000..$5FFF
        // and $7000..$7FFF.
    } else
    {   // assert(machine == INTERTON || machine == ELEKTOR);
        max = 0x1FFF;
    }

    T1;

    for (i = 0; i <= max; i += 16)
    {   hex2(astring, (UWORD) i);
        for (j = 0; j <= 15; j++)
        {   hex1(disp[j], memory[mirror_r[i + j]]);
        }
        printf("$%s: %s %s %s %s %s %s %s %s  %s %s %s %s %s %s %s %s\n",
                astring,
                disp[0], disp[1],  disp[2],  disp[3],  disp[4],  disp[5],  disp[6],  disp[7],
                disp[8], disp[9], disp[10], disp[11], disp[12], disp[13], disp[14], disp[15]
              );
    }
    printf("\n");

    T2;
}

EXPORT void view_mram(void)
{   int i, j,
        limit;

    T1;

    if (machine == ARCADIA)
    {   for (i = 0; i < 16; i++)
        {   hex1(disp[i], memory[0x18D0 + i]);
        }
        printf("$18D0: %s %s %s %s %s %s %s %s  %s %s %s %s %s %s %s %s\n",
               disp[0], disp[1],  disp[2],  disp[3],  disp[4],  disp[5],  disp[6],  disp[7],
               disp[8], disp[9], disp[10], disp[11], disp[12], disp[13], disp[14], disp[15]
              );

        for (i = 0; i < 16; i++)
        {   hex1(disp[i], memory[0x18E0 + i]);
        }
        printf("$18E0: %s %s %s %s %s %s %s %s  %s %s %s %s %s %s %s %s\n",
               disp[0], disp[1],  disp[2],  disp[3],  disp[4],  disp[5],  disp[6],  disp[7],
               disp[8], disp[9], disp[10], disp[11], disp[12], disp[13], disp[14], disp[15]
              );

        for (i = 0; i < 4; i++)
        {   hex1(disp[i], memory[0x18F8 + i]);
        }
        printf("$18F8:                         %s %s %s %s\n",
               disp[0], disp[1],  disp[2],  disp[3]
              );

        for (i = 0; i < 16; i++)
        {   hex1(disp[i], memory[0x1AD0 + i]);
        }
        printf("$1AD0: %s %s %s %s %s %s %s %s  %s %s %s %s %s %s %s %s\n",
               disp[0], disp[1],  disp[2],  disp[3],  disp[4],  disp[5],  disp[6],  disp[7],
               disp[8], disp[9], disp[10], disp[11], disp[12], disp[13], disp[14], disp[15]
              );

        for (i = 0; i < 16; i++)
        {   hex1(disp[i], memory[0x1AE0 + i]);
        }
        printf("$1AE0: %s %s %s %s %s %s %s %s  %s %s %s %s %s %s %s %s\n\n",
               disp[0], disp[1],  disp[2],  disp[3],  disp[4],  disp[5],  disp[6],  disp[7],
               disp[8], disp[9], disp[10], disp[11], disp[12], disp[13], disp[14], disp[15]
              );
    } else
    {   if (machine == ELEKTOR)
        {   // monitor RAM
            for (i = 0x800; i <= 0x8BF; i += 16)
            {   hex2(astring, (UWORD) i);
                for (j = 0; j <= 15; j++)
                {   hex1(disp[j], memory[i + j]);
                }
                printf("$%s: %s %s %s %s %s %s %s %s  %s %s %s %s %s %s %s %s\n",
                       astring,
                       disp[0], disp[1],  disp[2],  disp[3],  disp[4],  disp[5],  disp[6],  disp[7],
                       disp[8], disp[9], disp[10], disp[11], disp[12], disp[13], disp[14], disp[15]
                      );
            }
            printf("\n");

            if (memmap == MEMMAP_E)
            {   limit = 0x0FFF;
            } else
            {   // assert(memmap == MEMMAP_F);
                limit = 0x15FF;
            }

            // user RAM
            for (i = 0x8C0; i <= limit; i += 16)
            {   hex2(astring, (UWORD) i);
                for (j = 0; j <= 15; j++)
                {   hex1(disp[j], memory[i + j]);
                }
                printf("$%s: %s %s %s %s %s %s %s %s  %s %s %s %s %s %s %s %s\n",
                       astring,
                       disp[0], disp[1],  disp[2],  disp[3],  disp[4],  disp[5],  disp[6],  disp[7],
                       disp[8], disp[9], disp[10], disp[11], disp[12], disp[13], disp[14], disp[15]
                      );
            }
            printf("\n");

            if (memmap == MEMMAP_F)
            {   // user RAM
                // 8 bytes in this range can be actually obscured RAM, so
                // shouldn't really be shown (if old monitor ROM is in
                // use).
                for (i = 0x1800; i <= 0x1BFF; i += 16)
                {   hex2(astring, (UWORD) i);
                    for (j = 0; j <= 15; j++)
                    {   hex1(disp[j], memory[i + j]);
                    }
                    printf("$%s: %s %s %s %s %s %s %s %s  %s %s %s %s %s %s %s %s\n",
                           astring,
                           disp[0], disp[1],  disp[2],  disp[3],  disp[4],  disp[5],  disp[6],  disp[7],
                           disp[8], disp[9], disp[10], disp[11], disp[12], disp[13], disp[14], disp[15]
                          );
                }
                printf("\n");
        }   }

        hex1(disp[0], memory[0x1F0E]);
        hex1(disp[1], memory[0x1F0F]);
        printf("$1F0E:                                            %s %s\n",
                disp[0], disp[1]
              );

        hex1(disp[0], memory[0x1F1E]);
        hex1(disp[1], memory[0x1F1F]);
        printf("$1F1E:                                            %s %s\n",
                disp[0], disp[1]
              );

        hex1(disp[0], memory[0x1F4E]);
        hex1(disp[1], memory[0x1F4F]);
        printf("$1F4E:                                            %s %s\n",
                disp[0], disp[1]
               );

        for (i = 0; i < 16; i++)
        {   hex1(disp[i], memory[0x1F50 + i]);
        }
        printf("$1F50: %s %s %s %s %s %s %s %s  %s %s %s %s %s %s %s %s\n",
                disp[0], disp[1],  disp[2],  disp[3],  disp[4],  disp[5],  disp[6],  disp[7],
                disp[8], disp[9], disp[10], disp[11], disp[12], disp[13], disp[14], disp[15]
              );

        for (i = 0; i < 14; i++)
        {   hex1(disp[i], memory[0x1F60 + i]);
        }
        printf("$1F60: %s %s %s %s %s %s %s %s  %s %s %s %s %s %s\n",
                disp[0], disp[1],  disp[2],  disp[3],  disp[4],  disp[5],  disp[6],  disp[7],
                disp[8], disp[9], disp[10], disp[11], disp[12], disp[13]
              );

        hex1(disp[0], memory[0x1FAD]);
        printf("$1FAD:                                         %s\n\n",
                disp[0]
              );
    }

    T2;
}

EXPORT void view_psgs(void)
{   // assert(machine == ELEKTOR);

    T1;
    printf
    (   "---------------PSG #1---------------- "
        "---------------PSG #2----------------\n"
    );

    hex2(disp[0], (UWORD) ((memory[E_PITCHA1_H] * 256) + memory[E_PITCHA1_L]));
    hex1(disp[1],  memory[E_PITCHD1      ]);
    hex2(disp[2], (UWORD) ((memory[E_PITCHA2_H] * 256) + memory[E_PITCHA2_L]));
    hex1(disp[3],  memory[E_PITCHD2      ]);
    printf
    (   "PITCHA1:      %s PITCHD1:        %s PITCHA2:      %s PITCHD2:        %s\n",
         disp[0],         disp[1],           disp[2],         disp[3]
    );

    hex2(disp[0], (UWORD) ((memory[E_PITCHB1_H] * 256) + memory[E_PITCHB1_L]));
    hex1(disp[1],  memory[E_ENABLE1      ]);
    hex2(disp[2], (UWORD) ((memory[E_PITCHB2_H] * 256) + memory[E_PITCHB2_L]));
    hex1(disp[3],  memory[E_ENABLE2      ]);
    printf
    (   "PITCHB1:      %s ENABLE1:        %s PITCHB2:      %s ENABLE2:        %s\n",
         disp[0],         disp[1],           disp[2],         disp[3]
    );

    hex2(disp[0], (UWORD) ((memory[E_PITCHC1_H] * 256) + memory[E_PITCHC1_L]));
    hex2(disp[1], (UWORD) ((memory[E_PERIOD1_H] * 256) + memory[E_PERIOD1_L]));
    hex2(disp[2], (UWORD) ((memory[E_PITCHC2_H] * 256) + memory[E_PITCHC2_L]));
    hex2(disp[3], (UWORD) ((memory[E_PERIOD2_H] * 256) + memory[E_PERIOD2_L]));
    printf
    (   "PITCHC1:      %s PERIOD1:      %s PITCHC2:      %s PERIOD2:      %s\n",
         disp[0],         disp[1],         disp[2],         disp[3]
    );

    hex1(disp[0], memory[E_AMPLITUDEA1]);
    hex1(disp[1], memory[E_SHAPE1     ]);
    hex1(disp[2], memory[E_AMPLITUDEA2]);
    hex1(disp[3], memory[E_SHAPE2     ]);
    printf
    (   "AMPLITUDEA1:    %s SHAPE1:         %s AMPLITUDEA2:    %s SHAPE2:         %s\n",
         disp[0],           disp[1],           disp[2],           disp[3]
    );

    hex1(disp[0], memory[E_AMPLITUDEB1]);
    hex1(disp[1], memory[E_PORTA1     ]);
    hex1(disp[2], memory[E_AMPLITUDEB2]);
    hex1(disp[3], memory[E_PORTA2     ]);
    printf
    (   "AMPLITUDEB1:    %s PORTA1:         %s AMPLITUDEB2:    %s PORTA2:         %s\n",
         disp[0],           disp[1],           disp[2],           disp[3]
    );

    hex1(disp[0], memory[E_AMPLITUDEC1]);
    hex1(disp[1], memory[E_PORTB1     ]);
    hex1(disp[2], memory[E_AMPLITUDEC2]);
    hex1(disp[3], memory[E_PORTB2     ]);
    printf
    (   "AMPLITUDEC1:    %s PORTB1:         %s AMPLITUDEC2:    %s PORTB2:         %s\n\n",
         disp[0],           disp[1],           disp[2],           disp[3]
    );

    T2;
}

EXPORT void calcrunningtime(void)
{   PERSIST int hours,
                minutes,
                seconds,
                micros; // all PERSISTent for speed

    if (region == REGION_NTSC)
    {   seconds =          frames / 60 ; // how many seconds
        micros  = 16667 * (frames % 60); // microseconds
    } else
    {   // assert(region == REGION_PAL);
        seconds =          frames / 50 ; // how many seconds
        micros  = 20000 * (frames % 50); // microseconds
    }
    hours   = (seconds / 3600) % 100; // resets to 00 hours after 100 hours
    minutes = (seconds % 3600) / 60;
    seconds %= 60;

    timestring[0]  = '0' + (hours   / 10);
    timestring[1]  = '0' + (hours   % 10);
    timestring[3]  = '0' + (minutes / 10);
    timestring[4]  = '0' + (minutes % 10);
    timestring[6]  = '0' + (seconds / 10);
    timestring[7]  = '0' + (seconds % 10);
    timestring[9]  = '0' + (micros  / 100000);
    timestring[10] = '0' + (micros  % 100000) / 10000;
}

EXPORT void view_xvi(void)
{   decimal(xvistring[0], elapsed + cycles);
    decimal(xvistring[1], (unsigned long int) rast);
    decimal(xvistring[2], frames);
    calcrunningtime();

    T1;
    printf
    (   "%-7s %s %-7s %s %-7s %s %-6s %s\n",
        GetCatalogStr(CatalogPtr, MSG_CLOCK_SHORT,  "Clock:" ),
        xvistring[0],
        GetCatalogStr(CatalogPtr, MSG_RASTER_SHORT, "Raster:"),
        xvistring[1],
        GetCatalogStr(CatalogPtr, MSG_FRAME_SHORT,  "Frame:" ),
        xvistring[2],
        GetCatalogStr(CatalogPtr, MSG_TIME_SHORT,   "Time:"  ),
        timestring
    );

    if (psu & PSU_F)
    {   strcpy(astring, GetCatalogStr(CatalogPtr, MSG_ENGINE_ON2,  "On" ));
    } else
    {   strcpy(astring, GetCatalogStr(CatalogPtr, MSG_ENGINE_OFF2, "Off"));
    }
    if (psu & PSU_S)
    {   strcpy(bstring, GetCatalogStr(CatalogPtr, MSG_ENGINE_ON2,  "On" ));
    } else
    {   strcpy(bstring, GetCatalogStr(CatalogPtr, MSG_ENGINE_OFF2, "Off"));
    }
    printf
    (   "%-11s %6s %-11s %6s\n",
        GetCatalogStr(CatalogPtr, MSG_FLAG,  "Flag:" ),
        astring,
        GetCatalogStr(CatalogPtr, MSG_SENSE, "Sense:"),
        bstring
    );

    if (machine == ARCADIA)
    {   hex1(disp[0], memory[A_CONSOLE      ]);
        hex1(disp[1], memory[A_PITCH        ]);
        hex1(disp[2], memory[A_VOLUME       ]);
        printf
        (   "CONSOLE:        %s PITCH:          %s VOLUME:         %s\n",
            disp[0],           disp[1],           disp[2]
        );

        hex1(disp[0], memory[A_RESOLUTION   ]);
        hex1(disp[1], memory[A_BGCOLOUR     ]);
        hex1(disp[2], memory[A_VSCROLL      ]);
        hex1(disp[3], memory[A_CHARLINE     ]);
        printf
        (   "RESOLUTION:     %s BGCOLOUR:       %s VSCROLL:        %s CHARLINE:       %s\n",
            disp[0],           disp[1],           disp[2],           disp[3]
        );

        hex1(disp[0], memory[A_BGCOLLIDE    ]);
        hex1(disp[1], memory[A_SPRITECOLLIDE]);
        hex1(disp[2], memory[A_SPRITES01CTRL]);
        hex1(disp[3], memory[A_SPRITES23CTRL]);
        printf
        (   "BGCOLLIDE:      %s SPRITECOLLIDE:  %s SPRITES01CTRL:  %s SPRITES23CTRL:  %s\n",
             disp[0],           disp[1],           disp[2],           disp[3]
        );

            hex1(disp[0], memory[A_SPRITE0X     ]);
            hex1(disp[1], memory[A_SPRITE0Y     ]);
            hex1(disp[2], memory[A_SPRITE1X     ]);
            hex1(disp[3], memory[A_SPRITE1Y     ]);
            hex1(disp[4], memory[A_SPRITE2X     ]);
            hex1(disp[5], memory[A_SPRITE2Y     ]);
            hex1(disp[6], memory[A_SPRITE3X     ]);
            hex1(disp[7], memory[A_SPRITE3Y     ]);
            printf
            (   "SPRITE0X,Y:  %s,%s SPRITE1X,Y:  %s,%s SPRITE2X,Y:  %s,%s SPRITE3X,Y:  %s,%s\n",
                 disp[0], disp[1], disp[2], disp[3], disp[4], disp[5], disp[6], disp[7]
            );

            hex1(disp[0], memory[A_P1PADDLE     ]);
            hex1(disp[1], memory[A_P1LEFTKEYS   ]);
            hex1(disp[2], memory[A_P1MIDDLEKEYS ]);
            hex1(disp[3], memory[A_P1RIGHTKEYS  ]);
            printf
            (   "P1PADDLE:       %s P1LEFTKEYS:     %s P1MIDDLEKEYS:   %s P1RIGHTKEYS:    %s\n",
                 disp[0], disp[1], disp[2], disp[3]
            );

            hex1(disp[0], memory[A_P2PADDLE     ]);
            hex1(disp[1], memory[A_P2LEFTKEYS   ]);
            hex1(disp[2], memory[A_P2MIDDLEKEYS ]);
            hex1(disp[3], memory[A_P2RIGHTKEYS  ]);
            hex1(disp[4], memory[A_P2PALLADIUM  ]);
            printf
            (   "P2PADDLE:       %s P2LEFTKEYS:     %s P2MIDDLEKEYS:   %s P2RIGHTKEYS:    %s\n",
                 disp[0], disp[1], disp[2], disp[3]
            );

        hex1(disp[0], memory[A_P1PALLADIUM  ]);
        hex1(disp[1], memory[A_P2PALLADIUM  ]);
        printf
        (   "P1PALLADIUM:    %s P2PALLADIUM:    %s\n\n",
            disp[0],
            disp[1]
        );
    } else
    {   // assert(machine == INTERTON || machine == ELEKTOR);

        hex1(disp[0], memory[I_CONSOLE      ]);
        hex1(disp[1], memory[I_PITCH        ]);
        hex1(disp[3], memory[I_SIZES        ]);
        if (machine == ELEKTOR && memmap == MEMMAP_E)
        {   printf
            (   "CONSOLE:        %s PITCH:          %s                    SIZES:          %s\n",
                 disp[0],           disp[1],                              disp[3]
            );
        } else
        {   hex1(disp[2], memory[I_NOISE        ]);
            printf
            (   "CONSOLE:        %s PITCH:          %s NOISE:          %s SIZES:          %s\n",
                 disp[0],           disp[1],           disp[2],           disp[3]
            );
        }

        hex1(disp[0], memory[I_SCORECTRL    ]);
        hex1(disp[1], memory[I_SCORELT      ]);
        hex1(disp[2], memory[I_SCORERT      ]);
        hex1(disp[3], memory[I_BGCOLOUR     ]);
        printf
        (   "SCORECTRL:      %s SCORELT:        %s SCORERT:        %s BGCOLOUR:       %s\n",
            disp[0],           disp[1],           disp[2],           disp[3]
        );

        hex1(disp[0], memory[I_BGCOLLIDE    ]);
        hex1(disp[1], memory[I_SPRITECOLLIDE]);
        hex1(disp[2], memory[I_SPR01COLOURS ]);
        hex1(disp[3], memory[I_SPR23COLOURS ]);
        printf
        (   "BGCOLLIDE:      %s SPRITECOLLIDE:  %s SPR01COLOURS:   %s SPR23COLOURS:   %s\n",
            disp[0],           disp[1],           disp[2],           disp[3]
        );

        hex1(disp[0], memory[I_SPRITE0AX    ]);
        hex1(disp[1], memory[I_SPRITE0AY    ]);
        hex1(disp[2], memory[I_SPRITE1AX    ]);
        hex1(disp[3], memory[I_SPRITE1AY    ]);
        hex1(disp[4], memory[I_SPRITE2AX    ]);
        hex1(disp[5], memory[I_SPRITE2AY    ]);
        hex1(disp[6], memory[I_SPRITE3AX    ]);
        hex1(disp[7], memory[I_SPRITE3AY    ]);
        printf
        (   "SPRITE0AX,Y: %s,%s SPRITE1AX,Y: %s,%s SPRITE2AX,Y1 %s,%s SPRITE3AX,Y: %s,%s\n",
            disp[0], disp[1], disp[2], disp[3], disp[4], disp[5], disp[6], disp[7]
        );

        hex1(disp[0], memory[I_SPRITE0BX    ]);
        hex1(disp[1], memory[I_SPRITE0BY    ]);
        hex1(disp[2], memory[I_SPRITE1BX    ]);
        hex1(disp[3], memory[I_SPRITE1BY    ]);
        hex1(disp[4], memory[I_SPRITE2BX    ]);
        hex1(disp[5], memory[I_SPRITE2BY    ]);
        hex1(disp[6], memory[I_SPRITE3BX    ]);
        hex1(disp[7], memory[I_SPRITE3BY    ]);
        printf
        (   "SPRITE0BX,Y: %s,%s SPRITE1BX,Y: %s,%s SPRITE2BX,Y: %s,%s SPRITE3BX,Y: %s,%s\n",
            disp[0], disp[1], disp[2], disp[3], disp[4], disp[5], disp[6], disp[7]
        );

        hex1(disp[0], memory[I_P1PADDLE     ]);
        hex1(disp[1], memory[I_P1LEFTKEYS   ]);
        hex1(disp[2], memory[I_P1MIDDLEKEYS ]);
        hex1(disp[3], memory[I_P1RIGHTKEYS  ]);
        printf
        (   "P1PADDLE:       %s P1LEFTKEYS:     %s P1MIDDLEKEYS:   %s P1RIGHTKEYS:    %s\n",
            disp[0], disp[1], disp[2], disp[3]
        );

        hex1(disp[0], memory[I_P2PADDLE     ]);
        hex1(disp[1], memory[I_P2LEFTKEYS   ]);
        hex1(disp[2], memory[I_P2MIDDLEKEYS ]);
        hex1(disp[3], memory[I_P2RIGHTKEYS  ]);
        printf
        (   "P2PADDLE:       %s P2LEFTKEYS:     %s P2MIDDLEKEYS:   %s P2RIGHTKEYS:    %s\n\n",
             disp[0], disp[1], disp[2], disp[3]
        );
    }

    T2;
}

EXPORT void view_udcs(void)
{   AUTO          int    i, j, k;
    PERSIST       FLAG   first = TRUE;
    PERSIST       STRPTR colourname[8];
    PERSIST const STRPTR a_sizename[2] =
    {  "(8*16",
       " (8*8",
    }, ie_sizename[4] =
    { " (8*10",
      "(16*20",
      "(32*40",
      "(64*80"
    };

    if (first)
    {   colourname[0] = GetCatalogStr(CatalogPtr, MSG_COLOUR_WHITE,  "(white)" );
        colourname[1] = GetCatalogStr(CatalogPtr, MSG_COLOUR_YELLOW, "(yellow)");
        colourname[2] = GetCatalogStr(CatalogPtr, MSG_COLOUR_PURPLE, "(purple)");
        colourname[3] = GetCatalogStr(CatalogPtr, MSG_COLOUR_RED,    "(red)"   );
        colourname[4] = GetCatalogStr(CatalogPtr, MSG_COLOUR_CYAN,   "(cyan)"  );
        colourname[5] = GetCatalogStr(CatalogPtr, MSG_COLOUR_GREEN,  "(green)" );
        colourname[6] = GetCatalogStr(CatalogPtr, MSG_COLOUR_BLUE,   "(blue)"  );
        colourname[7] = GetCatalogStr(CatalogPtr, MSG_COLOUR_BLACK,  "(black)" );
    }

    T1;
    if (machine == ARCADIA)
    {   for (i = 0; i < 4; i++)
        {   for (j = 0; j < 8; j++)
            {   for (k = 0; k < 8; k++)
                {   if (memory[A_OFFSET_SPRITES + (i * 8) + j] & (0x80 >> k))
                    {   bits[i][j][k] = '#';
                    } else
                    {   bits[i][j][k] = '.';
                }   }
                bits[i][j][8] = 0;
        }   }

        printf
        (   "SPRITE0: %s SPRITE1: %s SPRITE2: %s SPRITE3: %s\n",
            bits[0][0], bits[1][0], bits[2][0], bits[3][0]
        );
        if (flagline && (psu & PSU_F))
        {   printf
            (   "%8s %s %8s %s %8s %s %8s %s",
                colourname[from_a[7 - ((memory[A_SPRITES01CTRL] & 0x38) >> 3)]],
                bits[0][1],
                colourname[from_a[7 - ( memory[A_SPRITES01CTRL] & 0x07      )]],
                bits[1][1],
                colourname[from_a[7 - ((memory[A_SPRITES23CTRL] & 0x38) >> 3)]],
                bits[2][1],
                colourname[from_a[7 - ( memory[A_SPRITES23CTRL] & 0x07      )]],
                bits[3][1]
            );
        } else
        {   printf
            (   "%8s %s %8s %s %8s %s %8s %s",
                colourname[from_a[(memory[A_SPRITES01CTRL] & 0x38) >> 3]],
                bits[0][1],
                colourname[from_a[ memory[A_SPRITES01CTRL] & 0x07      ]],
                bits[1][1],
                colourname[from_a[(memory[A_SPRITES23CTRL] & 0x38) >> 3]],
                bits[2][1],
                colourname[from_a[memory[A_SPRITES23CTRL] & 0x07       ]],
                bits[3][1]
            );
        }

        printf
        (   "\n  %s) %s   %s) %s   %s) %s   %s) %s\n",
            a_sizename[(memory[A_SPRITES01CTRL] & 0x80) >> 7],
            bits[0][2],
            a_sizename[(memory[A_SPRITES01CTRL] & 0x40) >> 6],
            bits[1][2],
            a_sizename[(memory[A_SPRITES23CTRL] & 0x80) >> 7],
            bits[2][2],
            a_sizename[(memory[A_SPRITES23CTRL] & 0x40) >> 6],
            bits[3][2]
        );

        for (i = 3; i < 8; i++)
        {   printf
            (   "         %s          %s          %s          %s\n",
                bits[0][i], bits[1][i], bits[2][i], bits[3][i]
            );
        }

        for (i = 4; i < 8; i++)
        {   for (j = 0; j < 8; j++)
            {   for (k = 0; k < 8; k++)
                {   if (memory[A_OFFSET_SPRITES + (i * 8) + j] & (0x80 >> k))
                    {   bits[i - 4][j][k] = (TEXT) '#';
                    } else
                    {   bits[i - 4][j][k] = (TEXT) '.';
                }   }
                bits[i - 4][j][8] = 0;
        }   }

        printf
        (   "\n   UDC0: %s    UDC1: %s    UDC2: %s    UDC3: %s\n",
            bits[0][0], bits[1][0], bits[2][0], bits[3][0]
        );
        if (memory[A_BGCOLOUR] & 0x80)
        {   printf
            (   "   (8*8) %s    (8*8) %s    (8*8) %s    (8*8) %s\n",
                bits[0][1],
                bits[1][1],
                bits[2][1],
                bits[3][1]
            );
        } else
        {   printf
            (   "  (8*16) %s   (8*16) %s   (8*16) %s   (8*16) %s\n",
                bits[0][1],
                bits[1][1],
                bits[2][1],
                bits[3][1]
            );
        }

        for (i = 2; i < 8; i++)
        {   printf
            (   "         %s          %s          %s          %s\n",
                bits[0][i], bits[1][i], bits[2][i], bits[3][i]
            );
    }   }
    else
    {   // assert(machine == INTERTON || machine == ELEKTOR);

        for (i = 0; i < 4; i++)
        {   for (j = 0; j < 10; j++)
            {   for (k = 0; k < 8; k++)
                {   if (memory[ie_spritedata[i] + j] & (0x80 >> k))
                    {   bits[i][j][k] = '#';
                    } else
                    {   bits[i][j][k] = '.';
                }   }
                bits[i][j][8] = 0;
        }   }

        printf
        (   "SPRITE0: %s SPRITE1: %s SPRITE2: %s SPRITE3: %s\n",
            bits[0][0], bits[1][0], bits[2][0], bits[3][0]
        );
        printf
        (   "%8s %s %8s %s %8s %s %8s %s\n",
            colourname[(memory[I_SPR01COLOURS] >> 3) & 7],
            bits[0][1],
            colourname[ memory[I_SPR01COLOURS]       & 7],
            bits[1][1],
            colourname[(memory[I_SPR23COLOURS] >> 3) & 7],
            bits[2][1],
            colourname[ memory[I_SPR23COLOURS]       & 7],
            bits[3][1]
        );
        printf
        (   " %s) %s  %s) %s  %s) %s  %s) %s\n",
            ie_sizename[ memory[I_SIZES]       & 3],
            bits[0][2],
            ie_sizename[(memory[I_SIZES] >> 2) & 3],
            bits[1][2],
            ie_sizename[(memory[I_SIZES] >> 4) & 3],
            bits[2][2],
            ie_sizename[(memory[I_SIZES] >> 6) & 3],
            bits[3][2]
        );
        for (i = 3; i < 10; i++)
        {   printf
            (   "         %s          %s          %s          %s\n",
                bits[0][i], bits[1][i], bits[2][i], bits[3][i]
            );
    }   }
    printf("\n");
    T2;
}

EXPORT FLAG parse_bytes(FLAG loading, FLAG silent)
{   AUTO    UBYTE kind;
    AUTO    UWORD loadaddress;
    AUTO    int   i,
                  version;
    PERSIST FLAG  found = FALSE;

    // assert(!crippled);
    // assert(!inframe);

    oldmemmap  = memmap;
    oldmachine = machine;

/* EAS/EAR V6 files are:
        0:  magic byte    (   1 byte ) ($10 for EAS, $11 for EAR)
   1.. 37:  CPU state     (  37 bytes)
  38..630:  UVI/RAM dump  ( 593 bytes)
      631+: ROM dump
    after+: recording     (EAR only  )
Supported EAS version is V6 (V4.62+).
Supported EAR version is V6 (V4.62+).

INS/INR V6 files are:
        0:  magic byte    (   1 byte ) ($B6 for INS, $B7 for INR)
   1.. 37:  CPU state     (  37 bytes)
 38..3266:  PVI/RAM dump  (3229 bytes)
     3267+: ROM dump
    after+: recording     (INR only  )
Supported INS version is V6 (V4.62+).
Supported INR version is V6 (V4.62+).

TVS/TVR V6 files are:
        0:  magic byte    (   1 byte ) ($90 for TVS, $91 for TVR)
   1.. 37:  CPU state     (  37 bytes)
  38.. 39:  whichgame     (   2 bytes)
 40..6183:  PVI/RAM dump  (6144 bytes)
     6184+: recording     (TVR only  )
Supported TVS version is V6 (V4.62+).
Supported TVR version is V6 (V4.62+).

TVC files are:
        0:  magic byte    (always $02)
     1..2:  load address  (where to load program to)
     3..4:  start address (where to point IAR to)
        5+: the program itself

    magic byte */
    switch(IOBuffer[0])
    {
    case 0x00:
#ifdef VERBOSE
        T1; printf("Loading %s as PGM.\n", fn_game); T2;
#endif
        kind = KIND_PGM;
        machine = ELEKTOR;
        startaddr_h = ReadByteAt(1);
        startaddr_l = ReadByteAt(2);
        romsize = filesize;
        offset = 0; // for autosense()
    acase 0x02:
#ifdef VERBOSE
        T1; printf("Loading %s as TVC.\n", fn_game); T2;
#endif
        kind = KIND_TVC;
        machine = ELEKTOR;
        startaddr_h = ReadByteAt(3);
        startaddr_l = ReadByteAt(4);
        romsize = filesize - 5 + (startaddr_h * 256) + startaddr_l;
        offset = 0; // for autosense()
    acase 0x10:
#ifdef VERBOSE
        T1; printf("Loading %s as EAS.\n", fn_game); T2;
#endif
        offset = 31;
        version = ReadByte();
        if (version != 6)
        {   say("Unsupported EAS version!", FALSE);
            return FALSE;
        }
        kind = KIND_EAS;
        machine = ARCADIA;
        romsize = (int) ReadWordAt(1);
        offset = 631; // for autosense()
    acase 0x11:
#ifdef VERBOSE
        T1; printf("Loading %s as EAR.\n", fn_game); T2;
#endif
        version = ReadByteAt(31);
        if (version != 6)
        {   say("Unsupported EAR version!", FALSE);
            return FALSE;
        }
        kind = KIND_EAR;
        machine = ARCADIA;
        romsize = (int) ReadWordAt(1);
        offset = 631; // for autosense()
    acase 0x90:
#ifdef VERBOSE
        T1; printf("Loading %s as TVS.\n", fn_game); T2;
#endif
        version = ReadByteAt(31);
        if (version != 6)
        {   say("Unsupported TVS version!", FALSE);
            return FALSE;
        }
        kind = KIND_TVS;
        machine = ELEKTOR;
        romsize = (int) ReadWordAt(1);
        startaddr_h = ReadByteAt(40 + 0xBE);
        startaddr_l = ReadByteAt(40 + 0xBF);
        offset = 38; // for autosense()
    acase 0x91:
#ifdef VERBOSE
        T1; printf("Loading %s as TVR.\n", fn_game); T2;
#endif
        version = ReadByteAt(31);
        if (version != 6)
        {   say("Unsupported TVR version!", FALSE);
            return FALSE;
        }
        kind = KIND_TVR;
        machine = ELEKTOR;
        romsize = (int) ReadWordAt(1);
        startaddr_h = ReadByteAt(40 + 0xBE);
        startaddr_l = ReadByteAt(40 + 0xBF);
        offset = 38; // for autosense()
    acase 0xB6:
#ifdef VERBOSE
        T1; printf("Loading %s as INS.\n", fn_game); T2;
#endif
        version = ReadByteAt(31);
        if (version != 6)
        {   say("Unsupported INS version!", FALSE);
            return FALSE;
        }
        kind = KIND_INS;
        machine = INTERTON;
        romsize = (int) ReadWordAt(1);
        offset = 3267; // for autosense()
    acase 0xB7:
#ifdef VERBOSE
        T1; printf("Loading %s as INR.\n", fn_game); T2;
#endif
        version = ReadByteAt(31);
        if (version != 6)
        {   say("Unsupported INR version!", FALSE);
            return FALSE;
        }
        kind = KIND_INR;
        machine = INTERTON;
        romsize = (int) ReadWordAt(1);
        offset = 3267; // for autosense()
    adefault:
#ifdef VERBOSE
        T1; printf("Loading %s as BIN.\n", fn_game); T2;
#endif
        kind = KIND_BIN;
        if (machine == ELEKTOR) // shouldn't normally happen
        {   machine = INTERTON; // hopefully it's an Interton ROM, not an Arcadia ROM :-|
        }
        romsize = filesize;
        offset = 0; // for autosense()
    break;
    }

    if (loading)
    {   unautosense();

        if (sensegame != SENSEGAME_NO)
        {   // assert(sensegame == SENSEGAME_VERBOSE || sensegame == SENSEGAME_QUIET);
            if
            (   kind == KIND_TVS
             || kind == KIND_TVR
            )
            {   whichgame = (int) ReadWordAt(38);
                if (whichgame >= 0 && whichgame < KNOWNGAMES)
                {   found = TRUE;
                    configure(silent);
                } else
                {   // found = FALSE;
#ifdef VERBOSE
                    T1; printf("Unknown game!\n"); T2;
#endif
                    whichgame = -1;
            }   }
            elif (kind == KIND_TVC)
            {   found = autosense(filesize, silent);
            } else
            {   /* assert
                (   kind == KIND_BIN
                 || kind == KIND_EAS
                 || kind == KIND_EAR
                 || kind == KIND_INS
                 || kind == KIND_INR
                 || kind == KIND_PGM
                ); */
                found = autosense(romsize, silent);
        }   }

        if (machine == INTERTON && !found)
        {   if (romsize <= 2048)
            {   memmap = MEMMAP_A;
            } elif (romsize <= 4096)
            {   memmap = MEMMAP_C; // could be B or C, C is generally safer
            } else
            {   memmap = MEMMAP_D;
        }   }

        if (machine == ARCADIA)
        {   arcadia(FALSE);
        } elif (machine == INTERTON)
        {   interton(FALSE);
        } else
        {   // assert(machine == ELEKTOR);
            elektor(FALSE);
    }   }

    if
    (   kind == KIND_EAS
     || kind == KIND_EAR
     || kind == KIND_INS
     || kind == KIND_INR
     || kind == KIND_TVS
     || kind == KIND_TVR
    )
    {   // CPU state (37 bytes)
        offset = 1;
        DISCARD   ReadWord();            //  1.. 2 (romsize)
        iar     = (UWORD) ReadWord();    //  3.. 4
        page    = (UWORD) ReadByte();   //      5
        psu     = ReadByte();           //      6
        psl     = ReadByte();           //      7
        for (i = 0; i < 8; i++)
        {   ras[i] = (UWORD) ReadWord(); //  8..23
        }
        for (i = 0; i < 7; i++)
        {   r[i] = ReadByte();          // 24..30
        }
        DISCARD ReadByte();             //     31
        overcalc  = ReadByte();         //     32
        interrupt = (FLAG)  ReadByte(); //     33
        ax[0]     = (SLONG) ReadByte(); //     34
        ay[0]     = (SLONG) ReadByte(); //     35
        ax[1]     = (SLONG) ReadByte(); //     36
        ay[1]     = (SLONG) ReadByte(); //     37

        // XVI/RAM areas
        if (machine == ARCADIA)
        {   for (i = 0x1800; i <= 0x1908; i++) /* 265 bytes */ memory[i] = ReadByte();
            for (i = 0x1980; i <= 0x19BF; i++) /*  64 bytes */ memory[i] = ReadByte();
            for (i = 0x19F8; i <= 0x1AFF; i++) /* 264 bytes */ memory[i] = ReadByte();
                                             /* = 593 bytes */
        } elif (machine == INTERTON)
        {   for (i = 0x1000; i <= 0x15FF; i++)                 memory[i] = ReadByte();
            for (i = 0x1800; i <= 0x1DFF; i++)                 memory[i] = ReadByte();
            for (i = 0x1E88; i <= 0x1E8E; i++) /*   7 bytes */ memory[i] = ReadByte();
            for (i = 0x1F00; i <= 0x1F2D; i++) /*  46 bytes */ memory[i] = ReadByte();
            for (i = 0x1F40; i <= 0x1F6D; i++) /*  46 bytes */ memory[i] = ReadByte();
            for (i = 0x1F80; i <= 0x1FAD; i++) /*  46 bytes */ memory[i] = ReadByte();
            for (i = 0x1FC0; i <= 0x1FC3; i++) /*   4 bytes */ memory[i] = ReadByte();
            for (i = 0x1FC6; i <= 0x1FCD; i++) /*   8 bytes */ memory[i] = ReadByte();
        } else
        {   // assert(machine == ELEKTOR);

            DISCARD ReadWord(); // whichgame (38..39)

            for (i =  0x800; i <= 0x1FFF; i++) /*   6K      */ memory[i] = ReadByte();
    }   }

    if
    (   kind == KIND_BIN
     || kind == KIND_EAS
     || kind == KIND_EAR
     || kind == KIND_INS
     || kind == KIND_INR
    )
    {   if (machine == ARCADIA)
        {   a_readrom();
        } else
        {   // assert(machine == INTERTON);
            i_readrom();
    }   }
    elif (kind == KIND_PGM)
    {   offset = 0x8C0;
        for (i = 0x8C0; i < romsize; i++)
        {   if (!(memflags[i] & NOWRITE))
            {   memory[i] = ReadByte();
            } else
            {   DISCARD ReadByte();
    }   }   }
    elif (kind == KIND_TVC)
    {   loadaddress = ReadWordAt(1);
        offset = 5;
        for (i = 0; i < filesize - 5; i++)
        {   if (!(memflags[loadaddress + i] & NOWRITE))
            {   memory[loadaddress + i] = ReadByte();
            } else
            {   DISCARD ReadByte();
    }   }   }

    if
    (   kind == KIND_BIN
     || kind == KIND_PGM
     || kind == KIND_TVC
    )
    {   engine_reset(FALSE);
    }
    else
    {
#ifdef PROFILE
        for (i = 0; i <= 255; i++)
        {   profile[i] = 0;
        }
#endif
#ifdef COVERAGE
        for (i = 0; i < 32 * KILOBYTE; i++)
        {   coverage[i] = 0;
        }
#endif
        ;
    }

    if (kind == KIND_PGM || kind == KIND_TVC)
    {   // memory[$1FC0..$1FC9] = memory[$177..$180]
        // ie.:
        //  SIZES        = $AA
        //  SPR01COLOURS = $09
        //  SPR23COLOURS = $09
        //  SCORECTRL    = $00
        //  $1FC4        = $00
        //  $1FC5        = $00
        //  BGCOLOUR     = $19
        //  PITCH        = $00
        //  SCORELT      = $AA
        //  SCORERT      = $AA
        for (i = 0; i < 10; i++)
        {   memory[0x1FC0 + i] = memory[0x177 + i];
        }

        // make $8B9..$8BA = $0903 (important!)
        memory[0x8B9] = 0x09;
        memory[0x8BA] = 0x03;
        memory[0x8BD] = 0x1F; // bcta,un
        memory[0x8BE] = startaddr_h;
        memory[0x8BF] = startaddr_l;
        iar  = 0x8BD;
        page = (UWORD) iar & PAGE;
        iar  = (UWORD) iar & ~(PAGE);
#ifdef VERBOSE
        T1; printf("Setting IAR to $%lX.\n", page | iar); T2;
#endif
    }

    if
    (   kind == KIND_BIN
     || kind == KIND_PGM
     || kind == KIND_TVC
     || kind == KIND_EAS
     || kind == KIND_INS
     || kind == KIND_TVS
    )
    {   free_iobuffer();
        recmode   = RECMODE_NORMAL;
    } else
    {   /* assert
        (   kind == KIND_EAR
         || kind == KIND_INR
         || kind == KIND_TVR
        ); */
        recmode   = RECMODE_PLAY;
        recsize   = filesize;
    } // else assert(0);

    game      = TRUE;
    interrupt = FALSE;
    frames    =
    oldframes =
    cycles    =
    elapsed   = 0; // unsigned

#ifdef MODE_MHZ
    oldelapsed = 0;
#endif

    updatemenus();
    updatesmlgads();

    return TRUE;
}

EXPORT void docommand(int which)
{   switch(which)
    {
    case MENUITEM_ABSOLUTE:
        updatemenu(MENUITEM_ABSOLUTE);
#ifdef AMIGA
        DISCARD CloseAJoystick(); // harmless if already closed
        if (absolute)
        {   if (joy1)
            {   DISCARD OpenAJoystick(AJOYUNIT0 | AJOYUNIT1);
            } else
            {   DISCARD OpenAJoystick(AJOYUNIT1);
        }   }
#endif
    acase MENUITEM_CLIENT:
        if (connected == NET_OFF)
        {   NetPlay(NET_CLIENT);
            if (connected != NET_OFF)
            {   project_reset(TRUE); /* reset machine when netplay connected */
        }   }
        else
        {   NetPlay(NET_OFF);
        }
        updatemenu(MENUITEM_CLIENT);
    acase MENUITEM_DEMULTIPLEX:
        updatesmlgad(GADPOS_DEMULTIPLEX, demultiplex, TRUE);
        updatemenu(MENUITEM_DEMULTIPLEX);
        if (paused)
        {   redrawscreen();
        }
#ifdef AMIGA
    acase MENUITEM_JOY1:
        if (absolute)
        {   DISCARD CloseAJoystick();
            if (joy1)
            {   DISCARD OpenAJoystick(AJOYUNIT0 | AJOYUNIT1);
            } else
            {   DISCARD OpenAJoystick(AJOYUNIT1);
        }   }
        updatemenu(MENUITEM_JOY1);
#endif
    acase MENUITEM_POINTER:
        updatemenu(MENUITEM_POINTER);
        if (showpointer)
#ifdef AMIGA
            ClearPointer(MainWindowPtr);
#endif
#ifdef WIN32
            ShowCursor(TRUE);
#endif
        else
#ifdef AMIGA
            SetPointer(MainWindowPtr, PointerData, 1, 1, 0, 0);
#endif
#ifdef WIN32
            ShowCursor(FALSE);
#endif
    acase MENUITEM_RESETTOMONITOR:
        // assert(machine == ELEKTOR);
        // assert(!crippled);
        iar = 0x23; // gives blue  background with yellow "IIII" at bottom
        NetPlay(NET_OFF);
        updatemenu(MENUITEM_SERVER);
        updatemenu(MENUITEM_CLIENT);
    acase MENUITEM_REVERT:
        if (crippled || !game) return;
        macro_stop();
        DISCARD engine_load(FALSE); // arguably TRUE might be better
    acase MENUITEM_QUICKLOAD:
        if (crippled) return;
        macro_stop();
        changefilepart(quickname[machine]);
        if (!(engine_load(TRUE)))
        {   beep();
        }
    acase MENUITEM_QUICKSAVE:
        if (crippled || !game) return;
        changefilepart(quickname[machine]);
        if (machine == ARCADIA)
        {   engine_save(KIND_EAS, TRUE);
        } elif (machine == INTERTON)
        {   engine_save(KIND_INS, TRUE);
        } else
        {   // assert(machine == ELEKTOR);
            engine_save(KIND_TVS, TRUE);
        }
    acase MENUITEM_SAVEROM:
        if (machine == ELEKTOR)
        {   project_save(KIND_PGM);
        } else
        {   // assert(machine == ARCADIA || machine == INTERTON);
            project_save(KIND_BIN);
        }
#ifdef WIN32
    acase MENUITEM_MENUBAR:
        // no point doing updatemenu(MENUITEM_MENUBAR),
        // as we're about to close the window
        closewindow();
        if (!fullscreen)
        {   winleftx -= 3;
        }
        openwindow();
        redrawscreen();
    acase MENUITEM_STRETCH:
        updatemenu(MENUITEM_STRETCH);
        if (fullscreen)
        {   calcsize(FALSE);
            clearscreen();
            redrawscreen();
        }
#endif
    acase MENUITEM_SERVER:
        if (connected == NET_OFF)
        {   NetPlay(NET_SERVER);
            if (connected != NET_OFF)
            {   project_reset(TRUE); /* reset machine when netplay connected */
        }   }
        else
        {   NetPlay(NET_OFF);
        }
        updatemenu(MENUITEM_SERVER);
    acase MENUITEM_SOUND:
        if (sound)
        {   sound_on();
        } else
        {   sound_off(TRUE);
            // We have to force it, because sound variable is already FALSE
        }
        updatesmlgad(GADPOS_SOUND, sound, TRUE);
        updatemenu(MENUITEM_SOUND);
    acase MENUITEM_STEP:
        step = TRUE;
        if (!trace)
        {   T1;
            printf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_STEPPING,
                    "Executing %s instruction at $%X.\n\n"
                ),
                opcodename[memory[iar]],
                (unsigned int) (page + iar)
            );
            T2;
        }
        unpause();
        // updatemenu(MENUITEM_STEP); not required here
    acase MENUITEM_TITLEBAR:
        updatemenu(MENUITEM_TITLEBAR);
#ifdef AMIGA
        ShowTitle(ScreenPtr, showtitlebar);
#endif
#ifdef WIN32
        closewindow();
        if (showtitlebar && !fullscreen)
        {   winleftx -= 4;
        }
        openwindow();
        redrawscreen();
    acase MENUITEM_TOOLBAR:
        updatemenu(MENUITEM_TOOLBAR);
        closewindow();
        if (showtoolbar && !fullscreen)
        {   winleftx -= 6;
        }
        openwindow();
        redrawscreen();
#endif
    acase COMMAND_ARCADIA:
        if (machine != ARCADIA)
        {   strcpy(fn_game, pathpart);
            arcadia(TRUE);
            updatemenus();
            updatebiggads();
            updatesmlgads();
            settitle();
            redrawscreen();
        }
    acase COMMAND_INTERTON:
        if (machine != INTERTON)
        {   strcpy(fn_game, pathpart);
            interton(TRUE);
            updatemenus();
            updatebiggads();
            updatesmlgads();
            settitle();
            redrawscreen();
        }
    acase COMMAND_ELEKTOR:
        if (machine != ELEKTOR)
        {   strcpy(fn_game, pathpart);
            elektor(TRUE);
            updatemenus();
            updatebiggads();
            updatesmlgads();
            settitle();
            redrawscreen();
        }
    acase MENUITEM_CLI:
        T1;
        if (cli)
        {   printf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_CLIENABLED,
                    "Debugger CLI enabled (\"H\" for help).\n\n"
            )   );
        } else
        {   printf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_CLIDISABLED,
                    "Debugger CLI disabled.\n\n"
            )   );
        }
        T2;
        updatemenu(MENUITEM_CLI);
        if (cli)
        {   pause();
        }
    acase COMMAND_IGNORE:
        T1;
        printf
        (   GetCatalogStr(CatalogPtr,
                MSG_ENGINE_IGNORING,
                "Ignoring %s instruction at $%X.\n\n"
            ),
            opcodename[memory[page + iar]],
            (unsigned int) (page + iar)
        );
        T2;
        iar = (iar + table_sz[memory[iar]]) & PMSK;
    adefault:
        // assert(0);
    break;
}   }

EXPORT void fixextension(STRPTR extension, STRPTR passedfn, FLAG force)
{   FLAG   found = FALSE;
    int    i;
    size_t length;

    // This routine should really only be used on the file component of the pathname.

    // force means we will replace an existing extension
    //  (typically used before showing an ASL save requester)
    // otherwise, we just add an extension if there is no recognised extension
    //  (typically used after showing an ASL load requester)

    length = strlen(passedfn);
    if (force)
    {   if (length >= 1)
        {   // find extension
            for (i = length - 1; i >= 0; i--)
            {   if (passedfn[i] == '.')
                {   found = TRUE;
                    break;
            }   }
            if (found)
            {   passedfn[i] = 0;
            }
            strcat(passedfn, extension);
    }   }
    else
    {   for (i = 0; i < KINDS; i++)
        {   if
            (   length > strlen(filekind[i].extension)
             && !stricmp(&passedfn[length - strlen(filekind[i].extension)], filekind[i].extension)
            )
            {   found = TRUE;
                break; // for speed
        }   }
        if (!found)
        {   // if there is no recognized extension, add an appropriate extension
            strcat(passedfn, extension); // add extension
}   }   }

// this routine reads but doesn't write offset
MODULE FLAG autosense(int thesize, FLAG silent)
{   AUTO    ULONG crc32;
    AUTO    int   i;

    golf = FALSE;
    crc32 = getcrc32(IOBuffer + offset, thesize);
#ifdef VERBOSE
    T1; printf("ROM size: %ld. CRC32: $%lX.\n", romsize, crc32); T2;
#endif
    for (i = 0; i < KNOWNGAMES; i++)
    {   if (thesize == known[i].size && crc32 == known[i].crc32)
        {   whichgame = i;
            configure(silent);
            return TRUE; // for speed
    }   }

#ifdef VERBOSE
    T1; printf("Unknown game.\n"); T2;
#endif

    return FALSE;
}

MODULE void configure(FLAG silent)
{   machine        = (int)   known[whichgame].machine;
    autofire       = (ULONG) known[whichgame].autofire;
    analog         = (ULONG) known[whichgame].analog;
    sensitivity    = (int)   known[whichgame].sensitivity;
    demultiplex    = (ULONG) known[whichgame].demultiplex;
    swapped        = (ULONG) known[whichgame].swapped;
    flagline       = (ULONG) known[whichgame].flagline;
    memmap         = (int)   known[whichgame].memmap;
    downframes     = (ULONG) known[whichgame].downframes;
    totalframes    = (ULONG) known[whichgame].totalframes;
    // upframes    = totalframes - downframes;
    key1           = (int)   known[whichgame].key1;
    key2           = (int)   known[whichgame].key2;
    key3           = (int)   known[whichgame].key3;
    key4           = (int)   known[whichgame].key4;
    p1sprite       = (int)   known[whichgame].p1sprite;
    p2sprite       = (int)   known[whichgame].p2sprite;
    if (usehacks && machine != ARCADIA)
    {   // assert(machine == INTERTON || machine == ELEKTOR);
        fudge      = (int)   known[whichgame].fudge;
        cpl        = (int)   known[whichgame].cpl;
        forceinterrupt =     known[whichgame].forceinterrupt;
    }

#ifdef VERBOSE
    T1; printf("Autosensed game #%ld.\n", whichgame); T2;
#endif

    if (whichgame == GOLFPOS)
    {   // assert(machine == ARCADIA);
        // assert(romsize == 6144);
        golf = TRUE;
    } else
    {   golf = FALSE;
    }

    if (silent)
    {   return;
    }

    if (sensegame != SENSEGAME_VERBOSE)
    {   return;
    }

    strcpy(autotext[0], known[whichgame].name);

    if (machine == ARCADIA)
    {   strcpy(autotext[1], "Emerson Arcadia 2001");
    } elif (machine == INTERTON)
    {   strcpy(autotext[1], "Interton VC 4000, type \"");
        if (memmap == MEMMAP_A)
        {   strcat(autotext[1], "A\" (2K ROM + 0");
        } elif (memmap == MEMMAP_B)
        {   strcat(autotext[1], "B\" (4K ROM + 0");
        } elif (memmap == MEMMAP_C)
        {   strcat(autotext[1], "C\" (4K ROM + 1");
        } else
        {   // assert(memmap == MEMMAP_D);
            strcat(autotext[1], "D\" (6K ROM + 1");
        }
        strcat(autotext[1], "K RAM)");
    } else
    {   // assert(machine == ELEKTOR);
        strcpy(autotext[1], "Elektor TV Games Computer (");
        if (memmap == MEMMAP_E)
        {   strcat(autotext[1], "basic)");
        } else
        {   // assert(memmap == MEMMAP_F);
            strcat(autotext[1], "expanded)");
    }   }

        if (known[whichgame].size % KILOBYTE)
        {   sprintf
            (   autotext[2],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_BYTES,
                    "%ld bytes"
                ),
                (long int) known[whichgame].size
            );
        } else
        {   sprintf
            (   autotext[2],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_KILOBYTES,
                    "%ldK"
                ),
                (long int) (known[whichgame].size / KILOBYTE)
            );
        }

        if (machine == ELEKTOR)
        {   sprintf
            (   autotext[3],
                "$%lX",
                (long unsigned int) ((startaddr_h * 256) + startaddr_l)
            );
        } else
        {   // assert(machine == ARCADIA || machine == INTERTON);
            strcpy(autotext[3], "$0");
        }

        if (autofire)
        {   strcpy
            (   autotext[4],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_YES,
                    "Yes"
            )   );
        } else
        {   strcpy
            (   autotext[4],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_NO,
                    "No"
            )   );
        }
        sprintf
        (   &autotext[4][strlen(autotext[4])],
            " (%ld ",
            (long int) downframes
        );

        if (downframes == 1)
        {   strcat
            (   autotext[4],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_FRAME,
                    "frame"
            )   );
        } else
        {   strcat
            (   autotext[4],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_FRAMES,
                    "frames"
            )   );
        }
        strcat
        (   autotext[4],
            GetCatalogStr(CatalogPtr,
                MSG_ENGINE_DOWN,
                "down"
        )   );
        sprintf
        (   &autotext[4][strlen(autotext[4])],
            ", %ld ",
            (long int) (totalframes - downframes)
        );
        if (totalframes - downframes == 1)
        {   strcat
            (   autotext[4],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_FRAME,
                    "frame"
            )   );
        } else
        {   strcat
            (   autotext[4],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_FRAMES,
                    "frames"
            )   );
        }
        strcat(autotext[4], " ");
        strcat
        (   autotext[4],
            GetCatalogStr(CatalogPtr,
                MSG_ENGINE_UP,
                "up"
        )   );
        strcat(autotext[4], ")");

        if (analog)
        {   strcpy
            (   autotext[5],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_ANALOG,
                    "Analog"
            )   );
        } else
        {   strcpy
            (   autotext[5],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_DIGITAL,
                    "Digital"
            )   );
        }
        strcat(autotext[5], ", ");
        if (swapped)
        {   strcat
            (   autotext[5],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_SWAPPED,
                    "swapped"
            )   );
        } else
        {   strcat
            (   autotext[5],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_UNSWAPPED,
                    "unswapped"
            )   );
        }
        strcat(autotext[5], " (");
        switch(sensitivity)
        {
        case SENSITIVITY_VLOW:
            strcat
            (   autotext[5],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_VERYLOW,
                    "very low"
            )   );
        acase SENSITIVITY_LOW:
            strcat
            (   autotext[5],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_LOW,
                    "low"
            )   );
        acase SENSITIVITY_MEDIUM:
            strcat
            (   autotext[5],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_MEDIUM,
                    "medium"
            )   );
        acase SENSITIVITY_HIGH:
            strcat
            (   autotext[5],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_HIGH,
                    "high"
            )   );
        adefault:
            // assert(sensitivity == 5);
            strcat
            (   autotext[5],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_VERYHIGH,
                    "very high"
            )   );
        }
        strcat(autotext[5], " ");
        strcat
        (   autotext[5],
            GetCatalogStr(CatalogPtr,
                MSG_ENGINE_SENSITIVITY,
                "sensitivity"
        )   );
        strcat(autotext[5], ")");

        if (machine != ARCADIA)
        {   strcpy
            (   autotext[6],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_NOTAPPLICABLE,
                    "n/a"
            )   );
        } elif (demultiplex)
        {   strcpy
            (   autotext[6],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_YES,
                    "Yes"
            )   );
        } else
        {   strcpy
            (   autotext[6],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_NO,
                    "No"
            )   );
        }

        if (flagline)
        {   strcpy
            (   autotext[7],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_YES,
                    "Yes"
            )   );
        } else
        {   strcpy
            (   autotext[7],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_NO,
                    "No"
            )   );
        }

        if (machine == ELEKTOR)
        {   sprintf
            (   autotext[8],
                "%s, %s, %s, %s",
                keytable[key1].e_name,
                keytable[key2].e_name,
                keytable[key3].e_name,
                keytable[key4].e_name
            );
        } else
        {   // assert(machine == ARCADIA || machine == INTERTON);
            sprintf
            (   autotext[8],
                "%s, %s, %s, %s",
                keytable[key1].ai_name,
                keytable[key2].ai_name,
                keytable[key3].ai_name,
                keytable[key4].ai_name
            );
        }

        if (machine != ARCADIA && usehacks)
        {   // assert(machine == INTERTON || machine == ELEKTOR);
            sprintf
            (   autotext[9],
                "%ld",
                (long int) fudge
            );
            if (fudge == FUDGE)
            {   strcat(autotext[9], " ");
                strcat
                (   autotext[9],
                    GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_DEFAULT,
                        "(default)"
                )   );
            }

            sprintf
            (   autotext[10],
                "%ld",
                (long int) cpl
            );
            if (cpl == PVI_CPL)
            {   strcat(autotext[10], " ");
                strcat
                (   autotext[10],
                    GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_DEFAULT,
                        "(default)"
                )   );
            }

            if (forceinterrupt)
            {   strcpy
                (   autotext[11],
                    GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_YES,
                        "Yes"
                )   );
            } else
            {   strcpy
                (   autotext[11],
                    GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_NO,
                        "No"
                )   );
                strcat(autotext[11], " ");
                strcat
                (   autotext[11],
                    GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_DEFAULT,
                        "(default)"
                )   );
        }   }
        else
        {   strcpy
            (   autotext[9],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_NOTAPPLICABLE,
                    "n/a"
            )   );
            strcpy
            (   autotext[10],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_NOTAPPLICABLE,
                    "n/a"
            )   );
            strcpy
            (   autotext[11],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_NOTAPPLICABLE,
                    "n/a"
            )   );
        }

        infowindow();
}

MODULE ULONG getcrc32(UBYTE* address, int thesize)
{   register ULONG crc;
             int   i;

    crc = 0xFFFFFFFF;
    for (i = 0; i < thesize; i++)
    {   crc = ((crc >> 8) & 0x00FFFFFF) ^ crcTable[(crc ^ *address++) & 0xFF];
    }
    return(crc ^ 0xFFFFFFFF);
}

EXPORT void generate_crctable(void)
{   ULONG crc;
    int   i, j;

    for (i = 0; i < 256; i++)
    {   crc = (ULONG) i;
        for (j = 8; j > 0; j--)
        {   if (crc & 1)
            {   crc = (crc >> 1) ^ 0xEDB88320L;
            } else
            {   crc >>= 1;
        }   }
        crcTable[i] = crc;
}   }

EXPORT void runto(void)
{   if (machine == ARCADIA)
    {   if (rast + voffset < BOXHEIGHT)
        {   screen[0][rast + voffset] = from_a[7 - from_a[screen[0][rast + voffset]]];
    }   }
    else
    {   // assert(machine == INTERTON || machine == ELEKTOR);
        if (rast < BOXHEIGHT)
        {   screen[0][rast] = 15 - screen[0][rast];
    }   }
    // maybe better to use a/ie_drawpixel() for these.
    if (inframe && !crippled)
    {   cripple();
    }
    pause();

    updatemonitor(FALSE);
    updatescreen();

    // assert(crippled);
    while (paused)
    {   checkinput();
}   }

EXPORT void hex1(char* out, UBYTE value)
{   ULONG digit;

    digit = value / 16;
    if (digit <= 9)
    {   *(out) = (char) ('0' + digit);
    } else
    {   *(out) = (char) ('A' + digit - 10);
    }

    digit = value % 16;
    if (digit <= 9)
    {   *(out + 1) = (char) ('0' + digit);
    } else
    {   *(out + 1) = (char) ('A' + digit - 10);
    }

    *(out + 2) = 0;
}

EXPORT void hex2(char* out, UWORD value)
{   ULONG digit;

    digit = value / 4096;
    if (digit <= 9)
    {   *(out) = (char) ('0' + digit);
    } else
    {   *(out) = (char) ('A' + digit - 10);
    }

    digit = (value % 4096) / 256;
    if (digit <= 9)
    {   *(out + 1) = (char) ('0' + digit);
    } else
    {   *(out + 1) = (char) ('A' + digit - 10);
    }

    digit = (value % 256) / 16;
    if (digit <= 9)
    {   *(out + 2) = (char) ('0' + digit);
    } else
    {   *(out + 2) = (char) ('A' + digit - 10);
    }

    digit = value % 16;
    if (digit <= 9)
    {   *(out + 3) = (char) ('0' + digit);
    } else
    {   *(out + 3) = (char) ('A' + digit - 10);
    }

    *(out + 4) = 0;
}

EXPORT void decimal(char* out, unsigned long value)
{   ULONG calc,
          i,
          where   = 0;
    FLAG  started = FALSE;

    for (i = ONE_BILLION; i >= 1; i /= 10)
    {   calc = value / i;
        // assert(calc < 10);
        if (calc || started || i == 1)
        {   *(out + where) = (char) ('0' + calc);
            started = TRUE;
        } else
        {   *(out + where) = ' ';
        }
        where++;
        value %= i;
    }

    *(out + where) = 0;
}

EXPORT void engine_setup(void)
{   int i, j;

    srand((unsigned int) time(NULL));

    generate_crctable();

    timestring[2]  =
    timestring[5]  = ':';
    timestring[8]  = '.';
    timestring[11] = 0;

    lastchange[0]  = 0;

    for (i = 0; i < 2; i++)
    {   for (j = 0; j < NUMKEYS; j++)
        {   keypads[i][j] = default_keypads[i][j];
    }   }
    for (i = 0; i < 4; i++)
    {   console[i] = default_console[i];
    }

    cpu_setup();
}

#ifdef PROFILE
EXPORT void printprofile(void)
{   int i;

    /* Possible enhancements:
    print percentage of total instructions executed, for each instruction.
    print percentage of total cycles spent, for each instruction.
    take any indirection penalties into account, instead of just saying
     "at least".
    option of whether to view as "fast" or "slow" cycles (currently only
     shows as "fast" cycles). Such an option should also affect the
     clock shown in the monitor, tracing, etc. (ie. everywhere "cycles"
     and/or "elapsed" are shown to the user) for consistency. */

    T1;
    for (i = 0; i <= 255; i++)
    {
#ifdef SUPPRESSUNPROFILED
        if (profile[i] != 0)
#endif
            if (profile[i] == 1)
            {   if (i & 8) // interestingly, this algorithm gives all indirectable opcodes
                {   printf("%-7s ($%02lX):          1 time  (at least %10u cycles)\n", opcodename[i], i, table_clockperiods[i]);
                } else
                {   printf("%-7s ($%02lX):          1 time  (exactly  %10u cycles)\n", opcodename[i], i, table_clockperiods[i]);
            }   }
            else
            {   if (i & 8) // interestingly, this algorithm gives all indirectable opcodes
                {   printf("%-7s ($%02lX): %10u times (at least %10u cycles)\n", opcodename[i], i, profile[i], profile[i] * table_clockperiods[i]);
                } else
                {   printf("%-7s ($%02lX): %10u times (exactly  %10u cycles)\n", opcodename[i], i, profile[i], profile[i] * table_clockperiods[i]);
            }   }
    }
    printf("\n");
    T2;
}
#endif
#ifdef COVERAGE
EXPORT void printcoverage(void)
{   AUTO          int    i = 0,
                         regionstart;
    AUTO          UBYTE  flags;
    PERSIST const STRPTR coveragename[16] =
    {   "----",
        "---W",
        "--R-",
        "--RW",
        "-E--",
        "-E-W",
        "-ER-",
        "-ERW",
        "A---",
        "A--W",
        "A-R-",
        "A-RW",
        "AE--",
        "AE-W",
        "AER-",
        "AERW"
    };

    /* Possible enhancements:
    keep a count of how many times they were each accessed.
    does coverage work properly with mirroring, especially for the "A"
     flag? perhaps not.
    read/writes by the UVI/PVI don't set any coverage bits; perhaps they
     should?
    make some distinction between relative/absolute/zero page accesses
     (but why?).

    How to interpret the coverage report:

    "A" (address):
    This is set for the two relevant address bytes during resolution of
    indirect addresses. Non-indirect accesses do not set it. Examples:

        bcta,un *$1234
     or bctr,un *$1234
     or loda,r0 *$1234
     or lodr,r0 *$1234
     or stra,r0 *$1234

    will set it for bytes $1234 and $1235.

        bcta,gt *$1234
     or bctr,lt *$1234

    will set it for bytes $1234 and $1235 if the branch is taken.

    "E" (execute):
    This is set for the opcode byte and also for any operand bytes, for
    any instruction that is executed.

    "R" (read):
    This is set for any byte that is attempted to be read by a game
    instruction.
    Examples:

        loda,r0 $1234
     or lodr,r0 $1234

    will set it for byte $1234.

    "W" (write):
    This is set for any byte that is attempted to be written by a game
    instruction.
    Examples:

        stra,r0 $1234
     or strr,r0 $1234

    will set it for byte $1234.

    "E" + "W" = self-modifying code.

    "E" + "R" = bytes that are used as both code and data.

    "A" + "E" = usually an operand which is also accessed from elsewhere
                with a relative indirect access. Eg. this code, assuming
                that it is executed:

    $1000: loda,r0 $1234
    $1003: addi,r0 1
    $1005: strr,r0 $*1001

    will give the following coverage report:

    -E--: $1000
    AE--: $1001..$1002
    -E--: $1003..$1006
    */

    T1;
    do
    {   regionstart = i;
        // assert(coverage[i] < 16);
        flags = coverage[i];
        do
        {   i++;
        } while (coverage[i] == flags && i < 32 * KILOBYTE);
#ifdef SUPPRESSUNCOVERED
        if (flags != 0)
#endif
            if (regionstart != i - 1)
            {   printf("%s: $%lX..$%lX\n", coveragename[flags], regionstart, i - 1);
            } else
            {   printf("%s: $%lX\n"      , coveragename[flags], regionstart);
            }
    } while (i < 32 * KILOBYTE);
    printf("\n");
    T2;
}
#endif

EXPORT void savescreenshot(int kind, FLAG wantasl)
{   AUTO UBYTE* TheBody;
    AUTO FILE*  LocalHandle;
    AUTO int    i, x, y;

PERSIST UBYTE a_iff_header[2][80] =
{ { 0x46, 0x4F, 0x52, 0x4D, 0x00, 0x00, 0x4B, 0xF0,   // "FORMxxxx"
    0x49, 0x4C, 0x42, 0x4D, 0x42, 0x4D, 0x48, 0x44,   // "ILBMBMHD"
    0x00, 0x00, 0x00, 0x14, 0x00, 0xB4, 0x01, 0x0D,   // "xxxx1234"
    0x00, 0x00, 0x00, 0x00, 0x03, 0x00, 0x00, 0x00,   // "56789012"
    0x00, 0x00, 0x02, 0x01, 0x00, 0xB4, 0x01, 0x0D,   // "34567890"
    0x43, 0x4D, 0x41, 0x50, 0x00, 0x00, 0x00, 0x18,   // "CMAPxxxx"
    0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0xFF, 0x00,   // "12345678"
    0xFF, 0xFF, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0x00,   // "90123456"
    0xFF, 0x00, 0x00, 0x00, 0xFF, 0x00, 0x00, 0x00,   // "78901234"
    0x42, 0x4F, 0x44, 0x59, 0x00, 0x00, 0x4B, 0xA8    // "BODYxxxx"
  },
  { 0x46, 0x4F, 0x52, 0x4D, 0x00, 0x00, 0x4B, 0xF0,   // "FORMxxxx"
    0x41, 0x43, 0x42, 0x4D, 0x42, 0x4D, 0x48, 0x44,   // "ACBMBMHD"
    0x00, 0x00, 0x00, 0x14, 0x00, 0xB4, 0x01, 0x0D,   // "xxxx1234"
    0x00, 0x00, 0x00, 0x00, 0x03, 0x00, 0x00, 0x00,   // "56789012"
    0x00, 0x00, 0x02, 0x01, 0x00, 0xB4, 0x01, 0x0D,   // "34567890"
    0x43, 0x4D, 0x41, 0x50, 0x00, 0x00, 0x00, 0x18,   // "CMAPxxxx"
    0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0xFF, 0x00,   // "12345678"
    0xFF, 0xFF, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0x00,   // "90123456"
    0xFF, 0x00, 0x00, 0x00, 0xFF, 0x00, 0x00, 0x00,   // "78901234"
    0x41, 0x42, 0x49, 0x54, 0x00, 0x00, 0x4B, 0xA8    // "ABITxxxx"
} }, ie_iff_header[2][104] =
{ { 0x46, 0x4F, 0x52, 0x4D, 0x00, 0x00, 0x65, 0x40,   // "FORMxxxx"
    0x49, 0x4C, 0x42, 0x4D, 0x42, 0x4D, 0x48, 0x44,   // "ILBMBMHD"
    0x00, 0x00, 0x00, 0x14, 0x00, 0xB4, 0x01, 0x0D,   // "xxxx1234"
    0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00,   // "56789012"
    0x00, 0x00, 0x02, 0x01, 0x00, 0xB4, 0x01, 0x0D,   // "34567890"
    0x43, 0x4D, 0x41, 0x50, 0x00, 0x00, 0x00, 0x30,   // "CMAPxxxx"
    0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0xFF, 0x00,   // "12345678"
    0xFF, 0xFF, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0x00,   // "90123456"
    0xFF, 0x00, 0x00, 0x00, 0xFF, 0x00, 0x00, 0x00,   // "78901234"
    0xBB, 0xBB, 0xBB, 0xCC, 0xCC, 0x00, 0xBB, 0x00,   // "56789012"
    0xBB, 0x99, 0x00, 0x00, 0x00, 0xAA, 0xAA, 0x00,   // "34567890"
    0xBB, 0x00, 0x00, 0x00, 0xAA, 0x00, 0x00, 0x00,   // "12345678"
    0x42, 0x4F, 0x44, 0x59, 0x00, 0x00, 0x64, 0xE0    // "BODYxxxx"
  },
  { 0x46, 0x4F, 0x52, 0x4D, 0x00, 0x00, 0x65, 0x40,   // "FORMxxxx"
    0x41, 0x43, 0x42, 0x4D, 0x42, 0x4D, 0x48, 0x44,   // "ACBMBMHD"
    0x00, 0x00, 0x00, 0x14, 0x00, 0xB4, 0x01, 0x0D,   // "xxxx1234"
    0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00,   // "56789012"
    0x00, 0x00, 0x02, 0x01, 0x00, 0xB4, 0x01, 0x0D,   // "34567890"
    0x43, 0x4D, 0x41, 0x50, 0x00, 0x00, 0x00, 0x30,   // "CMAPxxxx"
    0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0xFF, 0x00,   // "12345678"
    0xFF, 0xFF, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0x00,   // "90123456"
    0xFF, 0x00, 0x00, 0x00, 0xFF, 0x00, 0x00, 0x00,   // "78901234"
    0xBB, 0xBB, 0xBB, 0xCC, 0xCC, 0x00, 0xBB, 0x00,   // "56789012"
    0xBB, 0x99, 0x00, 0x00, 0x00, 0xAA, 0xAA, 0x00,   // "34567890"
    0xBB, 0x00, 0x00, 0x00, 0xAA, 0x00, 0x00, 0x00,   // "12345678"
    0x41, 0x42, 0x49, 0x54, 0x00, 0x00, 0x64, 0xE0    // "ABITxxxx"
} }, a_bmp_header[86] =
{   // bytes 2..5 is the size of the entire file (little-endian)
    // bytes 18..21 is the width of the bitmap (little-endian)
    // bytes 22..25 is the height of the bitmap (little-endian)
    0x42, 0x4D, 0x02, 0x61, 0x00, 0x00, 0x00, 0x00, // "BMsize00"
    0x00, 0x00, 0x56, 0x00, 0x00, 0x00, 0x28, 0x00, // "00yyyy*0"
    0x00, 0x00, 0xB4, 0x00, 0x00, 0x00, 0x0D, 0x01, // "00wwwwhh"
    0x00, 0x00, 0x01, 0x00, 0x04, 0x00, 0x00, 0x00, // "hh104000"
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x13, 0x0B, // "00ssssbb"
    0x00, 0x00, 0x13, 0x0B, 0x00, 0x00, 0x08, 0x00, // "bbbbbb#0"
    0x00, 0x00, 0x08, 0x00, 0x00, 0x00,             // "00#000"
    // B     G     R
    0xFF, 0xFF, 0xFF, 0x00, // white
    0x00, 0xFF, 0xFF, 0x00, // yellow
    0xFF, 0x00, 0xFF, 0x00, // purple
    0x00, 0x00, 0xFF, 0x00, // red
    0xFF, 0xFF, 0x00, 0x00, // cyan
    0x00, 0xFF, 0x00, 0x00, // green
    0xFF, 0x00, 0x00, 0x00, // blue
    0x00, 0x00, 0x00, 0x00  // black
}, ie_bmp_header[118] =
{   // bytes  2..5 is the size of the entire file (little-endian)
    // bytes 18..21 is the width of the bitmap (little-endian)
    // bytes 22..25 is the height of the bitmap (little-endian)
    0x42, 0x4D, 0x22, 0x61, 0x00, 0x00, 0x00, 0x00, // "BMzzzz00"
    0x00, 0x00, 0x76, 0x00, 0x00, 0x00, 0x28, 0x00, // "00yyyy*0"
    0x00, 0x00, 0xB4, 0x00, 0x00, 0x00, 0x0D, 0x01, // "00wwwwhh"
    0x00, 0x00, 0x01, 0x00, 0x04, 0x00, 0x00, 0x00, // "hh104000"
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x13, 0x0B, // "00ssssbb"
    0x00, 0x00, 0x13, 0x0B, 0x00, 0x00, 0x10, 0x00, // "bbbbbb#0"
    0x00, 0x00, 0x10, 0x00, 0x00, 0x00,             // "00#000"
    // B     G     R
    0xFF, 0xFF, 0xFF, 0x00, // white
    0x00, 0xFF, 0xFF, 0x00, // yellow
    0xFF, 0x00, 0xFF, 0x00, // purple
    0x00, 0x00, 0xFF, 0x00, // red
    0xFF, 0xFF, 0x00, 0x00, // cyan
    0x00, 0xFF, 0x00, 0x00, // green
    0xFF, 0x00, 0x00, 0x00, // blue
    0x00, 0x00, 0x00, 0x00, // black
    0xBB, 0xBB, 0xBB, 0x00,
    0x00, 0xCC, 0xCC, 0x00,
    0xBB, 0x00, 0xBB, 0x00,
    0x00, 0x00, 0x99, 0x00,
    0xAA, 0xAA, 0x00, 0x00,
    0x00, 0xBB, 0x00, 0x00,
    0xAA, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00
};

    if (wantasl)
    {   strcpy(fn_screenshot, filepart);
        fixextension(filekind[kind].extension, fn_screenshot, TRUE);

        if (asl
        (   filekind[kind].hail,
            filekind[kind].pattern,
            TRUE,
            DIR_SCREENSHOTS,
            fn_screenshot, // this fortunately doesn't contain the path right now
            fn_screenshot
        ) != EXIT_SUCCESS)
        {   return;
    }   }

    /* Hardcoded for BOXWIDTH of 180!
       180 / 8 = 22.5, but we must round up to the nearest word (ie. to 24)

       Arcadia BMPs are:
       54 bytes header
       32 bytes colour table
    24748 bytes imagery
   =24834 bytes

       Interton/Elektor BMPs are:
       54 bytes header
       64 bytes colour table
    24748 bytes imagery
   =24866 bytes

       Possible enhancements:
     (a) could use compression.
     (b) could use datatypes to support non-ILBM formats.
     (c) could optionally save them in "wide" (doubled horizontally)
         format.
     (d) more detailed/specific error messages.
     (e) check ALL possible errors (eg. calloc(), Close(), etc.).
     (f) perhaps omit left/right/bottom "overscan" margins from Arcadia
         screenshots.

    Known issues:
     (*) some Arcadia games have grey at the bottom of the screen when
         run under the emulator; these grey areas save as white. This
         is really an emulation/graphics issue rather than a screenshot
         issue.

    ACBM files do have pad bits at the end of each row (based on Personal
    Paint 6.4 behaviour). */

    // assert(BOXWIDTH  == 180);
    // assert(BOXHEIGHT == 269);

    if (!(LocalHandle = fopen(fn_screenshot, "wb")))
    {   beep();
        return;
    }

    if (machine == ARCADIA)
    {   if (kind == KIND_BMP)
        {   DISCARD fwrite(a_bmp_header,   (size_t) 86, 1, LocalHandle);
            TheBody = calloc(24748, 1); // should check return code!
            i = 0;
            for (y = BOXHEIGHT - 1; y >= 0; y--)
            {   for (x = 0; x < BOXWIDTH; x++)
                {   if (x % 2)
                    {   TheBody[i++] |= screen[x][y];
                    } else
                    {   TheBody[i] = (screen[x][y] << 4);
                }   }
                TheBody[i++] = 0; // these are needed because (BOXWIDTH / 2) % 4 != 0
                TheBody[i++] = 0; // actual value stored here is unimportant
            }
            DISCARD fwrite(TheBody, (size_t) 24748, 1, LocalHandle);
        } else
        {   // assert(kind == KIND_ILBM || kind == KIND_ACBM);
            DISCARD fwrite(a_iff_header[kind], (size_t) 80, 1, LocalHandle);
            TheBody = calloc(19368, 1); // should check return code!
            if (kind == KIND_ILBM)
            {   for (y = 0; y < BOXHEIGHT; y++)
                {   for (i = 0; i < 3; i++)
                    {   for (x = 0; x < BOXWIDTH; x++)
                        {   if (screen[x][y] & (1 << i))
                            {   TheBody[y * (3 * 24)
                                       +    (i * 24)
                                       +    (x /  8)] |= 0x80 >> (x % 8);
            }   }   }   }   }
            else
            {   // assert(kind == KIND_ACBM);
                for (i = 0; i < 3; i++)
                {   for (y = 0; y < BOXHEIGHT; y++)
                    {   for (x = 0; x < BOXWIDTH; x++)
                        {   if (screen[x][y] & (1 << i))
                            {   TheBody[(i * BOXHEIGHT * 24)
                                      + (y * 24)
                                      + (x / 8)] |= 0x80 >> (x % 8);
            }   }   }   }   }
            DISCARD fwrite(TheBody, (size_t) 19368, 1, LocalHandle);
    }   }
    else
    {   // assert(machine == INTERTON || machine == ELEKTOR);
        if (kind == KIND_BMP)
        {   DISCARD fwrite(ie_bmp_header,       (size_t) 118, 1, LocalHandle);
            TheBody = calloc(24748, 1); // should check return code!
            i = 0;
            for (y = BOXHEIGHT - 1; y >= 0; y--)
            {   for (x = 0; x < BOXWIDTH; x++)
                {   if (x % 2)
                    {   TheBody[i++] |= screen[x][y];
                    } else
                    {   TheBody[i] = (screen[x][y] << 4);
                }   }
                TheBody[i++] = 0; // these are needed because (BOXWIDTH / 2) % 4 != 0
                TheBody[i++] = 0; // actual value stored here is unimportant
            }
            DISCARD fwrite(TheBody, (size_t) 24748, 1, LocalHandle);
        } else
        {   // assert(kind == KIND_ILBM || kind == KIND_ACBM);
            DISCARD fwrite(ie_iff_header[kind], (size_t) 104, 1, LocalHandle);
            TheBody = calloc(25824, 1); // should check return code!
            if (kind == KIND_ILBM)
            {   for (y = 0; y < BOXHEIGHT; y++)
                {   for (i = 0; i < 4; i++)
                    {   for (x = 0; x < BOXWIDTH; x++)
                        {   if (screen[x][y] & (1 << i))
                            {   TheBody[y * (4 * 24)
                                       +    (i * 24)
                                       +    (x /  8)] |= 0x80 >> (x % 8);
            }   }   }   }   }
            else
            {   // assert(kind == KIND_ACBM);
                for (i = 0; i < 4; i++)
                {   for (y = 0; y < BOXHEIGHT; y++)
                    {   for (x = 0; x < BOXWIDTH; x++)
                        {   if (screen[x][y] & (1 << i))
                            {   TheBody[(i * BOXHEIGHT * 24)
                                      + (y * 24)
                                      + (x / 8)] |= 0x80 >> (x % 8);
            }   }   }   }   }
            DISCARD fwrite(TheBody, (size_t) 25824, 1, LocalHandle);
    }   }
    free(TheBody);
    DISCARD fclose(LocalHandle);
    // LocalHandle = NULL;

#ifdef AMIGA
    writeicon(kind, fn_screenshot);
#endif
}

MODULE void a_readrom(void) // Arcadia ROMs of up to 16K are supported
{   int i;

    // assert(machine == ARCADIA);

    if (romsize <= 4 * KILOBYTE)
    {   for (i = 0; i < romsize; i++)
        {   memory[i] = ReadByte();                  // $0000..$0FFF -> $0000..$0FFF
    }   }
    else
    {   for (i = 0; i < 4 * KILOBYTE; i++)
        {   memory[i] = ReadByte();                  // $0000..$0FFF -> $0000..$0FFF
        }
        if (romsize <= 8 * KILOBYTE)
        {   for (i = 0; i < romsize - (4 * KILOBYTE); i++)
            {   memory[0x2000 + i] = ReadByte();     // $1000..$1FFF -> $2000..$2FFF
            }
            if (golf)
            {   // This is for unpatched Arcadia Golf
                // It mirrors $2000..$27FF at $4000..$47FF
                for (i = 0; i < 2048; i++)
                {   memory[0x4000 + i] = memory[0x2000 + i];
        }   }   }
        else
        {   for (i = 0; i < 4 * KILOBYTE; i++)
            {   memory[0x2000 + i] = ReadByte();     // $1000..$1FFF -> $2000..$2FFF
            }
            if (romsize <= 12 * KILOBYTE)
            {   for (i = 0; i < romsize - (8 * KILOBYTE); i++)
                {   memory[0x4000 + i] = ReadByte(); // $2000..$2FFF -> $4000..$4FFF
            }   }
            else
            {   for (i = 0; i < 4 * KILOBYTE; i++)
                {   memory[0x4000 + i] = ReadByte(); // $2000..$2FFF -> $4000..$4FFF
                }
                for (i = 0; i < romsize - (12 * KILOBYTE); i++)
                {   memory[0x6000 + i] = ReadByte(); // $3000..$3FFF -> $6000..$6FFF
}   }   }   }   }

MODULE void i_readrom(void) // Interton ROMs of up to 6K are supported
{   int i;

    for (i = 0; i < romsize; i++)
    {   memory[i] = ReadByte();                      // $0000..$0FFF -> $0000..$0FFF
}   }

MODULE void unautosense(void)
{   downframes     = 4;
    totalframes    = 16;
    key1           = 2;
    key2           = 1;
    key3           = 3;
    key4           = 0;
    fudge          = FUDGE;
    cpl            = PVI_CPL;
    forceinterrupt = FALSE;
    whichgame      = -1;
    sensitivity    = SENSITIVITY_DEFAULT;
    p1sprite       =
    p2sprite       = 4;
    // found       = FALSE;
    if (machine == ARCADIA)
    {   memmap = MEMMAP_G;
    } elif (machine == INTERTON)
    {   memmap = MEMMAP_D;
    } else
    {   // assert(machine == ELEKTOR);
        memmap = MEMMAP_F;
}   }

EXPORT void debug_cli(void)
{   int  address1,
         address2,
         i, j;
    TEXT userinput[80 + 1], // there are buffer overflows possible
         arg1[80 + 1],
         arg2[80 + 1],
         arg3[80 + 1];
    FLAG done = FALSE;

    if (!cli)
    {   return;
    }

    while (!done)
    {   T1; // some function may have closed it
        printf(">");
        fflush(stdout); // maybe unnecessary?
        fflush(stdin);  // maybe unnecessary?
        lockgui();
        DISCARD gets(userinput);
        unlockgui();
        clearmsgs();
        // updatefps()?
        updatemonitor(FALSE);
        checkinput();

        arg1[0] =
        arg2[0] =
        arg3[0] = 0;
        DISCARD sscanf(userinput, "%s %s %s", arg1, arg2, arg3);

#ifdef VERBOSE
        printf("\"%s\", \"%s\", \"%s\", \"%s\"\n", userinput, arg1, arg2, arg3);
#endif

        if (!arg1[0] || !stricmp(arg1, "S"))
        {   done = TRUE;
            docommand(MENUITEM_STEP);
        } elif (!stricmp(arg1, "ALL"))
        {   view_allmemory();
        } elif (!stricmp(arg1, "BC"))
        {   if (bp == OUTOFRANGE)
            {   printf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_ALREADYCLEAREDBP,
                        "Code breakpoint was already cleared.\n\n"
                )   );
            } else
            {   debug_clearbp();
        }   }
        elif (!stricmp(arg1, "BP"))
        {   if (arg3[0])
            {   printf("Usage: BP [<address>]\n\n");
            } elif (arg2[0])
            {   address1 = friendly_to_number(arg2);
                getfriendly(address1);
                if (address1 == OUTOFRANGE)
                {   printf
                    (   GetCatalogStr(CatalogPtr,
                            MSG_INVALIDRANGE,
                            "Valid range is $0000..$7FFF!\n\n"
                    )   );
                } else
                {   bp = address1;
                    updatemenus();
                    printf
                    (   GetCatalogStr(CatalogPtr,
                            MSG_SETBP,
                            "Set code breakpoint to %s.\n\n"
                        ),
                        friendly
                    );
            }   }
            else
            {   debug_set(KIND_BP);
        }   }
        elif (!stricmp(arg1, "BRAM"))
        {   if (machine != ELEKTOR)
            {   // assert(machine == ARCADIA || machine == INTERTON);
                printf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_NOVIEWBRAM,
                        "View monitor RAM is unavailable (wrong machine).\n\n"
                )   );
            } else
            {   view_bram();
        }   }
        elif (!stricmp(arg1, "CPU"))
        {   view_cpu();
        } elif (!stricmp(arg1, "CRAM"))
        {   if
            (   machine == INTERTON
             && (memmap == MEMMAP_C || memmap == MEMMAP_D)
            )
            {   view_cram();
            } else
            {   printf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_NOCRAM,
                        "Cartridge RAM is unavailable.\n\n"
                )   );
        }   }
        elif (!stricmp(arg1, "D"))
        {   if (!arg2[0])
            {   printf("Usage: D <address1> [<address2>]\n\n");
            } else
            {   address1 = friendly_to_number(arg2);
                if (arg3[0])
                {   address2 = friendly_to_number(arg3);
                } else
                {   address2 = address1 + 255;
                }
                if (address2 % 16 != 15) // round up
                {   address2 = ((address2 / 16) * 16) + 15;
                }
                for (i = address1; i <= address2; i += 16)
                {   hex2(astring, (UWORD) i);
                    for (j = 0; j <= 15; j++)
                    {   hex1(disp[j], memory[i + j]);
                    }
                    printf("$%s: %s %s %s %s %s %s %s %s  %s %s %s %s %s %s %s %s\n",
                        astring,
                        disp[0], disp[1],  disp[2],  disp[3],  disp[4],  disp[5],  disp[6],  disp[7],
                        disp[8], disp[9], disp[10], disp[11], disp[12], disp[13], disp[14], disp[15]
                    );
                }
                printf("\n");
        }   }
        elif (!stricmp(arg1, "E"))
        {   if (arg2[0] && arg3[0])
            {   change_read(arg2); // sets tolimit, fromnum, etc.
                if (arg3[0] == '$')
                {   stch_l(&arg3[1], (long *) &tonum); // hexstring -> number
                } else
                {   stcd_l(arg3, (long *) &tonum); // decstring -> number
                }
                change_write(arg2, FALSE);
            } elif (!arg2[0] && !arg3[0])
            {   debug_edit();
            } else
            {   printf("Usage: E [<address> <value>]\n\n");
        }   }
        elif (!stricmp(arg1, "G"))
        {   if (arg3[0])
            {   printf("Usage: G [<address>]\n\n");
            } elif (arg2[0])
            {   address1 = friendly_to_number(arg2);
                getfriendly(address1);
                if (address1 == OUTOFRANGE)
                {   printf
                    (   GetCatalogStr(CatalogPtr,
                            MSG_INVALIDRANGE,
                            "Valid range is $0000..$7FFF!\n\n"
                    )   );
                } else
                {   bp = address1;
                    updatemenus();
                    printf
                    (   GetCatalogStr(CatalogPtr,
                            MSG_ENGINE_SETBPANDGOING,
                            "Setting breakpoint to %s and going.\n\n"
                        ),
                        friendly
                    );
                    done = TRUE;
                    unpause();
            }   }
            else
            {   printf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_GOING,
                        "Going.\n\n"
                )   );
                done = TRUE;
                unpause();
        }   }
        elif
        (   !stricmp(arg1, "H"   )
         || !stricmp(arg1, "HELP")
         || !stricmp(arg1, "?"   )
        )
        {   help_commands();
        } elif (!stricmp(arg1, "I"))
        {   docommand(COMMAND_IGNORE);
        } elif (!stricmp(arg1, "PSG"))
        {   if (machine == ELEKTOR && memmap == MEMMAP_F)
            {   view_psgs();
            } else
            {   printf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_NOPSGS,
                        "PSGs are unavailable.\n\n"
                )   );
        }   }
        elif (!stricmp(arg1, "LI"))
        {   printf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_LOGINSTRUCTIONS,
                    "Log illegal instructions"
            )   );
            printf(" ");
            if (loginstructions)
            {   loginstructions = FALSE;
                printf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_OFF,
                        "off"
                )   );
            } else
            {   loginstructions = TRUE;
                printf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_ON,
                        "on"
                )   );
            }
            printf(".\n\n");
            updatemenu(MENUITEM_LOGINSTRUCTIONS);
        } elif (!stricmp(arg1, "LM"))
        {   if (machine != ELEKTOR)
            {   // assert(machine == ARCADIA || machine == INTERTON);
                printf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_NOLOGMONITOR,
                        "Log monitor calls is unavailable (wrong machine).\n\n"
                )   );
            } else
            {   printf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_LOGMONITOR,
                        "Log monitor calls"
                )   );
                printf(" ");
                if (logmonitor)
                {   logmonitor = FALSE;
                    printf
                    (   GetCatalogStr(CatalogPtr,
                            MSG_ENGINE_OFF,
                            "off"
                    )   );
                } else
                {   logmonitor = TRUE;
                    printf
                    (   GetCatalogStr(CatalogPtr,
                            MSG_ENGINE_ON,
                            "on"
                    )   );
                }
                printf(".\n\n");
                updatemenu(MENUITEM_LOGMONITOR);
        }   }
        elif (!stricmp(arg1, "LOWER"))
        {   if (machine == ARCADIA)
            {   view_lowerscreen();
            } else
            {   printf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_NOLOWERSCRN,
                        "View lower screen is unavailable (wrong machine).\n\n"
                )   );
        }   }
        elif (!stricmp(arg1, "LR"))
        {   printf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_LOGREADS,
                    "Log illegal reads"
            )   );
            printf(" ");
            if (logreads)
            {   logreads = FALSE;
                printf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_OFF,
                        "off"
                )   );
            } else
            {   logreads = TRUE;
                printf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_ON,
                        "on"
                )   );
            }
            printf(".\n\n");
            updatemenu(MENUITEM_LOGREADS);
        } elif (!stricmp(arg1, "LW"))
        {   printf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_LOGWRITES,
                    "Log illegal writes"
            )   );
            printf(" ");
            if (logwrites)
            {   logwrites = FALSE;
                printf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_OFF,
                        "off"
                )   );
            } else
            {   logwrites = TRUE;
                printf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_ON,
                        "on"
                )   );
            }
            printf(".\n\n");
            updatemenu(MENUITEM_LOGWRITES);
        } elif (!stricmp(arg1, "MRAM"))
        {   view_mram();
        } elif (!stricmp(arg1, "PVI"))
        {   if (machine == INTERTON || machine == ELEKTOR)
            {   view_xvi();
            } else
            {   // assert(machine == ARCADIA);
                printf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_NOPVI,
                        "View PVI status is unavailable.\n\n"
                )   );
        }   }
        elif (!stricmp(arg1, "Q"))
        {   done = TRUE;
            cli = FALSE;
            docommand(MENUITEM_CLI);
        } elif (!stricmp(arg1, "R"))
        {   if (arg3[0])
            {   printf("Usage: R [D|F|I|R|<number>]\n\n");
            } elif (!stricmp(arg2, "D"))
            {   if (machine == ARCADIA)
                {   runtodma = TRUE;
                    unpause();
                    done = TRUE;
                } else
                {   printf
                    (   GetCatalogStr(CatalogPtr,
                            MSG_ENGINE_NORUNTODMA,
                            "Run to next DMA is unavailable (wrong machine).\n\n"
                    )   );
            }   }
            elif (!stricmp(arg2, "F"))
            {   runtoframe = TRUE;
                unpause();
                done = TRUE;
            } elif (!stricmp(arg2, "I"))
            {   if (machine != ARCADIA)
                {   runtointerrupt = TRUE;
                    unpause();
                    done = TRUE;
                } else
                {   printf
                    (   GetCatalogStr(CatalogPtr,
                            MSG_ENGINE_NORUNTOINTERRUPT,
                            "Run to next interrupt is unavailable (wrong machine).\n\n"
                    )   );
            }   }
            elif (!stricmp(arg2, "R"))
            {   runtorastline = TRUE;
                unpause();
                done = TRUE;
            } elif (arg2[0])
            {   stcd_l(arg2, (long *) &rastn); // string -> decimal number
                unpause();
                done = TRUE;
            } else
            {   debug_setrastn();
                unpause();
                done = TRUE;
        }   }
        elif (!stricmp(arg1, "RAM"))
        {   view_mram();
            if
            (   machine == INTERTON
             && (memmap == MEMMAP_C || memmap == MEMMAP_D)
            )
            {   view_cram();
        }   }
        elif (!stricmp(arg1, "ROM"))
        {   if (!game)
            {   printf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_NOROM,
                        "View ROM as imagery is unavailable (no game is loaded).\n\n"
                )   );
            } else
            {   view_imagery();
        }   }
        elif (!stricmp(arg1, "SCRN"))
        {   view_screen();
        } elif (!stricmp(arg1, "SYSTEM"))
        {   cleanexit(EXIT_SUCCESS);
        } elif (!stricmp(arg1, "T"))
        {   printf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_TRACECPU,
                    "CPU tracing"
            )   );
            printf(" ");
            if (trace)
            {   trace = FALSE;
                printf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_OFF,
                        "off"
                )   );
            } else
            {   trace = TRUE;
                printf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_ON,
                        "on"
                )   );
            }
            printf(".\n\n");
            updatemenu(MENUITEM_TRACECPU);
        } elif (!stricmp(arg1, "UDC"))
        {   view_udcs();
        } elif (!stricmp(arg1, "UVI"))
        {   if (machine == ARCADIA)
            {   view_xvi();
            } else
            {   // assert(machine == INTERTON || machine == ELEKTOR);
                printf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_NOUVI,
                        "View UVI status is unavailable (wrong machine).\n\n"
                )   );
        }   }
        elif (!stricmp(arg1, "UPPER"))
        {   view_upperscreen();
        } elif (!stricmp(arg1, "WC"))
        {   if (wp == OUTOFRANGE)
            {   printf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_ALREADYCLEAREDWP,
                        "Data watchpoint was already cleared.\n\n"
                )   );
            } else
            {   debug_clearwp();
        }   }
        elif (!stricmp(arg1, "WP"))
        {   if (arg3[0])
            {   printf("Usage: WP [<address>]\n\n");
            } elif (arg2[0])
            {   address1 = friendly_to_number(arg2);
                getfriendly(address1);
                if (address1 == OUTOFRANGE)
                {   printf
                    (   GetCatalogStr(CatalogPtr,
                            MSG_INVALIDRANGE,
                            "Valid range is $0000..$7FFF!\n\n"
                    )   );
                } else
                {   wp = address1;
                    updatemenus();
                    printf
                    (   GetCatalogStr(CatalogPtr,
                            MSG_SETWP,
                            "Set data watchpoint to %s.\n\n"
                        ),
                        friendly
                    );
            }   }
            else
            {   debug_set(KIND_WP);
        }   }
        elif (!stricmp(arg1, "WR"))
        {   if (watchreads)
            {   watchreads = FALSE;
                printf("Watch reads off.\n\n");
            } else
            {   watchreads = TRUE;
                printf("Watch reads on.\n\n");
            }
            updatemenu(MENUITEM_WATCHREADS);
        } else
        {   printf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_UNKNOWNCOMMAND,
                    "Unknown command \"%s\" (\"H\" for help).\n\n"
                ),
                strupr(arg1)
            );
    }   }

    T2;
    // we want to set focus to the Ami/WinArcadia window, but this doesn't work
    // (on IBM-PC at least).
    clearmsgs();
}

MODULE void help_commands(void)
{   PERSIST const STRPTR commandhelp =
"BC                           clear code breakpoint\n"                      \
"BP [<address>]               set code breakpoint to address\n"             \
"D <address1> [<address2>]    display address1 to address2\n"               \
"E [<address> <value>]        change address to value\n"                    \
"G [<address>]                set breakpoint to address and go\n"           \
"H|HELP|?                     view this help\n"                             \
"I                            ignore next instruction\n"                    \
"LI                           log illegal instructions on/off\n"            \
"LM                           log monitor calls on/off\n"                   \
"LR                           log illegal reads on/off\n"                   \
"LW                           log illegal writes on/off\n"                  \
"R R|D|I|F                    run to next rasterline/DMA/interrupt/frame\n" \
"R [<number>]                 run to rasterline number\n"                   \
"Q                            quit debugger CLI\n"                          \
"[S]                          step next instruction\n"                      \
"SYSTEM                       quit emulator\n"                              \
"T                            trace on/off\n"                               \
"WC                           clear data watchpoint\n"                      \
"WP [<address>]               set data watchpoint to address\n"             \
"WR                           watch reads on/off\n"                         \
"ALL, BRAM, CPU, CRAM, LOWER, PSG, PVI, MRAM, RAM, ROM, SCRN, UPPER, UDC,\n"\
" UVI                         view relevant data\n\n";

    printf("%s", commandhelp);
}

EXPORT void view_imagery(void)
{   int i, j;

    bits[0][0][8] = 0;
    T1;
    for (i = 0; i < romsize; i++)
    {   hex2(astring, (UWORD) i);
        hex1(disp[0], memory[i]);
        for (j = 0; j < 8; j++)
        {   if (memory[i] & (0x80 >> j))
            {   bits[0][0][j] = '#';
            } else
            {   bits[0][0][j] = '.';
        }   }

        printf("        db      $%s ;%s $%s\n", disp[0], bits[0][0], astring);
    }
    printf("\n");
    T2;
}

EXPORT void project_open(void)
{   if (crippled)
    {   return;
    }

    macro_stop();
    if (asl
    (   GetCatalogStr(CatalogPtr,
            MSG_ENGINE_OPEN,
            "Open Game/Snapshot/Recording"
        ),
        PATTERN_OPEN,
        FALSE,
        pathpart,
        filepart,
        fn_game
    ) != EXIT_SUCCESS)
    {   return;
    }

    if (machine == ELEKTOR)
    {   fixextension(".PGM", fn_game, FALSE);
    } else
    {   // assert(machine == ARCADIA || machine == INTERTON);
        fixextension(".BIN", fn_game, FALSE);
    }

    if (!engine_load(FALSE))
    {   beep();
    }

    sound_off(FALSE);
    sound_on();
}

EXPORT void project_save(int kind)
{   if (crippled || !game)
    {   return;
    }

    macro_stop();
    fixextension(filekind[kind].extension, fn_game, TRUE);
    breakparts();

    if (asl
    (   filekind[kind].hail,
        filekind[kind].pattern,
        TRUE,
        pathpart,
        filepart,
        fn_game
    ) != EXIT_SUCCESS)
    {   updatesmlgad(GADPOS_MACRO, (ULONG) recmode, TRUE);
        // on IBM-PC version, the above line is needed to "pop" the
        // "record" button back out
        return;
    }
    breakparts();

    engine_save(kind, TRUE);

    if (kind == KIND_EAR || kind == KIND_INR || kind == KIND_TVR)
    {   updatebiggads();
        updatesmlgads();
        updatemenus();
}   }

/* This routine returns its results in ax[dest] and ay[dest]. */
EXPORT void dopaddle(int source, int dest)
{   PERSIST       signed int xx, yy; // PERSISTent for speed
    PERSIST const UBYTE      digipos[3][3] =
    { { 1, 112, 254 }, // ARCADIA
      { 1, 112, 225 }, // INTERTON
      { 1, 111, 225 }  // ELEKTOR
    };
    // centred value of >= $70 is needed for Interton SUPERINV
    // centred value of <= $6F is needed for Elektor JOYSTICK
    // very low left X-value is needed for Interton INVADERS
#ifdef WIN32
    JOYINFOEX joyInfoEx;
#endif

    if   (KeyDown(keypads[source][17])) yy = -1;
    elif (KeyDown(keypads[source][18])) yy =  1;
    else                                yy =  0;
    if   (KeyDown(keypads[source][19])) xx = -1;
    elif (KeyDown(keypads[source][20])) xx =  1;
    else                                xx =  0;

#ifdef WIN32
    if
    (   (source == 0 && joy0)
     || (source == 1 && joy1)
    )
    {   ZeroMemory(&joyInfoEx, sizeof(joyInfoEx));
        joyInfoEx.dwSize  = sizeof(joyInfoEx);
        joyInfoEx.dwFlags = JOY_RETURNALL;
        if (joyGetPosEx(joystickIDMapping[1 - source], &joyInfoEx) == JOYERR_NOERROR)
        {   if (joyInfoEx.dwXpos > 65000)
            {   ax[dest] = 255;
            } else
            {   ax[dest] = (UBYTE) (joyInfoEx.dwXpos >> 8);
            }
            if (joyInfoEx.dwYpos > 65000)
            {   ay[dest] = 255;
            } else
            {   ay[dest] = (UBYTE) (joyInfoEx.dwYpos >> 8);
        }   }
#ifdef VERBOSE
        else { T1; printf("Failed to read paddles of joystick %ld!\n", source); T2; }
#endif
    }
#endif
#ifdef AMIGA
    if (absolute)
    {   if (source == 0 || joy1)
        {   ax[dest] = AJoy.x; // 0..255
            ay[dest] = AJoy.y; // 0..255
    }   }
    else // host digital
    {   if   (joystate & JOYUP)    yy = -1;
        elif (joystate & JOYDOWN)  yy =  1;
        if   (joystate & JOYLEFT)  xx = -1;
        elif (joystate & JOYRIGHT) xx =  1;
    }
#endif

    if (analog)
    {   if   (xx  == -1) { ax[dest] -= sensitivity; if (ax[dest] <   0) ax[dest] =   0; }
        elif (xx  ==  1) { ax[dest] += sensitivity; if (ax[dest] > 255) ax[dest] = 255; }
        if   (yy  == -1) { ay[dest] -= sensitivity; if (ay[dest] <   0) ay[dest] =   0; }
        elif (yy  ==  1) { ay[dest] += sensitivity; if (ay[dest] > 255) ay[dest] = 255; }
    } else
    {   if
        (   !absolute
         || xx != 0
#ifdef WIN32
         || (source == 0 && !joy0)
         || (source == 1 && !joy1)
#endif
        )
        {   ax[dest] = digipos[machine][xx + 1];
        }
        if
        (   !absolute
         || yy != 0
#ifdef WIN32
         || (source == 0 && !joy0)
         || (source == 1 && !joy1)
#endif
        )
        {   ay[dest] = digipos[machine][yy + 1];
}   }   }

EXPORT void update_controls(void)
{   int i;

    if (!ControlsWindowPtr)
    {   return;
    }

#ifdef WIN32
    ControlsRastPtr = GetDC(ControlsWindowPtr);
#endif

    if (machine == ARCADIA)
    {   newkeys[0]  = memory[A_CONSOLE]      & 0x01;
        newkeys[1]  = memory[A_CONSOLE]      & 0x02;
        newkeys[2]  = memory[A_CONSOLE]      & 0x04;
        newkeys[3]  = 0;
        newkeys[4]  = memory[A_P1LEFTKEYS]   & 0x08;
        newkeys[5]  = memory[A_P1LEFTKEYS]   & 0x04;
        newkeys[6]  = memory[A_P1LEFTKEYS]   & 0x02;
        newkeys[7]  = memory[A_P1LEFTKEYS]   & 0x01;
        newkeys[8]  =
        newkeys[36] =
        newkeys[37] = memory[A_P1MIDDLEKEYS] & 0x08;
        newkeys[9]  = memory[A_P1MIDDLEKEYS] & 0x04;
        newkeys[10] = memory[A_P1MIDDLEKEYS] & 0x02;
        newkeys[11] = memory[A_P1MIDDLEKEYS] & 0x01;
        newkeys[12] = memory[A_P1RIGHTKEYS]  & 0x08;
        newkeys[13] = memory[A_P1RIGHTKEYS]  & 0x04;
        newkeys[14] = memory[A_P1RIGHTKEYS]  & 0x02;
        newkeys[15] = memory[A_P1RIGHTKEYS]  & 0x01;
        newkeys[16] = memory[A_P2LEFTKEYS]   & 0x08;
        newkeys[17] = memory[A_P2LEFTKEYS]   & 0x04;
        newkeys[18] = memory[A_P2LEFTKEYS]   & 0x02;
        newkeys[19] = memory[A_P2LEFTKEYS]   & 0x01;
        newkeys[20] =
        newkeys[38] =
        newkeys[39] = memory[A_P2MIDDLEKEYS] & 0x08;
        newkeys[21] = memory[A_P2MIDDLEKEYS] & 0x04;
        newkeys[22] = memory[A_P2MIDDLEKEYS] & 0x02;
        newkeys[23] = memory[A_P2MIDDLEKEYS] & 0x01;
        newkeys[24] = memory[A_P2RIGHTKEYS]  & 0x08;
        newkeys[25] = memory[A_P2RIGHTKEYS]  & 0x04;
        newkeys[26] = memory[A_P2RIGHTKEYS]  & 0x02;
        newkeys[27] = memory[A_P2RIGHTKEYS]  & 0x01;
    } else
    {   // assert(machine == INTERTON || machine == ELEKTOR);
        newkeys[0]  = memory[I_CONSOLE]      & 0x40;
        newkeys[1]  = memory[I_CONSOLE]      & 0x80;
        newkeys[2]  = memory[I_CONSOLE]      & 0x20;
        newkeys[3]  = memory[I_CONSOLE]      & 0x10;
        newkeys[4]  = memory[I_P1LEFTKEYS]   & 0x80;
        newkeys[5]  = memory[I_P1LEFTKEYS]   & 0x40;
        newkeys[6]  = memory[I_P1LEFTKEYS]   & 0x20;
        newkeys[7]  = memory[I_P1LEFTKEYS]   & 0x10;
        newkeys[8]  =
        newkeys[36] =
        newkeys[37] = memory[I_P1MIDDLEKEYS] & 0x80;
        newkeys[9]  = memory[I_P1MIDDLEKEYS] & 0x40;
        newkeys[10] = memory[I_P1MIDDLEKEYS] & 0x20;
        newkeys[11] = memory[I_P1MIDDLEKEYS] & 0x10;
        newkeys[12] = memory[I_P1RIGHTKEYS]  & 0x80;
        newkeys[13] = memory[I_P1RIGHTKEYS]  & 0x40;
        newkeys[14] = memory[I_P1RIGHTKEYS]  & 0x20;
        newkeys[15] = memory[I_P1RIGHTKEYS]  & 0x10;
        newkeys[16] = memory[I_P2LEFTKEYS]   & 0x80;
        newkeys[17] = memory[I_P2LEFTKEYS]   & 0x40;
        newkeys[18] = memory[I_P2LEFTKEYS]   & 0x20;
        newkeys[19] = memory[I_P2LEFTKEYS]   & 0x10;
        newkeys[20] =
        newkeys[38] =
        newkeys[39] = memory[I_P2MIDDLEKEYS] & 0x80;
        newkeys[21] = memory[I_P2MIDDLEKEYS] & 0x40;
        newkeys[22] = memory[I_P2MIDDLEKEYS] & 0x20;
        newkeys[23] = memory[I_P2MIDDLEKEYS] & 0x10;
        newkeys[24] = memory[I_P2RIGHTKEYS]  & 0x80;
        newkeys[25] = memory[I_P2RIGHTKEYS]  & 0x40;
        newkeys[26] = memory[I_P2RIGHTKEYS]  & 0x20;
        newkeys[27] = memory[I_P2RIGHTKEYS]  & 0x10;
    }

    if (sx[0] < 32)
    {   newkeys[29] = 1;
        newkeys[31] = 0;
    } elif (sx[0] > 224)
    {   newkeys[29] = 0;
        newkeys[31] = 1;
    } else
    {   newkeys[29] =
        newkeys[31] = 0;
    }
    if (sx[1] < 32)
    {   newkeys[33] = 1;
        newkeys[35] = 0;
    } elif (sx[1] > 224)
    {   newkeys[33] = 0;
        newkeys[35] = 1;
    } else
    {   newkeys[33] =
        newkeys[35] = 0;
    }
    if (sy[0] < 32)
    {   newkeys[28] = 1;
        newkeys[30] = 0;
    } elif (sy[0] > 224)
    {   newkeys[28] = 0;
        newkeys[30] = 1;
    } else
    {   newkeys[28] =
        newkeys[30] = 0;
    }
    if (sy[1] < 32)
    {   newkeys[32] = 1;
        newkeys[34] = 0;
    } elif (sy[1] > 224)
    {   newkeys[32] = 0;
        newkeys[34] = 1;
    } else
    {   newkeys[32] =
        newkeys[34] = 0;
    }

    for (i = 0; i < KEYS; i++)
    {   if (machine == ELEKTOR && i == 36)
        {   break;
        }
        if
        (   oldkeys[i] != newkeys[i]
        )
        {   invert(i);
            oldkeys[i] = newkeys[i];
    }   }

#ifdef WIN32
    ReleaseDC(ControlsWindowPtr, ControlsRastPtr);
#endif
}

EXPORT void mousedown(int mousex, int mousey)
{   PERSIST int i; // PERSISTent for speed

    if (recmode == RECMODE_PLAY)
    {   return;
    }

    // assert(ControlsWindowPtr);

    wheremouse = -1;
    if (machine == ARCADIA)
    {   for (i = 0; i < KEYS; i++)
        {   if
            (   a_keyemul[i].address != 0
             && mousex >= keyrect[ARCADIA][i].left
             && mousex <  keyrect[ARCADIA][i].right
             && mousey >= keyrect[ARCADIA][i].top
             && mousey <  keyrect[ARCADIA][i].bottom
            )
            {   wheremouse = i;
                newkeys[i] |= memory[a_keyemul[i].mask];
                break; // for speed
    }   }   }
    elif (machine == INTERTON)
    {   for (i = 0; i < KEYS; i++)
        {   if
            (   i_keyemul[i].address != 0
             && mousex >= keyrect[INTERTON][i].left
             && mousex <  keyrect[INTERTON][i].right
             && mousey >= keyrect[INTERTON][i].top
             && mousey <  keyrect[INTERTON][i].bottom
            )
            {   wheremouse = i;
                newkeys[i] |= memory[i_keyemul[i].mask];
                break; // for speed
    }   }   }
    else
    {   // assert(machine == ELEKTOR);
        for (i = 0; i < 36; i++)
        {   if
            (   e_keyemul[i].address != 0
             && mousex >= keyrect[ELEKTOR][i].left
             && mousex <  keyrect[ELEKTOR][i].right
             && mousey >= keyrect[ELEKTOR][i].top
             && mousey <  keyrect[ELEKTOR][i].bottom
            )
            {   wheremouse = i;
                newkeys[i] |= memory[e_keyemul[i].mask];
                break; // for speed
}   }   }   }

EXPORT void mouseup(void)
{   wheremouse = -1;
}

EXPORT void domouse(void)
{   PERSIST UBYTE t; // PERSISTent for speed

    if (wheremouse != -1)
    {   // assert(ControlsWindowPtr);

        if (machine == ARCADIA)
        {   t = memory[a_keyemul[wheremouse].address] | a_keyemul[wheremouse].mask;
            uviwrite(a_keyemul[wheremouse].address, t);
        } elif (machine == INTERTON)
        {   t = memory[i_keyemul[wheremouse].address] | i_keyemul[wheremouse].mask;
            pviwrite(i_keyemul[wheremouse].address, t);
        } else
        {   // assert(machine == ELEKTOR);
            t = memory[e_keyemul[wheremouse].address] | e_keyemul[wheremouse].mask;
            pviwrite(e_keyemul[wheremouse].address, t);
}   }   }

EXPORT void view_controls_engine(void)
{   int i;

    for (i = 0; i < KEYS; i++)
    {   oldkeys[i] =
        newkeys[i] = 0;
    }

    sx[0] =
    sx[1] =
    sy[0] =
    sy[1] = 128; // so we don't have any permanently lit paddles

    mouseup();
}

/* Connect/disconnect netplay interface. */
EXPORT void NetPlay(int Switch)
{   int i;

#ifdef AMIGA
    if (!SocketBase)
    {   return;
    }
#endif

    switch (Switch)
    {
    case NET_OFF:
        // Disconnect netplay
        NetClose();
    acase NET_SERVER:
        // If already connected, drop out
        if (connected != NET_OFF)
        {   break;
        }

        if (!NetMyName(hostname, sizeof(hostname)))
        {   say
            (   GetCatalogStr(CatalogPtr,
                    MSG_NET_NORESOLVEHOSTNAME,
                    "Can't resolve hostname!"
                ),
                TRUE
            );
            return;
        }

        T1;
        printf
        (   GetCatalogStr(CatalogPtr,
                MSG_NET_ACTINGASSERVER,
               "Acting as server at %s...\n"
            ),
            hostname
        );
        T2;
        // Connect second player, report problems if present
        NetConnect(0, NETPORT);
        /* if (connected == NET_OFF)
        {   T1; printf("Can't act as server at %s!\n", hostname); T2;
        } */
    acase NET_CLIENT:
        // If already connected, drop out
        if (connected != NET_OFF)
        {   break;
        }

        sound_off(FALSE);
        hostnamewindow();
        sound_on();
        if (!hostname[0])
        {   return;
        }

        // convert to lower case
        for (i = 0; i < (int) strlen(hostname); i++)
        {   if (hostname[i] >= 'A' && hostname[i] <= 'Z')
            {   hostname[i] += 32;
        }   }

        T1;
        printf
        (   GetCatalogStr(CatalogPtr,
                MSG_NET_CONNECTINGASCLIENT,
               "Connecting as client to server at %s...\n"
            ),
            hostname
        );
        T2;
        // Connect second player, report problems if present
        NetConnect(hostname, NETPORT);
        /* if (connected == NET_OFF)
        {   T1; printf("Can't connect as client to server at %s!\n", hostname); T2;
        } */
    adefault:
    break;
    }
    settitle();
}

MODULE void NetConnect(const char* Server, unsigned short Port)
{   struct sockaddr_in Addr;
    struct hostent*    Host;
    int                LSocket;
#ifdef WIN32
    int                AddrLength;
    unsigned long      J;
#endif
#ifdef AMIGA
    LONG               AddrLength;
    char               J;
#endif

#ifdef AMIGA
    if (!SocketBase)
    {   return;
    }
#endif

    /* Close existing network connection */
    NetClose();

    /* Clear the address structure */
    memset(&Addr, 0, sizeof(Addr));

    /* If the server name is given, we first try to */
    /* connect as a client. */
    if (Server)
    {   /* Look up server address */

        if (!(Host = gethostbyname(Server)))
        {   T1;
            printf
            (   GetCatalogStr(CatalogPtr,
                    MSG_NET_NOGETHOSTNAME,
                    "Can't get hostname!\n"
            )   );
            T2;
            connected = NET_OFF;
            return;
        }

        /* Set fields of the address structure */
        memcpy(&Addr.sin_addr, Host->h_addr, Host->h_length);
        Addr.sin_family = AF_INET;
        Addr.sin_port   = htons(Port);

        /* Create a socket */
        if ((Socket = socket(AF_INET, SOCK_STREAM, 0)) < 0)
        {   T1;
            printf
            (   GetCatalogStr(CatalogPtr,
                    MSG_NET_NOCREATECLIENT,
                    "Can't create client socket!\n"
            )   );
            T2;
            connected = NET_OFF;
            return;
        }

        /* Connecting... */
        if (connect(Socket, (struct sockaddr *) &Addr, sizeof(Addr)) >= 0)
        {   /* Make communication socket blocking */
            J = 0;
            if (ioctlsocket(Socket, FIONBIO, &J) < 0)
            {   T1;
                printf
                (   GetCatalogStr(CatalogPtr,
                        MSG_NET_NOBLOCKCLIENT,
                        "Can't block client socket!\n"
                )   );
                T2;
                closesocket(Socket);
                Socket = -1;
                connected = NET_OFF;
                return;
            }

            connected = NET_CLIENT;
            T1;
            printf
            (   GetCatalogStr(CatalogPtr,
                    MSG_NET_CLIENTCONNECTED,
                    "Successfully connected as client!\n"
            )   );
            T2;
            return;
        }

        closesocket(Socket);
        T1;
        printf
        (   GetCatalogStr(CatalogPtr,
                MSG_NET_NOCONNECTCLIENT,
                "Can't connect as client!\n"
        )   );
        T2;
    } else
    {   /* Becoming a server and waiting for connection request. */

        /* Set fields of the address structure */
        Addr.sin_addr.s_addr = htonl(INADDR_ANY);
        Addr.sin_family      = AF_INET;
        Addr.sin_port        = htons(Port);

        /* Create a listening socket */
        if ((LSocket = socket(AF_INET, SOCK_STREAM, 0)) < 0)
        {   T1;
            printf
            (   GetCatalogStr(CatalogPtr,
                    MSG_NET_NOCREATESERVER,
                    "Can't create server socket!\n"
            )   );
            T2;
            connected = NET_OFF;
            return;
        }

        /* Bind listening socket */
        if (bind(LSocket, (struct sockaddr *) &Addr, sizeof(Addr)) < 0)
        {   T1;
            printf
            (   GetCatalogStr(CatalogPtr,
                    MSG_NET_NOBINDSOCKET,
                    "Can't bind server socket!\n"
            )   );
            T2;
            closesocket(LSocket);
            connected = NET_OFF;
            return;
        }

        /* Make listening socket non-blocking */
        J = 1;
        if (ioctlsocket(LSocket, FIONBIO, &J) < 0)
        {   T1;
            printf
            (   GetCatalogStr(CatalogPtr,
                    MSG_NET_NOUNBLOCKSOCKET,
                    "Can't unblock server socket!\n"
            )   );
            T2;
            closesocket(LSocket);
            connected = NET_OFF;
            return;
        }

        /* Listen for one client */
        if (listen(LSocket, 1) < 0)
        {   T1;
            printf
            (   GetCatalogStr(CatalogPtr,
                    MSG_NET_NOLISTEN,
                    "Can't listen!\n"
            )   );
            T2;
            closesocket(LSocket);
            connected = NET_OFF;
            return;
        }

        T1;
        printf
        (   GetCatalogStr(CatalogPtr,
                MSG_NET_LISTENING,
                "Listening...\n"
        )   );
        T2;

        lockgui();
        aborting = FALSE;
        listen_open();

        /* Accepting calls... */

        AddrLength = sizeof(Addr);
        do
        {   /* Process all messages */
#ifdef WIN32
            clearmsgs();
#endif
#ifdef AMIGA
            listen_loop();
#endif
            Socket = accept(LSocket, (struct sockaddr *) &Addr, &AddrLength);
#ifdef WIN32
            if (WSAGetLastError() && (WSAGetLastError() != WSAEWOULDBLOCK))
            {   break;
            }
#endif
        }
        while (Socket < 0 && !aborting);

     /* if (ListeningWindowPtr)
        {   // assert(!aborting);
            // assert(Socket >= 0);
         */ listen_close();
     // }

        closesocket(LSocket);
        clearkybd();
        unlockgui();

        /* Client failed to connect */
        if (Socket < 0)
        {   T1;
            printf
            (   GetCatalogStr(CatalogPtr,
                    MSG_NET_ABORTED,
                    "Aborted by user!\n"
            )   );
            T2;
            connected = NET_OFF;
            return;
        }

        /* Make communication socket blocking */
        J = 0;
        if (ioctlsocket(Socket, FIONBIO, &J) < 0)
        {   T1;
            printf
            (   GetCatalogStr(CatalogPtr,
                    MSG_NET_NOBLOCKSERVER,
                    "Can't block server socket!\n"
            )   );
            T2;
            closesocket(Socket);
            Socket = -1;
            connected = NET_OFF;
            return;
        }

        T1;
        printf
        (   GetCatalogStr(CatalogPtr,
                MSG_NET_SERVERCONNECTED,
                "The client has successfully connected to your server!\n"
        )   );
        T2;
        connected = NET_SERVER;
        // return;
}   }

/* Closes connection opened with NetConnect(). */
MODULE void NetClose(void)
{
#ifdef AMIGA
    if (!SocketBase)
    {   return;
    }
#endif

    if (Socket >= 0)
    {   // assert(connected != NET_OFF);

        closesocket(Socket);
        Socket = -1;
        connected = NET_OFF;
        T1;
        printf
        (   GetCatalogStr(CatalogPtr,
                MSG_NET_DISCONNECTED,
                "Disconnected.\n"
        )   );
        T2;
}   }

/* Returns local address or 0 on failure. */
MODULE const char* NetMyName(char* Buffer, int MaxChars)
{   struct hostent* Host;

#ifdef AMIGA
    if (!SocketBase)
    {   return(0);
    }
#endif

    /* Have to have enough characters */
    if (MaxChars < 16)
    {   return(0);
    }
    /* Get local hostname */
    gethostname(Buffer, MaxChars);
    /* Look up address */
    if (!(Host = gethostbyname(Buffer)))
    {   return(0);
    }
    /* Must have an address */
    if (!Host->h_addr_list || !Host->h_addr_list[0])
    {   return(0);
    }
    /* Copy address */
    sprintf
    (   Buffer,
        "%d.%d.%d.%d",
        (UBYTE) Host->h_addr_list[0][0],
        (UBYTE) Host->h_addr_list[0][1],
        (UBYTE) Host->h_addr_list[0][2],
        (UBYTE) Host->h_addr_list[0][3]
    );

    return(Buffer);
}

/* Exchanges given number of bytes with the other side and
   returns TRUE on success, FALSE on failure. */
EXPORT FLAG NetExchange(char* In, const char* Out, int N)
{   int I, J;

#ifdef AMIGA
    if (!SocketBase)
    {   return FALSE;
    }
#endif

    /* Have to have a socket */
    if (Socket < 0)
    {   return FALSE;
    }

    /* Send data */
    for (I = J = N; (J >= 0) && I;)
    {   J = send(Socket, Out, I, 0);
        if (J > 0)
        {   Out += J;
            I   -= J;
    }   }
    if (I)
    {   return FALSE;
    }

    /* Receive data */
    for (I = J = N; (J >= 0) && I;)
    {   J = recv(Socket, In, I, 0);
        if (J > 0)
        {   In += J;
            I  -= J;
    }   }
    if (I)
    {   return FALSE;
    }

    return TRUE;
}

EXPORT FLAG NetSendByte(UBYTE value)
{
#ifdef AMIGA
    if (!SocketBase)
    {   return FALSE;
    }
#endif

    /* Have to have a socket */
    if (Socket < 0)
    {   return FALSE;
    }

    /* Send data */
    if (send(Socket, &value, 1, 0) > 0)
    {   return TRUE;
    } // implied else
    return FALSE;
}

EXPORT UBYTE NetReceiveByte(void)
{   PERSIST UBYTE value; // PERSISTent for speed

#ifdef AMIGA
    if (!SocketBase)
    {   return 0;
    }
#endif

    /* Have to have a socket */
    if (Socket < 0)
    {   return 0;
    }

    /* Send data */
    if (recv(Socket, &value, 1, 0) > 0)
    {   return value;
    } // implied else
    return 0;

    /* When successful, returns 0..255.
       When failed, returns 0.
       Therefore, 0 can mean either success or failure :-( . */
}

EXPORT void closemonitors(FLAG both)
{   /* If both is TRUE,  monitor and controls are closed/reopened.
       If both is FALSE, only monitor is closed/reopened. */

    if (MonitorWindowPtr)
    {   close_monitor();
        view_monitor();
    }
    if (both && ControlsWindowPtr)
    {   close_controls();
        view_controls();
}   }

EXPORT void updatemonitor(FLAG force)
{   PERSIST TEXT bytestring[2 + 1],
                 wordstring[4 + 1],
                 coordstring[5 + 1] = "--,--",
                 decimalstring[10 + 1];
    PERSIST int  i, j;

    update_controls();

    if (!MonitorWindowPtr)
    {   return;
    }

 /* In some games, eg. Grand Slam Tennis, the monitor will always report
    PnPADDLE as being $FF; this is because after the game reads these
    registers, they are set to $FF (being read-once). */

    // CPU chip...

    for (i = 0; i < 7; i++)
    {   if (r[i] != monitorr[i] || force)
        {   monitorr[i] = r[i];
            hex1(bytestring, monitorr[i]);
            setmonitor(RGad[i], bytestring);
    }   }

    if (psu != monitorpsu || force)
    {   if ((psu & PSU_S) != (monitorpsu & PSU_S) || force)
        {   if (psu & PSU_S)
            {   setmonitor(SensePtr, "S");
            } else
            {   setmonitor(SensePtr, "s");
        }   }
        if ((psu & PSU_F) != (monitorpsu & PSU_F) || force)
        {   if (psu & PSU_F)
            {   setmonitor(FlagPtr, "F");
            } else
            {   setmonitor(FlagPtr, "f");
        }   }
        if ((psu & PSU_II) != (monitorpsu & PSU_II) || force)
        {   if (psu & PSU_II)
            {   setmonitor(InterruptPtr, "I");
            } else
            {   setmonitor(InterruptPtr, "i");
        }   }
        if ((psu & PSU_SP) != (monitorpsu & PSU_SP) || force)
        {   stcul_d(decimalstring, psu & PSU_SP);
            setmonitor(SPGad, decimalstring);
        }

        monitorpsu = psu;
    }

    if (psl != monitorpsl || force)
    {   if ((psl & PSL_CC) != (monitorpsl & PSL_CC) || force)
        {   setmonitor(CCPtr, (STRPTR) (ccstring[(psl & 0xC0) >> 6])); // %11000000 -> %00000011
        }

        if ((psl & PSL_IDC) != (monitorpsl & PSL_IDC) || force)
        {   if (psl & PSL_IDC)
            {   setmonitor(IDCPtr, "D");
            } else
            {   setmonitor(IDCPtr, "d");
        }   }

        if ((psl & PSL_RS) != (monitorpsl & PSL_RS) || force)
        {   if (psl & PSL_RS)
            {   setmonitor(RSPtr, "R");
            } else
            {   setmonitor(RSPtr, "r");
        }   }

        if ((psl & PSL_WC) != (monitorpsl & PSL_WC) || force)
        {   if (psl & PSL_WC)
            {   setmonitor(WCPtr, "W");
            } else
            {   setmonitor(WCPtr, "w");
        }   }

        if ((psl & PSL_OVF) != (monitorpsl & PSL_OVF) || force)
        {   if (psl & PSL_OVF)
            {   setmonitor(OVFPtr, "O");
            } else
            {   setmonitor(OVFPtr, "o");
        }   }

        if ((psl & PSL_COM) != (monitorpsl & PSL_COM) || force)
        {   if (psl & PSL_COM)
            {   setmonitor(COMPtr, "M");
            } else
            {   setmonitor(COMPtr, "m");
        }   }

        if ((psl & PSL_C) != (monitorpsl & PSL_C) || force)
        {   if (psl & PSL_C)
            {   setmonitor(CPtr, "C");
            } else
            {   setmonitor(CPtr, "c");
        }   }

        monitorpsl = psl;
    }

    for (i = 0; i < 8; i++)
    {   if (ras[i] != monitorras[i] || force)
        {   monitorras[i] = ras[i];
            hex2(wordstring, monitorras[i]);
            setmonitor(RASGad[i], wordstring);
    }   }

    if (page + iar != monitoriar || force)
    {   monitoriar = page + iar;
        hex2(wordstring, monitoriar);
        setmonitor(IARPtr, wordstring);
    }

    if (frames != monitorframes || force)
    {   monitorframes = frames;
        stcul_d(decimalstring, frames);
        setmonitor(FramePtr, decimalstring);

        calcrunningtime();
        setmonitor(TimePtr, timestring);
    }
    if (monitorelapsed != elapsed + cycles || force)
    {   monitorelapsed = elapsed + cycles;
        stcul_d(decimalstring, monitorelapsed);
        setmonitor(ClockPtr, decimalstring);
    }
    if (rast != monitorrast || force)
    {   monitorrast = rast;
        stcul_d(decimalstring, rast);
        setmonitor(RasterGad, decimalstring);
    }

    if (machine == ARCADIA)
    {   for (i = 73; i < MONITORGADS; i++)
        {   if
            (   memory[monitoraddr[i]] != monitorcontents[i]
             || force
            )
            {   monitorcontents[i] = memory[monitoraddr[i]];
                hex1(bytestring, monitorcontents[i]);
                setmonitor(monitorgad[i], bytestring);
        }   }

        for (i = 0; i < 4; i++)
        {   if
            (   monitorSPRITEnAX[i] != memory[A_SPRITE0X + (i * 2)] // SPRITEnX
             || monitorSPRITEnAY[i] != memory[A_SPRITE0Y + (i * 2)] // SPRITEnY
             || force
            )
            {   monitorSPRITEnAX[i] = memory[A_SPRITE0X + (i * 2)];
                monitorSPRITEnAY[i] = memory[A_SPRITE0Y + (i * 2)];
                hex1(coordstring, monitorSPRITEnAX[i]);
                coordstring[2] = ',';
                hex1(&coordstring[3], monitorSPRITEnAY[i]);
                setmonitor(SPRITEnAXYGad[i], coordstring);
    }   }   }
    else
    {   // assert(machine == INTERTON || machine == ELEKTOR);

        if (machine == ELEKTOR && memmap == MEMMAP_E)
        {   for (i = 0; i <= 18; i++)
            {   if (memory[monitoraddr[i]] != monitorcontents[i] || force)
                {   monitorcontents[i] = memory[monitoraddr[i]];
                    hex1(bytestring, monitorcontents[i]);
                    setmonitor(monitorgad[i], bytestring);
        }   }   }
        elif (machine == ELEKTOR && memmap == MEMMAP_F)
        {   for (i = 0; i <= 35; i++)
            {   if (memory[monitoraddr[i]] != monitorcontents[i] || force)
                {   monitorcontents[i] = memory[monitoraddr[i]];
                    hex1(bytestring, monitorcontents[i]);
                    setmonitor(monitorgad[i], bytestring);
            }   }

            for (i = j = 0; i <= 1; i++, j += 16)
            {   if (monitorPITCHA[i] != ((memory[E_PITCHA1_H + j] * 256) + memory[E_PITCHA1_L + j]) || force)
                {   monitorPITCHA[i] = (memory[E_PITCHA1_H + j] * 256) + memory[E_PITCHA1_L + j];
                    hex2(bytestring, monitorPITCHA[i]);
                    setmonitor(PITCHAGad[i], bytestring);
                }
                if (monitorPITCHB[i] != ((memory[E_PITCHB1_H + j] * 256) + memory[E_PITCHB1_L + j]) || force)
                {   monitorPITCHB[i] = (memory[E_PITCHB1_H + j] * 256) + memory[E_PITCHB1_L + j];
                    hex2(bytestring, monitorPITCHB[i]);
                    setmonitor(PITCHBGad[i], bytestring);
                }
                if (monitorPITCHC[i] != ((memory[E_PITCHC1_H + j] * 256) + memory[E_PITCHC1_L + j]) || force)
                {   monitorPITCHC[i] = (memory[E_PITCHC1_H + j] * 256) + memory[E_PITCHC1_L + j];
                    hex2(bytestring, monitorPITCHC[i]);
                    setmonitor(PITCHCGad[i], bytestring);
                }
                if (monitorPERIOD[i] != ((memory[E_PERIOD1_H + j] * 256) + memory[E_PERIOD1_L + j]) || force)
                {   monitorPERIOD[i] = (memory[E_PERIOD1_H + j] * 256) + memory[E_PERIOD1_L + j];
                    hex2(bytestring, monitorPERIOD[i]);
                    setmonitor(PERIODGad[i], bytestring);
        }   }   }
        else
        {   // assert(machine == INTERTON);
            for (i = 0; i <= 19; i++)
            {   if (memory[monitoraddr[i]] != monitorcontents[i] || force)
                {   monitorcontents[i] = memory[monitoraddr[i]];
                    hex1(bytestring, monitorcontents[i]);
                    setmonitor(monitorgad[i], bytestring);
            }   }
            for (i = 36; i <= 72; i++)
            {   if (memory[monitoraddr[i]] != monitorcontents[i] || force)
                {   monitorcontents[i] = memory[monitoraddr[i]];
                    hex1(bytestring, monitorcontents[i]);
                    setmonitor(monitorgad[i], bytestring);
        }   }   }

        for (i = 0; i < 4; i++)
        {   if
            (   monitorSPRITEnAX[i] != memory[ie_spritedata[i] + 10] // SPRITEnAX
             || monitorSPRITEnAY[i] != memory[ie_spritedata[i] + 12] // SPRITEnAY
             || force
            )
            {   monitorSPRITEnAX[i] = memory[ie_spritedata[i] + 10];
                monitorSPRITEnAY[i] = memory[ie_spritedata[i] + 12];
                hex1(coordstring, monitorSPRITEnAX[i]);
                coordstring[2] = ',';
                hex1(&coordstring[3], monitorSPRITEnAY[i]);
                setmonitor(SPRITEnAXYGad[i], coordstring);
            }
            if
            (   monitorSPRITEnBX[i] != memory[ie_spritedata[i] + 11] // SPRITEnBX
             || monitorSPRITEnBY[i] != memory[ie_spritedata[i] + 13] // SPRITEnBY
             || force
            )
            {   monitorSPRITEnBX[i] = memory[ie_spritedata[i] + 11];
                monitorSPRITEnBY[i] = memory[ie_spritedata[i] + 13];
                hex1(coordstring, monitorSPRITEnBX[i]);
                coordstring[2] = ',';
                hex1(&coordstring[3], monitorSPRITEnBY[i]);
                setmonitor(SPRITEnBXYGad[i], coordstring);
}   }   }   }

EXPORT void playsound(FLAG force)
{   PERSIST UWORD  guestpitch[14],
                   oldguestpitch[14]  = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
    PERSIST UBYTE  guestvolume[14],
                   oldguestvolume[14] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
    PERSIST int    i;     // PERSISTent for speed
    PERSIST float  hertz; // PERSISTent for speed

    /* guestvolume[0..13]: indexed by guest channel.
                           contents are stored in guest format.
       guestpitch[0..13]:  indexed by guest channel.
                           contents are stored in guest format.

    AmiArcadia: Paula audio is not currently working...
       (a) WinUAE compatibility problems.
       (b) Unable to allocate channels.

       Although Arcadia and Interton/Elektor volumes do not match, this
       is unimportant as playsound(TRUE) is called when changing machines.
       Thus, Arcadia volumes are never compared against Interton/Elektor
       volumes. */

    if (!sound)
    {   return;
    }

    if (machine == ARCADIA)
    {   if ((memory[A_VOLUME]) & 0x08)
        {   guestpitch[6]   = memory[A_PITCH]  & 0x7F;
            guestvolume[6]  = memory[A_VOLUME] & 0x07; // 0..7
        } else
        {   guestpitch[6]   =
            guestvolume[6]  = 0;
        }

        if ((memory[A_VOLUME]) & 0x10)
        {   guestpitch[13]  = memory[A_PITCH]  & 0x7F;
            guestvolume[13] = memory[A_VOLUME] & 0x07; // 0..7
        } else
        {   guestpitch[13]  =
            guestvolume[13] = 0;
    }   }
    else
    {   // assert(machine == INTERTON || machine == ELEKTOR);

        // PVI tones
        guestpitch[6] = memory[I_PITCH];
        if
        (   machine == ELEKTOR
         && memmap == MEMMAP_F
         && (memory[E_ENABLE1] & 0x40)
        )
        {   guestvolume[6] = memory[E_PORTA1] & 0xF;
        } else
        {   guestvolume[6] = 15;
        }
        // using I_NOISE to turn off PVI tone generator is not supported
        // if ((memory[I_NOISE]) & 0x0C) guestvolume[6] = 0;

        // should we still play noises when pitch is 0?

        if (machine == ELEKTOR && memmap == MEMMAP_E)
        {   guestpitch[13]  =  1; // must be nonzero!
            guestvolume[13] =  0;
        } else
        {   guestpitch[13]  =  1; // must be nonzero!
            if (memory[I_NOISE] & 0x10)
            {   guestvolume[13] = 15;
            } else
            {   guestvolume[13] = 0;
    }   }   }

    if (machine == ELEKTOR && memmap == MEMMAP_F)
    {   // 1st PSG, output channel A
        if (!(memory[E_AMPLITUDEA1] & 0x10))
        {   guestvolume[0] =
            guestvolume[7] = memory[E_AMPLITUDEA1] & 0x0F;
        } else
        {   guestvolume[0] =
            guestvolume[7] = 0;
        }
        if (!(memory[E_ENABLE1] &    1))
        {   guestpitch[0] = ((memory[E_PITCHA1_H] & 0x0F) * 256) + memory[E_PITCHA1_L];
        } else
        {   guestpitch[0] = 0;
        }
        if (!(memory[E_ENABLE1] &    8))
        {   guestpitch[7] = memory[E_PITCHD1] & 0x1F;
        } else
        {   guestpitch[7] = 0;
        }

        // 1st PSG, output channel B
        if (!(memory[E_AMPLITUDEB1] & 0x10))
        {   guestvolume[1] =
            guestvolume[8] = memory[E_AMPLITUDEB1] & 0x0F;
        } else
        {   guestvolume[1] =
            guestvolume[8] = 0;
        }
        if (!(memory[E_ENABLE1] &    2))
        {   guestpitch[1] = ((memory[E_PITCHB1_H] & 0x0F) * 256) + memory[E_PITCHB1_L];
        } else
        {   guestpitch[1] = 0;
        }
        if (!(memory[E_ENABLE1] & 0x10))
        {   guestpitch[8] = memory[E_PITCHD1] & 0x1F;
        } else
        {   guestpitch[8] = 0;
        }

        // 1st PSG, output channel C
        if (!(memory[E_AMPLITUDEC1] & 0x10))
        {   guestvolume[2] =
            guestvolume[9] = memory[E_AMPLITUDEC1] & 0x0F;
        } else
        {   guestvolume[2] =
            guestvolume[9] = 0;
        }
        if (!(memory[E_ENABLE1] &    4))
        {   guestpitch[2] = ((memory[E_PITCHC1_H] & 0x0F) * 256) + memory[E_PITCHC1_L];
        } else
        {   guestpitch[2] = 0;
        }
        if (!(memory[E_ENABLE1] & 0x20))
        {   guestpitch[9] = memory[E_PITCHD1] & 0x1F;
        } else
        {   guestpitch[9] = 0;
        }

        // 2nd PSG, output channel A
        if (!(memory[E_AMPLITUDEA2] & 0x10))
        {   guestvolume[3] =
            guestvolume[10] = memory[E_AMPLITUDEA2] & 0x0F;
        } else
        {   guestvolume[3] =
            guestvolume[10] = 0;
        }
        if (!(memory[E_ENABLE2] &    1))
        {   guestpitch[3] = ((memory[E_PITCHA2_H] & 0x0F) * 256) + memory[E_PITCHA2_L];
        } else
        {   guestpitch[3] = 0;
        }
        if (!(memory[E_ENABLE2] &    8))
        {   guestpitch[10] = memory[E_PITCHD2] & 0x1F;
        } else
        {   guestpitch[10] = 0;
        }

        // 2nd PSG, output channel B
        if (!(memory[E_AMPLITUDEB2] & 0x10))
        {   guestvolume[4]  =
            guestvolume[11] = memory[E_AMPLITUDEB2] & 0x0F;
        } else
        {   guestvolume[4]  =
            guestvolume[11] = 0;
        }
        if (!(memory[E_ENABLE2] &    2))
        {   guestpitch[4] = ((memory[E_PITCHB2_H] & 0x0F) * 256) + memory[E_PITCHB2_L];
        } else
        {   guestpitch[4] = 0;
        }
        if (!(memory[E_ENABLE2] & 0x10))
        {   guestpitch[11] = memory[E_PITCHD2] & 0x1F;
        } else
        {   guestpitch[11] = 0;
        }

        // 2nd PSG, output channel C
        if (!(memory[E_AMPLITUDEC2] & 0x10))
        {   guestvolume[5]  =
            guestvolume[12] = memory[E_AMPLITUDEC2] & 0x0F;
        } else
        {   guestvolume[5]  =
            guestvolume[12] = 0;
        }
        if (!(memory[E_ENABLE2] &    4))
        {   guestpitch[5] = ((memory[E_PITCHC2_H] & 0x0F) * 256) + memory[E_PITCHC2_L];
        } else
        {   guestpitch[5] = 0;
        }
        if (!(memory[E_ENABLE2] & 0x20))
        {   guestpitch[12] = memory[E_PITCHD2] & 0x1F;
        } else
        {   guestpitch[12] = 0;
    }   }
    else
    {   guestpitch[0]   =
        guestpitch[1]   =
        guestpitch[2]   =
        guestpitch[3]   =
        guestpitch[4]   =
        guestpitch[5]   =
        guestpitch[7]   =
        guestpitch[8]   =
        guestpitch[9]   =
        guestpitch[10]  =
        guestpitch[11]  =
        guestpitch[12]  = 0;

        guestvolume[0]  =
        guestvolume[1]  =
        guestvolume[2]  =
        guestvolume[3]  =
        guestvolume[4]  =
        guestvolume[5]  =
        guestvolume[7]  =
        guestvolume[8]  =
        guestvolume[9]  =
        guestvolume[10] =
        guestvolume[11] =
        guestvolume[12] = 0;
    }

/*   0: 1st PSG, tone  channel A
     1: 1st PSG, tone  channel B
     2: 1st PSG, tone  channel C
     3: 2nd PSG, tone  channel A
     4: 2nd PSG, tone  channel B
     5: 2nd PSG, tone  channel C
     6: PVI/UVI tone generator
     7: 1st PSG, noise channel A
     8: 1st PSG, noise channel B
     9: 1st PSG, noise channel C
    10: 2nd PSG, noise channel A
    11: 2nd PSG, noise channel B
    12: 2nd PSG, noise channel C
    13: $1E80 noise generator, or UVI noise generator */

    for (i = 0; i <= 6; i++)
    {   if (guestpitch[i] && guestvolume[i])
        {   if
            (   force
             || !guestplaying[i]
             || guestpitch[i]  != oldguestpitch[i]
             || guestvolume[i] != oldguestvolume[i]
            )
            {   if (i == 6)
                {   hertz = (float) (  7874.0 / (guestpitch[6] + 1)); // desired frequency of the waveform
                } else
                {   hertz = (float) (110837.0 / (guestpitch[i]    )); // desired frequency of the waveform
                }

                TonePlay(i, hertz, guestvolume[i]);
        }   }
        else
        {   SoundStop(i);
    }   }

    for (i = 7; i <= 13; i++)
    {   if (guestpitch[i] && guestvolume[i])
        {   if
            (   force
             || !guestplaying[i]
             || guestvolume[i] != oldguestvolume[i]
             || guestpitch[i] != oldguestpitch[i]
            )
            {   hertz     = (float) 7874.0 / (guestpitch[i] + 1);
                NoisePlay(i, hertz, guestvolume[i]);
        }   }
        else
        {   SoundStop(i);
    }   }

    for (i = 0; i < 14; i++)
    {   oldguestvolume[i] = guestvolume[i];
        oldguestpitch[i]  = guestpitch[i];
}   }

EXPORT void view_bram(void)
{   AUTO          int  i,
                       x, y;
    AUTO          TEXT whichletter;
    PERSIST const TEXT elektor_chars[256] =
#ifdef AMIGA
    "0123456789AbCdEFGLInPr= +-:x?_!NlTi:.";
#endif
#ifdef WIN32
    "0123456789AbCdEFGLInPr= +-:x?_!NlTi:.";
#endif

    if (machine != ELEKTOR)
    {   return;
    }
    T1;

    if (memory[0x8B4] & PSU_S  ) psubits[0] = 'S'; else psubits[0] = 's';
    if (memory[0x8B4] & PSU_F  ) psubits[1] = 'F'; else psubits[1] = 'f';
    if (memory[0x8B4] & PSU_II ) psubits[2] = 'I'; else psubits[2] = 'i';
    psubits[3] = 0;
    if (memory[0x8B3] & PSL_IDC) pslbits[0] = 'D'; else pslbits[0] = 'd';
    if (memory[0x8B3] & PSL_RS ) pslbits[1] = 'R'; else pslbits[1] = 'r';
    if (memory[0x8B3] & PSL_WC ) pslbits[2] = 'W'; else pslbits[2] = 'w';
    if (memory[0x8B3] & PSL_OVF) pslbits[3] = 'O'; else pslbits[3] = 'o';
    if (memory[0x8B3] & PSL_COM) pslbits[4] = 'M'; else pslbits[4] = 'm';
    if (memory[0x8B3] & PSL_C  ) pslbits[5] = 'C'; else pslbits[5] = 'c';
    pslbits[6] = 0;

    for (y = 0; y < 6; y++) // 6 character rows
    {   if (y == 0)
        {   printf("MONOB:  \"");
        } else
        {   printf("        \"");
        }

        for (x = 0; x < 4; x++) // 4 sprites per row
        {   whichletter = '#';
            for (i = 0; i < 256; i += 2)
            {   // check left letter against left ROM letters

                if
                (   ( memory[0x800 + (y * 24) + (x * 6)] & 0xE0)       == (memory[0x27B + ((i / 2) * 6)] & 0xE0)
                 && ( memory[0x801 + (y * 24) + (x * 6)] & 0xE0)       == (memory[0x27C + ((i / 2) * 6)] & 0xE0)
                 && ( memory[0x802 + (y * 24) + (x * 6)] & 0xE0)       == (memory[0x27D + ((i / 2) * 6)] & 0xE0)
                 && ( memory[0x803 + (y * 24) + (x * 6)] & 0xE0)       == (memory[0x27E + ((i / 2) * 6)] & 0xE0)
                 && ( memory[0x804 + (y * 24) + (x * 6)] & 0xE0)       == (memory[0x27F + ((i / 2) * 6)] & 0xE0)
                 && ( memory[0x805 + (y * 24) + (x * 6)] & 0xE0)       == (memory[0x280 + ((i / 2) * 6)] & 0xE0)
                )
                {   whichletter = elektor_chars[i];
                    break;
                }
                // check left letter against right ROM letters
                if
                (   ((memory[0x800 + (y * 24) + (x * 6)] & 0xE0) >> 4) == (memory[0x27B + ((i / 2) * 6)] & 0x0E)
                 && ((memory[0x801 + (y * 24) + (x * 6)] & 0xE0) >> 4) == (memory[0x27C + ((i / 2) * 6)] & 0x0E)
                 && ((memory[0x802 + (y * 24) + (x * 6)] & 0xE0) >> 4) == (memory[0x27D + ((i / 2) * 6)] & 0x0E)
                 && ((memory[0x803 + (y * 24) + (x * 6)] & 0xE0) >> 4) == (memory[0x27E + ((i / 2) * 6)] & 0x0E)
                 && ((memory[0x804 + (y * 24) + (x * 6)] & 0xE0) >> 4) == (memory[0x27F + ((i / 2) * 6)] & 0x0E)
                 && ((memory[0x805 + (y * 24) + (x * 6)] & 0xE0) >> 4) == (memory[0x280 + ((i / 2) * 6)] & 0x0E)
                )
                {   whichletter = elektor_chars[i + 1];
                    break;
            }   }
            printf("%c", whichletter);

            whichletter = '#';
            for (i = 0; i < 256; i += 2)
            {   // check right letter against left ROM letters
                if
                (   ( memory[0x800 + (y * 24) + (x * 6)] & 0x0E)       == (memory[0x27B + ((i / 2) * 6)] & 0xE0) >> 4
                 && ( memory[0x801 + (y * 24) + (x * 6)] & 0x0E)       == (memory[0x27C + ((i / 2) * 6)] & 0xE0) >> 4
                 && ( memory[0x802 + (y * 24) + (x * 6)] & 0x0E)       == (memory[0x27D + ((i / 2) * 6)] & 0xE0) >> 4
                 && ( memory[0x803 + (y * 24) + (x * 6)] & 0x0E)       == (memory[0x27E + ((i / 2) * 6)] & 0xE0) >> 4
                 && ( memory[0x804 + (y * 24) + (x * 6)] & 0x0E)       == (memory[0x27F + ((i / 2) * 6)] & 0xE0) >> 4
                 && ( memory[0x805 + (y * 24) + (x * 6)] & 0x0E)       == (memory[0x280 + ((i / 2) * 6)] & 0xE0) >> 4
                )
                {   whichletter = elektor_chars[i];
                    break;
                }
                // check right letter against right ROM letters
                if
                (   ( memory[0x800 + (y * 24) + (x * 6)] & 0x0E)       == (memory[0x27B + ((i / 2) * 6)] & 0x0E)
                 && ( memory[0x801 + (y * 24) + (x * 6)] & 0x0E)       == (memory[0x27C + ((i / 2) * 6)] & 0x0E)
                 && ( memory[0x802 + (y * 24) + (x * 6)] & 0x0E)       == (memory[0x27D + ((i / 2) * 6)] & 0x0E)
                 && ( memory[0x803 + (y * 24) + (x * 6)] & 0x0E)       == (memory[0x27E + ((i / 2) * 6)] & 0x0E)
                 && ( memory[0x804 + (y * 24) + (x * 6)] & 0x0E)       == (memory[0x27F + ((i / 2) * 6)] & 0x0E)
                 && ( memory[0x805 + (y * 24) + (x * 6)] & 0x0E)       == (memory[0x280 + ((i / 2) * 6)] & 0x0E)
                )
                {   whichletter = elektor_chars[i + 1];
                    break;
            }   }
            printf("%c", whichletter);
        }

        if (y == 0)
        {   printf
            (   "\" $8A0..$8AB: $%02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X\n",
                memory[0x8A0],
                memory[0x8A1],
                memory[0x8A2],
                memory[0x8A3],
                memory[0x8A4],
                memory[0x8A5],
                memory[0x8A6],
                memory[0x8A7],
                memory[0x8A8],
                memory[0x8A9],
                memory[0x8AA],
                memory[0x8AB]
            );
        } elif (y == 1)
        {   printf
            (   "\" $8B9..$8BD: $%02X %02X %02X %02X %02X\n",
                memory[0x8B9],
                memory[0x8BA],
                memory[0x8BB],
                memory[0x8BC],
                memory[0x8BD]
            ); // this area would be better shown as code rather than data
        } elif (y == 3)
        {   printf
            (   "\" LKBST:         $%02X RKBST:         $%02X MKBST:         $%02X\n",
                memory[E_LKBST],
                memory[E_RKBST],
                memory[E_MKBST]
            );
        } elif (y == 4)
        {   printf
            (   "\" BK1:         $%04X BK2:         $%04X PC:          $%04X\n",
                (256 * memory[E_BK1]) + memory[E_BK1 + 1],
                (256 * memory[E_BK2]) + memory[E_BK2 + 1],
                (256 * memory[E_PC ]) + memory[E_PC  + 1]
            );
        } elif (y == 5)
        {   printf
            (   "\" R0:            $%02X PSU:           %s PSL:        %s\n",
                memory[0x8AC],
                psubits,
                pslbits
            );
        } else
        {   printf("\"\n");
    }   }

    printf("MLINE:  \"");
    for (i = 0; i < 8; i++)
    {   printf("%c", elektor_chars[memory[0x890 + i]]);
    }
    printf
    (   "\" R1:            $%02X R2:            $%02X R3:            $%02X\n",
        memory[0x8AD],
        memory[0x8AE],
        memory[0x8AF]
    );
    printf
    (   "                   R4:            $%02X R5:            $%02X R6:            $%02X\n",
        memory[0x8B0],
        memory[0x8B1],
        memory[0x8B2]
    );
    printf
    (   "FSEQ:          $%02X ENTM:          $%02X MFUNC:         $%02X MSCR:        $%04X\n",
        memory[E_FSEQ], // FSEQ is aka SILENCE
        memory[E_ENTM],
        memory[E_MFUNC],
        (256 * memory[E_MSCR]) + memory[E_MSCR + 1]
    );
    
    printf("\n");
    T2;
}
